//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include <string>
#include <sstream>
#include <iomanip>

#include "g2pcera_api.hh"
#include "g2pcerad.hh"
#include "g2pcera_dm.hh"
#include "g2pcera_algovis.hh"

#include "vty.h"

//
//
//    PCERA API
//
//
pceraErrorCode_t
PCERA_API::g2pceraStart(std::string &	resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance();
	if (pce) {
		resp = "Default PCE instance already exists";
		METOUT();
		return PCERA_ERROR_INSTANCE_ALREADY_EXISTS;
	}

	G2PCERA_NOTICE("Creating default G2.PCE-RA instance...");

	pce = new PCERA::G2PCERA(G2PCERA_MASTER.routerId());
	if (!pce) {
		G2PCERA_ERROR("Cannot allocate enough memory "
			      "for a G2.PCE-RA instance");
		METOUT();
		return PCERA_ERROR_NO_MEMORY;
	}

	if (!G2PCERA_MASTER.attach(pce)) {
		G2PCERA_ERROR("Cannot attach G2.PCE-RA instance "
			      "to its master manager");
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_NOTICE("G2.PCE-RA instance started");

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_API::g2pceraStop(std::string &	resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance();
	if (!pce) {
		resp = "Default PCE instance does not exist";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	if (!G2PCERA_MASTER.detach(pce)) {
		G2PCERA_ERROR("Cannot detach G2.PCE-RA instance "
			      "to its master manager");
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	delete pce;
	pce = 0;

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_API::g2pceraStart(uint32_t	rootId,
			std::string &	resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(rootId);
	if (pce) {
		resp  = "PCE instance with root router ID ";
		resp += numb_htoa(rootId);
		resp += " already exists";
		METOUT();
		return PCERA_ERROR_INSTANCE_ALREADY_EXISTS;
	}

	G2PCERA_NOTICE("Creating G2.PCE-RA instance with "
		       "root router ID %s ...",
		       numb_htoa(rootId));

	pce = new PCERA::G2PCERA(rootId);
	if (!pce) {
		G2PCERA_ERROR("Cannot allocate enough memory "
			      "for a G2.PCE-RA instance");
		METOUT();
		return PCERA_ERROR_NO_MEMORY;
	}

	if (!G2PCERA_MASTER.attach(pce)) {
		G2PCERA_ERROR("Cannot attach G2.PCE-RA instance "
			      "to its master manager");
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_NOTICE("G2.PCE-RA instance started");

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_API::g2pceraStop(uint32_t		rootId,
		       std::string &	resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_NOTICE("Stopping G2.PCE-RA instance with "
		       "root router ID %s ...",
		       numb_htoa(rootId));

	pce = G2PCERA_MASTER.getPceInstance(rootId);
	if (!pce) {
		resp  = "PCE instance with root router ID ";
		resp += numb_htoa(rootId);
		resp += " not found";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	if (!G2PCERA_MASTER.detach(pce)) {
		G2PCERA_ERROR("Cannot detach G2.PCE-RA instance "
			      "to its master manager");
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	delete pce;
	pce = 0;

	G2PCERA_NOTICE("G2.PCE-RA instance stopped");

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_API::g2pceraLoadFromFile(std::string &	filename,
			       std::string &	resp)
{
	METIN();

	//read topology details from speciifed file using vty library
	vty_read_config ((char *)filename.c_str(), NULL);

	METOUT();
	return PCERA_ERROR_NONE;
}

std::list<uint32_t>
PCERA_API::g2pceraGetAll(void)
{
	METIN();

	std::list<uint32_t>				tmp;
	std::list<PCERA::G2PCERA *>			pces;
	std::list<PCERA::G2PCERA *>::const_iterator	iter;

	tmp.clear();

	pces = G2PCERA_MASTER.getPceList();

	for (iter = pces.begin(); iter != pces.end(); iter++) {
		assert(*iter);
		tmp.push_front((*iter)->rootNodeId());
	}

	METOUT();
	return tmp;
}

//
//
//  PCERA TOPOLOGY API
//
//
pceraErrorCode_t
PCERA_TOPO_API::topologyGet(uint32_t			g2pceraId,
			    topology_summary_data_t &	data,
			    std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	if (!pce->getTopology().getData(data)) {
		resp = "Cannot retrive topology data";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	std::string prefix = "   ";
	pce->getTopology().dump(prefix, true);

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::nodeAdd(uint32_t		g2pceraId,
			const node_ident_t &	ident,
			std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Adding node %s to G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node * ptr;

	ptr = pce->getTopology().getNode(ident);
	if (ptr) {
		resp  = "Node ";
		resp += nodeIdent2string(ident);
		resp += " just added into Topology";
		METOUT();
		return PCERA_ERROR_NODE_ALREADY_EXISTS;
	}

	switch (ident.type) {
		case TOPO_NODE_UNKNOWN: {
			ptr = new PCERA::Node(ident.type, ident.id);
			if (!ptr) {
				resp  = "Cannot allocate a Node object "
					"for node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_NO_MEMORY;
			}
		}
			break;
		case TOPO_NODE_NETWORK: {
			ptr = new PCERA::NetNode(ident.id);
			if (!ptr) {
				resp = "Cannot allocate a NetNode object "
					"for node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_NO_MEMORY;
			}
		}
			break;
		case TOPO_NODE_GRID: {
			ptr = new PCERA::GridNode(ident.id);
			if (!ptr) {
				resp = "Cannot allocate a GridNode object "
					"for node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_NO_MEMORY;
			}
		}
			break;
		default:
			G2PCERA_ERROR("Unrecognized node type (%d)...",
				      ident.type);
			assert(0);
			break;
	}

	if (!pce->getTopology().attachNode(ptr)) {
		delete ptr;
		resp  = "Cannot attach Node ";
		resp += nodeIdent2string(ident);
		resp += " object to Topology";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("Node %s added to G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_TOPO_API::nodeDel(uint32_t		g2pceraId,
			const node_ident_t &	ident,
			std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Deleting node %s from G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node * ptr;

	ptr = pce->getTopology().getNode(ident);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	if (ptr->type() == TOPO_NODE_NETWORK) {

		PCERA::NetNode * tmp;
		tmp = static_cast<PCERA::NetNode *>(ptr);

		assert(tmp);

		size_t links, tnas;

		links = tmp->getOutLinksCount();
		tnas  = tmp->getTnasCount();

		if (links) {
			resp  = "Cannot delete node ";
			resp += nodeIdent2string(ident);
			resp += " from Topology: ";
			resp += uint32string(links);
			resp += " TE-links still attached";
			METOUT();
			return PCERA_ERROR_GENERIC;
		}
		if (tnas) {
			resp  = "Cannot delete node ";
			resp += nodeIdent2string(ident);
			resp += " from Topology: ";
			resp += uint32string(tnas);
			resp += " TNAs still attached";
			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	} else if (ptr->type() == TOPO_NODE_GRID) {

		PCERA::GridNode * tmp;
		tmp = static_cast<PCERA::GridNode *>(ptr);

		assert(tmp);

		grid_subnodes_t  subnodes;

		subnodes = tmp->getSubnodesIdents();

		size_t services, compElems, subClusters, storageElems;

		services     = subnodes.services.size();
		compElems    = subnodes.comp_elems.size();
		subClusters  = subnodes.sub_clusters.size();
		storageElems = subnodes.storage_elems.size();

		if (services) {
			resp  = "Cannot delete node ";
			resp += nodeIdent2string(ident);
			resp += " from Topology: ";
			resp += uint32string(services);
			resp += " Grid services still attached";
			METOUT();
			return PCERA_ERROR_GENERIC;
		}
		if (compElems) {
			resp  = "Cannot delete node ";
			resp += nodeIdent2string(ident);
			resp += " from Topology: ";
			resp += uint32string(compElems);
			resp += " Grid Computing Elements still attached";
			METOUT();
			return PCERA_ERROR_GENERIC;
		}
		if (subClusters) {
			resp  = "Cannot delete node ";
			resp += nodeIdent2string(ident);
			resp += " from Topology: ";
			resp += uint32string(subClusters);
			resp += " Grid Sub-Clusters still attached";
			METOUT();
			return PCERA_ERROR_GENERIC;
		}
		if (storageElems) {
			resp  = "Cannot delete node ";
			resp += nodeIdent2string(ident);
			resp += " from Topology: ";
			resp += uint32string(storageElems);
			resp += " Grid Storage Elements still attached";
			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	}

	if (!pce->getTopology().detachNode(ptr)) {
		resp  = "Cannot detach node ";
		resp += nodeIdent2string(ident);
		resp += " from Topology";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	delete ptr;

	G2PCERA_DEBUG("Node %s deleted from G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


std::list<node_ident_t>
PCERA_TOPO_API::nodeGetAll(uint32_t		g2pceraId,
			   std::string &	resp)
{
	METIN();

	PCERA::G2PCERA *	pce;
	std::list<node_ident_t>	tmp;

	resp = "";
	tmp.clear();

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return tmp;
	}

	tmp = pce->getTopology().getNodesIdents();

	METOUT();
	return tmp;
}


pceraErrorCode_t
PCERA_TOPO_API::netNodeUpdate(uint32_t			g2pceraId,
			      uint32_t			rId,
			      const net_node_data_t &	data,
			      std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	ptr;
	node_ident_t		ident;

	ident.id   = rId;
	ident.type = TOPO_NODE_NETWORK;

	G2PCERA_DEBUG("Updating node %s in G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::NetNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	if (!ptr->setData(data)) {
		resp  = "Cannot set data in node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_PARAMS_MISMATCH;
	}

	G2PCERA_DEBUG("Node %s updated in G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::netNodeGet(uint32_t			g2pceraId,
			   uint32_t			rId,
			   net_node_data_t &		data,
			   std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	ptr;
	node_ident_t		ident;

	ident.id   = rId;
	ident.type = TOPO_NODE_NETWORK;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::NetNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	if (!ptr->getData(data)) {
		resp  = "Cannot get data from node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


#ifdef G2MPLS
pceraErrorCode_t
PCERA_TOPO_API::gridSiteUpdate(uint32_t			g2pceraId,
			       uint32_t			siteId,
			       const grid_site_data_t&	data,
			       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	G2PCERA_DEBUG("Updating node %s in G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	if (!ptr->setData(data)) {
		resp  = "Cannot set data in node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_PARAMS_MISMATCH;
	}

	G2PCERA_DEBUG("Node %s updated in G2PCERA %s",
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridSiteGet(uint32_t			g2pceraId,
			    uint32_t			siteId,
			    grid_site_data_t &		data,
			    grid_subnodes_t &		snodes,
			    std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	if (!ptr->getData(data)) {
		resp  = "Cannot get data from node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_PARAMS_MISMATCH;
	}

	snodes = ptr->getSubnodesIdents();

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridSubNodeDel(uint32_t			g2pceraId,
			       uint32_t			siteId,
			       grid_subnode_ident_t &	snIdent,
			       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	G2PCERA_DEBUG("Deleting sub-node %s from node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	switch (snIdent.type) {
		case GRIDSUBNODE_SERVICE: {
			PCERA::GridService *  snPtr;

			snPtr = ptr->getGridService(snIdent.id);
			if (!snPtr) {
				resp  = "Cannot get sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " into node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_SUBNODE_NOT_FOUND;
			}
			if (!ptr->detachGridService(snPtr)) {
				resp  = "Cannot detach sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " from node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_GENERIC;
			}

			delete snPtr;
		}
			break;
		case GRIDSUBNODE_COMPUTINGELEMENT: {
			PCERA::GridCompElem *  snPtr;

			snPtr = ptr->getGridCompElem(snIdent.id);
			if (!snPtr) {
				resp  = "Cannot get sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " into node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_SUBNODE_NOT_FOUND;
			}
			if (!ptr->detachGridCompElem(snPtr)) {
				resp  = "Cannot detach sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " from node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_GENERIC;
			}

			delete snPtr;
		}
			break;
		case GRIDSUBNODE_SUBCLUSTER: {
			PCERA::GridSubCluster *  snPtr;

			snPtr = ptr->getGridSubCluster(snIdent.id);
			if (!snPtr) {
				resp  = "Cannot get sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " into node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_SUBNODE_NOT_FOUND;
			}
			if (!ptr->detachGridSubCluster(snPtr)) {
				resp  = "Cannot detach sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " from node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_GENERIC;
			}

			delete snPtr;
		}
			break;
		case GRIDSUBNODE_STORAGEELEMENT: {
			PCERA::GridStorageElem *  snPtr;

			snPtr = ptr->getGridStorageElem(snIdent.id);
			if (!snPtr) {
				resp  = "Cannot get sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " into node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_SUBNODE_NOT_FOUND;
			}
			if (!ptr->detachGridStorageElem(snPtr)) {
				resp  = "Cannot detach sub-node ";
				resp += subnodeIdent2string(snIdent);
				resp += " from node ";
				resp += nodeIdent2string(ident);
				METOUT();
				return PCERA_ERROR_GENERIC;
			}

			delete snPtr;
		}
			break;
		default:
			G2PCERA_ERROR("Unrecognized sub-node type (%d)...",
				      snIdent.type);
			assert(0);
			break;
	}

	G2PCERA_DEBUG("Sub-node %s deleted from node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridServiceUpdate(uint32_t			g2pceraId,
				  uint32_t			siteId,
				  uint32_t			id,
				  const grid_service_data_t &	data,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridService *	snPtr;
	grid_subnode_ident_t	snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_SERVICE;

	G2PCERA_DEBUG("Updating GRID SERVICE %s in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	snPtr = ptr->getGridService(snIdent.id);

	if (!snPtr) {
		// subnode does not exist yet, gonna create it
		snPtr = new PCERA::GridService(snIdent, data);
		if (!snPtr) {
			resp  = "Cannot create sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		// and finally attach to GridNode
		if (!ptr->attachGridService(snPtr)) {
			resp  = "Cannot attach sub-node ";
			resp += subnodeIdent2string(snIdent);
			resp += " from node ";
			resp += nodeIdent2string(ident);

			delete snPtr;

			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	} else {
		// just update/set internal data
		if (!snPtr->setData(data)) {
			resp  = "Cannot update data into sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_SUBNODE_PARAMS_MISMATCH;
		}
	}

	G2PCERA_DEBUG("GRID SERVICE %s updated in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());


	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridServiceGet(uint32_t				g2pceraId,
			       uint32_t				siteId,
			       uint32_t				id,
			       grid_service_data_t &		data,
			       std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridService *	snPtr;
	grid_subnode_ident_t	snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_SERVICE;

	snPtr = ptr->getGridService(snIdent.id);

	if (!snPtr) {
		resp  = "Cannot get sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_SUBNODE_NOT_FOUND;
	}

	if (!snPtr->getData(data)) {
		resp  = "Cannot get data from sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridCompElemUpdate(uint32_t			g2pceraId,
				   uint32_t			siteId,
				   uint32_t			id,
				   const grid_ce_data_t &	data,
				   std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridCompElem *	snPtr;
	grid_subnode_ident_t	snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_COMPUTINGELEMENT;

	G2PCERA_DEBUG("Updating GRID COMPELEM %s in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	snPtr = ptr->getGridCompElem(snIdent.id);

	if (!snPtr) {
		// subnode does not exist yet, gonna create it
		snPtr = new PCERA::GridCompElem(snIdent, data);
		if (!snPtr) {
			resp  = "Cannot create sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		// and finally attach to GridNode
		if (!ptr->attachGridCompElem(snPtr)) {
			resp  = "Cannot attach sub-node ";
			resp += subnodeIdent2string(snIdent);
			resp += " from node ";
			resp += nodeIdent2string(ident);

			delete snPtr;

			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	} else {
		// just update/set internal data
		if (!snPtr->setData(data)) {
			resp  = "Cannot update data into sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_SUBNODE_PARAMS_MISMATCH;
		}
	}

	G2PCERA_DEBUG("GRID COMPELEM %s updated in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridCompElemGet(uint32_t			g2pceraId,
				uint32_t			siteId,
				uint32_t			id,
				grid_ce_data_t &		data,
				std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridCompElem *	snPtr;
	grid_subnode_ident_t	snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_COMPUTINGELEMENT;

	snPtr = ptr->getGridCompElem(snIdent.id);

	if (!snPtr) {
		resp  = "Cannot get sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_SUBNODE_NOT_FOUND;
	}

	if (!snPtr->getData(data)) {
		resp  = "Cannot get data from sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridSubClusterUpdate(uint32_t			g2pceraId,
				     uint32_t				siteId,
				     uint32_t				id,
				     const grid_subcluster_data_t &	data,
				     std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridSubCluster *	snPtr;
	grid_subnode_ident_t	snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_SUBCLUSTER;

	G2PCERA_DEBUG("Updating GRID SUBCLUSTER %s in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	snPtr = ptr->getGridSubCluster(snIdent.id);

	if (!snPtr) {
		// subnode does not exist yet, gonna create it
		snPtr = new PCERA::GridSubCluster(snIdent, data);
		if (!snPtr) {
			resp  = "Cannot create sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		// and finally attach to GridNode
		if (!ptr->attachGridSubCluster(snPtr)) {
			resp  = "Cannot attach sub-node ";
			resp += subnodeIdent2string(snIdent);
			resp += " from node ";
			resp += nodeIdent2string(ident);

			delete snPtr;

			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	} else {
		// just update/set internal data
		if (!snPtr->setData(data)) {
			resp  = "Cannot update data into sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_SUBNODE_PARAMS_MISMATCH;
		}
	}

	G2PCERA_DEBUG("GRID SUBCLUSTER %s updated in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridSubClusterGet(uint32_t			g2pceraId,
				  uint32_t			siteId,
				  uint32_t			id,
				  grid_subcluster_data_t &	data,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridSubCluster *	snPtr;
	grid_subnode_ident_t	snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_SUBCLUSTER;

	snPtr = ptr->getGridSubCluster(snIdent.id);

	if (!snPtr) {
		resp  = "Cannot get sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_SUBNODE_NOT_FOUND;
	}

	if (!snPtr->getData(data)) {
		resp  = "Cannot get data from sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridStorageElemUpdate(uint32_t			g2pceraId,
				      uint32_t			siteId,
				      uint32_t			id,
				      const grid_se_data_t &	data,
				      std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridStorageElem *	snPtr;
	grid_subnode_ident_t		snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_STORAGEELEMENT;

	G2PCERA_DEBUG("Updating GRID STOELEM %s in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	snPtr = ptr->getGridStorageElem(snIdent.id);

	if (!snPtr) {
		// subnode does not exist yet, gonna create it
		snPtr = new PCERA::GridStorageElem(snIdent, data);
		if (!snPtr) {
			resp  = "Cannot create sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		// and finally attach to GridNode
		if (!ptr->attachGridStorageElem(snPtr)) {
			resp  = "Cannot attach sub-node ";
			resp += subnodeIdent2string(snIdent);
			resp += " from node ";
			resp += nodeIdent2string(ident);

			delete snPtr;

			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	} else {
		// just update/set internal data
		if (!snPtr->setData(data)) {
			resp  = "Cannot update data into sub-node ";
			resp += subnodeIdent2string(snIdent);
			METOUT();
			return PCERA_ERROR_SUBNODE_PARAMS_MISMATCH;
		}
	}

	G2PCERA_DEBUG("GRID STOELEM %s updated in node %s in G2PCERA %s",
		      subnodeIdent2string(snIdent).c_str(),
		      nodeIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::gridStorageElemGet(uint32_t		g2pceraId,
				   uint32_t		siteId,
				   uint32_t		id,
				   grid_se_data_t &	data,
				   std::string &	resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::GridNode *	ptr;
	node_ident_t		ident;

	ident.id   = siteId;
	ident.type = TOPO_NODE_GRID;

	tmp = pce->getTopology().getNode(ident);
	ptr = static_cast<PCERA::GridNode *>(tmp);
	if (!ptr) {
		resp  = "Cannot get node ";
		resp += nodeIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::GridStorageElem *	snPtr;
	grid_subnode_ident_t		snIdent;

	snIdent.id   = id;
	snIdent.type = GRIDSUBNODE_STORAGEELEMENT;

	snPtr = ptr->getGridStorageElem(snIdent.id);

	if (!snPtr) {
		resp  = "Cannot get sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_SUBNODE_NOT_FOUND;
	}

	if (!snPtr->getData(data)) {
		resp  = "Cannot get data from sub-node ";
		resp += subnodeIdent2string(snIdent);
		resp += " into node ";
		resp += nodeIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}

#endif // G2MPLS

pceraErrorCode_t
PCERA_TOPO_API::tnaAdd(uint32_t			g2pceraId,
		       const tna_ident_t &	ident,
		       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Adding TNA %s in G2PCERA %s",
		      tnaIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode && ident.rc_id) {
		// make a second atempt on domains

		lclNodeIdent.id   = ident.rc_id;
		lclNodeIdent.type = TOPO_NODE_NETWORK;

		tmp = pce->getTopology().getNode(lclNodeIdent);
		lclNode = static_cast<PCERA::NetNode *>(tmp);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TNA ";
		resp += tnaIdent2string(ident).c_str();
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::Tna * ptr;

	ptr = lclNode->getTna(ident);
	if (ptr) {
		resp  = "TNA ";
		resp += tnaIdent2string(ident).c_str();
		resp += " just added into node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " or just included in an existing TNA pref";
		METOUT();
		return PCERA_ERROR_TNA_ALREADY_EXISTS;
	}

	ptr = new PCERA::Tna(ident);
	if (!ptr) {
		resp  = "Cannot allocate a Tna "
			"object for TNA ";
		resp += tnaIdent2string(ident).c_str();
		METOUT();
		return PCERA_ERROR_NO_MEMORY;
	}

	if (!lclNode->attachTna(ptr)) {
		delete ptr;
		resp  = "Cannot attach Tna ";
		resp += tnaIdent2string(ident).c_str();
		resp += " object to node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("TNA %s added in G2PCERA %s",
		      tnaIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_TOPO_API::tnaDel(uint32_t			g2pceraId,
		       const tna_ident_t &	ident,
		       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Deleting TNA %s from G2PCERA %s",
		      tnaIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode && ident.rc_id) {
		// make a second atempt on domains

		lclNodeIdent.id   = ident.rc_id;
		lclNodeIdent.type = TOPO_NODE_NETWORK;

		tmp = pce->getTopology().getNode(lclNodeIdent);
		lclNode = static_cast<PCERA::NetNode *>(tmp);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TNA ";
		resp += tnaIdent2string(ident).c_str();
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::Tna * ptr;

	ptr = lclNode->getTna(ident);
	if (!ptr) {
		resp  = "Cannot get ";
		resp += tnaIdent2string(ident).c_str();
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_TNA_NOT_FOUND;
	}


	if (!lclNode->detachTna(ptr)) {
		resp  = "Cannot detach TNA ";
		resp += tnaIdent2string(ident).c_str();
		resp += " object from node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}


	delete ptr;

	G2PCERA_DEBUG("TNA %s deleted from G2PCERA %s",
		      tnaIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}

std::list<tna_ident_t>
PCERA_TOPO_API::tnaGetAllFromNode(uint32_t		g2pceraId,
				  uint32_t		rId,
				  bool			isDomain,
				  std::string &		resp)
{
	METIN();

	PCERA::G2PCERA *	pce;
	std::list<tna_ident_t>	tmp;

	resp = "";
	tmp.clear();

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return tmp;
	}

	PCERA::Node *		gNode;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;
	uint32_t		borderRid;

	borderRid         = 0;
	lclNodeIdent.id   = rId;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	gNode = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(gNode);

	if (!lclNode && isDomain) {
		resp  = "Cannot get domain node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " into Topology";
		METOUT();
		return tmp;
	}

	if (lclNode && isDomain && !lclNode->isDomain()) {
		resp  = "Cannot get domain node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " into Topology, it is a non-domain netNode";
		METOUT();
		return tmp;
	}

	if (lclNode) {
		// node is in my topology, get TNAs directly
		tmp = lclNode->getTnasIdents(0);
	} else {
		// node could be under a domain, get TNAs directly
		std::list<node_ident_t>			nodeList;
		std::list<node_ident_t>::iterator	iter;

		nodeList = pce->getTopology().getNodesIdents();

		for (iter = nodeList.begin(); iter != nodeList.end(); iter++) {
			node_ident_t	id;

			id = (*iter);

			if (id.type != TOPO_NODE_NETWORK) {
				continue;
			}

			gNode = pce->getTopology().getNode(id);
			lclNode = static_cast<PCERA::NetNode *>(gNode);

			if (lclNode && lclNode->isDomain()) {
				tmp = lclNode->getTnasIdents(rId);

				if (!tmp.empty()) {
					break;
				}
			}
		}

		if (tmp.empty()) {
			resp  = "Cannot get a domain node with net-node ";
			resp += nodeIdent2string(lclNodeIdent);
			resp += " into Topology";
		}
	}

	METOUT();
	return tmp;
}

pceraErrorCode_t
PCERA_TOPO_API::linkAdd(uint32_t			g2pceraId,
			const telink_ident_t &		ident,
			std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Adding TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (ptr) {
		resp  = "Link ";
		resp += teLinkIdent2string(ident);
		resp += " just added into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_ALREADY_EXISTS;
	}

	ptr = new PCERA::TeLink(ident);
	if (!ptr) {
		resp  = "Cannot allocate a TeLink "
			"object for TE-Link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_NO_MEMORY;
	}

	if (!lclNode->attachOutLink(ptr)) {
		delete ptr;
		resp  = "Cannot attach TE-Link ";
		resp += teLinkIdent2string(ident);
		resp += " object to node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("TE-link %s added in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::linkDel(uint32_t			g2pceraId,
			const telink_ident_t &		ident,
			std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Deleting TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into no";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!lclNode->detachOutLink(ptr)) {
		resp  = "Cannot detach link ";
		resp += teLinkIdent2string(ident);
		resp += " from node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	delete ptr;

	G2PCERA_DEBUG("TE-link %s deleted from G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


std::list<telink_ident_t>
PCERA_TOPO_API::teLinkGetAllFromNode(uint32_t		g2pceraId,
				     uint32_t		rId,
				     std::string &	resp)
{
	METIN();

	PCERA::G2PCERA *		pce;
	std::list<telink_ident_t>	tmp;

	resp = "";
	tmp.clear();

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return tmp;
	}

	PCERA::Node *		gNode;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = rId;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	gNode = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(gNode);
	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " into Topology";
		METOUT();
		return tmp;
	}

	tmp = lclNode->getOutLinksIdents();

	METOUT();
	return tmp;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateCom(uint32_t			g2pceraId,
				const telink_ident_t &		ident,
				const telink_com_data_t &	data,
				std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating common data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->setData(data)) {
		resp  = "Cannot update common data into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("Common data updated for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetCom(uint32_t			g2pceraId,
			     const telink_ident_t &	ident,
			     telink_com_data_t &	data,
			     std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->getData(data)) {
		resp  = "Cannot get common data from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateTdm(uint32_t			g2pceraId,
				const telink_ident_t &		ident,
				const telink_tdm_data_t &	data,
				std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating TDM data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	lclNode = static_cast<PCERA::NetNode *>(tmp);
	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}


	PCERA::IscTdmSdhSonetInfo * childPtr;

	childPtr = ptr->getIscTdmSdhSonet();

	if (!childPtr) {
		resp  = "Cannot update TDM data into link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no TDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->setData(data)) {
		resp  = "Cannot update TDM data into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("TDM data updated for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetTdm(uint32_t			g2pceraId,
			     const telink_ident_t &	ident,
			     telink_tdm_data_t &	data,
			     std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscTdmSdhSonetInfo * childPtr;

	childPtr = ptr->getIscTdmSdhSonet();

	if (!childPtr) {
		resp  = "Cannot get TDM data from link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no TDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->getData(data)) {
		resp  = "Cannot get TDM data from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateLscG709(uint32_t			g2pceraId,
				    const telink_ident_t &		ident,
				    const telink_lscg709_data_t &	data,
				    std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating G.709 data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscG709Info * childPtr;

	childPtr = ptr->getIscLscG709();

	if (!childPtr) {
		resp  = "Cannot update LSC G.709 data in link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC G.709 ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->setData(data)) {
		resp  = "Cannot update G.709 data into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("G.709 data updated for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetLscG709(uint32_t			g2pceraId,
				 const telink_ident_t &		ident,
				 telink_lscg709_data_t &	data,
				 std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscG709Info * childPtr;

	childPtr = ptr->getIscLscG709();

	if (!childPtr) {
		resp  = "Cannot get LSC G.709 data from link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC G.709 ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->getData(data)) {
		resp  = "Cannot get G.709 data from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateLscWdm(uint32_t			g2pceraId,
				   const telink_ident_t &	ident,
				   const telink_lscwdm_data_t &	data,
				   std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating WDM data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscLambdaInfo * childPtr;

	childPtr = ptr->getIscLscLambda();

	if (!childPtr) {
		resp  = "Cannot update LSC WDM data in link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC WDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->setData(data)) {
		resp  = "Cannot update WDM data into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("WDM data updated for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetLscWdm(uint32_t		g2pceraId,
				const telink_ident_t &	ident,
				telink_lscwdm_data_t &	data,
				std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscLambdaInfo * childPtr;

	childPtr = ptr->getIscLscLambda();

	if (!childPtr) {
		resp  = "Cannot get LSC WDM data from link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC WDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->getData(data)) {
		resp  = "Cannot get WDM data from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateStates(uint32_t			g2pceraId,
				   const telink_ident_t &	ident,
				   opstate_t			opState,
				   admstate_t			admState,
				   std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating states for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->setStates(opState, admState)) {
		resp  = "Cannot update states into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("States updated for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetStates(uint32_t			g2pceraId,
				const telink_ident_t &		ident,
				opstate_t &			opState,
				admstate_t &			admState,
				std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->getStates(opState, admState)) {
		resp  = "Cannot get states from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateGenBw(uint32_t			g2pceraId,
				  const telink_ident_t &	ident,
				  const bw_per_prio_t &		avail_bw,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating available bw data for TE-link %s "
		      "in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->setGenAvailBw(avail_bw)) {
		resp  = "Cannot update generaliz. avail. bw info into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("Available bw updated data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetGenBw(uint32_t			g2pceraId,
			       const telink_ident_t &	ident,
			       bw_per_prio_t &		avail_bw,
			       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->getGenAvailBw(avail_bw)) {
		resp  = "Cannot get generaliz. avail. bw info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateTdmBw(uint32_t			g2pceraId,
				  const telink_ident_t &	ident,
				  const std::list<uint32_t> &	freeTS,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating TDM bw data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscTdmSdhSonetInfo * childPtr;

	childPtr = ptr->getIscTdmSdhSonet();

	if (!childPtr) {
		resp  = "Cannot update TDM bw data into link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no TDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->setTdmAvailBw(freeTS)) {
		resp  = "Cannot update TDM free timeslots info into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	bw_per_prio_t			abw;

	if (!ptr->getGenAvailBw(abw)) {
		resp  = "Cannot get generaliz. avail. bw info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	std::list<uint32_t>::const_iterator	iter;
	float					fbw;
	uint32_t				ubw;

	abw.bw[0] = 0;		// reset
	fbw       = 0;

	for (iter = freeTS.begin(); iter != freeTS.end(); iter++) {

		ubw = GMPLS_BW_FROM_SDH_ST(GMPLS_SIGNAL_TYPE(*iter));

		fbw += (((float) GMPLS_SDH_UNUSED_TIME_SLOT(*iter)) *
			BW_HEX2BPS(ubw));
	}
	abw.bw[0] = BW_BPS2HEX(fbw);

	if (!ptr->setGenAvailBw(abw)) {
		resp  = "Cannot get generaliz. avail. bw info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("TDM bw updated data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetTdmBw(uint32_t			g2pceraId,
			       const telink_ident_t &	ident,
			       std::list<uint32_t> &	freeTS,
			       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscTdmSdhSonetInfo * childPtr;

	childPtr = ptr->getIscTdmSdhSonet();

	if (!childPtr) {
		resp  = "Cannot get TDM bw data from link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no TDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->getTdmAvailBw(freeTS)) {
		resp  = "Cannot get TDM free timeslots info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateLscG709Bw(uint32_t			   g2pceraId,
				      const telink_ident_t &	   ident,
				      const std::list<uint32_t> &  freeODUk,
				      const std::list<uint32_t> &  freeOCh,
				      std::string &		   resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating G.709 bw data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp  = " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscG709Info * childPtr;

	childPtr = ptr->getIscLscG709();

	if (!childPtr) {
		resp  = "Cannot update LSC G.709 bw data in link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC G.709 ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->setLscG709AvailBw(freeODUk, freeOCh)) {
		resp  = "Cannot update G.709 free timeslots info into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	bw_per_prio_t			abw;

	if (!ptr->getGenAvailBw(abw)) {
		resp  = "Cannot get generaliz. avail. bw info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	std::list<uint32_t>::const_iterator	iter;
	float					fbw;
	uint32_t				ubw;

	abw.bw[0] = 0;		// reset
	fbw       = 0;

	for (iter = freeODUk.begin(); iter != freeODUk.end(); iter++) {

		ubw = GMPLS_BW_FROM_G709_ST(GMPLS_SIGNAL_TYPE(*iter));

		fbw += (((float) GMPLS_G709_UNUSED_ODUK(*iter)) *
			BW_HEX2BPS(ubw));
	}

	for (iter = freeOCh.begin(); iter != freeOCh.end(); iter++) {

		ubw = GMPLS_BW_FROM_G709_ST(GMPLS_SIGNAL_TYPE(*iter));

		fbw += (((float) GMPLS_G709_UNUSED_OCH(*iter)) *
			BW_HEX2BPS(ubw));
	}

	abw.bw[0] = BW_BPS2HEX(fbw);

	if (!ptr->setGenAvailBw(abw)) {
		resp  = "Cannot get generaliz. avail. bw info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("G.709 bw updated data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetLscG709Bw(uint32_t			g2pceraId,
				   const telink_ident_t &	ident,
				   std::list<uint32_t> &	freeODUk,
				   std::list<uint32_t> &	freeOCh,
				   std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscG709Info * childPtr;

	childPtr = ptr->getIscLscG709();

	if (!childPtr) {
		resp  = "Cannot get LSC G.709 bw data from link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC G.709 ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->getLscG709AvailBw(freeODUk, freeOCh)) {
		resp  = "Cannot get G.709 free timeslots info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkUpdateLscWdmBw(uint32_t			  g2pceraId,
				     const telink_ident_t &	  ident,
				     wdm_link_lambdas_bitmap_t &  bm,
				     std::string &		  resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Updating WDM bw data for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscLambdaInfo * childPtr;

	childPtr = ptr->getIscLscLambda();

	if (!childPtr) {
		resp  = "Cannot update LSC WDM bw data in link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC WDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->setLscWdmAvailBw(bm)) {
		resp  = "Cannot update WDM wavelength bitmap into link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	bw_per_prio_t			abw;

	if (!ptr->getGenAvailBw(abw)) {
		resp  = "Cannot get generaliz. avail. bw info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}
	abw.bw[0] = 0;		// reset

	std::list<uint32_t>::iterator	iter;

	float		fbw;
	uint32_t	wv_bw_unit;
	uint32_t	free_waves;
	size_t		i;

	free_waves = 0;
	wv_bw_unit = BwEnc_10GigELAN;

	for (i = 0; i < bm.bitmap_size; i++) {
		size_t	j;

		for (j = 0; j < 32; j++) {
			uint32_t mask;

			mask = 1 << j;

			if (bm.bitmap_word[i] & mask) {
				free_waves++;
			}
		}
	}

	fbw       = (((float) free_waves) *
		     BW_HEX2BPS(wv_bw_unit));
	abw.bw[0] = BW_BPS2HEX(fbw);

	if (!ptr->setGenAvailBw(abw)) {
		resp  = "Cannot get generaliz. avail. bw info from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("WDM bw data updated for TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetLscWdmBw(uint32_t			g2pceraId,
				  const telink_ident_t &	ident,
				  wdm_link_lambdas_bitmap_t &	bm,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	PCERA::IscLscLambdaInfo * childPtr;

	childPtr = ptr->getIscLscLambda();

	if (!childPtr) {
		resp  = "Cannot get LSC WDM bw data from link ";
		resp += teLinkIdent2string(ident);
		resp += " which has no LSC WDM ISC";
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	if (!childPtr->getLscWdmAvailBw(bm)) {
		resp  = "Cannot get WDM wavelength bitmap from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkAppendSrlgs(uint32_t			g2pceraId,
				  const telink_ident_t &	ident,
				  const std::list<uint32_t> &	srlgs,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Appending SRLGs to TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->appendSrlgs(srlgs)) {
		resp  = "Cannot append an srlg list to link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("SRLGs appended to TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetSrlgs(uint32_t			g2pceraId,
			       const telink_ident_t &	ident,
			       std::list<uint32_t> &	srlgs,
			       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->getSrlgs(srlgs)) {
		resp  = "Cannot get the srlg list from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkAppendCalendar(uint32_t			g2pceraId,
				     const telink_ident_t &	ident,
				     const
				     std::map<uint32_t, bw_per_prio_t> & cal,
				     std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Appending calendar to TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->appendCalEvents(cal)) {
		resp  = "Cannot append a calendar to link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	G2PCERA_DEBUG("Calendar appended to TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetCalendar(uint32_t			g2pceraId,
				  const telink_ident_t &	ident,
				  std::map<uint32_t, bw_per_prio_t> & cal,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->getCalEvents(cal)) {
		resp  = "Cannot get calendar from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}



pceraErrorCode_t
PCERA_TOPO_API::teLinkAppendIsc(uint32_t			g2pceraId,
				const telink_ident_t &		ident,
				const std::list<isc_t> &	iscs,
				std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Appending ISCs to TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * lptr;

	lptr = lclNode->getOutLink(ident);
	if (!lptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	std::list<isc_t>::const_iterator iter;
	for (iter = iscs.begin(); iter != iscs.end(); iter++) {
		PCERA::IscInfo *	iptr;
		sw_cap_t		sw;
		enc_type_t		enc;
		bw_per_prio_t		mxBw;

		sw     = (*iter).switching;
		enc    = (*iter).encoding;
		mxBw = (*iter).max_lsp_bw;

		switch (sw) {
			case SWCAP_PSC_1:
			case SWCAP_PSC_2:
			case SWCAP_PSC_3:
			case SWCAP_PSC_4: {
				uint32_t	mnBw;
				uint16_t	ifMtu;

				mnBw = (*iter).spec.psc.min_lsp_bw;
				ifMtu    = (*iter).spec.psc.if_mtu;

				if (enc == ENCT_PACKET) {
					iptr = new
						PCERA::IscPscPacketInfo(sw,
									enc,
									mxBw,
									mnBw,
									ifMtu);
				} else if ((enc == ENCT_ETHERNET) ||
					   (enc == ENCT_FIBERCHANNEL)) {
					iptr = new PCERA::IscInfo(sw,
								  enc,
								  mxBw);
				} else {
					resp += "Swcap/Enctype mismatch: ";
					resp += "cannot append an ISC ";
					resp += "list to link ";
					resp += teLinkIdent2string(ident);

					METOUT();
					return PCERA_ERROR_GENERIC;
				}
			}
				break;

			case SWCAP_TDM: {
				uint32_t			mnBw;
				gmpls_sdhsonet_stdarbcap_t	ind;

				mnBw = (*iter).spec.tdm.min_lsp_bw;
				ind = (gmpls_sdhsonet_stdarbcap_t)
					((*iter).spec.tdm.indication);

				if (enc == ENCT_SDH_SONET) {
					iptr = new
						PCERA::IscTdmSdhSonetInfo(sw,
									  enc,
									  mxBw,
									  mnBw,
									  ind);
				} else if (enc == ENCT_ANSI_ETSI_PDH){
					iptr = new PCERA::IscInfo(sw,
								  enc,
								  mxBw);
				} else {
					resp += "Swcap/Enctype mismatch: ";
					resp += "cannot append an ISC ";
					resp += "list to link ";
					resp += teLinkIdent2string(ident);

					METOUT();
					return PCERA_ERROR_GENERIC;
				}

			}
				break;

			case SWCAP_LSC: {
				if ((enc == ENCT_G709_ODU) ||
				    (enc == ENCT_G709_OC)) {
					iptr = new
						PCERA::IscLscG709Info(sw,
								      enc,
								      mxBw);
				} else if (enc == ENCT_LAMBDA) {
					iptr = new
						PCERA::IscLscLambdaInfo(sw,
									enc,
									mxBw);
				} else if (enc == ENCT_DIGITAL_WRAPPER) {
					iptr = new PCERA::IscInfo(sw,
								  enc,
								  mxBw);
				} else {
					resp += "Swcap/Enctype mismatch: ";
					resp += "cannot append an ISC ";
					resp += "list to link ";
					resp += teLinkIdent2string(ident);

					METOUT();
					return PCERA_ERROR_GENERIC;
				}
			}
				break;
			case SWCAP_L2SC: {
				if (enc == ENCT_ETHERNET) {
					iptr = new PCERA::IscInfo(sw,
								  enc,
								  mxBw);
				} else {
					resp += "Swcap/Enctype mismatch: ";
					resp += "cannot append an ISC ";
					resp += "list to link ";
					resp += teLinkIdent2string(ident);

					METOUT();
					return PCERA_ERROR_GENERIC;
				}
			}
				break;
			case SWCAP_FSC : {
				if (enc == ENCT_FIBER) {
					iptr = new PCERA::IscInfo(sw,
								  enc,
								  mxBw);
				} else {
					resp += "Swcap/Enctype mismatch: ";
					resp += "cannot append an ISC ";
					resp += "list to link ";
					resp += teLinkIdent2string(ident);

					METOUT();
					return PCERA_ERROR_GENERIC;
				}
			}
				break;
			default: {
				G2PCERA_ERROR("Unrecognized swcap (%d)...",
					      (*iter).switching);
				BUG();
				assert(0);
			}
				break;
		}

		if (!iptr) {
			resp += "Cannot allocate a ISC info object";
			resp += "to append to ISC TE-link ";
			resp += teLinkIdent2string(ident);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		if (!lptr->attachIscInfo(iptr)) {
			resp  = "Cannot append an ISC to TE-link ";
			resp += teLinkIdent2string(ident);
			delete iptr;
			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	}

	G2PCERA_DEBUG("ISCs appended to TE-link %s in G2PCERA %s",
		      teLinkIdent2string(ident).c_str(),
		      rid2string(g2pceraId).c_str());
	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_TOPO_API::teLinkGetIsc(uint32_t			g2pceraId,
			     const telink_ident_t &	ident,
			     std::list<isc> &		iscs,
			     std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::Node *		tmp;
	PCERA::NetNode *	lclNode;
	node_ident_t		lclNodeIdent;

	lclNodeIdent.id   = ident.lcl_rid;
	lclNodeIdent.type = TOPO_NODE_NETWORK;

	if (ident.mode == LINK_MODE_ENNI_INTRADOMAIN) {
		if (!ident.rem_rc_id ||
		    (ident.rem_rc_id != ident.lcl_rc_id)) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTRADOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.rem_rc_id;

	} else if (ident.mode == LINK_MODE_ENNI_INTERDOMAIN) {
		if (ident.rem_rc_id == ident.lcl_rc_id) {
				resp  = "Inconsisten identity ";
				resp += "for LINK_MODE_ENNI_INTERDOMAIN ";
				resp += teLinkIdent2string(ident);
				METOUT();
				return PCERA_ERROR_INVALID_PARAMETER;
		}

		lclNodeIdent.id = ident.lcl_rc_id;

		if (!ident.lcl_rc_id) {
			lclNodeIdent.id = ident.lcl_rid;
		}
	}

	tmp = pce->getTopology().getNode(lclNodeIdent);
	lclNode = static_cast<PCERA::NetNode *>(tmp);

	if (!lclNode) {
		// make a second attempt on domains
		lclNode = pce->getTopology().getAncestorNode(lclNodeIdent.id);
	}

	if (!lclNode) {
		resp  = "Cannot get local node ";
		resp += nodeIdent2string(lclNodeIdent);
		resp += " for TE-link ";
		resp += teLinkIdent2string(ident);
		resp += " into Topology";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND;
	}

	PCERA::TeLink * ptr;

	ptr = lclNode->getOutLink(ident);
	if (!ptr) {
		resp  = "Cannot get link ";
		resp += teLinkIdent2string(ident);
		resp += " into node ";
		resp += nodeIdent2string(lclNodeIdent);
		METOUT();
		return PCERA_ERROR_LINK_NOT_FOUND;
	}

	if (!ptr->getIscs(iscs)) {
		resp  = "Cannot get ISCs list from link ";
		resp += teLinkIdent2string(ident);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	METOUT();
	return PCERA_ERROR_NONE;
}



//
//
//  PCERA CALCULATION API
//
//

#define CHECK_INPUT(P, PSTR, ECODE, RSTR)	\
{						\
	std::string err;			\
	if (!isValid(P, err)) {			\
		(RSTR)  = "Input parameter ";	\
		(RSTR) += (PSTR);		\
		(RSTR) += " is not valid (";	\
		(RSTR) += err;			\
		(RSTR) += ")";			\
		METOUT();			\
		return (ECODE);			\
	}					\
}

pceraErrorCode_t
PCERA_CALC_API::nodeGetFromTnaRes(uint32_t			g2pceraId,
				  const  net_res_spec_t &	tnaRes,
				  uint32_t	&		rId,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";
	rId  = 0;

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	CHECK_INPUT(tnaRes, "tnaRes", PCERA_ERROR_INVALID_PARAMETER, resp);

	g2mpls_addr_t tna;

	memcpy(&tna, &(tnaRes.tna), sizeof(tna));

	std::list<node_ident_t>			nodeList;
	std::list<node_ident_t>::iterator	iterN;

	nodeList = pce->getTopology().getNodesIdents();

	for (iterN = nodeList.begin(); iterN != nodeList.end(); iterN++) {
		std::list<tna_ident_t>			tnas;
		std::list<tna_ident_t>::iterator	iterT;
		node_ident_t				id;
		PCERA::Node *				gNode;
		PCERA::NetNode *			lclNode;

		id = (*iterN);

		if (id.type != TOPO_NODE_NETWORK) {
			continue;
		}

		gNode = pce->getTopology().getNode(id);
		lclNode = static_cast<PCERA::NetNode *>(gNode);

		assert(lclNode);

		tnas = lclNode->getTnasIdents(0); // get all from this node

		for (iterT = tnas.begin(); iterT != tnas.end(); iterT++) {
			if (addr_equal((*iterT).id, tna) != 0) {
				rId = (*iterT).rid;
				METOUT();
				return PCERA_ERROR_NONE;
			}
		}
	}

	resp  = "TNA ";
	resp += addr_ntoa(tna);
	resp += " not found under any network node in G2PCERA ";
	resp += rid2string(g2pceraId);
	METOUT();
	return PCERA_ERROR_TNA_NOT_FOUND;
}

pceraErrorCode_t
PCERA_CALC_API::nodeGetFromTna(uint32_t			g2pceraId,
			       const  g2mpls_addr_t &	tna,
			       uint32_t	&		rId,
			       std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";
	rId  = 0;

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	std::list<node_ident_t>			nodeList;
	std::list<node_ident_t>::iterator	iterN;

	nodeList = pce->getTopology().getNodesIdents();

	for (iterN = nodeList.begin(); iterN != nodeList.end(); iterN++) {
		std::list<tna_ident_t>			tnas;
		std::list<tna_ident_t>::iterator	iterT;
		node_ident_t				id;
		PCERA::Node *				gNode;
		PCERA::NetNode *			lclNode;

		id = (*iterN);

		if (id.type != TOPO_NODE_NETWORK) {
			continue;
		}

		gNode = pce->getTopology().getNode(id);
		lclNode = static_cast<PCERA::NetNode *>(gNode);

		assert(lclNode);

		tnas = lclNode->getTnasIdents(0);

		for (iterT = tnas.begin(); iterT != tnas.end(); iterT++) {
			if (addr_equal((*iterT).id, tna) != 0) {
				rId = (*iterT).rid;
				METOUT();
				return PCERA_ERROR_NONE;
			}
		}
	}

	resp  = "TNA ";
	resp += addr_ntoa(tna);
	resp += " not found under any network node in G2PCERA ";
	resp += rid2string(g2pceraId);
	METOUT();
	return PCERA_ERROR_TNA_NOT_FOUND;
}


#ifdef G2MPLS
pceraErrorCode_t
PCERA_CALC_API::nodeGetFromGnsTna(uint32_t			g2pceraId,
				  const grid_res_spec_t &	tnaGnsRes,
				  const std::list<uint32_t> &	excludeSet,
				  uint32_t &			netNodeId,
				  uint32_t &			gridSiteId,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	CHECK_INPUT(tnaGnsRes,"tnaGnsRes",PCERA_ERROR_INVALID_PARAMETER,resp);

	netNodeId  = 0;
	gridSiteId = 0;

	std::list<node_ident_t>			nodeList;
	std::list<node_ident_t>::iterator	iterN;
	PCERA::GridNode *			selNode;
	uint32_t				ranking;

	selNode = 0;
	ranking = 0;

	nodeList = pce->getTopology().getNodesIdents();
	for (iterN = nodeList.begin(); iterN != nodeList.end(); iterN++) {
		node_ident_t				id;
		PCERA::Node *				gNode;
		PCERA::GridNode *			lclNode;
		uint32_t				lclRanking;

		id = (*iterN);

		if (id.type != TOPO_NODE_GRID) {
			continue;
		}

		gNode = pce->getTopology().getNode(id);
		lclNode = static_cast<PCERA::GridNode *>(gNode);

		assert(lclNode);

		lclRanking = lclNode->match(tnaGnsRes);

		if (!lclRanking || (lclRanking < ranking)) {
			continue;
		}

		selNode = lclNode;
		ranking = lclRanking;
	}

	if (!selNode) {
		resp  = "Cannot find a Grid node matching ";
		resp += "the passed GNS grid_res_spec_t";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND; // or NO_ROUTE ??
	}

	node_ident_t netId;
	PCERA::Node *		tmp;
	PCERA::NetNode *	ptr;

	netId.id   = selNode->getPeRouterId();
	netId.type = TOPO_NODE_NETWORK;

	tmp = pce->getTopology().getNode(netId);
	ptr = static_cast<PCERA::NetNode *>(tmp);

	if (!ptr) {
		// peRouter could be under a domain
		resp  = "Cannot get peering netNode ";
		resp += nodeIdent2string(netId);
		resp += "... it might not exist yet or ";
		resp +="be out of this Routing Domain";
		METOUT();
		return PCERA_ERROR_NODE_NOT_FOUND; // or NO_ROUTE ??
	}

	netNodeId  = selNode->getPeRouterId();
	gridSiteId = selNode->ident().id;

	METOUT();
	return PCERA_ERROR_NONE;
}

#endif // G2MPLS

pceraErrorCode_t
PCERA_CALC_API::nodeGetFromTeLink(uint32_t			g2pceraId,
				  const g2mpls_addr_t &		telink,
				  uint32_t &			rId,
				  uint32_t &			adjRid,
				  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp   = "";
	rId    = 0;
	adjRid = 0;

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	std::list<node_ident_t>			nodeList;
	std::list<node_ident_t>::iterator	iterN;

	nodeList = pce->getTopology().getNodesIdents();

	for (iterN = nodeList.begin(); iterN != nodeList.end(); iterN++) {
		std::list<telink_ident_t>		tels;
		std::list<telink_ident_t>::iterator	iterT;
		node_ident_t				id;
		PCERA::Node *				gNode;
		PCERA::NetNode *			lclNode;

		id = (*iterN);

		if (id.type != TOPO_NODE_NETWORK) {
			continue;
		}

		gNode = pce->getTopology().getNode(id);
		lclNode = static_cast<PCERA::NetNode *>(gNode);

		assert(lclNode);

		tels = lclNode->getOutLinksIdents();

		for (iterT = tels.begin(); iterT != tels.end(); iterT++) {
			if (addr_equal((*iterT).lcl_id, telink) != 0) {
				rId    = (*iterT).lcl_rid;
				adjRid = (*iterT).rem_rid;
				METOUT();
				return PCERA_ERROR_NONE;
			}
		}
	}

	resp  = "TE-link ";
	resp += addr_ntoa(telink);
	resp += " not found under any network node in G2PCERA ";
	resp += rid2string(g2pceraId);
	METOUT();
	return PCERA_ERROR_LINK_NOT_FOUND;
}

pceraErrorCode_t
PCERA_CALC_API::callRoute(uint32_t			g2pceraId,
			  const net_res_spec_t &	src,
			  const net_res_spec_t &	dst,
			  const call_ident_t &		callIdent,
			  const call_info_t &		callInfo,
			  const recovery_info_t &	recInfo,
			  const lsp_info_t &		lspInfo,
			  std::list<ero_hop_t> &	wEro,
			  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Routing GNS call %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	CHECK_INPUT(src      ,"src"      ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(dst      ,"dst"      ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(callIdent,"callIdent",PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(callInfo ,"callInfo" ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(recInfo  ,"recInfo"  ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(lspInfo  ,"lspInfo"  ,PCERA_ERROR_INVALID_PARAMETER, resp);

	PCERA::GnsCall *	ptr;
	res_spec_t		glbdst;

	memset(&glbdst, 0, sizeof(glbdst));
	glbdst.net = dst;
	BITMASK_BITSET(glbdst.mask_, net);
	BITMASK_BITRESET(glbdst.mask_, grid);

	ptr = pce->getGnsCall(callIdent);
	if (ptr && !ptr->match(src, glbdst, callInfo, recInfo, lspInfo)) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " already exist into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " with different parameters";
		METOUT();
		return PCERA_ERROR_CALL_PARAMS_MISMATCH;
	}

	if (!ptr) {
		// Create a GnsCall with connections from scratch
		ptr = new PCERA::GnsCall(pce,
					 callIdent,
					 src,
					 glbdst,
					 callInfo,
					 recInfo,
					 lspInfo);
		if (!ptr) {
			resp  = "Cannot allocate a GnsCall object "
				"for call ";
			resp += callIdent2string(callIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		if (!ptr->route()) {
			resp  = "Cannot route GNS call ";
			resp += callIdent2string(callIdent);
			resp += " into G2PCERA ";
			resp += rid2string(g2pceraId);
			delete ptr;

			METOUT();
			return PCERA_ERROR_NO_ROUTE;
		}

		ptr->dump("[callRoute] ", true);

		if (!pce->attachGnsCall(ptr)) {
			resp  = "Cannot attach GNS call ";
			resp += callIdent2string(callIdent);
			resp += " into G2PCERA ";
			resp += rid2string(g2pceraId);
			delete ptr;

			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	}

	assert(ptr);

	wEro = ptr->getGnsCallEro();

	if (wEro.empty()) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " exists into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " without connections";
		METOUT();
		return PCERA_ERROR_NO_ROUTE;
	}

	G2PCERA_DEBUG("GNS call %s routed in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_CALC_API::callRouteAny(uint32_t			g2pceraId,
			     const net_res_spec_t &	src,
			     const grid_res_spec_t &	dst,
			     const call_ident_t &	callIdent,
			     const call_info_t &	callInfo,
			     const recovery_info_t &	recInfo,
			     const lsp_info_t &		lspInfo,
			     std::list<ero_hop_t> &	wEro,
			     net_res_spec_t &		eTna,
			     std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Routing anycast GNS call %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	CHECK_INPUT(src      ,"src"      ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(dst      ,"dst"      ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(callIdent,"callIdent",PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(callInfo ,"callInfo" ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(recInfo  ,"recInfo"  ,PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(lspInfo  ,"lspInfo"  ,PCERA_ERROR_INVALID_PARAMETER, resp);

	PCERA::GnsCall *	ptr;
	res_spec_t		glbdst;

	memset(&glbdst.net, 0, sizeof(glbdst.net));
	glbdst.grid = dst;
	BITMASK_BITRESET(glbdst.mask_, net);
	BITMASK_BITSET(glbdst.mask_, grid);

	ptr = pce->getGnsCall(callIdent);
	if (ptr && !ptr->match(src, glbdst, callInfo, recInfo, lspInfo)) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " already exist into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " with different parameters";
		METOUT();
		return PCERA_ERROR_CALL_PARAMS_MISMATCH;
	}

	if (!ptr) {
		// Create a GnsCall with connections from scratch
		ptr = new PCERA::GnsCall(pce,
					 callIdent,
					 src,
					 glbdst,
					 callInfo,
					 recInfo,
					 lspInfo);
		if (!ptr) {
			resp  = "Cannot allocate a GnsCall object "
				"for call ";
			resp += callIdent2string(callIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		if (!ptr->route()) {
			resp  = "Cannot route GNS call ";
			resp += callIdent2string(callIdent);
			resp += " into G2PCERA ";
			resp += rid2string(g2pceraId);
			delete ptr;

			METOUT();
			return PCERA_ERROR_NO_ROUTE;
		}

		if (!pce->attachGnsCall(ptr)) {
			resp  = "Cannot attach GNS call ";
			resp += callIdent2string(callIdent);
			resp += " into G2PCERA ";
			resp += rid2string(g2pceraId);
			delete ptr;

			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	}

	assert(ptr);

	wEro = ptr->getGnsCallEro();

	if (wEro.empty()) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " already exist into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " without connections";
		METOUT();
		return PCERA_ERROR_NO_ROUTE;
	}

	eTna = ptr->getGnsCallEgressTnaRes();

	G2PCERA_DEBUG("GNS call %s routed in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_CALC_API::callFlush(uint32_t			g2pceraId,
			  const call_ident_t &		callIdent,
			  std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Flushing GNS call %s from G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::GnsCall *	ptr;
	PCERA::Connection *	priptr;
	PCERA::Connection *	secptr;

	ptr = pce->getGnsCall(callIdent);
	if (!ptr) {
		resp  = "Cannot get GNS call ";
		resp += callIdent2string(callIdent);
		resp += " into G2PCERA ";
		resp += rid2string(g2pceraId);
		METOUT();
		return PCERA_ERROR_CALL_NOT_FOUND;
	}

	if (!pce->detachGnsCall(ptr)) {
		resp  = "Cannot detach GNS call ";
		resp += callIdent2string(callIdent);
		resp += " from G2PCERA ";
		resp += rid2string(g2pceraId);
		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	delete ptr;

	G2PCERA_DEBUG("GNS call %s flushed from G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_CALC_API::callConfirm(uint32_t			g2pceraId,
			    const call_ident_t &	callIdent,
			    std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Confirming GNS call %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());


	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	PCERA::GnsCall * ptr;

	ptr = pce->getGnsCall(callIdent);
	if (!ptr) {
		resp  = "Cannot get GNS call ";
		resp += callIdent2string(callIdent);
		resp += " into G2PCERA ";
		resp += rid2string(g2pceraId);
		METOUT();
		return PCERA_ERROR_CALL_NOT_FOUND;
	}

	ptr->confirm(true);

	G2PCERA_DEBUG("GNS call confirmed %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}


pceraErrorCode_t
PCERA_CALC_API::connectionRoute(uint32_t			g2pceraId,
				const ero_hop_t &		srcHop,
				const ero_hop_t &		dstHop,
				const call_ident_t &		callIdent,
				const lsp_info_t &		lspInfo,
				std::list<ero_hop_t> &		pEro,
				std::list<ero_hop_t> &		sEro,
				std::string &			resp)
{
	METIN();

	PCERA::G2PCERA * pce;

	resp = "";

	G2PCERA_DEBUG("Routing connections for GNS call %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	CHECK_INPUT(callIdent,"callIdent",PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(lspInfo  ,"lspInfo"  ,PCERA_ERROR_INVALID_PARAMETER, resp);

	pEro.clear();
	sEro.clear();

	PCERA::GnsCall *	ptr;

	ptr = pce->getGnsCall(callIdent);
	if (ptr && !ptr->match(srcHop, dstHop, lspInfo)) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " already exist into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " with different parameters";
		METOUT();
		return PCERA_ERROR_CALL_PARAMS_MISMATCH;
	}

	if (!ptr) {
		G2PCERA_DEBUG("Creating from scratch GNS call %s "
			      "in G2PCERA %s",
			      callIdent2string(callIdent).c_str(),
			      rid2string(g2pceraId).c_str());

		CHECK_INPUT(srcHop, "srcHop",
			    PCERA_ERROR_INVALID_PARAMETER, resp);
		CHECK_INPUT(dstHop, "dstHop",
			    PCERA_ERROR_INVALID_PARAMETER, resp);

		// Create a GnsCall with connections from scratch
		ptr = new PCERA::GnsCall(pce,
					 callIdent,
					 srcHop,
					 dstHop,
					 lspInfo);
		if (!ptr) {
			resp  = "Cannot allocate a GnsCall object "
				"for call ";
			resp += callIdent2string(callIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}

		if (!ptr->route()) {
			resp  = "Cannot route GNS call ";
			resp += callIdent2string(callIdent);
			resp += " into G2PCERA ";
			resp += rid2string(g2pceraId);
			delete ptr;

			METOUT();
			return PCERA_ERROR_NO_ROUTE;
		}

		if (!pce->attachGnsCall(ptr)) {
			resp  = "Cannot attach GNS call ";
			resp += callIdent2string(callIdent);
			resp += " into G2PCERA ";
			resp += rid2string(g2pceraId);
			delete ptr;

			METOUT();
			return PCERA_ERROR_GENERIC;
		}
	}

	assert(ptr);

	pEro = ptr->getConnectionEro(LSP_ROLE_PRIMARY, true /*justMydmn*/);
	sEro = ptr->getConnectionEro(LSP_ROLE_SECONDARY, true /*justMydmn*/);

	if (pEro.empty()) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " exists into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " without connections";
		METOUT();
		return PCERA_ERROR_NO_ROUTE;
	}

	G2PCERA_DEBUG("Connections routed for GNS call %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}

pceraErrorCode_t
PCERA_CALC_API::connectionRouteDisjoint(uint32_t		g2pceraId,
					const ero_hop_t &	srcHop,
					const ero_hop_t &	dstHop,
					const call_ident_t &	callIdent,
					const lsp_info_t &	lspInfo,
					const std::list< std::list<ero_hop_t> > & multiXro,
					std::list<ero_hop_t> &	pEro,
					std::list<ero_hop_t> &	sEro,
					std::string &		resp)
{
	METIN();

	PCERA::G2PCERA * pce;
	bool             onCreate;

	resp = "";

	G2PCERA_DEBUG("Routing disjoint connections for GNS call %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	pce = G2PCERA_MASTER.getPceInstance(g2pceraId);
	if (!pce) {
		resp = "INSTANCE NOT FOUND";
		METOUT();
		return PCERA_ERROR_INSTANCE_NOT_FOUND;
	}

	onCreate = false;

	CHECK_INPUT(callIdent,"callIdent",PCERA_ERROR_INVALID_PARAMETER, resp);
	CHECK_INPUT(lspInfo  ,"lspInfo"  ,PCERA_ERROR_INVALID_PARAMETER, resp);

	pEro.clear();
	sEro.clear();

	PCERA::GnsCall *	ptr;

	ptr = pce->getGnsCall(callIdent);
	if (ptr && !ptr->match(srcHop, dstHop, lspInfo)) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " already exist into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " with different parameters";
		METOUT();
		return PCERA_ERROR_CALL_PARAMS_MISMATCH;
	}

	if (!ptr) {
		onCreate = true;

		G2PCERA_DEBUG("Creating from scratch GNS call %s "
			      "in G2PCERA %s",
			      callIdent2string(callIdent).c_str(),
			      rid2string(g2pceraId).c_str());

		CHECK_INPUT(srcHop, "srcHop",
			    PCERA_ERROR_INVALID_PARAMETER, resp);
		CHECK_INPUT(dstHop, "dstHop",
			    PCERA_ERROR_INVALID_PARAMETER, resp);

		// Create a GnsCall with connections from scratch
		ptr = new PCERA::GnsCall(pce,
					 callIdent,
					 srcHop,
					 dstHop,
					 lspInfo);
		if (!ptr) {
			resp  = "Cannot allocate a GnsCall object "
				"for call ";
			resp += callIdent2string(callIdent);
			METOUT();
			return PCERA_ERROR_NO_MEMORY;
		}
	}

	if (!pce->getTopology().transform(multiXro, true /*forward*/)) {
		resp  = "Cannot make a forward topology transformation ";
		resp += "with multiple exclude routes (multiXro) ";
		resp += "for GNS call ";
		resp += callIdent2string(callIdent);
		resp += " into G2PCERA ";
		resp += rid2string(g2pceraId);
		delete ptr;

		METOUT();
		return PCERA_ERROR_ERO_MISMATCH;
	}

	if (!ptr->route()) {
		pce->getTopology().transform(multiXro);

		resp  = "Cannot route disjointly the GNS call ";
		resp += callIdent2string(callIdent);
		resp += " into G2PCERA ";
		resp += rid2string(g2pceraId);
		delete ptr;

		METOUT();
		return PCERA_ERROR_NO_ROUTE;
	}

	if (!pce->getTopology().transform(multiXro)) {
		resp  = "Cannot make a reverse topology transformation ";
		resp += "with multiple exclude routes (multiXro) ";
		resp += "for GNS call ";
		resp += callIdent2string(callIdent);
		resp += " into G2PCERA ";
		resp += rid2string(g2pceraId);
		delete ptr;

		METOUT();
		return PCERA_ERROR_ERO_MISMATCH; // or assert(0) ???
	}

	if (onCreate && !pce->attachGnsCall(ptr)) {
		resp  = "Cannot attach GNS call ";
		resp += callIdent2string(callIdent);
		resp += " into G2PCERA ";
		resp += rid2string(g2pceraId);
		delete ptr;

		METOUT();
		return PCERA_ERROR_GENERIC;
	}

	assert(ptr);

	pEro = ptr->getConnectionEro(LSP_ROLE_PRIMARY, true /*justMydmn*/);
	sEro = ptr->getConnectionEro(LSP_ROLE_SECONDARY, true /*justMydmn*/);

	if (pEro.empty()) {
		resp  = "A GNS call ";
		resp += callIdent2string(callIdent);
		resp += " exists into G2PCERA ";
		resp += rid2string(g2pceraId);
		resp += " without connections";
		METOUT();
		return PCERA_ERROR_NO_ROUTE;
	}

	G2PCERA_DEBUG("Connections disjointly routed for "
		      "GNS call %s in G2PCERA %s",
		      callIdent2string(callIdent).c_str(),
		      rid2string(g2pceraId).c_str());

	METOUT();
	return PCERA_ERROR_NONE;
}
