//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include "g2pcera_dumper.hh"
#include "g2pcera_trace.hh"
#include <g2mpls_types.h>

#include <vty.h>

#include <string>
#include <sstream>
#include <iomanip>
#include <iostream>
#include <stdlib.h>


//
// log dump policy
//
void
logDump(const std::string &	prefix,
	const call_ident_t &	cid)
{
	G2PCERA_DEBUG("%sCALL identity [%s]",
		      prefix.c_str(),
		      SHOW_CALL_ID_TYPE(cid.type));
	G2PCERA_DEBUG("%s- srcAddr             : %s",
		      prefix.c_str(),
		      addr_ntoa(cid.src_addr));
	G2PCERA_DEBUG("%s- localId             : %s",
		      prefix.c_str(),
		      key2string(cid.local_id).c_str());

	if (cid.type == CALLID_GLOBUNIQ) {
		G2PCERA_DEBUG("%s- Globally unique attributes");
		G2PCERA_DEBUG("%s    - ituCountryCode: 0x%06x",
			      prefix.c_str(),
			      cid.itu_country_code);
		G2PCERA_DEBUG("%s    - ituCarrierCode    : %s",
			      prefix.c_str(),
			      key2string(cid.itu_carrier_code).c_str());
		G2PCERA_DEBUG("%s    - uniqueAP          : %s",
			      prefix.c_str(),
			      key2string(cid.unique_ap).c_str());
	}
}

void
logDump(const std::string &	prefix,
	const lsp_ident_t &	lid)
{
	G2PCERA_DEBUG("%sLSP identity",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- src router addr     : %s",
		      prefix.c_str(),
		      numb_htoa(lid.src_nid));
	G2PCERA_DEBUG("%s- dst router addr     : %s",
		      prefix.c_str(),
		      numb_htoa(lid.dst_nid));
	G2PCERA_DEBUG("%s- tunnel ID           : 0x%08x",
		      prefix.c_str(),
		      lid.tun_id);
	G2PCERA_DEBUG("%s- extended tunnel ID  : 0x%08x",
		      prefix.c_str(),
		      lid.ext_tun_id);
	G2PCERA_DEBUG("%s- LSP ID              : 0x%08x",
		      prefix.c_str(),
		      lid.lsp_id);

}

void
logDump(const std::string &	prefix,
	const net_res_spec_t &	res)
{
	G2PCERA_DEBUG("%s  - Network data      : is present",
		      prefix.c_str());
	G2PCERA_DEBUG("%s      - tna           : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.mask_, tna) ?
		      addr_ntoa(res.tna) : "undeclared");
	G2PCERA_DEBUG("%s      - data link     : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.mask_, data_link) ?
		      addr_ntoa(res.data_link) : "undeclared");
	G2PCERA_DEBUG("%s      - label         : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.mask_, label) ?
		      label2string(res.label).c_str() :
		      "undeclared");
}

void
logDump(const std::string &	prefix,
	const res_spec_t &	res)
{
	G2PCERA_DEBUG("%s  - Network data      : %s present",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.mask_, net) ?
		      "is" : "not");
	G2PCERA_DEBUG("%s      - tna           : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.net.mask_, tna) ?
		      addr_ntoa(res.net.tna) : "undeclared");
	G2PCERA_DEBUG("%s      - data link     : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.net.mask_, data_link) ?
		      addr_ntoa(res.net.data_link) : "undeclared");
	G2PCERA_DEBUG("%s      - label         : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.net.mask_, label) ?
		      label2string(res.net.label).c_str() :
		      "undeclared");
#ifdef G2MPLS
	G2PCERA_DEBUG("%s  - Grid data         : %s present",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.mask_, grid) ?
		      "is" : "not");
	// APPLICATION
	G2PCERA_DEBUG("%s      - APP  type     : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, application) ?
		      SHOW_GRID_APPLICATION(res.grid.application.type) :
		      "undeclared");
	G2PCERA_DEBUG("%s      - APP  version  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, application) ?
		      version2string(res.grid.application.version).c_str() :
		      "undeclared");

	// CAND HOST
	G2PCERA_DEBUG("%s      - CAND HOST     : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, cand_host) ?
		      addr_ntoa(res.grid.cand_host) : "undeclared");

	// FS_RESOURCES
	G2PCERA_DEBUG("%s      - FS name       : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		      SHOW_GRID_FS_NAME(res.grid.fs_resources.fs_name) :
		      "undeclared");
	G2PCERA_DEBUG("%s      - FS type       : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		      SHOW_GRID_FS_TYPE(res.grid.fs_resources.fs_type) :
		      "undeclared");
	G2PCERA_DEBUG("%s      - FS disk space : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		      range2string(res.grid.fs_resources.disk_space).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - FS mount point: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		      (res.grid.fs_resources.mount_point)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - FS mount src  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		      (res.grid.fs_resources.mount_source)->c_str() :
		      "undeclared");

	// SYSTEM CAPS
	G2PCERA_DEBUG("%s      - SYS OS        : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		      SHOW_GRID_OS(res.grid.sys_caps.os_type) :
		      "undeclared");
	G2PCERA_DEBUG("%s      - SYS OS version: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		      version2string(res.grid.sys_caps.os_version).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - SYS CPU arch  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		      SHOW_GRID_CPU_ARCH(res.grid.sys_caps.cpu_arch) :
		      "undeclared");
	G2PCERA_DEBUG("%s      - SYS exclAccess: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		      ((res.grid.sys_caps.exclusive_access == 1) ?
		       "yes" : "no")  :
		      "undeclared");
	// ranges
	G2PCERA_DEBUG("%s      - ind CPU speed : %s MHz",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, ind_cpu_speed) ?
		      range2string(res.grid.ind_cpu_speed).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - ind CPU time  : %s seconds",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, ind_cpu_time) ?
		      range2string(res.grid.ind_cpu_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - ind CPU count : %s cores",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, ind_cpu_count) ?
		      range2string(res.grid.ind_cpu_count).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - ind NET bw    : %s Mbps",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, ind_net_bw) ?
		      range2string(res.grid.ind_net_bw).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - ind MEM phy   : %s MB",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, ind_phy_mem) ?
		      range2string(res.grid.ind_phy_mem).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - ind MEM vir   : %s MB",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, ind_phy_mem) ?
		      range2string(res.grid.ind_phy_mem).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - ind DISK space: %s MB",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, ind_disk_space) ?
		      range2string(res.grid.ind_disk_space).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - tot CPU time  : %s seconds",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, tot_cpu_time) ?
		      range2string(res.grid.tot_cpu_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - tot CPU count : %s cores",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, tot_cpu_count) ?
		      range2string(res.grid.tot_cpu_count).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - tot MEM phy   : %s MB",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, tot_phy_mem) ?
		      range2string(res.grid.tot_phy_mem).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - tot MEM vir   : %s MB",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, tot_phy_mem) ?
		      range2string(res.grid.tot_phy_mem).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - tot DISK space: %s MB",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, tot_disk_space) ?
		      range2string(res.grid.tot_disk_space).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - tot RES count : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, tot_res_count) ?
		      range2string(res.grid.tot_res_count).c_str() :
		      "undeclared");
	// DATA STAGING
	G2PCERA_DEBUG("%s      - STAG FS name  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		      SHOW_GRID_FS_NAME(res.grid.data_staging.fs_name) :
		      "undeclared");
	G2PCERA_DEBUG("%s      - STAG createFlg: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		      SHOW_GRID_CREATION_FLAG(res.grid.data_staging.
					      creation_flag):
		      "undeclared");
	G2PCERA_DEBUG("%s      - STAG delOnEnd : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		      ((res.grid.data_staging.del_on_termination == 1) ?
		       "yes" : "no")  :  "undeclared");
	G2PCERA_DEBUG("%s      - STAG filename : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		      (res.grid.data_staging.filename)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - STAG source   : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		      (res.grid.data_staging.source)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s      - STAG target   : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		      (res.grid.data_staging.target)->c_str() :
		      "undeclared");

	// GRID SITE ID
	G2PCERA_DEBUG("%s      - GRID SITE addr: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(res.grid.mask_, grid_site_id) ?
		      numb_htoa(res.grid.grid_site_id) : "undeclared");

#endif // G2MPLS
}

void
logDump(const std::string &	prefix,
	const call_info_t &	cdata)
{
	tm *               curTime;
	const char *       timeStringFormat = "%a, %d %b %Y %H:%M:%S %zGMT";
	const int          timeStringLength = 40;
	char               timeString[timeStringLength];

	G2PCERA_DEBUG("%sCALL data",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- Symbolic name       : %s",
		      prefix.c_str(),
		      (cdata.call_name) ?
		      (cdata.call_name)->c_str() : "==UNKNOWN==");
	curTime = localtime((time_t *) &cdata.times.start_time);
	strftime(timeString, timeStringLength, timeStringFormat, curTime);
	G2PCERA_DEBUG("%s- startTime           : %s",
		      prefix.c_str(),
		      timeString);
	curTime = localtime((time_t *) &cdata.times.end_time);
	strftime(timeString, timeStringLength, timeStringFormat, curTime);
	G2PCERA_DEBUG("%s- endTime             : %s",
		      prefix.c_str(),
		      timeString);
	double duration = difftime(cdata.times.end_time,
				   cdata.times.start_time);
	G2PCERA_DEBUG("%s    - duration    : %.0f seconds",
		      prefix.c_str(),
		      (float)duration);
#ifdef G2MPLS
	G2PCERA_DEBUG("%s- jobName             : %s",
		      prefix.c_str(),
		      (cdata.job_name) ?
		      (cdata.job_name)->c_str() : "==UNKNOWN==");
	G2PCERA_DEBUG("%s- jobProject          : %s",
		      prefix.c_str(),
		      (cdata.job_project) ?
		      (cdata.job_project)->c_str() : "==UNKNOWN==");
#endif // G2MPLS

	G2PCERA_DEBUG("%s- iTNA resources",
		      prefix.c_str());
	logDump(prefix, cdata.iTNA_res);

	G2PCERA_DEBUG("%s- eTNA resources",
		      prefix.c_str());
	logDump(prefix, cdata.eTNA_res);
}

void
logDump(const std::string &	prefix,
	const recovery_info_t &	rdata)
{
	G2PCERA_DEBUG("%sRecovery data",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- recovery type       : %s",
		      prefix.c_str(),
		      SHOW_GMPLS_RECOVERYTYPE(rdata.rec_type));
	G2PCERA_DEBUG("%s- disjointness        : %s",
		      prefix.c_str(),
		      SHOW_GMPLS_DISJOINTNESS(rdata.disj_type));
}

void
logDump(const std::string &	prefix,
	const lsp_info_t &	ldata)
{
	G2PCERA_DEBUG("%sLSP data",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- swcap               : %s",
		      prefix.c_str(),
		      SHOW_SW_CAP(ldata.sw_cap));
	G2PCERA_DEBUG("%s- enctype             : %s",
		      prefix.c_str(),
		      SHOW_ENC_TYPE(ldata.enc_type));
	G2PCERA_DEBUG("%s- gpid                : %s",
		      prefix.c_str(),
		      SHOW_GPID(ldata.gpid));
	G2PCERA_DEBUG("%s- bandwidth           : %.3fMbps (signal %s)",
		      prefix.c_str(),
		      BW_HEX2MBPS(ldata.bw),
		      SHOW_GMPLS_BWENC(ldata.bw));
	G2PCERA_DEBUG("%s- setupPrio           : %d",
		      prefix.c_str(),
		      ldata.setup_prio);
	G2PCERA_DEBUG("%s- holdingPrio         : %d",
		      prefix.c_str(),
		      ldata.holding_prio);
	G2PCERA_DEBUG("%s- excludeAny          : 0x%08x",
		      prefix.c_str(),
		      ldata.exclude_any);
	G2PCERA_DEBUG("%s- includeAny          : 0x%08x",
		      prefix.c_str(),
		      ldata.include_any);
	G2PCERA_DEBUG("%s- includeAll          : 0x%08x",
		      prefix.c_str(),
		      ldata.include_all);
	G2PCERA_DEBUG("%s- linkProtMask        : %s",
		      prefix.c_str(),
		      SHOW_GMPLS_PROTTYPE(ldata.link_prot_mask));
	G2PCERA_DEBUG("%s- crankback scope     : %s",
		      prefix.c_str(),
		      SHOW_CBACK_SCOPE(ldata.crankback));
	G2PCERA_DEBUG("%s    - maxRetrSrc      : %d",
		      prefix.c_str(),
		      ldata.max_cback_retries_src);
	G2PCERA_DEBUG("%s    - maxRetrInt      : %d",
		      prefix.c_str(),
		      ldata.max_cback_retries_intmd);
	G2PCERA_DEBUG("%s- type                : %s",
		      prefix.c_str(),
		      SHOW_LSP_TYPE(ldata.type));
	G2PCERA_DEBUG("%s- action              : %s",
		      prefix.c_str(),
		      SHOW_LSP_RES_ACTION(ldata.action));
	G2PCERA_DEBUG("%s- rroMode             : %s",
		      prefix.c_str(),
		      SHOW_LSP_RRO_MODE(ldata.rro_mode));
	G2PCERA_DEBUG("%s- refresh             : %d ms",
		      prefix.c_str(),
		      ldata.refresh_interval);
	G2PCERA_DEBUG("%s- activateAck         : %s",
		      prefix.c_str(),
		      ldata.activate_ack ? "YES" : "NO");
	G2PCERA_DEBUG("%s- rapidRetrInt        : %d ms",
		      prefix.c_str(),
		      ldata.rapid_retransm_interval);
	G2PCERA_DEBUG("%s- rapidRetryLimit     : %d",
		      prefix.c_str(),
		      ldata.rapid_retry_limit);
	G2PCERA_DEBUG("%s- incrValueDelta      : %d",
		      prefix.c_str(),
		      ldata.increment_value_delta);
}


void
logDump(const std::string &	prefix,
	const ero_hop_t &	hop)
{
	G2PCERA_DEBUG("%sEro Hop [%s]",
		      prefix.c_str(),
		      (hop.loose == 1) ? "LOOSE" : "STRICT");
	G2PCERA_DEBUG("%s- node                : %s",
		      prefix.c_str(),
		      numb_htoa(hop.node));
	G2PCERA_DEBUG("%s- TE-link             : %s",
		      prefix.c_str(),
		      addr_ntoa(hop.te_link));
	G2PCERA_DEBUG("%s- upstream D-link     : %s",
		      prefix.c_str(),
		      addr_ntoa(hop.upstream_datalink));
	G2PCERA_DEBUG("%s- downstream D-link   : %s",
		      prefix.c_str(),
		      addr_ntoa(hop.downstream_datalink));
	G2PCERA_DEBUG("%s- upstream label      : %s",
		      prefix.c_str(),
		      label2string(hop.upstream_label).c_str());
	G2PCERA_DEBUG("%s- downstream label    : %s",
		      prefix.c_str(),
		      label2string(hop.downstream_label).c_str());
}


void
logDump(const std::string &		prefix,
	const topology_summary_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- root router_id      : %s",
		      prefix.c_str(),
		      numb_htoa(data.root_rid));
	G2PCERA_DEBUG("%s- total nodes         : %d",
		      prefix.c_str(),
		      data.tot_nodes);
	G2PCERA_DEBUG("%s- total links         : %d",
		      prefix.c_str(),
		      data.tot_links);
	G2PCERA_DEBUG("%s- total TNAs          : %d",
		      prefix.c_str(),
		      data.tot_tnas);
	G2PCERA_DEBUG("%s- total GNS calls     : %d",
		      prefix.c_str(),
		      data.tot_gns_calls);

	G2PCERA_DEBUG("%s- algorithm info",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- max link cost    : %d",
		      prefix1.c_str(),
		      data.max_link_cost);
	G2PCERA_DEBUG("%s- SPF selector     : %s",
		      prefix1.c_str(),
		      SPF_TYPE_TO_STRING(data.spf_selector));
	G2PCERA_DEBUG("%s- SPF revision     : %d",
		      prefix1.c_str(),
		      data.spf_revision);
}

void
logDump(const std::string &		prefix,
	const node_ident_t &		data)
{
	G2PCERA_DEBUG("%s- type                : %s",
		      prefix.c_str(),
		      NODE_TYPE_TO_STRING(data.type));

	G2PCERA_DEBUG("%s- id                  : %s",
		      prefix.c_str(),
		      numb_htoa(data.id));
}

void
logDump(const std::string &		prefix,
	const net_node_data_t &		data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- is domain           : %s",
		      prefix.c_str(),
		      data.is_domain ? "YES" : "NO");
	G2PCERA_DEBUG("%s- operational state   : %s",
		      prefix.c_str(),
		      SHOW_OPSTATE(data.op_state));
	G2PCERA_DEBUG("%s- administrative state: %s",
		      prefix.c_str(),
		      SHOW_ADMSTATE(data.adm_state));
	G2PCERA_DEBUG("%s- TE color mask       : 0x%08x",
		      prefix.c_str(), data.te_colors);

	G2PCERA_DEBUG("%s- Areas count         : %d",
		      prefix.c_str(),
		      data.areas.size());

	std::list<uint16_t>::const_iterator areaIter;
	for (areaIter  = data.areas.begin();
	     areaIter != data.areas.end();
	     areaIter++) {
		G2PCERA_DEBUG("%s- Area = %d",
			      prefix1.c_str(),
			      *areaIter);
	}
}

void
logDump(const std::string &		prefix,
	const grid_site_data_t &	data)
{
	G2PCERA_DEBUG("%s- name                : %s",
		      prefix.c_str(),
		      data.name ? data.name->c_str() : "<empty>");
	G2PCERA_DEBUG("%s- location            : %s",
		      prefix.c_str(),
		      geoCoord2string(data.location).c_str());
	G2PCERA_DEBUG("%s- PE router id        : %s",
		      prefix.c_str(),
		      numb_htoa(data.pe_router_id));
}

void
logDump(const std::string &		prefix,
	const grid_subnode_ident_t &	data)
{
	G2PCERA_DEBUG("%s- type                   : %s",
		      prefix.c_str(),
		      GRID_SUBNODE_TO_STRING(data.type));

	G2PCERA_DEBUG("%s- id                     : %s",
		      prefix.c_str(),
		      numb_htoa(data.id));
}

void
logDump(const std::string &		prefix,
	const grid_service_data_t &	data)
{
	G2PCERA_DEBUG("%s- Service name           : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, name) ?
		      (data.name)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- type                   : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, data) ?
		      SHOW_GRID_SERVICETYPE(data.data.type) :
		      "undeclared");
	G2PCERA_DEBUG("%s- version                : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, data) ?
		      version2string(data.data.version).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- state                  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, state) ?
		      SHOW_GRID_SERVICESTATE(data.state) :
		      "undeclared");
	G2PCERA_DEBUG("%s- endpoint address       : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, endpoint_addr) ?
		      addr_ntoa(data.endpoint_addr) : "undeclared");
}

void
logDump(const std::string &		prefix,
	const grid_ce_data_t &		data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- CompElem name          : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, name) ?
		      (data.name)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- LRMS type              : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, lrms_info) ?
		      SHOW_GRID_LRMS(data.lrms_info.type) :
		      "undeclared");
	G2PCERA_DEBUG("%s- LRMS version           : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, lrms_info) ?
		      version2string(data.lrms_info.version).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- host address           : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, host_addr) ?
		      addr_ntoa(data.host_addr) :
		      "undeclared");
	G2PCERA_DEBUG("%s- gatekeeper port        : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, gatekeeper_port) ?
		      uint32string(data.gatekeeper_port).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- job manager            : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, job_manager) ?
		      (data.job_manager)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- data directory         : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, data_dir) ?
		      (data.data_dir)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- dflt storage elem      : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, default_storage_elem_id) ?
		      uint32string(data.default_storage_elem_id).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- jobs state             : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_state) ?
		      SHOW_GRID_CESE_STATE(data.jobs_state.state) :
		      "undeclared");
	G2PCERA_DEBUG("%s- free job slots      : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_state) ?
		      uint32string(data.jobs_state.free_job_slots).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- running jobs        : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_stats) ?
		      uint32string(data.jobs_stats.running_jobs).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- waiting jobs        : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_stats) ?
		      uint32string(data.jobs_stats.waiting_jobs).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- total jobs          : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_stats) ?
		      uint32string(data.jobs_stats.total_jobs).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- jobs time performances",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- est resp time       : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_timeperf) ?
		      uint32string(data.jobs_timeperf.
				   estimated_response_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- worst resp time     : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_timeperf) ?
		      uint32string(data.jobs_timeperf.
				   worst_response_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- jobs time policies",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- max wclock time     : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		      uint32string(data.jobs_timepolicy.
				   max_wallclock_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- max obt wclock time : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		      uint32string(data.jobs_timepolicy.
				   max_obtainable_wallclock_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- max cpu time        : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		      uint32string(data.jobs_timepolicy.
				   max_cpu_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- max obt cpu time    : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		      uint32string(data.jobs_timepolicy.
				   max_obtainable_cpu_time).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- jobs load policies",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- max total jobs      : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		      uint32string(data.jobs_loadpolicy.
				   max_total_jobs).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- max running jobs    : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		      uint32string(data.jobs_loadpolicy.
				   max_running_jobs).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- max waiting jobs    : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		      uint32string(data.jobs_loadpolicy.
				   max_waiting_jobs).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- assigned job slots  : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		      uint16string(data.jobs_loadpolicy.
				   assigned_job_slots).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- max slots per job   : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		      uint16string(data.jobs_loadpolicy.
				   max_slots_per_job).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- priority            : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		      uint8string(data.jobs_loadpolicy.
				  priority).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- preemption flag     : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		      bool2string(data.jobs_loadpolicy.
				  preemption_flag).c_str() :
		      "undeclared");

	G2PCERA_DEBUG("%s- free job slots cal     : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, free_job_slots_calendar) ?
		      "" : "undeclared");

	if (data.free_job_slots_calendar.size()) {

		std::map<uint32_t, uint16_t>::const_iterator	iter;

		for (iter  = data.free_job_slots_calendar.begin();
		     iter != data.free_job_slots_calendar.end();
		     iter++) {
			G2PCERA_DEBUG("%s- @[%s] --> fjs = %s",
				      prefix1.c_str(),
				      unixTime2string((*iter).first).c_str(),
				      uint16string((*iter).second).c_str());
		}
	}
}

void
logDump(const std::string &		prefix,
	const grid_subcluster_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- SubCluster name        : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, name) ?
		      (data.name)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- CPUs arch              : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, cpu) ?
		      SHOW_GRID_CPU_ARCH(data.cpu.arch) :
		      "undeclared");
	G2PCERA_DEBUG("%s- CPUs count physical    : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, cpu) ?
		      uint32string(data.cpu.count.physical).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- CPUs count logical     : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, cpu) ?
		      uint32string(data.cpu.count.logical).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- OS type                : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, os) ?
		      SHOW_GRID_OS(data.os.type) :
		      "undeclared");
	G2PCERA_DEBUG("%s- OS version             : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, os) ?
		      version2string(data.os.version).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- RAM size               : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, memory) ?
		      uint32string(data.memory.ram_size).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- virtual memory size    : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, memory) ?
		      uint32string(data.memory.virtual_memory_size).c_str() :
		      "undeclared");

	G2PCERA_DEBUG("%s- Software packages      : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, software_packages) ?
		      "" : "undeclared");
	if (data.software_packages.size()) {

		std::list<grid_software_package_t *>::const_iterator	iter;

		for (iter  = data.software_packages.begin();
		     iter != data.software_packages.end();
		     iter++) {
			G2PCERA_DEBUG("%s-----------------------------",
				      prefix1.c_str());
			G2PCERA_DEBUG("%s- Application type       : %s",
				      prefix1.c_str(),
				      SHOW_GRID_APPLICATION((*iter)->
							    software.type));
			G2PCERA_DEBUG("%s- Application version    : %s",
				      prefix1.c_str(),
				      version2string((*iter)->
						     software.version).
				      c_str());
			G2PCERA_DEBUG("%s- software env. setup    : %s",
				      prefix1.c_str(),
				      (*iter)->software_env_setup->c_str());
		}
	}

	G2PCERA_DEBUG("%s- CPU count cal          : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, subcluster_calendar) ?
		      "" : "undeclared");

	if (data.subcluster_calendar.size()) {

		std::map<uint32_t, grid_cpu_count_t>::const_iterator	iter;

		for (iter  = data.subcluster_calendar.begin();
		     iter != data.subcluster_calendar.end();
		     iter++) {
			G2PCERA_DEBUG("%s- @[%s] --> "
				      "CPU phy = %d, CPU log = %d",
				      prefix1.c_str(),
				      unixTime2string((*iter).first).c_str(),
				      (*iter).second.physical,
				      (*iter).second.logical);
		}
	}
}

void
logDump(const std::string &		prefix,
	const grid_se_data_t &		data)
{
	std::string prefix1, prefix2;
	prefix1 = prefix + DUMP_TAB;
	prefix2 = prefix1 + DUMP_TAB;

	G2PCERA_DEBUG("%s- StoElem name           : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, name) ?
		      (data.name)->c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- arch                   : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, storage_info) ?
		      SHOW_GRID_STORAGE_ARCH(data.storage_info.arch) :
		      "undeclared");
	G2PCERA_DEBUG("%s- state                  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, storage_info) ?
		      SHOW_GRID_CESE_STATE(data.storage_info.state) :
		      "undeclared");
	uint16_t tmp;
	tmp = *((uint16_t *)(&(data.storage_info.access_protocols)));
	G2PCERA_DEBUG("%s- access protocols       : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, storage_info) ?
		      uint16string(tmp).c_str() :
		      "undeclared");
	tmp = *((uint16_t *)(&(data.storage_info.control_protocols)));
	G2PCERA_DEBUG("%s- control protocols      : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, storage_info) ?
		      uint16string(tmp).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- online size            : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, online_size) ?
		      "" : "undeclared");
	if (BITMASK_BITTEST(data.mask_, online_size)) {
		G2PCERA_DEBUG("%s- total               : %d MB",
			      prefix1.c_str(),
			      data.online_size.total);
		G2PCERA_DEBUG("%s- used                : %d MB",
			      prefix1.c_str(),
			      data.online_size.used);
	}
	G2PCERA_DEBUG("%s- nearline size          : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, nearline_size) ?
		      "" : "undeclared");
	if (BITMASK_BITTEST(data.mask_, nearline_size)) {
		G2PCERA_DEBUG("%s- total               : %d MB",
			      prefix1.c_str(),
			      data.nearline_size.total);
		G2PCERA_DEBUG("%s- used                : %d MB",
			      prefix1.c_str(),
			      data.nearline_size.used);
	}

	G2PCERA_DEBUG("%s- storage areas          : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, storage_areas) ?
		      "" : "undeclared");

	if (data.storage_areas.size()) {

		std::list<grid_storage_area_t *>::const_iterator iter;

		for (iter  = data.storage_areas.begin();
		     iter != data.storage_areas.end();
		     iter++) {
			G2PCERA_DEBUG("%s-----------------------------",
				prefix1.c_str());
			G2PCERA_DEBUG("%s- storage area name      : %s",
				      prefix1.c_str(),
				      ((*iter)->storage_area_name)->c_str());
			G2PCERA_DEBUG("%s- storage area path      : %s",
				      prefix1.c_str(),
				      ((*iter)->storage_area_path)->c_str());
			G2PCERA_DEBUG("%s- storage area info      : ",
				      prefix1.c_str());
			G2PCERA_DEBUG("%s- total online size   : %d MB",
				      prefix2.c_str(),
				      (*iter)->storage_area_info.
				      total_online_size);
			G2PCERA_DEBUG("%s- free online size    : %d MB",
				      prefix2.c_str(),
				      (*iter)->storage_area_info.
				      free_online_size);
			G2PCERA_DEBUG("%s- resv total online sz: %d MB",
				      prefix2.c_str(),
				      (*iter)->storage_area_info.
				      reserved_total_online_size);
			G2PCERA_DEBUG("%s- total nearline size : %d MB",
				      prefix2.c_str(),
				      (*iter)->storage_area_info.
				      total_nearline_size);
			G2PCERA_DEBUG("%s- free nearline size  : %d MB",
				      prefix2.c_str(),
				      (*iter)->storage_area_info.
				      free_nearline_size);
			G2PCERA_DEBUG("%s- resv total nrline sz: %d MB",
				      prefix2.c_str(),
				      (*iter)->storage_area_info.
				      reserved_nearline_size);
			G2PCERA_DEBUG("%s- retention policy    : %s",
				      prefix2.c_str(),
				      SHOW_GRID_STORAGE_RETENTION_POLICY((*iter)->storage_area_info.retention_policy));
			G2PCERA_DEBUG("%s- access latency      : %s",
				      prefix2.c_str(),
				      SHOW_GRID_STORAGE_ACCESS_LATENCY((*iter)->storage_area_info.access_latency));
			G2PCERA_DEBUG("%s- expiration mode     : %s",
				      prefix2.c_str(),
				      SHOW_GRID_STORAGE_EXPIRATION_MODE((*iter)->storage_area_info.expiration_mode));
		}
	}

	G2PCERA_DEBUG("%s- storage count cal      : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, se_calendar) ?
		      "" : "undeclared");

	if (data.se_calendar.size()) {

		std::map<uint32_t, grid_storage_count_t>::const_iterator iter;

		for (iter  = data.se_calendar.begin();
		     iter != data.se_calendar.end();
		     iter++) {
			G2PCERA_DEBUG("%s- @[%s] --> "
				      "free online size = %d MB, CPU log = %d",
				      prefix1.c_str(),
				      unixTime2string((*iter).first).c_str(),
				      (*iter).second.free_online_size,
				      (*iter).second.logical_cpus);
		}
	}
}

void
logDump(const std::string &		prefix,
	const tna_ident_t &		ident)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- ancestor RC id         : %s",
		      prefix.c_str(),
		      numb_htoa(ident.rc_id));
	G2PCERA_DEBUG("%s- border node            : %s",
		      prefix.c_str(),
		      numb_htoa(ident.rid));
	G2PCERA_DEBUG("%s- TNA address            : %s",
		      prefix.c_str(),
		      addr_ntoa(ident.id));
}

void
logDump(const std::string &		prefix,
	const telink_ident_t &		ident)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- local id               : %s",
		      prefix.c_str(),
		      addr_ntoa(ident.lcl_id));
	G2PCERA_DEBUG("%s- local router id        : %s",
		      prefix.c_str(),
		      numb_htoa(ident.lcl_rid));
	G2PCERA_DEBUG("%s- remote id              : %s",
		      prefix.c_str(),
		      addr_ntoa(ident.rem_id));
	G2PCERA_DEBUG("%s- remote router id       : %s",
		      prefix.c_str(),
		      numb_htoa(ident.rem_rid));
	G2PCERA_DEBUG("%s- mode                   : %s",
		      prefix.c_str(),
		      LINK_MODE_TO_STRING(ident.mode));
	G2PCERA_DEBUG("%s- ancestor RC id         : %s",
		      prefix.c_str(),
		      numb_htoa(ident.lcl_rc_id));
	G2PCERA_DEBUG("%s- neighbor RC id         : %s",
		      prefix.c_str(),
		      numb_htoa(ident.rem_rc_id));
}

void
logDump(const std::string &		prefix,
	const telink_com_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- admin metric           : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, admin_metric) ?
		      uint32string(data.admin_metric).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- TE info",
		      prefix.c_str());
	G2PCERA_DEBUG("%s- te metric           : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, te_metric) ?
		      uint32string(data.te_metric).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- te color mask       : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, te_metric) ?
		      uint32string(data.te_color_mask).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- te prot type mask   : %s",
		      prefix1.c_str(),
		      BITMASK_BITTEST(data.mask_, te_prot_type_mask) ?
		      uint8string(data.te_prot_type_mask).c_str() :
		      "undeclared");
	uint32_t bw;

	bw = BITMASK_BITTEST(data.mask_, te_max_bw) ?
		data.te_max_bw : 0;
	G2PCERA_DEBUG("%s- te max link bw      : %.3fMbps",
		      prefix1.c_str(),
		      BW_HEX2MBPS(bw));
	bw = BITMASK_BITTEST(data.mask_, te_max_resv_bw) ?
		data.te_max_resv_bw : 0;
	G2PCERA_DEBUG("%s- te max resv link bw : %.3fMbps",
		      prefix1.c_str(),
		      BW_HEX2MBPS(bw));
}

void
logDump(const std::string &		prefix,
	const opstate_t &		op,
	const admstate_t &		adm)
{
	G2PCERA_DEBUG("%s- operational state      : %s",
		      prefix.c_str(),
		      SHOW_OPSTATE(op));
	G2PCERA_DEBUG("%s- administrative state   : %s",
		      prefix.c_str(),
		      SHOW_ADMSTATE(adm));
}

void
logDump(const std::string &		prefix,
	const std::list<isc_t> &	iscs)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- ISC descriptor",
		      prefix.c_str());

	std::list<isc_t>::const_iterator iter;
	for (iter = iscs.begin(); iter != iscs.end(); iter++) {
		isc_t  isc;
		size_t i;

		isc = *iter;
		G2PCERA_DEBUG("%s--------------------------------------------",
			      prefix1.c_str());
		G2PCERA_DEBUG("%s- swcap            : %s",
			      prefix1.c_str(),
			      SHOW_SW_CAP(isc.switching));
		G2PCERA_DEBUG("%s- enctype          : %s",
			      prefix1.c_str(),
			      SHOW_ENC_TYPE(isc.encoding));
		for (i = 0; i < GMPLS_BW_PRIORITIES; i++) {
			G2PCERA_DEBUG("%s- te max LSP bw[%d] : %.3fMbps",
				      prefix1.c_str(),
				      i,
				      BW_HEX2MBPS(isc.max_lsp_bw.bw[i]));
		}
		switch (isc.switching) {
			case SWCAP_PSC_1:
			case SWCAP_PSC_2:
			case SWCAP_PSC_3:
			case SWCAP_PSC_4:
				G2PCERA_DEBUG("%s- te min LSP bw    : %.3fMbps",
					      prefix1.c_str(),
					      BW_HEX2MBPS(isc.spec.psc.min_lsp_bw));
				G2PCERA_DEBUG("%s- interface MTU    : %d",
					      prefix1.c_str(),
					      isc.spec.psc.if_mtu);
				break;
			case SWCAP_TDM:
				G2PCERA_DEBUG("%s- te min LSP bw    : %.3fMbps",
					      prefix1.c_str(),
					      BW_HEX2MBPS(isc.spec.tdm.min_lsp_bw));
				G2PCERA_DEBUG("%s- indication       : 0x%02x",
					      prefix1.c_str(),
					      isc.spec.tdm.indication);
				break;
			case SWCAP_L2SC:
			case SWCAP_LSC:
			case SWCAP_FSC:
				break;
			default:
				BUG();
				break;
		}
	}
	G2PCERA_DEBUG("%s--------------------------------------------",
		      prefix1.c_str());

}

void
logDump(const std::string &	prefix,
	const bw_per_prio_t &	bw)
{
	size_t i;
	std::string prefix1;

	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- available bw per prio",
		      prefix.c_str());

	for (i = 0; i < GMPLS_BW_PRIORITIES; i++) {
		G2PCERA_DEBUG("%s- bw @ prio %d         : %.3fMbps",
			      prefix1.c_str(),
			      i, BW_HEX2MBPS(bw.bw[i]));
	}
}

void
logDump(const std::string &		prefix,
	const std::list<uint32_t> &	ll,
	bool				isTdm,
	bool				isG709ODUk,
	bool				isG709OCh)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	std::list<uint32_t>::const_iterator iter;
	for (iter = ll.begin(); iter != ll.end(); iter++) {
		uint32_t  elem;
		uint8_t   st;

		elem = *iter;

		st = GMPLS_SIGNAL_TYPE(elem);

		if (isTdm) {
			G2PCERA_DEBUG("%s- signal = %s,\t free elems = %d",
				      prefix1.c_str(),
				      SHOW_GMPLS_SDH_SIGTYPE(st),
				      GMPLS_SDH_UNUSED_TIME_SLOT(elem));
		} else if (isG709ODUk) {
			G2PCERA_DEBUG("%s- signal = %s,\t free elems = %d",
				      prefix1.c_str(),
				      SHOW_GMPLS_G709_SIGTYPE(st),
				      GMPLS_G709_UNUSED_ODUK(elem));
		} else if (isG709OCh) {
			G2PCERA_DEBUG("%s- signal = %s,\t free elems = %d,"
				      "\t reduced = %s",
				      prefix1.c_str(),
				      SHOW_GMPLS_G709_SIGTYPE(st),
				      GMPLS_G709_UNUSED_OCH(elem),
				      GMPLS_G709_OCH_REDUCED(elem)?
				      "YES" : "NO");
		} else {
			G2PCERA_DEBUG("%s- id = 0x%08x",
				      prefix1.c_str(),
				      elem);
		}
	}
}

void
logDump(const std::string &				prefix,
	const std::map<uint32_t, bw_per_prio_t> &	cal)

{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- Available bw calendar",
		      prefix.c_str());

	std::map<uint32_t, bw_per_prio_t>::const_iterator iter;
	for (iter = cal.begin(); iter != cal.end(); iter++) {
		uint32_t	uTime;
		bw_per_prio_t	bw;

		uTime = (*iter).first;
		bw    = (*iter).second;

		G2PCERA_DEBUG("%s--------------------------------------------",
			      prefix1.c_str());
		G2PCERA_DEBUG("%s- Time event             : %s",
			      prefix1.c_str(),
			      unixTime2string(uTime).c_str());
		logDump(prefix1, bw);
	}
	G2PCERA_DEBUG("%s--------------------------------------------",
		      prefix1.c_str());
}

void
logDump(const std::string &		prefix,
	const telink_psc_data_t &	data)
{
	G2PCERA_DEBUG("%s- PSC min LSP bw   : %.3fMbps",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, min_lsp_bw) ?
		      BW_HEX2MBPS(data.min_lsp_bw) : 0);
	G2PCERA_DEBUG("%s- PSC interface MTU: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, if_mtu) ?
		      uint16string(data.if_mtu).c_str() :
		      "undeclared");
}
void
logDump(const std::string &		prefix,
	const telink_tdm_data_t &	data)
{
	G2PCERA_DEBUG("%s- TDM min LSP bw   : %.3fMbps",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, min_lsp_bw) ?
		      BW_HEX2MBPS(data.min_lsp_bw) : 0);
	G2PCERA_DEBUG("%s- TDM indication   : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, min_lsp_bw) ?
		      SHOW_GMPLS_SDH_STD_ARB_CAP(data.stdarb_cc_indication) :
		      "undeclared");
	G2PCERA_DEBUG("%s- HO mux cap mask  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, ho_muxcap_mask) ?
		      uint8string(data.ho_muxcap_mask).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- LO mux cap mask  : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, lo_muxcap_mask) ?
		      uint8string(data.lo_muxcap_mask).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- transparency mask: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, transparency_mask) ?
		      uint32string(data.transparency_mask).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- BLSR ring id     : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, blsr_ring_id) ?
		      uint32string(data.blsr_ring_id).c_str() :
		      "undeclared");
}

void
logDump(const std::string &		prefix,
	const telink_lscg709_data_t &	data)
{
	G2PCERA_DEBUG("%s- ODUk mux cap mask: %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, oduk_muxcap_mask) ?
		      uint8string(data.oduk_muxcap_mask).c_str() :
		      "undeclared");
}

void
logDump(const std::string &		prefix,
	const telink_lscwdm_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- dispersion PMD   : %s",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, dispersion_PMD) ?
		      uint32string(data.dispersion_PMD).c_str() :
		      "undeclared");
	G2PCERA_DEBUG("%s- span length      : %s m",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, span_length) ?
		      uint32string(data.span_length).c_str() :
		      "undeclared");

	G2PCERA_DEBUG("%s- amplifiers count : %d",
		      prefix.c_str(),
		      BITMASK_BITTEST(data.mask_, amplifiers) ?
		      data.amplifiers.size() : 0);

	if (BITMASK_BITTEST(data.mask_, amplifiers)) {
		std::list<wdm_amplifier_data_t>::const_iterator iter;
		for (iter  = data.amplifiers.begin();
		     iter != data.amplifiers.end();
		     iter++) {
			G2PCERA_DEBUG("%s-----------------------",
				      prefix1.c_str());
			G2PCERA_DEBUG("%s- gain          : %d dB",
				      prefix1.c_str(),
				      (*iter).gain);
			G2PCERA_DEBUG("%s- noise figure  : %d",
				      prefix1.c_str(),
				      (*iter).noise_figure);
		}
		G2PCERA_DEBUG("%s-----------------------",
			      prefix1.c_str());
	}
}

void
logDump(const std::string &			prefix,
	const wdm_link_lambdas_bitmap_t &	data)
{
	size_t		i;
	std::string	prefix1;

	prefix1 = prefix + DUMP_TAB;

	G2PCERA_DEBUG("%s- base label       : %s",
		      prefix.c_str(),
		      wdmLabel2string(data.base_lambda_label).c_str());
	G2PCERA_DEBUG("%s- wavelength count : %d",
		      prefix.c_str(),
		      data.num_wavelengths);

	std::string		bm;
	std::ostringstream	word;
	bm = "Ox";
	for (i = 0; i < data.bitmap_size; i++) {
		word << std::hex << std::setw(8) << std::setfill('0')
		     << data.bitmap_word[i] << "|";
	}
	bm += word.str();

	if (data.bitmap_size == 0) {
		bm = "==NULL==";
	}

	G2PCERA_DEBUG("%s- wavel. bitmask   : %s",
		      prefix.c_str(),
		      bm.c_str());

}



//
// VTY dump policy
//
void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const call_ident_t &	cid)
{
	vty_out(vty,
		"%sCALL identity [%s]%s",
		prefix.c_str(),
		SHOW_CALL_ID_TYPE(cid.type),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- srcAddr             : %s%s",
		prefix.c_str(),
		addr_ntoa(cid.src_addr),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- localId             : %s%s",
		prefix.c_str(),
		key2string(cid.local_id).c_str(),
		VTY_NEWLINE);

	if (cid.type == CALLID_GLOBUNIQ) {
		vty_out(vty,
			"%s- Globally unique attributes%s",
			VTY_NEWLINE);
		vty_out(vty,
			"%s    - ituCountryCode: 0x%06x%s",
			prefix.c_str(),
			cid.itu_country_code,
			VTY_NEWLINE);
		vty_out(vty,
			"%s    - ituCarrierCode    : %s%s",
			prefix.c_str(),
			key2string(cid.itu_carrier_code).c_str(),
			VTY_NEWLINE);
		vty_out(vty,
			"%s    - uniqueAP          : %s%s",
			prefix.c_str(),
			key2string(cid.unique_ap).c_str(),
			VTY_NEWLINE);
	}
}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const lsp_ident_t &	lid)
{
	vty_out(vty,
		"%sLSP identity%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- src router addr     : %s%s",
		prefix.c_str(),
		numb_htoa(lid.src_nid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- dst router addr     : %s%s",
		prefix.c_str(),
		numb_htoa(lid.dst_nid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- tunnel ID           : 0x%08x%s",
		prefix.c_str(),
		lid.tun_id,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- extended tunnel ID  : 0x%08x%s",
		prefix.c_str(),
		lid.ext_tun_id,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- LSP ID              : 0x%08x%s",
		prefix.c_str(),
		lid.lsp_id,
		VTY_NEWLINE);

}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const net_res_spec_t &	res)
{
	vty_out(vty,
		"%s  - Network data      : is present%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tna           : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.mask_, tna) ?
		addr_ntoa(res.tna) : "undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - data link     : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.mask_, data_link) ?
		addr_ntoa(res.data_link) : "undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - label         : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.mask_, label) ?
		label2string(res.label).c_str() : "undeclared",
		VTY_NEWLINE);
}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const res_spec_t &	res)
{
	vty_out(vty,
		"%s  - Network data      : %s present%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.mask_, net) ?
		"is" : "not",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tna           : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.net.mask_, tna) ?
		addr_ntoa(res.net.tna) : "undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - data link     : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.net.mask_, data_link) ?
		addr_ntoa(res.net.data_link) : "undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - label         : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.net.mask_, label) ?
		label2string(res.net.label).c_str() :
		"undeclared",
		VTY_NEWLINE);
#ifdef G2MPLS
	vty_out(vty,
		"%s  - Grid data         : %s present%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.mask_, grid) ?
		"is" : "not",
		VTY_NEWLINE);
	// APPLICATION
	vty_out(vty,
		"%s      - APP  type     : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, application) ?
		SHOW_GRID_APPLICATION(res.grid.application.type) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - APP  version  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, application) ?
		version2string(res.grid.application.version).c_str() :
		"undeclared",
		VTY_NEWLINE);

	// CAND HOST
	vty_out(vty,
		"%s      - CAND HOST     : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, cand_host) ?
		addr_ntoa(res.grid.cand_host) : "undeclared",
		VTY_NEWLINE);

	// FS_RESOURCES
	vty_out(vty,
		"%s      - FS name       : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		SHOW_GRID_FS_NAME(res.grid.fs_resources.fs_name) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - FS type       : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		SHOW_GRID_FS_TYPE(res.grid.fs_resources.fs_type) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - FS disk space : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		range2string(res.grid.fs_resources.disk_space).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - FS mount point: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		(res.grid.fs_resources.mount_point)->c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - FS mount src  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, fs_resources) ?
		(res.grid.fs_resources.mount_source)->c_str() :
		"undeclared",
		VTY_NEWLINE);

	// SYSTEM CAPS
	vty_out(vty,
		"%s      - SYS OS        : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		SHOW_GRID_OS(res.grid.sys_caps.os_type) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - SYS OS version: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		version2string(res.grid.sys_caps.os_version).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - SYS CPU arch  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		SHOW_GRID_CPU_ARCH(res.grid.sys_caps.cpu_arch) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - SYS exclAccess: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, sys_caps) ?
		((res.grid.sys_caps.exclusive_access == 1) ?
		 "yes" : "no")  :
		"undeclared",
		VTY_NEWLINE);
	// ranges
	vty_out(vty,
		"%s      - ind CPU speed : %s MHz%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, ind_cpu_speed) ?
		range2string(res.grid.ind_cpu_speed).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - ind CPU time  : %s seconds%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, ind_cpu_time) ?
		range2string(res.grid.ind_cpu_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - ind CPU count : %s cores%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, ind_cpu_count) ?
		range2string(res.grid.ind_cpu_count).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - ind NET bw    : %s Mbps%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, ind_net_bw) ?
		range2string(res.grid.ind_net_bw).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - ind MEM phy   : %s MB%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, ind_phy_mem) ?
		range2string(res.grid.ind_phy_mem).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - ind MEM vir   : %s MB%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, ind_phy_mem) ?
		range2string(res.grid.ind_phy_mem).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - ind DISK space: %s MB%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, ind_disk_space) ?
		range2string(res.grid.ind_disk_space).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tot CPU time  : %s seconds%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, tot_cpu_time) ?
		range2string(res.grid.tot_cpu_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tot CPU count : %s cores%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, tot_cpu_count) ?
		range2string(res.grid.tot_cpu_count).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tot MEM phy   : %s MB%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, tot_phy_mem) ?
		range2string(res.grid.tot_phy_mem).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tot MEM vir   : %s MB%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, tot_phy_mem) ?
		range2string(res.grid.tot_phy_mem).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tot DISK space: %s MB%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, tot_disk_space) ?
		range2string(res.grid.tot_disk_space).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - tot RES count : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, tot_res_count) ?
		range2string(res.grid.tot_res_count).c_str() :
		"undeclared",
		VTY_NEWLINE);
	// DATA STAGING
	vty_out(vty,
		"%s      - STAG FS name  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		SHOW_GRID_FS_NAME(res.grid.data_staging.fs_name) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - STAG createFlg: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		SHOW_GRID_CREATION_FLAG(res.grid.data_staging.
					creation_flag):
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - STAG delOnEnd : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		((res.grid.data_staging.del_on_termination == 1) ?
		 "yes" : "no")  :  "undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - STAG filename : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		(res.grid.data_staging.filename)->c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - STAG source   : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		(res.grid.data_staging.source)->c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s      - STAG target   : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, data_staging) ?
		(res.grid.data_staging.target)->c_str() :
		"undeclared",
		VTY_NEWLINE);

	// GRID SITE ID
	vty_out(vty,
		"%s      - GRID SITE addr: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(res.grid.mask_, grid_site_id) ?
		numb_htoa(res.grid.grid_site_id) : "undeclared",
		VTY_NEWLINE);

#endif // G2MPLS
}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const call_info_t &	cdata)
{
	tm *               curTime;
	const char *       timeStringFormat = "%a, %d %b %Y %H:%M:%S %zGMT";
	const int          timeStringLength = 40;
	char               timeString[timeStringLength];

	vty_out(vty,
		"%sCALL data%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- Symbolic name       : %s%s",
		prefix.c_str(),
		(cdata.call_name) ?
		(cdata.call_name)->c_str() : "==UNKNOWN==",
		VTY_NEWLINE);
	curTime = localtime((time_t *) &cdata.times.start_time);
	strftime(timeString, timeStringLength, timeStringFormat, curTime);
	vty_out(vty,
		"%s- startTime           : %s%s",
		prefix.c_str(),
		timeString,
		VTY_NEWLINE);
	curTime = localtime((time_t *) &cdata.times.end_time);
	strftime(timeString, timeStringLength, timeStringFormat, curTime);
	vty_out(vty,
		"%s- endTime             : %s%s",
		prefix.c_str(),
		timeString,
		VTY_NEWLINE);
	double duration = difftime(cdata.times.end_time,
				   cdata.times.start_time);
	vty_out(vty,
		"%s    - duration    : %.0f seconds%s",
		prefix.c_str(),
		(float)duration,
		VTY_NEWLINE);
#ifdef G2MPLS
	vty_out(vty,
		"%s- jobName             : %s%s",
		prefix.c_str(),
		(cdata.job_name) ?
		(cdata.job_name)->c_str() : "==UNKNOWN==",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- jobProject          : %s%s",
		prefix.c_str(),
		(cdata.job_project) ?
		(cdata.job_project)->c_str() : "==UNKNOWN==",
		VTY_NEWLINE);
#endif // G2MPLS

	vty_out(vty,
		"%s- iTNA resources%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vtyDump(vty, prefix, cdata.iTNA_res);

	vty_out(vty,
		"%s- eTNA resources%s",
		VTY_NEWLINE);
	vtyDump(vty, prefix, cdata.eTNA_res);
}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const recovery_info_t &	rdata)
{
	vty_out(vty,
		"%sRecovery data%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- recovery type       : %s%s",
		prefix.c_str(),
		SHOW_GMPLS_RECOVERYTYPE(rdata.rec_type),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- disjointness        : %s%s",
		prefix.c_str(),
		SHOW_GMPLS_DISJOINTNESS(rdata.disj_type),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const lsp_info_t &	ldata)
{
	vty_out(vty,
		"%sLSP data%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- swcap               : %s%s",
		prefix.c_str(),
		SHOW_SW_CAP(ldata.sw_cap),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- enctype             : %s%s",
		prefix.c_str(),
		SHOW_ENC_TYPE(ldata.enc_type),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- gpid                : %s%s",
		prefix.c_str(),
		SHOW_GPID(ldata.gpid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- bandwidth           : %.3fMbps (signal %s)%s",
		prefix.c_str(),
		BW_HEX2MBPS(ldata.bw),
		SHOW_GMPLS_BWENC(ldata.bw),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- setupPrio           : %d%s",
		prefix.c_str(),
		ldata.setup_prio,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- holdingPrio         : %d%s",
		prefix.c_str(),
		ldata.holding_prio,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- excludeAny          : 0x%08x%s",
		prefix.c_str(),
		ldata.exclude_any,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- includeAny          : 0x%08x%s",
		prefix.c_str(),
		ldata.include_any,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- includeAll          : 0x%08x%s",
		prefix.c_str(),
		ldata.include_all,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- linkProtMask        : %s%s",
		prefix.c_str(),
		SHOW_GMPLS_PROTTYPE(ldata.link_prot_mask),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- crankback scope     : %s%s",
		prefix.c_str(),
		SHOW_CBACK_SCOPE(ldata.crankback),
		VTY_NEWLINE);
	vty_out(vty,
		"%s    - maxRetrSrc      : %d%s",
		prefix.c_str(),
		ldata.max_cback_retries_src,
		VTY_NEWLINE);
	vty_out(vty,
		"%s    - maxRetrInt      : %d%s",
		prefix.c_str(),
		ldata.max_cback_retries_intmd,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- type                : %s%s",
		prefix.c_str(),
		SHOW_LSP_TYPE(ldata.type),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- action              : %s%s",
		prefix.c_str(),
		SHOW_LSP_RES_ACTION(ldata.action),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- rroMode             : %s%s",
		prefix.c_str(),
		SHOW_LSP_RRO_MODE(ldata.rro_mode),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- refresh             : %d ms%s",
		prefix.c_str(),
		ldata.refresh_interval,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- activateAck         : %s%s",
		prefix.c_str(),
		ldata.activate_ack ? "YES" : "NO",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- rapidRetrInt        : %d ms%s",
		prefix.c_str(),
		ldata.rapid_retransm_interval,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- rapidRetryLimit     : %d%s",
		prefix.c_str(),
		ldata.rapid_retry_limit,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- incrValueDelta      : %d%s",
		prefix.c_str(),
		ldata.increment_value_delta,
		VTY_NEWLINE);
}


void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const ero_hop_t &	hop)
{
	vty_out(vty,
		"%sEro Hop [%s]%s",
		prefix.c_str(),
		(hop.loose == 1) ? "LOOSE" : "STRICT",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- node                : %s%s",
		prefix.c_str(),
		numb_htoa(hop.node),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- TE-link             : %s%s",
		prefix.c_str(),
		addr_ntoa(hop.te_link),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- upstream D-link     : %s%s",
		prefix.c_str(),
		addr_ntoa(hop.upstream_datalink),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- downstream D-link   : %s%s",
		prefix.c_str(),
		addr_ntoa(hop.downstream_datalink),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- upstream label      : %s%s",
		prefix.c_str(),
		label2string(hop.upstream_label).c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- downstream label    : %s%s",
		prefix.c_str(),
		label2string(hop.downstream_label).c_str(),
		VTY_NEWLINE);
}


void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const topology_summary_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- root router_id      : %s%s",
		prefix.c_str(),
		numb_htoa(data.root_rid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total nodes         : %d%s",
		prefix.c_str(),
		data.tot_nodes,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total links         : %d%s",
		prefix.c_str(),
		data.tot_links,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total TNAs          : %d%s",
		prefix.c_str(),
		data.tot_tnas,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total GNS calls     : %d%s",
		prefix.c_str(),
		data.tot_gns_calls,
		VTY_NEWLINE);

	vty_out(vty,
		"%s- algorithm info%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max link cost    : %d%s",
		prefix1.c_str(),
		data.max_link_cost,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- SPF selector     : %s%s",
		prefix1.c_str(),
		SPF_TYPE_TO_STRING(data.spf_selector),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- SPF revision     : %d%s",
		prefix1.c_str(),
		data.spf_revision,
		VTY_NEWLINE);
}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const node_ident_t &	data)
{
	vty_out(vty,
		"%s- type                : %s%s",
		prefix.c_str(),
		NODE_TYPE_TO_STRING(data.type),
		VTY_NEWLINE);

	vty_out(vty,
		"%s- id                  : %s%s",
		prefix.c_str(),
		numb_htoa(data.id),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const net_node_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- is domain           : %s%s",
		prefix.c_str(),
		data.is_domain ? "YES" : "NO",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- operational state   : %s%s",
		prefix.c_str(),
		SHOW_OPSTATE(data.op_state),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- administrative state: %s%s",
		prefix.c_str(),
		SHOW_ADMSTATE(data.adm_state),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- TE color mask       : 0x%08x%s",
		prefix.c_str(), data.te_colors,
		VTY_NEWLINE);

	vty_out(vty,
		"%s- Areas count         : %d%s",
		prefix.c_str(),
		data.areas.size(),
		VTY_NEWLINE);

	std::list<uint16_t>::const_iterator areaIter;
	for (areaIter  = data.areas.begin();
	     areaIter != data.areas.end();
	     areaIter++) {
		vty_out(vty,
			"%s- Area = %d%s",
			prefix1.c_str(),
			*areaIter,
			VTY_NEWLINE);
	}
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const grid_site_data_t &	data)
{
	vty_out(vty,
		"%s- name                : %s%s",
		prefix.c_str(),
		data.name ? data.name->c_str() : "<empty>",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- location            : %s%s",
		prefix.c_str(),
		geoCoord2string(data.location).c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- PE router id        : %s%s",
		prefix.c_str(),
		numb_htoa(data.pe_router_id),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const grid_subnode_ident_t &	data)
{
	vty_out(vty,
		"%s- type                   : %s%s",
		prefix.c_str(),
		GRID_SUBNODE_TO_STRING(data.type),
		VTY_NEWLINE);

	vty_out(vty,
		"%s- id                     : %s%s",
		prefix.c_str(),
		numb_htoa(data.id),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const grid_service_data_t &	data)
{
	vty_out(vty,
		"%s- Service name           : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, name) ?
		(data.name)->c_str() :
		"undeclared");
	vty_out(vty,
		"%s- type                   : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, data) ?
		SHOW_GRID_SERVICETYPE(data.data.type) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- version                : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, data) ?
		version2string(data.data.version).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- state                  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, state) ?
		SHOW_GRID_SERVICESTATE(data.state) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- endpoint address       : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, endpoint_addr) ?
		addr_ntoa(data.endpoint_addr) : "undeclared",
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const grid_ce_data_t &		data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- CompElem name          : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, name) ?
		(data.name)->c_str() :
		"undeclared");
	vty_out(vty,
		"%s- LRMS type              : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, lrms_info) ?
		SHOW_GRID_LRMS(data.lrms_info.type) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- LRMS version           : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, lrms_info) ?
		version2string(data.lrms_info.version).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- host address           : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, host_addr) ?
		addr_ntoa(data.host_addr) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- gatekeeper port        : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, gatekeeper_port) ?
		uint32string(data.gatekeeper_port).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- job manager            : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, job_manager) ?
		(data.job_manager)->c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- data directory         : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, data_dir) ?
		(data.data_dir)->c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- dflt storage elem      : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, default_storage_elem_id) ?
		uint32string(data.default_storage_elem_id).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- jobs state             : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_state) ?
		SHOW_GRID_CESE_STATE(data.jobs_state.state) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- free job slots      : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_state) ?
		uint32string(data.jobs_state.free_job_slots).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- running jobs        : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_stats) ?
		uint32string(data.jobs_stats.running_jobs).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- waiting jobs        : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_stats) ?
		uint32string(data.jobs_stats.waiting_jobs).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- total jobs          : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_stats) ?
		uint32string(data.jobs_stats.total_jobs).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- jobs time performances%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- est resp time       : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_timeperf) ?
		uint32string(data.jobs_timeperf.
			     estimated_response_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- worst resp time     : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_timeperf) ?
		uint32string(data.jobs_timeperf.
			     worst_response_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- jobs time policies%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max wclock time     : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		uint32string(data.jobs_timepolicy.
			     max_wallclock_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max obt wclock time : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		uint32string(data.jobs_timepolicy.
			     max_obtainable_wallclock_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max cpu time        : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		uint32string(data.jobs_timepolicy.
			     max_cpu_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max obt cpu time    : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_timepolicy) ?
		uint32string(data.jobs_timepolicy.
			     max_obtainable_cpu_time).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- jobs load policies%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max total jobs      : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		uint32string(data.jobs_loadpolicy.
			     max_total_jobs).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max running jobs    : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		uint32string(data.jobs_loadpolicy.
			     max_running_jobs).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max waiting jobs    : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		uint32string(data.jobs_loadpolicy.
			     max_waiting_jobs).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- assigned job slots  : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		uint16string(data.jobs_loadpolicy.
			     assigned_job_slots).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- max slots per job   : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		uint16string(data.jobs_loadpolicy.
			     max_slots_per_job).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- priority            : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		uint8string(data.jobs_loadpolicy.
			    priority).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- preemption flag     : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, jobs_loadpolicy) ?
		bool2string(data.jobs_loadpolicy.
			    preemption_flag).c_str() :
		"undeclared",
		VTY_NEWLINE);

	vty_out(vty,
		"%s- free job slots cal     : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, free_job_slots_calendar) ?
		"" : "undeclared",
		VTY_NEWLINE);

	if (data.free_job_slots_calendar.size()) {

		std::map<uint32_t, uint16_t>::const_iterator	iter;

		for (iter  = data.free_job_slots_calendar.begin();
		     iter != data.free_job_slots_calendar.end();
		     iter++) {
			vty_out(vty,
				"%s- @[%s] --> fjs = %s%s",
				prefix1.c_str(),
				unixTime2string((*iter).first).c_str(),
				uint16string((*iter).second).c_str(),
				VTY_NEWLINE);
		}
	}
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const grid_subcluster_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- SubCluster name        : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, name) ?
		(data.name)->c_str() :
		"undeclared");
	vty_out(vty,
		"%s- CPUs arch              : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, cpu) ?
		SHOW_GRID_CPU_ARCH(data.cpu.arch) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- CPUs count physical    : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, cpu) ?
		uint32string(data.cpu.count.physical).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- CPUs count logical     : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, cpu) ?
		uint32string(data.cpu.count.logical).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- OS type                : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, os) ?
		SHOW_GRID_OS(data.os.type) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- OS version             : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, os) ?
		version2string(data.os.version).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- RAM size               : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, memory) ?
		uint32string(data.memory.ram_size).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- virtual memory size    : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, memory) ?
		uint32string(data.memory.virtual_memory_size).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- Software packages      : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, software_packages) ?
		"" : "undeclared",
		VTY_NEWLINE);
	if (data.software_packages.size()) {

		std::list<grid_software_package_t *>::const_iterator	iter;

		for (iter  = data.software_packages.begin();
		     iter != data.software_packages.end();
		     iter++) {
			vty_out(vty,
				"%s-----------------------------%s",
				prefix1.c_str(),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- Application type       : %s%s",
				prefix1.c_str(),
				SHOW_GRID_APPLICATION((*iter)->software.type),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- Application version    : %s%s",
				prefix1.c_str(),
				version2string((*iter)->
					       software.version).c_str(),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- software env. setup    : %s%s",
				prefix1.c_str(),
				((*iter)->software_env_setup)->c_str(),
				VTY_NEWLINE);
		}
	}

	vty_out(vty,
		"%s- CPU count cal          : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, subcluster_calendar) ?
		"" : "undeclared",
		VTY_NEWLINE);

	if (data.subcluster_calendar.size()) {

		std::map<uint32_t, grid_cpu_count_t>::const_iterator	iter;

		for (iter  = data.subcluster_calendar.begin();
		     iter != data.subcluster_calendar.end();
		     iter++) {
			vty_out(vty,
				"%s- @[%s] --> "
				"CPU phy = %d, CPU log = %d%s",
				prefix1.c_str(),
				unixTime2string((*iter).first).c_str(),
				(*iter).second.physical,
				(*iter).second.logical,
				VTY_NEWLINE);
		}
	}
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const grid_se_data_t &		data)
{
	std::string prefix1, prefix2;
	prefix1 = prefix + DUMP_TAB;
	prefix2 = prefix1 + DUMP_TAB;

	vty_out(vty,
		"%s- StoElem name           : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, name) ?
		(data.name)->c_str() :
		"undeclared");
	vty_out(vty,
		"%s- arch                   : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, storage_info) ?
		SHOW_GRID_STORAGE_ARCH(data.storage_info.arch) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- state                  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, storage_info) ?
		SHOW_GRID_CESE_STATE(data.storage_info.state) :
		"undeclared",
		VTY_NEWLINE);
	uint16_t tmp;
	tmp = *((uint16_t *)(&(data.storage_info.access_protocols)),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- access protocols       : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, storage_info) ?
		uint16string(tmp).c_str() :
		"undeclared",
		VTY_NEWLINE);
	tmp = *((uint16_t *)(&(data.storage_info.control_protocols)),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- control protocols      : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, storage_info) ?
		uint16string(tmp).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- online size            : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, online_size) ?
		"" : "undeclared",
		VTY_NEWLINE);
	if (BITMASK_BITTEST(data.mask_, online_size)) {
		vty_out(vty,
			"%s- total               : %d MB%s",
			prefix1.c_str(),
			data.online_size.total,
			VTY_NEWLINE);
		vty_out(vty,
			"%s- used                : %d MB%s",
			prefix1.c_str(),
			data.online_size.used,
			VTY_NEWLINE);
	}
	vty_out(vty,
		"%s- nearline size          : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, nearline_size) ?
		"" : "undeclared",
		VTY_NEWLINE);
	if (BITMASK_BITTEST(data.mask_, nearline_size)) {
		vty_out(vty,
			"%s- total               : %d MB%s",
			prefix1.c_str(),
			data.nearline_size.total,
			VTY_NEWLINE);
		vty_out(vty,
			"%s- used                : %d MB%s",
			prefix1.c_str(),
			data.nearline_size.used,
			VTY_NEWLINE);
	}

	vty_out(vty,
		"%s- storage areas          : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, storage_areas) ?
		"" : "undeclared",
		VTY_NEWLINE);

	if (data.storage_areas.size()) {

		std::list<grid_storage_area_t *>::const_iterator iter;

		for (iter  = data.storage_areas.begin();
		     iter != data.storage_areas.end();
		     iter++) {
			vty_out(vty,
				"%s-----------------------------%s",
				prefix1.c_str(),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- storage area name      : %s%s",
				prefix1.c_str(),
				((*iter)->storage_area_name)->c_str(),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- storage area path      : %s%s",
				prefix1.c_str(),
				((*iter)->storage_area_path)->c_str(),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- storage area info      : %s",
				prefix1.c_str(),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- total online size   : %d MB%s",
				prefix2.c_str(),
				(*iter)->storage_area_info.total_online_size,
				VTY_NEWLINE);
			vty_out(vty,
				"%s- free online size    : %d MB%s",
				prefix2.c_str(),
				(*iter)->storage_area_info.free_online_size,
				VTY_NEWLINE);
			vty_out(vty,
				"%s- resv total online sz: %d MB%s",
				prefix2.c_str(),
				(*iter)->storage_area_info.
				reserved_total_online_size,
				VTY_NEWLINE);
			vty_out(vty,
				"%s- total nearline size : %d MB%s",
				prefix2.c_str(),
				(*iter)->storage_area_info.total_nearline_size,
				VTY_NEWLINE);
			vty_out(vty,
				"%s- free nearline size  : %d MB%s",
				prefix2.c_str(),
				(*iter)->storage_area_info.free_nearline_size,
				VTY_NEWLINE);
			vty_out(vty,
				"%s- resv total nrline sz: %d MB%s",
				prefix2.c_str(),
				(*iter)->storage_area_info.
				reserved_nearline_size,
				VTY_NEWLINE);
			vty_out(vty,
				"%s- retention policy    : %s%s",
				prefix2.c_str(),
				SHOW_GRID_STORAGE_RETENTION_POLICY((*iter)->storage_area_info.retention_policy),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- access latency      : %s%s",
				prefix2.c_str(),
				SHOW_GRID_STORAGE_ACCESS_LATENCY((*iter)->storage_area_info.access_latency),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- expiration mode     : %s%s",
				prefix2.c_str(),
				SHOW_GRID_STORAGE_EXPIRATION_MODE((*iter)->storage_area_info.expiration_mode),
				VTY_NEWLINE);
		}
	}

	vty_out(vty,
		"%s- storage count cal      : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, se_calendar) ?
		"" : "undeclared",
		VTY_NEWLINE);

	if (data.se_calendar.size()) {

		std::map<uint32_t, grid_storage_count_t>::const_iterator iter;

		for (iter  = data.se_calendar.begin();
		     iter != data.se_calendar.end();
		     iter++) {
			vty_out(vty,
				"%s- @[%s] --> "
				"free online size = %d MB, CPU log = %d%s",
				prefix1.c_str(),
				unixTime2string((*iter).first).c_str(),
				(*iter).second.free_online_size,
				(*iter).second.logical_cpus,
				VTY_NEWLINE);
		}
	}
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const tna_ident_t &		ident)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- ancestor RC id         : %s%s",
		prefix.c_str(),
		numb_htoa(ident.rc_id),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- border node            : %s%s",
		prefix.c_str(),
		numb_htoa(ident.rid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- TNA address            : %s%s",
		prefix.c_str(),
		addr_ntoa(ident.id),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const telink_ident_t &		ident)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- local id               : %s%s",
		prefix.c_str(),
		addr_ntoa(ident.lcl_id),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- local router id        : %s%s",
		prefix.c_str(),
		numb_htoa(ident.lcl_rid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- remote id              : %s%s",
		prefix.c_str(),
		addr_ntoa(ident.rem_id),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- remote router id       : %s%s",
		prefix.c_str(),
		numb_htoa(ident.rem_rid),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- mode                   : %s%s",
		prefix.c_str(),
		LINK_MODE_TO_STRING(ident.mode),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- ancestor RC id         : %s%s",
		prefix.c_str(),
		numb_htoa(ident.lcl_rc_id),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- neighbor RC id         : %s%s",
		prefix.c_str(),
		numb_htoa(ident.rem_rc_id),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const telink_com_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- admin metric           : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, admin_metric) ?
		uint32string(data.admin_metric).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- TE info%s",
		prefix.c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- te metric           : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, te_metric) ?
		uint32string(data.te_metric).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- te color mask       : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, te_metric) ?
		uint32string(data.te_color_mask).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- te prot type mask   : %s%s",
		prefix1.c_str(),
		BITMASK_BITTEST(data.mask_, te_prot_type_mask) ?
		uint8string(data.te_prot_type_mask).c_str() :
		"undeclared",
		VTY_NEWLINE);
	uint32_t bw;

	bw = BITMASK_BITTEST(data.mask_, te_max_bw) ?
		data.te_max_bw : 0;
	vty_out(vty,
		"%s- te max link bw      : %.3fMbps%s",
		prefix1.c_str(),
		BW_HEX2MBPS(bw),
		VTY_NEWLINE);
	bw = BITMASK_BITTEST(data.mask_, te_max_resv_bw) ?
		data.te_max_resv_bw : 0;
	vty_out(vty,
		"%s- te max resv link bw : %.3fMbps%s",
		prefix1.c_str(),
		BW_HEX2MBPS(bw),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const opstate_t &		op,
	const admstate_t &		adm)
{
	vty_out(vty,
		"%s- operational state      : %s%s",
		prefix.c_str(),
		SHOW_OPSTATE(op),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- administrative state   : %s%s",
		prefix.c_str(),
		SHOW_ADMSTATE(adm),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const std::list<isc_t> &	iscs)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- ISC descriptor%s",
		prefix.c_str(),
		VTY_NEWLINE);

	std::list<isc_t>::const_iterator iter;
	for (iter = iscs.begin(); iter != iscs.end(); iter++) {
		isc_t  isc;
		size_t i;

		isc = *iter;
		vty_out(vty,
			"%s--------------------------------------------%s",
			prefix1.c_str(),
			VTY_NEWLINE);
		vty_out(vty,
			"%s- swcap            : %s%s",
			prefix1.c_str(),
			SHOW_SW_CAP(isc.switching),
			VTY_NEWLINE);
		vty_out(vty,
			"%s- enctype          : %s%s",
			prefix1.c_str(),
			SHOW_ENC_TYPE(isc.encoding),
			VTY_NEWLINE);
		for (i = 0; i < GMPLS_BW_PRIORITIES; i++) {
			vty_out(vty,
				"%s- te max LSP bw[%d] : %.3fMbps%s",
				prefix1.c_str(),
				i,
				BW_HEX2MBPS(isc.max_lsp_bw.bw[i]),
				VTY_NEWLINE);
		}
		switch (isc.switching) {
			case SWCAP_PSC_1:
			case SWCAP_PSC_2:
			case SWCAP_PSC_3:
			case SWCAP_PSC_4:
				vty_out(vty,
					"%s- te min LSP bw    : %.3fMbps%s",
					prefix1.c_str(),
					BW_HEX2MBPS(isc.spec.psc.min_lsp_bw),
					VTY_NEWLINE);
				vty_out(vty,
					"%s- interface MTU    : %d%s",
					prefix1.c_str(),
					isc.spec.psc.if_mtu,
					VTY_NEWLINE);
				break;
			case SWCAP_TDM:
				vty_out(vty,
					"%s- te min LSP bw    : %.3fMbps%s",
					prefix1.c_str(),
					BW_HEX2MBPS(isc.spec.tdm.min_lsp_bw),
					VTY_NEWLINE);
				vty_out(vty,
					"%s- indication       : 0x%02x%s",
					prefix1.c_str(),
					isc.spec.tdm.indication,
					VTY_NEWLINE);
				break;
			case SWCAP_L2SC:
			case SWCAP_LSC:
			case SWCAP_FSC:
				break;
			default:
				BUG();
				break;
		}
	}
	vty_out(vty,
		"%s--------------------------------------------%s",
		prefix1.c_str(),
		VTY_NEWLINE);

}

void
vtyDump(struct vty *		vty,
	const std::string &	prefix,
	const bw_per_prio_t &	bw)
{
	size_t i;
	std::string prefix1;

	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- available bw per prio%s",
		prefix.c_str(),
		VTY_NEWLINE);

	for (i = 0; i < GMPLS_BW_PRIORITIES; i++) {
		vty_out(vty,
			"%s- bw @ prio %d         : %.3fMbps%s",
			prefix1.c_str(),
			i, BW_HEX2MBPS(bw.bw[i]),
			VTY_NEWLINE);
	}
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const std::list<uint32_t> &	ll,
	bool				isTdm,
	bool				isG709ODUk,
	bool				isG709OCh)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	std::list<uint32_t>::const_iterator iter;
	for (iter = ll.begin(); iter != ll.end(); iter++) {
		uint32_t  elem;
		uint8_t   st;

		elem = *iter;

		st = GMPLS_SIGNAL_TYPE(elem);

		if (isTdm) {
			vty_out(vty,
				"%s- signal = %s,\t free elems = %d%s",
				prefix1.c_str(),
				SHOW_GMPLS_SDH_SIGTYPE(st),
				GMPLS_SDH_UNUSED_TIME_SLOT(elem),
				VTY_NEWLINE);
		} else if (isG709ODUk) {
			vty_out(vty,
				"%s- signal = %s,\t free elems = %d%s",
				prefix1.c_str(),
				SHOW_GMPLS_G709_SIGTYPE(st),
				GMPLS_G709_UNUSED_ODUK(elem),
				VTY_NEWLINE);
		} else if (isG709OCh) {
			vty_out(vty,
				"%s- signal = %s,\t free elems = %d,"
				"\t reduced = %s%s",
				prefix1.c_str(),
				SHOW_GMPLS_G709_SIGTYPE(st),
				GMPLS_G709_UNUSED_OCH(elem),
				GMPLS_G709_OCH_REDUCED(elem)?
				"YES" : "NO",
				VTY_NEWLINE);
		} else {
			vty_out(vty,
				"%s- id = 0x%08x%s",
				prefix1.c_str(),
				elem,
				VTY_NEWLINE);
		}
	}
}

void
vtyDump(struct vty *					vty,
	const std::string &				prefix,
	const std::map<uint32_t, bw_per_prio_t> &	cal)

{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- Available bw calendar%s",
		prefix.c_str(),
		VTY_NEWLINE);

	std::map<uint32_t, bw_per_prio_t>::const_iterator iter;
	for (iter = cal.begin(); iter != cal.end(); iter++) {
		uint32_t	uTime;
		bw_per_prio_t	bw;

		uTime = (*iter).first;
		bw    = (*iter).second;

		vty_out(vty,
			"%s--------------------------------------------%s",
			prefix1.c_str(),
			VTY_NEWLINE);
		vty_out(vty,
			"%s- Time event             : %s%s",
			prefix1.c_str(),
			unixTime2string(uTime).c_str(),
			VTY_NEWLINE);
		vtyDump(vty, prefix1, bw);
	}
	vty_out(vty,
		"%s--------------------------------------------%s",
		prefix1.c_str(),
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const telink_psc_data_t &	data)
{
	vty_out(vty,
		"%s- PSC min LSP bw   : %.3fMbps%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, min_lsp_bw) ?
		BW_HEX2MBPS(data.min_lsp_bw) : 0,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- PSC interface MTU: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, if_mtu) ?
		uint16string(data.if_mtu).c_str() :
		"undeclared",
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const telink_tdm_data_t &	data)
{
	vty_out(vty,
		"%s- TDM min LSP bw   : %.3fMbps%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, min_lsp_bw) ?
		BW_HEX2MBPS(data.min_lsp_bw) : 0,
		VTY_NEWLINE);
	vty_out(vty,
		"%s- TDM indication   : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, min_lsp_bw) ?
		SHOW_GMPLS_SDH_STD_ARB_CAP(data.stdarb_cc_indication) :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- HO mux cap mask  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, ho_muxcap_mask) ?
		uint8string(data.ho_muxcap_mask).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- LO mux cap mask  : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, lo_muxcap_mask) ?
		uint8string(data.lo_muxcap_mask).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- transparency mask: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, transparency_mask) ?
		uint32string(data.transparency_mask).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- BLSR ring id     : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, blsr_ring_id) ?
		uint32string(data.blsr_ring_id).c_str() :
		"undeclared",
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const telink_lscg709_data_t &	data)
{
	vty_out(vty,
		"%s- ODUk mux cap mask: %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, oduk_muxcap_mask) ?
		uint8string(data.oduk_muxcap_mask).c_str() :
		"undeclared",
		VTY_NEWLINE);
}

void
vtyDump(struct vty *			vty,
	const std::string &		prefix,
	const telink_lscwdm_data_t &	data)
{
	std::string prefix1;
	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- dispersion PMD   : %s%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, dispersion_PMD) ?
		uint32string(data.dispersion_PMD).c_str() :
		"undeclared",
		VTY_NEWLINE);
	vty_out(vty,
		"%s- span length      : %s m%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, span_length) ?
		uint32string(data.span_length).c_str() :
		"undeclared",
		VTY_NEWLINE);

	vty_out(vty,
		"%s- amplifiers count : %d%s",
		prefix.c_str(),
		BITMASK_BITTEST(data.mask_, amplifiers) ?
		data.amplifiers.size() : 0,
		VTY_NEWLINE);

	if (BITMASK_BITTEST(data.mask_, amplifiers)) {
		std::list<wdm_amplifier_data_t>::const_iterator iter;
		for (iter  = data.amplifiers.begin();
		     iter != data.amplifiers.end();
		     iter++) {
			vty_out(vty,
				"%s-----------------------%s",
				prefix1.c_str(),
				VTY_NEWLINE);
			vty_out(vty,
				"%s- gain          : %d dB%s",
				prefix1.c_str(),
				(*iter).gain,
				VTY_NEWLINE);
			vty_out(vty,
				"%s- noise figure  : %d%s",
				prefix1.c_str(),
				(*iter).noise_figure,
				VTY_NEWLINE);
		}
		vty_out(vty,
			"%s-----------------------%s",
			prefix1.c_str(),
			VTY_NEWLINE);
	}
}

void
vtyDump(struct vty *				vty,
	const std::string &			prefix,
	const wdm_link_lambdas_bitmap_t &	data)
{
	size_t		i;
	std::string	prefix1;

	prefix1 = prefix + DUMP_TAB;

	vty_out(vty,
		"%s- base label       : %s%s",
		prefix.c_str(),
		wdmLabel2string(data.base_lambda_label).c_str(),
		VTY_NEWLINE);
	vty_out(vty,
		"%s- wavelength count : %d%s",
		prefix.c_str(),
		data.num_wavelengths,
		VTY_NEWLINE);

	std::string		bm;
	std::ostringstream	word;
	bm = "Ox";
	for (i = 0; i < data.bitmap_size; i++) {
		word << std::hex << std::setw(8) << std::setfill('0')
		     << data.bitmap_word[i] << "|";
	}
	bm += word.str();

	if (data.bitmap_size == 0) {
		bm = "==NULL==";
	}

	vty_out(vty,
		"%s- wavel. bitmask   : %s%s",
		prefix.c_str(),
		bm.c_str(),
		VTY_NEWLINE);

}
