//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//




#include <zebra.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <getopt.h>

#include "log.h"
#include "version.h"
#include "command.h"
#include "thread.h"
#include "sigevent.h"
#include "privs.h"
#include "memory.h"
#include "thread.h"
#include "stream.h"
#include "sockopt.h"

#include "g2pcerad.hh"
#include "g2pcera_trace.hh"
#include "g2pcera_config.hh"

//#include "g2pcera_corba.hh"
//#include "lib/corba.h"

using namespace PCERA;

struct thread_master * master = NULL;

// g2pcerad privileges
zebra_capabilities_t _caps_p [] = {
	ZCAP_BIND,
	ZCAP_NET_ADMIN,
};

// GLOBALS
struct zebra_privs_t		g2pcerad_privs;
struct quagga_signal_t		g2pcera_signals[4];
char *				progname;
const char *			pid_file = PATH_G2PCERAD_PID;
char				config_default[]= G2PCERAD_DEFAULT_CONFIG;

struct option longopts[] = {
	{ "daemon",		no_argument,		NULL,	'd'},
	{ "config_file",	required_argument,	NULL,	'f'},
	{ "pid_file",		required_argument,	NULL,	'i'},
	{ "dryrun",		no_argument,		NULL,	'C'},
	{ "alone",		no_argument,		NULL,	'A'},
	{ "iors_dir",		required_argument,	NULL,	'o'},
	{ "help",		no_argument,		NULL,	'h'},
	{ "vty_port",		required_argument,	NULL,	'P'},
	{ "user",		required_argument,	NULL,	'u'},
	{ "group",		required_argument,	NULL,	'g'},
	{ "version",		no_argument,		NULL,	'v'},
	{ 0 }
};

// SIG handlers
static void
sighup(void)
{
	G2PCERA_INFO("SIGHUP received");
}

static int terminate_execution = 0;
static void
sigint(void)
{
	G2PCERA_NOTICE("Terminating on SIGINT");
	terminate_execution = 1;

	if (G2PCERA_MASTER.dryrun()) {
		G2PCERA_MASTER.destroy();
		exit(-1);
	}
}

// SIGUSR1 handler
static void
sigusr1 (void)
{
	zlog_rotate(NULL);
}

// Help information display
static void __attribute__ ((noreturn))
     usage (char *progname, int status)
{
	if (status != 0) {
		fprintf(stderr,
			"Try `%s --help' for more information.\n",
			progname);
	} else {
		printf("Usage : %s [OPTION...]\n"
		       "Daemon which manages G2PCERA.\n\n"
		       "-d, --daemon       Runs in daemon mode\n"
		       "-f, --config_file  Set configuration file name\n"
		       "-i, --pid_file     Set process identifier file name\n"
		       "-C, --dryrun       Check configuration and exit\n"
		       "-o, --iors_dir     Set IORs directory\n"
		       "-P, --vty_port     Set vty's port number\n"
		       "-u, --user         User to run as\n"
		       "-g, --group        Group to run as\n"
		       "-v, --version      Print program version\n"
		       "-h, --help         Display this help and exit\n"
		       "\n"
		       "Report bugs to %s\n",
		       progname, ZEBRA_BUG_ADDRESS);
	}

	exit(status);
}


int main(int argc, char ** argv)
{
	char *		p;
	int		vty_port;
	char *		vty_addr;
	int		daemon_mode;
	char *		iors_dir;
	char *		config_file;
	struct thread	thread;
	bool		alone;

	vty_port    = G2PCERAD_VTY_PORT;
	vty_addr    = NULL;
	daemon_mode = 0;
	config_file = NULL;
	iors_dir    = NULL;
	alone       = false;

	// INITIALIZATION
#if defined(QUAGGA_USER) && defined(QUAGGA_GROUP)
	g2pcerad_privs.user	   = QUAGGA_USER;
	g2pcerad_privs.group	   = QUAGGA_GROUP;
#endif
#if defined(VTY_GROUP)
	g2pcerad_privs.vty_group   = VTY_GROUP;
#endif
	g2pcerad_privs.caps_p	   = _caps_p;
	g2pcerad_privs.cap_num_p   = sizeof(_caps_p)/sizeof(_caps_p[0]);
	g2pcerad_privs.cap_num_i   = 0;

	g2pcera_signals[0].signal  = SIGHUP;
	g2pcera_signals[0].handler = &sighup;
	g2pcera_signals[1].signal  = SIGUSR1;
	g2pcera_signals[1].handler = &sigusr1;
	g2pcera_signals[2].signal  = SIGINT;
	g2pcera_signals[2].handler = &sigint;
	g2pcera_signals[3].signal  = SIGTERM;
	g2pcera_signals[3].handler = &sigint;

	// Set umask before anything for security
	umask(0027);

	// Get program name
	progname = ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);

	// Initialize the log subsystem
	zlog_default = openzlog(progname,
				ZLOG_G2PCERA,
				LOG_CONS | LOG_NDELAY | LOG_PID,
				LOG_DAEMON);
	// Create Master
	if (!G2PCERA_Master::init()) {
		G2PCERA_NOTICE("Cannot allocate a PCERA master object");
		exit (-1);
	}
	master = G2PCERA_MASTER.getMaster();

	// just reads the options from stdin
	while (1) {
		int opt;

		opt = getopt_long(argc, argv, "dCAhvf:P:u:g:i:o:",
				  longopts, 0);

		if (opt == EOF) {
			break;
		}

		switch (opt) {
			case 0:
				break;
			case 'd':
				daemon_mode = 1;
				break;
			case 'C':
				G2PCERA_MASTER.dryrun(true);
				break;
			case 'A':
				alone = true;
				break;
			case 'f':
				config_file = optarg;
				break;
			case 'i':
				pid_file = optarg;
				break;
			case 'o':
				iors_dir = optarg;
				break;
			case 'P':
				vty_port = atoi(optarg);
				break;
			case 'u':
				g2pcerad_privs.user = optarg;
				break;
			case 'g':
				g2pcerad_privs.group = optarg;
				break;
			case 'v':
				printf("G2PCERA daemon v.%s: vty@%d\n",
				       G2PCERA_CODE_VERSION, vty_port);
				printf("This program is part of "
				       "phosphorus-g2mpls\n");
				exit (0);
				break;
			case 'h':
				usage(progname, 0);
				break;
			default:
				usage(progname, 1);
				break;
		}
	}

	// Library inits.
	zprivs_init(&g2pcerad_privs);
	signal_init(G2PCERA_MASTER.getMaster(),
		    Q_SIGC(g2pcera_signals),
		    g2pcera_signals);
	cmd_init(1);

	// these are all from libzebra
	vty_init(G2PCERA_MASTER.getMaster());
	memory_init();

	G2PCERA_MASTER.init_vty();

	// sorting is needed by the command subsystem to finalize init
	sort_node();

	// Change to the daemon program.
	if (daemon_mode) {
		daemon(0, 0);
	}

	if (!G2PCERA_MASTER.dryrun()) {
#if HAVE_OMNIORB
		int res;
		try {
			res = corba_init(iors_dir,
					 CORBA_SERVANT_G2MPLS_TOPOLOGY,
					 G2PCERA_MASTER.getMaster());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem "
				"for G2MPLS_TOPOLOGY... reason %s\n",
				e.c_str());
			exit(-1);
		} catch (...) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem "
				"for G2MPLS_TOPOLOGY... unknown expection\n");
			exit(-1);
		}

		if (!res) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem "
				"for G2MPLS_TOPOLOGY\n");
			exit(-1);
		}

		try {
			corba_server_setup_g2topo();
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot setup CORBA subsystem "
				"for G2MPLS_TOPOLOGY (err=%s)\n",
				e.c_str());
			exit(-1);
		} catch (CORBA::BAD_PARAM & e) {
			fprintf(stderr,
				"Cannot setup server object "
				"(CORBA::BAD_PARAM)\n");
			exit(-1);
		} catch (...) {
			fprintf(stderr,
				"Cannot setup server object "
				"(exception other)\n");
			exit(-1);
		}

		fprintf(stdout,
			"--> CORBA_SERVANT_G2MPLS_TOPOLOGY set up\n\n\n");

		try {
			res = corba_init(iors_dir,
					 CORBA_SERVANT_G2PCERA,
					 G2PCERA_MASTER.getMaster());
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem "
				"for G2PCERA... reason %s\n",
				e.c_str());
			exit(-1);
		} catch (...) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem "
				"for G2PCERA... unknown expection \n");
			exit(-1);
		}

		if (!res) {
			fprintf(stderr,
				"Cannot initialize CORBA subsystem "
				"for G2PCERA\n");
			exit(-1);
		}

		try {
			corba_server_setup_g2pcera();
		} catch (std::string & e) {
			fprintf(stderr,
				"Cannot setup CORBA subsystem "
				"for G2PCERA (err=%s)\n",
				e.c_str());
			exit(-1);
		} catch (CORBA::BAD_PARAM & e) {
			fprintf(stderr,
				"Cannot setup server object "
				"(CORBA::BAD_PARAM)\n");
			exit(-1);
		} catch (...) {
			fprintf(stderr,
				"Cannot setup server object "
				"(exception other)\n");
			exit(-1);
		}

		fprintf(stdout,
			"--> CORBA_SERVANT_G2PCERA set up\n\n\n");

		if (!alone) {
			try {
				corba_init(iors_dir,
					   CORBA_SERVANT_LRM,
					   G2PCERA_MASTER.getMaster());
			} catch (...) {
				fprintf(stderr,
					"Cannot initialize CORBA subsystem "
					"for LRM\n");
				exit(-1);
			}

			try {
				corba_client_setup(G2PCERA_MASTER.getLrmProxy());
			} catch (std::string & e) {
				fprintf(stderr,
					"Cannot setup client object "
					"(err=%s)\n",
					e.c_str());
				exit(-1);
			} catch (CORBA::BAD_PARAM & e) {
				fprintf(stderr,
					"Cannot setup client object "
					"(CORBA::BAD_PARAM)\n");
				exit(-1);
			} catch (...) {
				fprintf(stderr,
					"Cannot setup client object "
					"(exception other)\n");
				exit(-1);
			}

			fprintf(stdout,
				"--> CORBA_CLIENT_LRM set up\n\n\n");

			try {
				LRM::Info_var lrmProxy;

				lrmProxy = G2PCERA_MASTER.getLrmProxy();

				if (CORBA::is_nil(lrmProxy)) {
					fprintf(stderr,
						"Cannot get Corba LRM client "
						"object "
						"previously set up\n");
					assert(0);
				}

				g2mplsTypes::nodeId myNid;

				myNid = lrmProxy->nodeId();

				G2PCERA_MASTER.routerId(myNid);
			} catch (...) {
				fprintf(stderr,
					"Cannot get  Router ID from LRM..."
					"running with default one "
					"(0x%08x/%s)\n\n",
					G2PCERA_MASTER.routerId(),
					numb_htoa(G2PCERA_MASTER.routerId()));
				//exit(-1);
			}
		}
#endif
	}

	vty_read_config(config_file, config_default);	// Get .conf file

	// Create VTY socket and start the TCP and unix socket listeners
	vty_serv_sock(vty_addr, vty_port, G2PCERA_VTYSH_PATH);

	// Process id file create.
	pid_output(pid_file);

	// Print banner.
	G2PCERA_NOTICE("G2PCERA daemon v.%s: starting... vty@%d",
		       G2PCERA_CODE_VERSION, vty_port);

	while (!terminate_execution) {
		if (thread_fetch(G2PCERA_MASTER.getMaster(), &thread)) {
			thread_call(&thread);
		}
	}

	G2PCERA_DEBUG("Termination in progress ...");

	if (!G2PCERA_MASTER.dryrun()) {
#if HAVE_OMNIORB
		try {
			(void) corba_server_shutdown(CORBA_SERVANT_G2MPLS_TOPOLOGY);
			(void) corba_server_shutdown(CORBA_SERVANT_G2PCERA);
			(void) corba_client_shutdown(CORBA_SERVANT_LRM);

			// shutdown all corba clients
			(void) corba_fini();
		} catch(...) {
			fprintf(stderr,
				"Somethig was wrong with corba shutdown\n");
			exit(-1);
		}
#endif
	}

	G2PCERA_MASTER.destroy();

	return 0;
}
