//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



//
//

#include "g2rsvpte_common_types.h"
#include "g2rsvpte_utils.h"
#include <string>
#include <sstream>
#include <iomanip>

//
//
//    Utilities
//
//
bool
operator==(lsp_ident_t src, lsp_ident_t dst)
{
	METIN();
	METOUT();
	return ((src.tun_id  == dst.tun_id)  &&
		(src.dst_nid == dst.dst_nid) &&
		(src.src_nid == dst.src_nid) &&
		(src.lsp_id ==  dst.lsp_id));
}

bool
operator==(lsp_ero_sobj_t src, lsp_ero_sobj_t dst)
{
	METIN();

	if ((src.loose != dst.loose) ||
	    (src.type  != dst.type )) {
		METOUT();
		return false;
	}

	switch (src.type) {
		case ERO_SUBOBJ_IPV4:
			if ((src.hop.ipv4.addr.s_addr !=
			     dst.hop.ipv4.addr.s_addr) ||
			    (src.hop.ipv4.prefix !=
			     dst.hop.ipv4.prefix)) {
				METOUT();
				return false;
			}
			break;
		case ERO_SUBOBJ_IPV6:
			if (src.hop.ipv6.prefix != dst.hop.ipv6.prefix) {
				METOUT();
				return false;
			}

			for (int i = 0; i < 4; i++) {
				if (src.hop.ipv6.addr.s6_addr32[i] !=
				    dst.hop.ipv6.addr.s6_addr32[i]) {
					METOUT();
					return false;
				}
			}
			break;
		case ERO_SUBOBJ_LABEL:
			if ((src.hop.label.upstream !=
			     dst.hop.label.upstream) ||
			    (src.hop.label.value   !=
			     dst.hop.label.value) ) {
				METOUT();
				return false;
			}
			break;
		case ERO_SUBOBJ_UNNUM:
			if ((src.hop.unnum.router_id !=
			     dst.hop.unnum.router_id) ||
			    (src.hop.unnum.intf_id   !=
			     dst.hop.unnum.intf_id) ) {
				METOUT();
				return false;
			}
			break;
		case ERO_SUBOBJ_AS:
			if ((src.hop.as.number != dst.hop.as.number)) {
				METOUT();
				return false;
			}
			break;
		case ERO_SUBOBJ_UNKNOWN:
		default:
			METOUT();
			return false;
	}

	return true;
}

std::string
hop2string(const lsp_ero_sobj_t & sobj)
{
	std::string  tmp;

	tmp = std::string("");
	tmp += sobj.loose ? "Loose  | " : "Strict | ";

	switch (sobj.type) {
		case ERO_SUBOBJ_IPV4: {
			std::ostringstream saddr;
			std::ostringstream spref;

			saddr << inet_ntoa(sobj.hop.ipv4.addr);
			spref << (uint32_t) sobj.hop.ipv4.prefix;

			tmp += "IPv4  | ";
			tmp += saddr.str();
			tmp += std::string("/");
			tmp += spref.str();
		}
			break;
		case ERO_SUBOBJ_IPV6: {
			std::ostringstream saddr;
			std::ostringstream spref;

			for (int i = 0; i < 16; i++) {
				saddr << std::hex
				      << sobj.hop.ipv6.addr.s6_addr[i];
				if (i / 4) saddr << ":";
			}
			spref << (uint32_t) sobj.hop.ipv6.prefix;

			tmp += "IPv6  | ";
			tmp += saddr.str();
			tmp += std::string("/");
			tmp += spref.str();
		}
			break;
		case ERO_SUBOBJ_LABEL: {
			std::ostringstream label;

			label << "0x"
			      << std::hex
			      << std::setw(8)
			      << std::setfill('0')
			      << sobj.hop.label.value;

			tmp += "LABEL | ";
			tmp += label.str();
			tmp +=  ((sobj.hop.label.upstream) ?
				 std::string("/Upstream"):
				 std::string("/Downstream"));

		}
			break;
		case ERO_SUBOBJ_UNNUM: {
			std::ostringstream srid;
			std::ostringstream sifid;
			ipv4_t             rid;

			rid.s_addr = htonl(sobj.hop.unnum.router_id);

			srid  << inet_ntoa(rid);
			sifid << "0x"
			      << std::hex
			      << std::setw(4)
			      << std::setfill('0')
			      << sobj.hop.unnum.intf_id;

			tmp += "UNNUM | ";
			tmp += srid.str();
			tmp += std::string("/");
			tmp += sifid.str();
		}
			break;
		case ERO_SUBOBJ_AS: {
			std::ostringstream asnum;

			asnum << "0x"
			      << std::hex
			      << std::setw(4)
			      << std::setfill('0')
			      << sobj.hop.as.number;

			tmp += "AS    | ";
			tmp += asnum.str();
		}
			break;
		case ERO_SUBOBJ_UNKNOWN:
		default: {
			tmp += "==UNKNOWN==";
		}
			break;

	}

	return tmp;
}


RSVP::Label_Class * &
operator << (RSVP::Label_Class * & dst, const label_id_t & src)
{
	switch (src.type) {
		case LABEL_32BIT: {
			uint32_t tmp = src.value.label32.id;
			dst = new RSVP::GLabel_Object(&tmp, sizeof(tmp));
			break;
		}
		case LABEL_60BIT: {
			uint64_t l60v;
			int i;

			l60v = 0x0ull;
			for (i = 0; i < 6; i++) {
				int disp = ((5 - i) * 8 + 12);
				l60v |= src.value.label60.mac[i] << disp;
			}
			l60v |= src.value.label60.vlan_id;

			dst = new RSVP::GLabel_Object((uint32_t *)&l60v,
						      sizeof(l60v));
			break;
		}
		default:
			throw(std::string("G2.LABEL type out-of-range"));
			break;
	}

	return dst;
}

label_id_t &
operator << (label_id_t & dst, const RSVP::Label_Class * src)
{
	assert(src);


	switch (src->cType()) {
		case CTYPE_LABEL: {
			RSVP::LabelMpls_Object * tmp;

			// pigsty workaround not to justify C++ knowledge
			tmp = static_cast<const RSVP::LabelMpls_Object *>
				(src->clone());
			assert(tmp);

			dst.type             = LABEL_32BIT;
			dst.value.label32.id = tmp->get().label;
			break;
		}
		case CTYPE_GENERALISED_LABEL: {
			RSVP::GLabel_Object * tmp;

			// pigsty workaround not to justify C++ knowledge
			tmp = static_cast<const RSVP::GLabel_Object *>
				(src->clone());
			assert(tmp);

			uint32_t size = tmp->get().size;
			if (size != 4 && size != 8) {
				throw(std::string("only 32/64 bit labels are "
						  "supported"));
			}
			uint32_t * label;
			label = tmp->get().label;
			if (size == 4) {
				dst.type             = LABEL_32BIT;
				dst.value.label32.id = *label;
			} else {
				dst.type         = LABEL_60BIT;
				dst.value.raw_id = *((uint64_t *)label);

				//uint64_t l60v, l60m;
				//int      i;
				//
				//l60v = label60 & 0x0FFFFFFFFFFFFFFFull;
				//
				//dst.type = LABEL_60BIT;
				//for (i = 0; i < 6; i++) {
				//	int disp = ((5 - i) * 8 + 12);
				//	l60m = (0xFFull) << disp;
				//	dst.value.label60.mac[i] =
				//	      (uint8_t) ((l60v & l60m) >> disp);
				//}
				//dst.value.label60.vlan_id =
				//	(l60v & 0x0000000000000FFFull);
			}
			break;
		}
		default:
			throw(std::string("G2.LABEL type not supported"));
			break;
	}

	return dst;
}
