//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//



#include <zebra.h>

#include "version.h"
#include <getopt.h>
#include "command.h"
#include "thread.h"
#include "sigevent.h"
#include "privs.h"
#include "memory.h"

#include "g2rsvpte_config.h"
#include "g2rsvpted.h"
#include "g2rsvpte_trace.h"
#include "g2rsvpte_dm.h"
#include "g2rsvpte_utils.h"
#include "g2rsvpte_scngw_api.h"
#include "lib/corba.h"
#include "g2rsvpte_corba.h"

//#include "g2rsvpte_messages.h"
#include "g2mpls_corba_utils.h"

using namespace RSVP;

struct thread_master * master = NULL;

// rsvpd privileges
zebra_capabilities_t _caps_p [] = {
	ZCAP_NET_RAW,
	ZCAP_BIND,
	ZCAP_NET_ADMIN,
};

// GLOBALS
struct zebra_privs_t   g2rsvpted_privs;
struct quagga_signal_t g2rsvpte_signals[4];
char *                 progname;

// Master of threads

const char *           pid_file         = PATH_G2RSVPTED_PID;
char                   config_default[] = G2RSVPTED_DEFAULT_CONFIG;


struct option longopts[] = {
	{ "daemon",        no_argument,       NULL, 'd'},
	{ "config_file",   required_argument, NULL, 'f'},
	{ "pid_file",      required_argument, NULL, 'i'},
	{ "iors_dir",      required_argument, NULL, 'o'},
	{ "help",          no_argument,       NULL, 'h'},
	{ "instance_type", required_argument, NULL, 't'},
	{ "vty_port",      required_argument, NULL, 'P'},
	{ "user",          required_argument, NULL, 'u'},
	{ "group",         required_argument, NULL, 'g'},
	{ "version",       no_argument,       NULL, 'v'},
	{ 0 }
};


// SIG handlers
static void sighup (void)
{
	G2RSVPTE_INF("SIGHUP received");
}

static int terminate_execution = 0;

static void sigint (void)
{
	G2RSVPTE_NOT("Terminating on SIGINT");

	// rsvp_terminate();
	terminate_execution = 1;
}

// SIGUSR1 handler.
static void sigusr1 (void)
{
	zlog_rotate (NULL);
}


// Help information display.
static void __attribute__ ((noreturn))
usage (char *progname, int status)
{
	if (status != 0) {
		fprintf(stderr,
			"Try `%s --help' for more information.\n",
			progname);
	} else {
		printf("Usage : %s [OPTION...]\n"
		       "Daemon which manages RSVP.\n\n"
		       "-d, --daemon        Runs in daemon mode\n"
		       "-f, --config_file   Set configuration file name\n"
		       "-i, --pid_file      Set process identifier file name\n"
		       "-o, --iors_dir      Set IORs directory\n"
		       "-t, --instance_type Instance type to run(inni-c|inni-b|"
		       "uni-c|uni-n|enni)\n"
		       "-P, --vty_port      Set vty's port number\n"
		       "-u, --user          User to run as\n"
		       "-g, --group         Group to run as\n"
		       "-v, --version       Print program version\n"
		       "-h, --help          Display this help and exit\n"
		       "\n"
		       "Report bugs to %s\n",
		       progname, ZEBRA_BUG_ADDRESS);
	}

	exit(status);
}

int main(int argc, char ** argv)
{
	char          * p;
	int             vty_port;
	char          * vty_addr;
	int             daemon_mode;
	char          * config_file;
	char          * iors_dir;
	char          * instance_type;
	struct thread   thread;

	vty_port      = G2RSVPTED_VTY_PORT;
	vty_addr      = NULL;
	daemon_mode   = 0;
	config_file   = NULL;
	iors_dir      = NULL;
	instance_type = NULL;

	//RSVP_Master   * rm;

	// INITIALIZATION
#if defined(QUAGGA_USER) && defined(QUAGGA_GROUP)
	g2rsvpted_privs.user      = QUAGGA_USER;
	g2rsvpted_privs.group     = QUAGGA_GROUP;
#endif
#if defined(VTY_GROUP)
	g2rsvpted_privs.vty_group = VTY_GROUP;
#endif
	g2rsvpted_privs.caps_p    = _caps_p;
	g2rsvpted_privs.cap_num_p = sizeof(_caps_p)/sizeof(_caps_p[0]);
	g2rsvpted_privs.cap_num_i = 0;

	g2rsvpte_signals[0].signal = SIGHUP;
	g2rsvpte_signals[0].handler = &sighup;
	g2rsvpte_signals[1].signal = SIGUSR1;
	g2rsvpte_signals[1].handler = &sigusr1;
	g2rsvpte_signals[2].signal = SIGINT;
	g2rsvpte_signals[2].handler = &sigint;
	g2rsvpte_signals[3].signal = SIGTERM;
	g2rsvpte_signals[3].handler = &sigint;

	// Set umask before anything for security
	umask(0027);

	// Get program name
	progname = ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);

	// Initialize the log subsystem
	zlog_default = openzlog(progname,
				ZLOG_G2RSVPTE,
				LOG_CONS | LOG_NDELAY | LOG_PID,
				LOG_DAEMON);

	//rm = new RSVP_Master;
	if (!RSVP_Master::init()) {
		fprintf(stderr, "Cannot allocate an RSVP master object\n");
		exit(-1);
	}
	master = RSVP_MASTER.getMaster();

	// just reads the options from stdin
	while (1) {
		int opt;

		opt = getopt_long(argc, argv, "df:h:t:P:u:g:v:i:o:", longopts, 0);

		if (opt == EOF) {
			break;
		}

		switch (opt) {
			case 0:
				break;
			case 'd':
				daemon_mode = 1;
				break;
			case 'f':
				config_file = optarg;
				break;
			case 'i':
				pid_file = optarg;
				break;
			case 'o':
				iors_dir = optarg;
				break;
			case 't':
				instance_type = optarg;
				break;
			case 'P':
				vty_port = atoi(optarg);
				break;
			case 'u':
				g2rsvpted_privs.user = optarg;
				break;
			case 'g':
				g2rsvpted_privs.group = optarg;
				break;
			case 'v':
				printf("G2.RSVP-TE daemon v.%s: vty@%d\n",
				       G2RSVPTE_CODE_VERSION, vty_port);
				printf("This program is part of "
				       "phosphorus-g2mpls\n");
				//print_version(progname);
				exit(0);
				break;
			case 'h':
				usage(progname, 0);
				break;
			default:
				usage(progname, 1);
				break;
		}
	}

	// Library inits.
	zprivs_init(&g2rsvpted_privs);
	signal_init(RSVP_MASTER.getMaster(),
		    Q_SIGC(g2rsvpte_signals),
		    g2rsvpte_signals);
	cmd_init(1);

	// these are all from libzebra
	vty_init(RSVP_MASTER.getMaster());
	memory_init();

	RSVP_MASTER.init_vty();

	// sorting is needed by the command subsystem to finalize init
	sort_node();

	//TMP Try to move read_cofig file, because the INSTANCE need CORBA
	//vty_read_config(config_file, config_default);	// Get .conf file

	// Change to the daemon program.
	if (daemon_mode) {
		daemon(0, 0);
	}

	if (instance_type == NULL) {
		fprintf(stderr, "Required instance_type parameter not specified. Exit.\n");
		exit(-1);
	}

	if (!strcmp(instance_type, "inni-c") ||
	    !strcmp(instance_type, "INNI-c") ||
	    !strcmp(instance_type, "Inni-c") ||
	    !strcmp(instance_type, "inni-b") ||
	    !strcmp(instance_type, "INNI-b") ||
	    !strcmp(instance_type, "Inni-b")) {
#if HAVE_OMNIORB
			// CORBA Servants setup
			if (!corba_init(iors_dir,
					CORBA_SERVANT_TNRCONTROLLER_G2RSVPTE,
					RSVP_MASTER.getMaster())) {
				fprintf(stderr, "Cannot initialize TNRC CORBA subsystem\n");
				exit(-1);
			}
			if (!tnrc_server_setup()) {
				fprintf(stderr, "Cannot setup G2RSVPTE TNRCONTROLLER "
					"CORBA subsystem\n");
				exit(-1);
			}
			printf("G2RSVPTE TNRCONTROLLER CORBA server setup ... OK\n");

			if (!corba_init(iors_dir,
					CORBA_SERVANT_NORTHBOUND_G2RSVPTE,
					RSVP_MASTER.getMaster())) {
				fprintf(stderr, "Cannot initialize NB CORBA subsystem\n");
				exit(-1);
			}
			if (!nb_server_setup()) {
				fprintf(stderr, "Cannot setup G2RSVPTE NORTHBOUND "
					"CORBA subsystem\n");
				exit(-1);
			}
			printf("G2RSVPTE NORTHBOUND CORBA server setup ... OK\n");

#ifndef G2RSVPTE2RC_DEBUG
			if (!strcmp(instance_type, "inni-b") ||
			    !strcmp(instance_type, "INNI-b") ||
			    !strcmp(instance_type, "Inni-b")) {
				if (!corba_client_setup(rc_proxy)) {
					fprintf(stderr, "Cannot setup "
						"RC-SBI client object\n");
					exit(-1);
				}
				printf("RC-SBI client object setup "
				       "... OK\n");
			}
#endif // G2RSVPTE2RC_DEBUG
#ifdef PERSISTENCY_ON
			if (!corba_client_setup(pc_innirsvplsp_proxy)) {
				fprintf(stderr, "Cannot setup PersistencyController InniRsvpLsp client object\n");
				exit(-1);
			}
			printf("PersistencyController InniRsvpLsp client object setup ... OK\n");
#endif // PERSISTENCY_ON
#ifdef CRANKBACK_ON
			if (!corba_client_setup(g2pcera_proxy)) {
				fprintf(stderr, "Cannot setup G2PCERA RoutingServices client object\n");
				exit(-1);
			}
			printf("G2PCERA RoutingServices client object setup ... OK\n");
#endif // CRANKBACK_ON
#endif
	} else if (!strcmp(instance_type, "uni-c") ||
		   !strcmp(instance_type, "UNI-c") ||
		   !strcmp(instance_type, "Uni-c") ||
		   !strcmp(instance_type, "uni-n") ||
		   !strcmp(instance_type, "UNI-n") ||
		   !strcmp(instance_type, "Uni-n")) {
#if HAVE_OMNIORB
			//CORBA Servants setup
			if (!corba_init(iors_dir,
					CORBA_SERVANT_LSPSIGNORTH_UNIRSVP,
					RSVP_MASTER.getMaster())) {
				fprintf(stderr, "Cannot initialize LSPSIGNORTH CORBA subsystem\n");
				exit(-1);
			}
			if (!unirsvp_lspsignorth_server_setup()) {
				fprintf(stderr, "Cannot setup UNI G2RSVPTE LSPSIGNORTH "
					"CORBA subsystem\n");
				exit(-1);
			}
			printf("UNI G2RSVPTE LSPSIGNORTH CORBA server setup ... OK\n");

			//CORBA Clients setup
#ifndef G2RSVPTE2CC_DEBUG
			bool isUniClient = (!strcmp(instance_type, "uni-c") ||
					    !strcmp(instance_type, "UNI-c") ||
					    !strcmp(instance_type, "Uni-c") ? true : false);
			if (!corba_client_setup(cc_ls_proxy, isUniClient)) {
				fprintf(stderr, "Cannot setup CALLCONTROLLER "
					"LSPSIGSOUTH client object\n");
				exit(-1);
			}
			printf("CALLCONTROLLER LSPSIGSOUTH client object setup "
			       "... OK\n");
			if (!corba_client_setup(cc_mgmt_proxy, isUniClient)) {
				fprintf(stderr, "Cannot setup CALLCONTROLLER "
					"MGMT client object\n");
				exit(-1);
			}
			printf("CALLCONTROLLER MGMT client object setup "
			       "... OK\n");
#endif // G2RSVPTE2CC_DEBUG
#ifdef PERSISTENCY_ON
			if (!corba_client_setup(pc_unirsvplsp_proxy)) {
				fprintf(stderr, "Cannot setup PersistencyController UniRsvpLsp client object\n");
				exit(-1);
			}
			printf("PersistencyController UniRsvpLsp client object setup ... OK\n");
#endif // PERSISTENCY_ON
#endif
	} else if (!strcmp(instance_type, "enni") ||
		   !strcmp(instance_type, "ENNI") ||
		   !strcmp(instance_type, "Enni")) {
#if HAVE_OMNIORB
			//CORBA Servants setup
			if (!corba_init(iors_dir,
					CORBA_SERVANT_LSPSIGNORTH_ENNIRSVP,
					RSVP_MASTER.getMaster())) {
				fprintf(stderr, "Cannot initialize LSPSIGNORTH CORBA subsystem\n");
				exit(-1);
			}
			if (!ennirsvp_lspsignorth_server_setup()) {
				fprintf(stderr, "Cannot setup ENNI G2RSVPTE LSPSIGNORTH "
					"CORBA subsystem\n");
				exit(-1);
			}
			printf("ENNI G2RSVPTE LSPSIGNORTH CORBA server setup ... OK\n");

			//CORBA Clients setup
#ifndef G2RSVPTE2CC_DEBUG
			if (!corba_client_setup(cc_ls_proxy)) {
				fprintf(stderr, "Cannot setup CALLCONTROLLER "
					"LSPSIGSOUTH client object\n");
				exit(-1);
			}
			printf("CALLCONTROLLER LSPSIGSOUTH client object setup "
			       "... OK\n");

			if (!corba_client_setup(cc_mgmt_proxy, false)) {
				fprintf(stderr, "Cannot setup CALLCONTROLLER "
					"MGMT client object\n");
				exit(-1);
			}
			printf("CALLCONTROLLER MGMT client object setup "
			       "... OK\n");
#endif // G2RSVPTE2CC_DEBUG
#ifdef PERSISTENCY_ON
			if (!corba_client_setup(pc_ennirsvplsp_proxy)) {
				fprintf(stderr, "Cannot setup PersistencyController EnniRsvpLsp client object\n");
				exit(-1);
			}
			printf("PersistencyController EnniRsvpLsp client object setup ... OK\n");
#endif // PERSISTENCY_ON
#endif
	} else {
		fprintf(stderr, "Invalid instance_type parmater (%s)\n", instance_type);
		exit(-1);
	}

	//CORBA setup
#if HAVE_OMNIORB
	if (!corba_client_setup(lrm_proxy)) {
		fprintf(stderr, "Cannot setup LRM client object\n");
		exit(-1);
	}
	printf("LRM client object setup ... OK\n");

	if (!corba_client_setup(tnrc_proxy)) {
		fprintf(stderr, "Cannot setup TNRC client object\n");
		exit(-1);
	}
	printf("TNRC client object setup ... OK\n");

	if (!strcmp(instance_type, "inni-c") ||
	    !strcmp(instance_type, "INNI-c") ||
	    !strcmp(instance_type, "Inni-c") ||
	    !strcmp(instance_type, "inni-b") ||
	    !strcmp(instance_type, "INNI-b") ||
	    !strcmp(instance_type, "Inni-b")) {
			try {
				tnrc_proxy->g2rsvpteIsUp();
			} catch (TNRC::InternalProblems) {
				fprintf(stderr, "TNRC internal problems on g2rsvpteIsUp\n");
				exit(-1);
			} catch (...) {
				fprintf(stderr, "Unexpected error on TNRC::g2rsvpteIsUp\n");
				exit(-1);
			}
	}
#ifdef PERSISTENCY_ON
	if (!corba_client_setup(pc_mgmt_proxy)) {
		fprintf(stderr, "Cannot setup PersistencyController Mgmt client object\n");
		exit(-1);
	}
	printf("PersistencyController Mgmt client object setup ... OK\n");
#endif // PERSISTENCY_ON
#endif

	vty_read_config(config_file, config_default); // Get .conf file

	// Create VTY socket and start the TCP and unix socket listeners
	vty_serv_sock(vty_addr, vty_port, G2RSVPTE_VTYSH_PATH);

	// Process id file create.
	pid_output(pid_file);

	// Print banner.
	G2RSVPTE_NOT("G2.RSVP-TE daemon v.%s: starting... vty@%d",
		     G2RSVPTE_CODE_VERSION, vty_port);

	while (!terminate_execution) {
		if (thread_fetch(RSVP_MASTER.getMaster(), &thread)) {
			thread_call(&thread);
		}
	}

	G2RSVPTE_DBG("Termination in progress ...");

#if HAVE_OMNIORB
	(void) corba_server_shutdown();
	(void) corba_fini();
#endif
	return 0;
}
