/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#ifndef G2RSVPTE_MESSAGES_H
#define G2RSVPTE_MESSAGES_H

#include <list>
#include <map>

#include "g2rsvpte_config.h"
#include "g2rsvpte_utils.h"
#include "g2rsvpte_message_objects.h"
#include "g2rsvpte_trace.h"
#include "g2rsvpte_common_types.h"
#include "g2rsvpte_buffer.h"
#include "g2rsvpte_message_objects_utils.h"

// Print Utils
#define MSGTYPE2STR(T)						\
  (((T) == RSVP::Message::PATH          ) ? "PATH"          :	\
  (((T) == RSVP::Message::RESV          ) ? "RESV"          :	\
  (((T) == RSVP::Message::PATH_ERR      ) ? "PATH_ERR"      :	\
  (((T) == RSVP::Message::RESV_ERR      ) ? "RESV_ERR"      :	\
  (((T) == RSVP::Message::PATH_TEAR     ) ? "PATH_TEAR"     :	\
  (((T) == RSVP::Message::RESV_TEAR     ) ? "RESV_TEAR"     :	\
  (((T) == RSVP::Message::RESV_CONF     ) ? "RESV_CONF"     :	\
  (((T) == RSVP::Message::ACK           ) ? "ACK"           :	\
  (((T) == RSVP::Message::SREFRESH      ) ? "SREFRESH"      :	\
  (((T) == RSVP::Message::HELLO         ) ? "HELLO"         :	\
  (((T) == RSVP::Message::NOTIFY        ) ? "NOTIFY"        :	\
  (((T) == RSVP::Message::RECOVERY_PATH ) ? "RECOVERY_PATH" :	\
					    "UNKNOWN"))))))))))))
//
// References
//     base:   RFC2205, RFC2210, RFC2215, RFC2747, RFC2750, RFC2961
//     mpls:   RFC3209,
//     gmpls:  RFC3473, RFC3477
//     ason:   RFC3474,
//     oifuni: RFC3476
//
//

namespace RSVP {

//
// RSVP Common Header
//
//           0             1              2             3
//    +-------------+-------------+-------------+-------------+
//    | Vers | Flags|  Msg Type   |       RSVP Checksum       |
//    +-------------+-------------+-------------+-------------+
//    |  Send_TTL   | (Reserved)  |        RSVP Length        |
//    +-------------+-------------+-------------+-------------+
//

	class Message {
		friend MemoryCursor & operator >> (MemoryCursor &  mc,
						   Message * &     msg);

		friend MemoryCursor & operator << (MemoryCursor &  mc,
						   const Message & msg);

		friend std::ostream & operator << (std::ostream &  os,
						   const Message & msg);
	public:
		typedef enum {
			PATH          = 1,
			RESV          = 2,
			PATH_ERR      = 3,
			RESV_ERR      = 4,
			PATH_TEAR     = 5,
			RESV_TEAR     = 6,
			RESV_CONF     = 7,
			ACK           = 13,
			SREFRESH      = 15,
			HELLO         = 20,
			NOTIFY        = 21,
			RECOVERY_PATH = 30
		} Type;

	        Message(uint8_t t) :
		        versFlags_(1<<4),
			ttl_(255),
			type_(t) { }

	        Message(uint8_t v, uint8_t t) :
		        versFlags_(v<<4),
			ttl_(255),
			type_(t) { }

		virtual ~Message(void) { }

		uint8_t  getType(void) const     { return type_;             }
		uint8_t  getTTL(void) const      { return ttl_;              }
		uint8_t  getVersion(void) const  { return versFlags_ >> 4;   }
		uint8_t  getFlags(void) const    { return versFlags_ & 0x0f; }

		void     setTTL(uint8_t ttl)     { ttl_ = ttl;               }
		void     setFlags(uint8_t flags) {
			versFlags_ = (versFlags_ & 0xf0) | (flags & 0x0f);
		}

		// Must be overloaded!!!
		virtual bool getLSPinfo(lsp_ident_t & info) { return false; }
		virtual bool      isConsistent(void) const = 0;
		virtual uint16_t  getLength(void) const = 0;
		virtual Message * clone(void) const = 0;

	protected:
		Message(void);

	public:
		static const uint16_t msgHdrLength = 8;

	protected:
		uint8_t  versFlags_;  // version and flags
		uint8_t  ttl_;
		uint8_t  type_;
	};

	class PathMessage : public Message {
		friend MemoryCursor & operator >> (MemoryCursor & mc,
						   PathMessage &  path);

		friend MemoryCursor & operator << (MemoryCursor &      mc,
						   const PathMessage & path);

		friend std::ostream & operator << (std::ostream &      os,
						   const PathMessage & path);
	private:
		void createPathMessage(const PathMessage & path, bool down , bool reflect);
	public:
		PathMessage(void);
		PathMessage(uint8_t version);
		PathMessage(uint8_t            version,
			    uint32_t           srcNid,
			    uint32_t           dstNid,
			    uint32_t           lspId,
			    uint32_t           tunId,
			    uint32_t           extTunId,
			    uint32_t           nodeId,
			    sw_cap_t           swcap,
			    enc_type_t         enctype,
			    uint32_t           bw,
			    gpid_t             gpid,
			    uint32_t           setupPrio,
			    uint32_t           holdingPrio,
#ifdef CRANKBACK_ON
			    crankback_scope_t  cbScope,
#endif // CRANKBACK_ON
			    lsp_rro_mode_t     rroMode,
			    uint32_t           refreshTime,
			    std::string        sessionName,
			    //const g2mpls_addr_t *  iTna,
			    //const g2mpls_addr_t *  eTna,
			    const net_res_spec_t * iRes,
			    const net_res_spec_t * eRes,
			    const call_ident_t *   callId,
			    const times_t          times);
		PathMessage(const PathMessage & path, bool down, bool reflect = true);
		PathMessage(const PathMessage & path);
		~PathMessage(void);

		PathMessage * partialClone(void) const;
		Message *     clone(void) const;

		bool setGnsCallExt(const call_info_t & callInfo);
		bool setGenUni(const net_res_spec & tna,
			       bool                 ingress,
			       call_type_t          type);
		bool updateGenUni(const net_res_spec & iTna,
				  bool                 source,
				  call_type_t          type);
		bool setGnsTna(const grid_res_spec_t & gnsTna,
			       bool                    ingress);
		bool getGnsTna(grid_res_spec_t & gnsTna,
			       bool              ingress);
		bool setPolicyData(const std::string aaaInfo);
		bool getEro(std::list<lsp_ero_sobj_t> & ero);

		bool      isDown(void) const;
		bool      isConsistent(void) const;
		bool      getLSPinfo(lsp_ident_t & info);
		uint16_t  getLength(void) const;

		//Optional objects
		Integrity_Class         *integrity_;
		MessageIdAckNack_Class  *message_id_ack_nack_;
		MessageId_Class         *message_id_;
		ExplicitRoute_Class     *ero_;
		Protection_Class        *protection_;
		LabelSet_Class          *label_set_;
		Session_Attr_Class      *session_attribute_;
#ifdef CRANKBACK_ON
		LSPAttr_Class           *lsp_attributes_;
#endif // CRANKBACK_ON
		NotifyRequest_Class     *notify_request_;
		AdminStatus_Class       *admin_status_;
		Policy_Class            *policy_data_;

		CallOptions_Class       *call_ops_;
		CallId_Class            *call_id_;
		GnsCallExt_Class        *gns_call_ext_;
		GeneralisedUni_Class    *gen_uni_;
		GnsUni_Class            *gns_uni_eg_;
		GnsUni_Class            *gns_uni_in_;
		Times_Class             *times_;

		//Mandatory objects
		Session_Class           *session_;
		Rsvp_Hop_Class          *rsvp_hop_;
		TimeValues_Class        *time_values_;
		LabelRequest_Class      *label_request_;
		// Sender descriptor (unidirectional LSP)
		SenderTemplate_Class    *sender_template_;
		SenderTSpec_Class       *sender_tspec_;
		ADSpec_Class            *adspec_;          // optional
		RecordedRoute_Class     *rro_;             // optional
		SuggestedLabel_Class    *suggested_label_; // optional
		RecoveryLabel_Class     *recovery_label_;  // optional
		// Sender descriptor (bidirectional LSP)
		UpstreamLabel_Class     *upstream_label_;

	private:
	};


	class ResvMessage : public Message {
		friend MemoryCursor & operator >> (MemoryCursor & mc,
						   ResvMessage &  resv);

		friend MemoryCursor & operator << (MemoryCursor &      mc,
						   const ResvMessage & resv);

		friend std::ostream & operator << (std::ostream &      os,
						   const ResvMessage & resv);
	private:
		void createResvMessage(const ResvMessage & resv, bool down);
	public:
		ResvMessage(void);
		ResvMessage(uint8_t version);
		ResvMessage(const PathMessage & path);
		ResvMessage(const ResvMessage & resv, bool down);
		ResvMessage(const ResvMessage & resv);
		~ResvMessage(void);

		ResvMessage * partialClone(void) const;
		Message *     clone(void) const;

		bool      isConsistent(void) const;
		bool      isDown(void) const;
		bool      getLSPinfo(lsp_ident_t & info);
		uint16_t  getLength(void) const;
		bool setGnsTna(const grid_res_spec_t & gnsTna,
			       bool                    ingress);
		bool getGnsTna(grid_res_spec_t & gnsTna,
			       bool              ingress);
		bool getRro(std::list<lsp_ero_sobj_t> & rro);

		//Optional objects
		Integrity_Class         *integrity_;
		MessageIdAckNack_Class  *message_id_ack_nack_;
		MessageId_Class         *message_id_;
		ResvConfirm_Class       *resv_confirm_;
		Scope_Class             *scope_;
		NotifyRequest_Class     *notify_request_;
		AdminStatus_Class       *admin_status_;
		Policy_Class            *policy_data_;

		//Mandatory objects
		Session_Class           *session_;
		Rsvp_Hop_Class          *rsvp_hop_;
		TimeValues_Class        *time_values_;
		CallOptions_Class       *call_ops_;
		CallId_Class            *call_id_;
		GnsCallExt_Class        *gns_call_ext_;
		GnsUni_Class            *gns_uni_eg_;
		GnsUni_Class            *gns_uni_in_;
		Style_Class             *style_;
		// flow description list
		Flow_Desc_List          *flow_desc_;

	private:
	};

	class PathErrMessage : public Message {
		friend MemoryCursor & operator >> (MemoryCursor &   mc,
						   PathErrMessage & pe);

		friend MemoryCursor & operator << (MemoryCursor &         mc,
						   const PathErrMessage & pe);

		friend std::ostream & operator << (std::ostream &         os,
						   const PathErrMessage & pe);
	private:
		PathErrMessage(const PathErrMessage & pathErr); //TODO???
	public:
		PathErrMessage(void);
		PathErrMessage(uint8_t version);
		PathErrMessage(const PathMessage & path);
		~PathErrMessage(void);

		bool getErrorDetails(error_info_t &              error,
				     std::list<lsp_ero_sobj_t> & excludeEro);
		bool getRro(std::list<lsp_ero_sobj_t> & rro);

		Message * clone(void) const;
		bool      isConsistent(void) const;
		uint16_t  getLength(void) const;
		bool      getLSPinfo(lsp_ident_t & info);

		//Mandatory objects
		Session_Class           *session_;
		Error_Spec_Class        *error_spec_;
		CallOptions_Class       *call_ops_;
		CallId_Class            *call_id_;
		GnsCallExt_Class        *gns_call_ext_;

		//Optional objects
		Integrity_Class         *integrity_;
		Policy_Class            *policy_data_;
		// Sender descriptor (unidirectional LSP)
		SenderTemplate_Class    *sender_template_;
		SenderTSpec_Class       *sender_tspec_;
		ADSpec_Class            *adspec_;          // optional
		RecordedRoute_Class     *rro_;             // optional
		SuggestedLabel_Class    *suggested_label_; // optional
		RecoveryLabel_Class     *recovery_label_;  // optional
		// Sender descriptor (bidirectional LSP)
		UpstreamLabel_Class     *upstream_label_;

	private:
	};

	class ResvErrMessage : public Message {
		friend MemoryCursor & operator >> (MemoryCursor &    mc,
						   ResvErrMessage &  rErr);

		friend MemoryCursor & operator << (MemoryCursor &         mc,
						   const ResvErrMessage & rErr);

		friend std::ostream & operator << (std::ostream &         os,
						   const ResvErrMessage & rErr);
	private:
		ResvErrMessage(const ResvErrMessage & resvErr); //TODO???
	public:
		ResvErrMessage(void);
		ResvErrMessage(uint8_t version);
		ResvErrMessage(const ResvMessage & resv);
		~ResvErrMessage(void);

		bool getErrorDetails(error_info_t & error);

		Message * clone(void) const;
		bool      isConsistent(void) const;
		uint16_t  getLength(void) const;
		bool      getLSPinfo(lsp_ident_t & info);

		//Mandatory objects
		Session_Class            *session_;
		Error_Spec_Class         *error_spec_;
		Style_Class              *style_;
		Rsvp_Hop_Class           *rsvp_hop_;
		// flow description list
		Flow_Desc_List           *flow_desc_;

		//Optional objects
		Integrity_Class          *integrity_;
		MessageIdAckNack_Class   *message_id_ack_nack_;
		MessageId_Class          *message_id_;
		Policy_Class             *policy_data_;
		Scope_Class              *scope_;
		AcceptableLabelSet_Class *acceptable_label_set_;
	private:
	};

	class PathTearMessage : public Message {
		friend MemoryCursor & operator >> (MemoryCursor &    mc,
						   PathTearMessage & pt);

		friend MemoryCursor & operator << (MemoryCursor &          mc,
						   const PathTearMessage & pt);

		friend std::ostream & operator << (std::ostream &          os,
						   const PathTearMessage & pt);
	private:
		PathTearMessage(const PathTearMessage & pathTear); //TODO???
	public:
		PathTearMessage(void);
		PathTearMessage(uint8_t version);
		PathTearMessage(const PathMessage & path);
		~PathTearMessage(void);

		Message * clone(void) const;
		bool      isConsistent(void) const;
		uint16_t  getLength(void) const;
		bool      getLSPinfo(lsp_ident_t & info);

		//Mandatory objects
		Session_Class           *session_;
		Rsvp_Hop_Class          *rsvp_hop_;
		CallOptions_Class       *call_ops_;
		CallId_Class            *call_id_;
		GnsCallExt_Class        *gns_call_ext_;

		//Optional objects
		Integrity_Class         *integrity_;
		// Sender descriptor (unidirectional LSP) optional
		SenderTemplate_Class    *sender_template_;
		SenderTSpec_Class       *sender_tspec_;
		ADSpec_Class            *adspec_;          // optional
		RecordedRoute_Class     *rro_;             // optional
		SuggestedLabel_Class    *suggested_label_; // optional
		RecoveryLabel_Class     *recovery_label_;  // optional
		// Sender descriptor (bidirectional LSP)
		UpstreamLabel_Class     *upstream_label_;
	};

	class ResvTearMessage : public Message {
		friend MemoryCursor & operator >> (MemoryCursor &     mc,
						   ResvTearMessage &  rt);

		friend MemoryCursor & operator << (MemoryCursor &          mc,
						   const ResvTearMessage & rt);

		friend std::ostream & operator << (std::ostream &          os,
						   const ResvTearMessage & rt);
	private:
		ResvTearMessage(const ResvTearMessage & resvTear); //TODO???
	public:
		ResvTearMessage(void);
		ResvTearMessage(uint8_t version);
		ResvTearMessage(const ResvMessage & resv);
		~ResvTearMessage(void);

		Message * clone(void) const;
		bool      isConsistent(void) const;
		uint16_t  getLength(void) const;
		bool      getLSPinfo(lsp_ident_t & info);

		//Mandatory objects
		Session_Class   *session_;
		Rsvp_Hop_Class  *rsvp_hop_;
		Style_Class     *style_;
		// flow description list
		Flow_Desc_List  *flow_desc_;

		//Optional objects
		Integrity_Class *integrity_;
		Scope_Class     *scope_;

	private:
	};

	class ResvConfMessage : public Message {
		friend MemoryCursor & operator >> (MemoryCursor &     mc,
						   ResvConfMessage &  rc);

		friend MemoryCursor & operator << (MemoryCursor &          mc,
						   const ResvConfMessage & rc);

		friend std::ostream & operator << (std::ostream &          os,
						   const ResvConfMessage & rc);
	private:
		ResvConfMessage(const ResvConfMessage & resvConf); //TODO???
	public:
		ResvConfMessage(void);
		ResvConfMessage(uint8_t version);
		ResvConfMessage(const PathMessage & path,
				const ResvMessage & resv);
		~ResvConfMessage(void);

		Message * clone(void) const;
		bool      isConsistent(void) const;
		uint16_t  getLength(void) const;
		bool getLSPinfo(lsp_ident_t & info);
		//Mandatory objects
		Session_Class     *session_;
		Error_Spec_Class  *error_spec_;
		ResvConfirm_Class *resv_confirm_;
		Style_Class       *style_;
		// flow description list
		Flow_Desc_List    *flow_desc_;

		//Optional objects
		Integrity_Class   *integrity_;
	};


	//
	// TO BE COMPLETED
	//
	class HelloMessage : public Message {
	public:
		HelloMessage(void) : Message(HELLO) {
			hello_       = 0;

			integrity_   = 0;
			restart_cap_ = 0;
		}

		HelloMessage(const HelloMessage & hello) : Message(HELLO) {
			// Copy constructor
		}

		~HelloMessage(void) { }

		Message * clone(void) const        { return 0; }
		bool      isConsistent(void) const { return true; }
		uint16_t  getLength(void) const    { return msgHdrLength; }

		//Mandatory objects
		Hello_Class       *hello_;

		//Optional objects
		Integrity_Class   *integrity_;
		RestartCap_Object *restart_cap_;
	private:
	};

	class AckMessage : public Message {
	public:
		AckMessage(void) : Message(ACK) {
			integrity_    = 0;
			msg_id_ack_nack_ = 0;
		}

		AckMessage(const AckMessage & ack):
			Message(ACK) {
			// Copy constructor
		}

		~AckMessage(void) { }

		bool isConsistent(void)	const { return true; }

		uint16_t getLength(void) const { return msgHdrLength; }

		//Mandatory
		MessageIdAckNack_Class    *msg_id_ack_nack_;
		//Optional
		Integrity_Class *integrity_;
		std::list<MessageIdAckNack_Class *> msg_n_acks_;
	};

	class SRefreshMessage : public Message {
	public:
		SRefreshMessage(void) : Message(SREFRESH) {
			integrity_  = 0;
			message_id_ = 0;
		}
		SRefreshMessage(const SRefreshMessage & ack) :
			Message(SREFRESH) {
			// Copy constructor
		}
		~SRefreshMessage(void) { }

		bool isConsistent(void)	const { return true; }

		uint16_t getLength(void) const { return msgHdrLength; }

		//Mandatory
		//<srefresh list> ::= <MESSAGE_ID LIST> | <MESSAGE_ID MCAST_LIST> [ <srefresh list> ]
		std::list<MessageIdList_Class *>    srefresh_list;
		//<source srefresh list> ::= <MESSAGE_ID SRC_LIST>
		std::list<MessageIdList_Class *>    source_srefresh_list;

		//optional
		Integrity_Class                    *integrity_;
		std::list<MessageIdAckNack_Class *> msg_n_acks_;
		MessageId_Class                    *message_id_;
	};

	class NotifyMessage : public Message {
	public:
		NotifyMessage(void) : Message(NOTIFY) {
			//Mandatory
			error_spec_             = 0;
			upstream_session_       = 0;
			upstream_call_id_       = 0;
			gns_call_ext_up_        = 0;
			sender_template_        = 0;
			sender_tspec_           = 0;
			adspec_                 = 0;
			rro_                    = 0;
			suggested_label_        = 0;
			recovery_label_         = 0;
			upstream_label_         = 0;

			//Optinal
			admin_status_           = 0;
			//downstream notify session
			downstream_session_     = 0;
			downstream_call_id_     = 0;
			gns_call_ext_down_      = 0;

			//optional
			downstream_policy_data_ = 0;
			//Optional
			integrity_              = 0;
			message_id_             = 0;
		}

		NotifyMessage(const NotifyMessage & notify) :
			Message(NOTIFY) {
			// Copy constructor
		}

		~NotifyMessage(void) { }

		bool isConsistent(void)	const { return true; }

		uint16_t getLength(void) const { return msgHdrLength; }

		//Mandatory
		Error_Spec_Class                 *error_spec_;
		//upstream notify session
		//Mandatory
		Session_Class                    *upstream_session_;
		CallId_Class                     *upstream_call_id_;
		GnsCallExt_Class                 *gns_call_ext_up_;
		//Sender description (unidirectional LSP)
		SenderTemplate_Class             *sender_template_;
		SenderTSpec_Class                *sender_tspec_;
		ADSpec_Class                     *adspec_;          // optional
		RecordedRoute_Class              *rro_;             // optional
		SuggestedLabel_Class             *suggested_label_; // optional
		RecoveryLabel_Class              *recovery_label_;  // optional
		// Sender descriptor (bidirectional LSP)
		UpstreamLabel_Class              *upstream_label_;
		// Optional
		AdminStatus_Class                *admin_status_;
		std::list<Policy_Class *>        upstream_policy_data_;
		//downstream notify session
		Session_Class                    *downstream_session_;
		CallId_Class                     *downstream_call_id_;
		GnsCallExt_Class                 *gns_call_ext_down_;
		//flow descriptor list
		Flow_Desc_List                   *flow_desc_;

		//Optional
		std::list<Policy_Class *>           *downstream_policy_data_;
		Integrity_Class                     *integrity_;
		std::list<MessageIdAckNack_Class *> msg_n_acks_;
		MessageId_Class                     *message_id_;
	};
}

#endif // G2RSVPTE_MESSAGES_H
