//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
//  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
//  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
//  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
//  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
//




#include <zebra.h>
#include "version.h"
#include <getopt.h>
#include "command.h"
#include "thread.h"
#include "sigevent.h"
#include "memory.h"
#include "thread.h"
#include "stream.h"
#include "sockopt.h"
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include "privs.h"
#include <sstream>

#include "g2rsvpte_config.h"
#include "g2rsvpte_dm.h"
#include "g2rsvpted.h"
#include "g2rsvpte_trace.h"
#include "g2rsvpte_utils.h"
#include "g2rsvpte_messages.h"
#include "g2rsvpte_scn_socket.h"
#include "g2rsvpte_scngw_api.h"
#include "g2rsvpte_buffer.h"
#include "scngwd/scngwc.h"

using namespace RSVP;

static char buffer[G2RSVPTE_MAX_PACKET_SIZE];

int init_scngw(adj_type_t adjType, bool tunneled)
{
	int res;

	switch (adjType) {
		case INNI:
			res = scngwc_init(adjType, IPPROTO_RSVP, tunneled ? TUNNEL : NO_TUNNEL, WANT_NO_ACK);
			break;
		case UNI:
		case ENNI:
			res = scngwc_init(adjType, IPPROTO_RSVP,  tunneled ? TUNNEL : NO_TUNNEL, WANT_NO_ACK);
			break;
		default:
			return -1;
			break;
	}

	return res;
}

void close_scngw(int fd)
{
	scngwc_close(fd);

}

int reconnect_scngw(struct thread *thread)
{
	int              fd;
	RSVP::G2RSVPTE * proto;

	proto = (RSVP::G2RSVPTE *) THREAD_ARG(thread);
	assert(proto);

	fd = init_scngw(proto->adjType(), proto->tunneledPackets());

	if (fd < 0) {
		G2RSVPTE_ERR("Cannot reconnect to SCNGW ... retry after 1s");
		thread_add_timer_msec(RSVP_MASTER.getMaster(),
				      reconnect_scngw,
				      proto,
				      1000);
		return 0;
	}

	proto->scngwFd(fd);

	G2RSVPTE_INF("Scheduling again the read thread after the "
		     "reconnection with SCNGW");

	//schedule the read thread again
	struct thread * thr;
	thr = thread_add_read(RSVP_MASTER.getMaster(),
			      listen_from_scngw,
			      proto,
			      proto->scngwFd());
	if (thr) {
		proto->readThread(thr);
	}

	return 0;
}

int listen_from_scngw(struct thread *thread)
{
	struct ip *      iph;
	Message *        msg;
	RSVP::G2RSVPTE * proto;

	proto = (RSVP::G2RSVPTE *) THREAD_ARG(thread);
	assert(proto);

	iph = 0;
	msg = 0;

	// reset static buffer
	memset(buffer, 0, G2RSVPTE_MAX_PACKET_SIZE);

	int len;
	len = scngwc_stream_recvmsg(buffer,
				    proto->scngwFd(),
				    &iph,
				    G2RSVPTE_MAX_PACKET_SIZE + 1);
	if (len <= 0) {
		G2RSVPTE_ERR("scngwc_stream_recvmsg failed...try to reconnect");

		thread_add_timer_msec(RSVP_MASTER.getMaster(),
				      reconnect_scngw,
				      proto,
				      1000);
		free(iph);
		return 0;
	}

	//schedule the read thread again
	struct thread * thr;
	thr = thread_add_read(RSVP_MASTER.getMaster(),
			      listen_from_scngw,
			      proto,
			      proto->scngwFd());
	if (thr) {
		proto->readThread(thr);
	}

	if (!iph) {
		G2RSVPTE_ERR("scngwc_stream_recvmsg failed ... "
			     "iph not initialized!");
		return 0;
	}

	try {
		MemoryStream ms((uint8_t*)buffer, len);
		ms >> msg;

		std::ostringstream str;
		str << *msg;
		G2RSVPTE_DBG("---------- RECEIVED PACKET ----------\n %s",
			     str.str().c_str());

	} catch (std::string e) {
		G2RSVPTE_ERR("Discarding G2RSVPTE msg, error: %s",
			     e.c_str());
		free(iph);
		return 0;
	} catch (...) {
		G2RSVPTE_ERR("Unknown exception, impossible to create G2RSVPTE "
			     "msg ... discarding packet!");
		free(iph);
		return 0;
	}

	if (!msg) {
		G2RSVPTE_ERR("Received invalid message!");
		free(iph);
		return 0;
	} else {

		//FIX for SCNGW
		g2mpls_addr_t addr;
		if (proto->tunneledPackets()) {
			addr.type        = IPv4;
			addr.preflen     = 32;
			addr.value.ipv4  = iph->ip_dst;
		} else {
			memset(&addr, 0, sizeof(addr));
		}
		RSVP_MASTER.manageMsg(proto->tunneledPackets(), addr, msg);

		// Delete message??????
		//delete msg;
	}

	free(iph);

	return 0;
}

int send_to_scngw(struct thread *thread)
{
	assert(thread);
	msg_data_t * data;
	data = (msg_data_t *)THREAD_ARG(thread);
	assert(data);
	int scngwFd;
        scngwFd = THREAD_VAL(thread);
	Message * msg = data->msg;
	assert(msg);

	MemoryStream ms;
	int          ret;

	try {
		std::ostringstream str;
		str << *msg;
		G2RSVPTE_DBG("---------- PACKET TO SEND ----------\n %s",
			     str.str().c_str());

		ms << *msg;

	} catch (std::string e) {
		G2RSVPTE_ERR("G2RSVPTE msg not sent, error: %s",
			     e.c_str());
		delete data;
		return 0;
	} catch (...) {
		G2RSVPTE_ERR("Unknown exception, impossible to create Memory"
			     "Stream ... G2RSVPTE msg not sent!");
		delete data;
		return 0;
	}

	// TMP: only IPv4
	if (data->src.type != IPv4 || data->dst.type != IPv4) {
		G2RSVPTE_ERR("SRC and DST interface MUST be IPv4!");
		delete data;
		return 0;
	}

#ifdef G2RSVPTE2SCNGW_DEBUG
	ret = send_buffer(ms.getBufferData(),
			  ms.getBufferDataSize(),
			  data->src.value.ipv4,
			  data->dst.value.ipv4);
#else
	int unread_pckts = 0;
	int flags        = 0;
	RSVP::G2RSVPTE * proto;

	proto = RSVP_MASTER.getProtoFromScngwFd(scngwFd);
	if (!proto) {
		G2RSVPTE_ERR("Cannot retrieve instance from SCNgw fd");
		return 0;
	}

	ret = scngwc_sendmsg(scngwFd,
			     ms.getBufferData(),
			     ms.getBufferDataSize(),
			     data->src.value.ipv4,
			     data->dst.value.ipv4,
			     flags,
			     &unread_pckts);
	if (unread_pckts) {
		struct thread * thr;
		thr = thread_add_read(RSVP_MASTER.getMaster(),
				      listen_from_scngw,
				      proto,
				      proto->scngwFd());
		if (thr) {
			proto->readThread(thr);
		}
	}
#endif

	if (ret < 0) {
		G2RSVPTE_ERR("send_to_scngw failed");
		delete data;
		return 0;
	}

	//delete data->msg;
	delete data;
	return 0;
}

#ifdef G2RSVPTE2SCNGW_DEBUG

#define RSVP_HEADER_LEN	      8
#define MAX_IP_HEADER_SIZE    60
#define MAX_PACKET_SIZE       65535U+MAX_IP_HEADER_SIZE

extern struct zebra_privs_t g2rsvpted_privs;

static int testFd = -1;

int g2rsvpte_sock_get(void)
{
	assert(testFd > 0);
	return testFd;
}

void server_start(void)
{
	testFd = sock_init();
	assert(testFd > 0); // XXX FIXME

	G2RSVPTE_DBG("adding read thread");
	thread_add_read(RSVP_MASTER.getMaster(),
			net_packet_read,
			NULL,
			g2rsvpte_sock_get());
}

int sock_init(void)
{
	if (testFd > 0) {
		return testFd;
	}

	int sock;
	int ret, hincl = 1;

	if (g2rsvpted_privs.change(ZPRIVS_RAISE)) {
		G2RSVPTE_ERR("net_sock_open: could not raise privs, %s",
			     safe_strerror (errno));
	}

	sock = socket (AF_INET, SOCK_RAW, IPPROTO_RSVP);
	int socket_errno = errno;
	if (sock < 0) {
		if (g2rsvpted_privs.change (ZPRIVS_LOWER)) {
			G2RSVPTE_ERR("could not lower privs, %s",
				     safe_strerror (errno));
		}
		G2RSVPTE_DBG("could not create raw socket: %s",
			     safe_strerror(socket_errno));
		exit(1);
	}

#ifdef IP_HDRINCL
	// we will include IP header with packet
	ret = setsockopt (sock, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof(hincl));
	G2RSVPTE_DBG("the option IP_HDRINCL was set on the socket");
	if (ret < 0) {
		int save_errno = errno;
		if (g2rsvpted_privs.change (ZPRIVS_LOWER)) {
			G2RSVPTE_ERR("could not lower privs, %s",
				  safe_strerror (errno));
		}
		G2RSVPTE_WRN("Can't set IP_HDRINCL option for fd %d: %s",
			     sock, safe_strerror(save_errno));
	}
#elif defined (IPTOS_PREC_INTERNETCONTROL)
#warning "IP_HDRINCL not available on this system"
#warning "using IPTOS_PREC_INTERNETCONTROL"
	int tos;
	// Set precedence field.
	tos = IPTOS_PREC_INTERNETCONTROL;
	ret = setsockopt (sock, IPPROTO_IP, IP_TOS,
			  (char *) &tos, sizeof (int));
	if (ret < 0) {
		int save_errno = errno;
		if (g2rsvpted_privs.change (ZPRIVS_LOWER)) {
			G2RSVPTE_ERR("could not lower privs, %s",
				  safe_strerror (errno));
		}
		G2RSVPTE_WRN("can't set sockopt IP_TOS %d to socket %d: %s",
			     tos, sock, safe_strerror(save_errno));
		close(sock);	// Prevent sd leak.
		return ret;
	}
#else // !IPTOS_PREC_INTERNETCONTROL
#warning "IP_HDRINCL not available, nor is IPTOS_PREC_INTERNETCONTROL"
	G2RSVPTE_WRN("IP_HDRINCL option not available");
#endif // IP_HDRINCL

	ret = setsockopt_ifindex(AF_INET, sock, 1);

	if (ret < 0) {
		G2RSVPTE_WRN("Can't set pktinfo option for fd %d", sock);
	}

	if (g2rsvpted_privs.change (ZPRIVS_LOWER)) {
		G2RSVPTE_ERR("could not lower privs, %s",
			     safe_strerror (errno));
	}
	return sock;
}

struct stream *
recv_net_packet (int fd, struct stream *ibuf)
{
	int ret;
	struct ip *iph;
	u_int16_t ip_len;
	struct iovec iov;
	struct msghdr msgh;

	memset (&msgh, 0, sizeof (struct msghdr));
	msgh.msg_iov = &iov;
	msgh.msg_iovlen = 1;

	ret = stream_recvmsg (ibuf, fd, &msgh, 0, MAX_PACKET_SIZE);

	if (ret < 0) {
		zlog_warn("stream_recvmsg failed: %s", safe_strerror(errno));
		return NULL;
	}
	if ((unsigned int)ret < sizeof(iph)) { // ret must be > 0 now
		zlog_warn("recv_net_packet: discarding runt packet of length %d"
			  " (ip header size is %u)",
			  ret, (u_int)sizeof(iph));
		return NULL;
	}

	// Note that there should not be alignment problems with this assignment
	   because this is at the beginning of the stream data buffer.
	iph = (struct ip *) STREAM_DATA(ibuf);
	sockopt_iphdrincl_swab_systoh (iph);

	ip_len = iph->ip_len;

#if !defined(GNU_LINUX) && (OpenBSD < 200311)
	//
	// Kernel network code touches incoming IP header parameters,
	// before protocol specific processing.
	//
	//   1) Convert byteorder to host representation.
	//      --> ip_len, ip_id, ip_off
	//
	//   2) Adjust ip_len to strip IP header size!
	//      --> If user process receives entire IP packet via RAW
	//          socket, it must consider adding IP header size to
	//          the "ip_len" field of "ip" structure.
	//
	// For more details, see <netinet/ip_input.c>.
	//

	ip_len = ip_len + (iph->ip_hl << 2);
#endif
	if (ret != ip_len) {
		zlog_warn ("recv_net_packet read length mismatch: ip_len is %d,"
			   " but recvmsg returned %d", ip_len, ret);
		return NULL;
	}

	return ibuf;
}

int net_packet_read (struct thread *thread)
{
	struct stream *ibuf;
	struct ip *iph;
	size_t sdu_size = 0;
	size_t iph_size = 0;
	char * buff = NULL;
	struct rsvpte_header rsvp_h;
	Message * msg;

	ibuf = stream_new(MAX_PACKET_SIZE+1);
	stream_reset(ibuf);

	if (!(ibuf = recv_net_packet(g2rsvpte_sock_get(), ibuf)))
		return -1;

	iph = (struct ip *) STREAM_DATA (ibuf);
	sdu_size = (iph->ip_len - (iph->ip_hl << 2));
	zlog_debug ("---------------- IP header dump ------------------------");
	zlog_debug ("net_packet_read: loc addr: %s", inet_ntoa(iph->ip_src));
	zlog_debug ("net_packet_read: rem addr: %s", inet_ntoa(iph->ip_dst));
	zlog_debug ("net_packet_read: SDU size: %u", sdu_size);
	zlog_debug ("net_packet_read: stream endp = %d", ibuf->endp);
	zlog_debug ("net_packet_read: stream getp = %d", ibuf->getp);
	zlog_debug ("--------------------------------------------------------");
	iph_size = iph->ip_hl << 2;
	zlog_debug("net_packet_read:  ip header size = %d", iph_size);
	stream_forward_getp (ibuf, iph_size); //we get the stream pointer at the beginning of the SDU
	zlog_debug ("net_packet_read: stream getp = %d", ibuf->getp);

	rsvp_h.version = stream_getc(ibuf);
	rsvp_h.flags = rsvp_h.version;
	rsvp_h.msg_type = stream_getc(ibuf);
	rsvp_h.check_sum = stream_getw(ibuf);
	rsvp_h.send_ttl = stream_getc(ibuf);
	stream_getc(ibuf);
	rsvp_h.length = stream_getw(ibuf);

	zlog_debug ("--------------------------------------------------------");
	zlog_debug ("----------------------rsvp header dump------------------");
	zlog_debug ("net_packet_read: stream getp = %d", ibuf->getp);
	zlog_debug ("net_packet_read: stream getp = %d", ibuf->endp);
	zlog_debug ("rsvp message type: %d", rsvp_h.msg_type);
	zlog_debug ("rsvp message length: %d", rsvp_h.length);
	zlog_debug ("rsvp message length: %d", rsvp_h.length);
	zlog_debug ("--------------------------------------------------------");
	zlog_debug ("--------------------------------------------------------");

	buff = (char *)malloc(rsvp_h.length);
	assert(buff);
	memcpy(buff, ibuf->data + ibuf->getp - RSVP_HEADER_LEN, rsvp_h.length);

	MemoryStream ms((uint8_t*)buff, rsvp_h.length);
	ms >> msg;

	if (!msg) {
		G2RSVPTE_ERR("Received invalid message!!!");
		//TODO sth????
	} else {

		std::cout << "---------- RECEIVED PACKET ----------" << std::endl;
		std::cout << "############## TXT ###############" << std::endl;
		std::cout << *msg;

		//FIX for SCNGW
		g2mpls_addr_t addr;
		addr.type        = IPv4;
		addr.preflen     = 32;
		addr.value.ipv4  = iph->ip_dst;

		RSVP_MASTER.manageMsg(true, addr, msg);

		// Delete message??????
		//delete msg;
	}

	stream_free (ibuf);
	free(buff);

	//schedule the read thread again
	thread_add_read(RSVP_MASTER.getMaster(),
			net_packet_read,
			NULL,
			g2rsvpte_sock_get());

	return 0;
}

int send_buffer(const void *   buff, size_t         size,
		struct in_addr src,  struct in_addr dst)
{
	testFd = sock_init();
	assert(testFd > 0);

	assert(buff);
	assert(size);

	int ret;
	struct msghdr msg;
	struct ip iph;
	struct iovec iov[2];
	struct sockaddr_in sa_dst;
#ifdef WANT_OSPF_WRITE_FRAGMENT
	static u_int16_t ipid = 0;
#endif // WANT_OSPF_WRITE_FRAGMENT
	//u_int16_t maxdatasize;

	memset (&iph, 0, sizeof (struct ip));
	memset (&sa_dst, 0, sizeof (sa_dst));

	sa_dst.sin_family = AF_INET;
#ifdef HAVE_SIN_LEN
	sa_dst.sin_len = sizeof(sa_dst);
#endif // HAVE_SIN_LEN
	sa_dst.sin_port = htons (0);

	//sa_dst.sin_addr.s_addr = msg_data->dst_addr.s_addr;
	//GGinet_aton(DST, &(sa_dst.sin_addr));
	sa_dst.sin_addr.s_addr = dst.s_addr;

	//maxdatasize = scnif.mtu - sizeof (struct ip);

	iph.ip_hl = sizeof (struct ip) >> 2;
	iph.ip_v = IPVERSION;
	iph.ip_tos = IPTOS_PREC_INTERNETCONTROL;

	iph.ip_len = (iph.ip_hl << 2) + size;
	//iph.ip_len = (iph.ip_hl << 2) + packet_data_size;

	iph.ip_off = 0;
	iph.ip_ttl = 100;
	iph.ip_p = IPPROTO_RSVP; // IPIP encapsulation
	iph.ip_sum = 0;

	//GGinet_aton(SRC, &(iph.ip_src));
	//iph.ip_src.s_addr = msg_data->src_addr.s_addr;
	iph.ip_src.s_addr = src.s_addr;

	//GGinet_aton(DST, &(iph.ip_dst));
	//iph.ip_dst.s_addr = msg_data->dst_addr.s_addr;
	iph.ip_dst.s_addr = dst.s_addr;

	zlog_debug(" - src IP %s\n", inet_ntoa(iph.ip_src));
	zlog_debug(" - dst IP %s\n", inet_ntoa(iph.ip_dst));

#ifdef WANT_OSPF_WRITE_FRAGMENT
	iph.ip_id = ++ipid;
#endif // WANT_OSPF_WRITE_FRAGMENT

	memset (&msg, 0, sizeof (msg));
	msg.msg_name = (caddr_t) &sa_dst;
	msg.msg_namelen = sizeof (sa_dst);
	msg.msg_iov = iov;
	msg.msg_iovlen = 2;
	iov[0].iov_base = (char*)&iph;
	iov[0].iov_len = iph.ip_hl << 2;

	//GG
	char * tmp;
	tmp = (char *)malloc(size);
	memcpy(tmp, buff, size);

	iov[1].iov_base = tmp;//buff;
	iov[1].iov_len =  size;
	//iov[1].iov_base = packet_data;
	//iov[1].iov_len = packet_data_size;

	sockopt_iphdrincl_swab_htosys (&iph);
	ret = sendmsg(testFd, &msg, 0);
	sockopt_iphdrincl_swab_systoh (&iph);

	if (ret < 0) {
		G2RSVPTE_WRN("*** sendmsg in net_packet_send failed to %s, "
			     "id %d, off %d, len %d : %s",
			     inet_ntoa (iph.ip_dst),
			     iph.ip_id,
			     iph.ip_off,
			     iph.ip_len,
			     safe_strerror (errno));
	}

	free(tmp);
	return ret;
}
#endif // G2RSVPTE2SCNGW_DEBUG
