/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */


#ifndef __USE_MISC
#define __USE_MISC
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/ioctl.h>
#ifndef __USE_MISC
#define __USE_MISC
#endif
#include <net/if.h>


/*Zebra includes*/
#include <zebra.h>
#include "linklist.h"
#include "lrm_core.h"
#include "log.h"
#include "lrmd.h"

#if HAVE_OMNIORB
#include "lrm_corba.h"
#endif /* HAVE_OMNIORB */


/*
 * This function checks up if the ip address ip_addr is actually configured
 * on a kernel interface
 * */
static int
check_kernel_for_scnif(g2mpls_addr_t addr)
{
	int                  sock;
	int                  n;
	int                  found_if;
	int                  numifreqs;
	struct ifreq *       ifr;
	struct ifreq         ifr_match;
	struct ifconf        ifc;
	struct sockaddr_in * ifaddr;

	found_if  = 0;
	numifreqs = 30;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		int save_errno = errno;
		zlog_err("Unable to check ctrl i/f: "
			 "cannot create UDP socket: %s",
			 strerror(save_errno));
		return 0;
	}

	ifc.ifc_buf = NULL;

	for (;;) {

		ifc.ifc_len = sizeof(ifr) * numifreqs;
		ifc.ifc_buf = realloc(ifc.ifc_buf, ifc.ifc_len);

		if (ioctl(sock, SIOCGIFCONF, &ifc) < 0) {
			zlog_err("Can't lookup kernel interfaces "
				 "by ioctl(SIOCGIFCONF): %s",
				 strerror (errno));
			close (sock);
			free(ifc.ifc_buf);
			return 0;
		}

		if ((u_int32_t)ifc.ifc_len == sizeof(ifr) * numifreqs) {
			/* assume it overflowed and try again */
			numifreqs += 10;
			continue;
		}

		break;
	}

	ifr = ifc.ifc_req;
	for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {

		ifaddr = (struct sockaddr_in *)(&(ifr->ifr_addr));

		zlog_debug("Examining kernel interface: %s",
			   inet_ntoa(ifaddr->sin_addr));


		/*test against the provided ip_addr*/
		if (ifaddr->sin_addr.s_addr == addr.value.ipv4.s_addr) {

			strcpy (ifr_match.ifr_name, ifr->ifr_name);
			if (ioctl(sock, SIOCGIFFLAGS, &ifr_match) < 0 ) {
				zlog_err("Can't lookup interface flags "
					 "by ioctl(SIOCGIFFLAGS): %s",
					 strerror (errno));
				close (sock);
				free(ifc.ifc_buf);
				return 0;
			}
			found_if = 1;
			break;
		}
		ifr++;
	}

	if(found_if) {
		zlog_debug("Found kernel interface %s %s and %s",
			   inet_ntoa(ifaddr->sin_addr),
			   (ifr_match.ifr_flags & IFF_UP) ? "UP" : "DOWN",
			   (ifr_match.ifr_flags & IFF_RUNNING) ? "RUNNING" :
			   "NOT RUNNING");
		if((ifr_match.ifr_flags & IFF_UP) &&
		   (ifr_match.ifr_flags & IFF_RUNNING)) {

			close (sock);
			free(ifc.ifc_buf);
			return 1;
		}
	}

	close (sock);
	free(ifc.ifc_buf);
	return 1;//0;
}

lrm_t *
lrm_lookup(void)
{
	if (listcount(lrm_m->lrm) == 0) {
		return NULL;
	}

	return listgetdata(listhead(lrm_m->lrm));
}

/* Allocate new lrm structure. */
static lrm_t *
lrm_new(void)
{
	lrm_t * lrm = XCALLOC(MTYPE_LRM, sizeof(lrm_t));

	if (lrm == NULL) {
		zlog_err("Cannot allocate memory for LRM instance");
		assert(0);
	}

	ASSERT(lrm);

	zlog_debug("Initializing LRM data structures");

	lrm->router_id        = 0;
	lrm->router_enni_id   = 0;
	lrm->router_uni_id    = 0;
	lrm->scn_if_list      = list_new();
	lrm->cc_list          = list_new();
	lrm->datalink_list    = list_new();
	lrm->telink_list      = list_new();
	lrm->adj_list         = list_new();
	lrm->telink_count     = 0;

	ASSERT(lrm);

	return lrm;
}

static void
lrm_add (lrm_t * lrm)
{
	ASSERT(lrm);
	listnode_add(lrm_m->lrm, lrm);
}

static void
lrm_delete (lrm_t * lrm)
{
	listnode_delete(lrm_m->lrm, lrm);
}

lrm_t *
lrm_get(void)
{
	lrm_t * lrm;

	lrm = lrm_lookup();
	if (lrm == NULL) {
		lrm = lrm_new();
		lrm_add(lrm);
	}

	return lrm;
}

/* Finalize lrm structure. */
void
lrm_finish(lrm_t * lrm)
{
	zlog_debug("Finalising LRM data structures");

	struct zlistnode * node;
	struct zlistnode * nnode;
	ctrl_intf_t     * scnif;
	cc_t            * cc;
	datalink_t      * dl;
	te_link_t       * tel;


	for (ALL_LIST_ELEMENTS(lrm->scn_if_list, node, nnode, scnif)) {
		XFREE(MTYPE_LRM_SCNIF, scnif);
		list_delete_node(lrm->scn_if_list, node);
	}
	list_free(lrm->scn_if_list);

	for (ALL_LIST_ELEMENTS(lrm->cc_list, node, nnode, cc)) {
		XFREE(MTYPE_LRM_CC,cc);
		list_delete_node(lrm->cc_list, node);
	}
	list_free(lrm->cc_list);

	for (ALL_LIST_ELEMENTS(lrm->datalink_list, node, nnode, dl)) {
		XFREE(MTYPE_LRM_DL, dl);
		list_delete_node(lrm->datalink_list, node);
	}
	list_free(lrm->datalink_list);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		struct zlistnode * s_node;
		struct zlistnode * s_nnode;
		u_int32_t       * id;

		for (ALL_LIST_ELEMENTS(tel->SRLG_ids, s_node, s_nnode, id)) {
			XFREE(MTYPE_LRM_OTHER, id);
			list_delete_node(tel->SRLG_ids, s_node);
		}
		list_free(tel->SRLG_ids);


		XFREE(MTYPE_LRM_TEL, tel);
		list_delete_node(lrm->telink_list, node);
	}
	list_free(lrm->telink_list);

	lrm_delete(lrm);

	XFREE(MTYPE_LRM, lrm);
}

int
lrm_set_rid(lrm_t * lrm, adj_type_t adjacency, u_int32_t  rid)
{
	ASSERT(lrm);


	switch(adjacency)
	{
		case INNI:
		if (lrm->router_id != 0) {
			zlog_err("Router INNI ID %s just configured in LRM", numb_htoa(rid));
			return LRM_ERROR;
		}

		if (rid == 0) {
			zlog_err("Set up a null (%s) Router INNI ID in LRM is non-sense", numb_htoa(rid));
			return LRM_ERROR;
		}

		lrm->router_id = rid;

		zlog_debug("Notifying new Router INNI ID %s to attached clients", numb_htoa(rid));
		lrm_router_id_notify(rid, INNI);
		break;

		case ENNI:
		if (lrm->router_enni_id != 0) {
			zlog_err("Router ENNI ID %s just configured in LRM", numb_htoa(rid));
			return LRM_ERROR;
		}

		if (rid == 0) {
			zlog_err("Set up a null (%s) Router ENNI ID in LRM is non-sense", numb_htoa(rid));
			return LRM_ERROR;
		}

		lrm->router_enni_id = rid;

		zlog_debug("Notifying new Router ENNI ID %s to attached clients", numb_htoa(rid));
		lrm_router_id_notify(rid, ENNI);
		break;

		case UNI:
		if (lrm->router_uni_id != 0) {
			zlog_err("Router UNI ID %s just configured in LRM", numb_htoa(rid));
			return LRM_ERROR;
		}

		if (rid == 0) {
			zlog_err("Set up a null (%s) Router UNI ID in LRM is non-sense", numb_htoa(rid));
			return LRM_ERROR;
		}

		lrm->router_uni_id = rid;

		zlog_debug("Notifying new Router UNI ID %s to attached clients", numb_htoa(rid));
		lrm_router_id_notify(rid, UNI);
		break;
	}

	return LRM_OK;
}

/*
 * for the moment this function only adds ipv4 ctrl interfaces
 */
int
scn_if_add(lrm_t * lrm, g2mpls_addr_t addr, if_type_t intf_type)
{
	struct zlistnode  * node;
	ctrl_intf_t      * ctrl_if;

	node    = 0;
	ctrl_if = 0;

	ASSERT(lrm);

	if(!check_kernel_for_scnif(addr)){
		zlog_err("Attempting to add an interface (%s) "
			 "that does not exist in the kernel",
			 addr_ntoa(addr));
		return LRM_ERROR;
	}

	/* At first check if this ctrl_if exists */
	for (ALL_LIST_ELEMENTS_RO(lrm->scn_if_list, node, ctrl_if)) {
		ASSERT(ctrl_if);
		if (addr_equal(ctrl_if->addr, addr) == 1) {
			zlog_err("Control interface %s just present in LRM",
				 addr_ntoa(addr));
			return LRM_ERROR;
		}
	}

	ctrl_if = XCALLOC(MTYPE_LRM_SCNIF, sizeof(*ctrl_if));
	if (ctrl_if == NULL) {
		zlog_err("Cannot allocate a control interface data structure");
		return LRM_ERROR;
	}

	memset(ctrl_if, 0, sizeof(ctrl_intf_t));

	memcpy(&(ctrl_if->addr), &addr, sizeof(addr));

	ctrl_if->type      = intf_type;
	ctrl_if->sync_status = NOT_SENT;

	if(check_kernel_for_scnif(addr)){
		zlog_debug("the control interface %s is UP and running",
			   addr_ntoa(addr));
		ctrl_if->op_state = UP;
	} else {
		ctrl_if->op_state = DOWN;
	}

	ctrl_if->adm_state = DISABLED;

	listnode_add(lrm->scn_if_list, ctrl_if);

	return LRM_OK;
}

int
scn_if_del(lrm_t * lrm, g2mpls_addr_t addr)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	ctrl_intf_t     * ctrl_if;
	int found = 0;

	node    = 0;
	nnode   = 0;
	ctrl_if = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->scn_if_list, node, nnode, ctrl_if)) {
		ASSERT(ctrl_if);
		if (addr_equal(ctrl_if->addr, addr) == 1) {
			if (ctrl_if->adm_state == ENABLED) {
				zlog_err("Cannot delete control interface %s, "
					 "administrative state is "
					 "ENABLED",
					 addr_ntoa(addr));
				return  LRM_ERROR;
			}
			found = 1;
			break;
		}
	}

	if (found) {
		//lrm_scnif_del_notify(ctrl_if);
		listnode_delete(lrm->scn_if_list, ctrl_if);
		free(ctrl_if);
		return LRM_OK;
	}

	zlog_err("Control interface %s not found", addr_ntoa(addr));
	return LRM_ERROR;
}

int
scn_if_ena(lrm_t * lrm, g2mpls_addr_t addr)
{
	struct zlistnode * node;
	ctrl_intf_t     * ctrl_if;

	node    = 0;
	ctrl_if = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS_RO(lrm->scn_if_list, node, ctrl_if)) {
		ASSERT(ctrl_if);
		if (addr_equal(ctrl_if->addr, addr) == 1) {
			ctrl_if->adm_state = ENABLED;
			lrm_scnif_add_notify(ctrl_if);
			return LRM_OK;
		}
	}

	zlog_err("Control interface %s not found", addr_ntoa(addr));
	return LRM_ERROR;
}

int
scn_if_dis(lrm_t * lrm, g2mpls_addr_t addr)
{
	struct zlistnode * node;
	ctrl_intf_t     * ctrl_if;

	node    = 0;
	ctrl_if = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS_RO(lrm->scn_if_list, node, ctrl_if)) {
		ASSERT(ctrl_if);
		if (addr_equal(ctrl_if->addr, addr) == 1) {
			lrm_scnif_del_notify(ctrl_if);
			ctrl_if->adm_state = DISABLED;
			return LRM_OK;
		}
	}

	zlog_err("Control interface %s not found", addr_ntoa(addr));
	return LRM_ERROR;
}

int
control_channel_add(lrm_t     *        lrm,
		    u_int32_t          cc_id,
		    g2mpls_addr_t      lcl_scn_addr,
		    g2mpls_addr_t      rem_scn_addr)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	cc_t            * cc;
	ctrl_intf_t     * ctrl_if;
	int found_ctrl_if = 0;

	node    = 0;
	nnode   = 0;
	cc      = 0;
	ctrl_if = 0;

	ASSERT(lrm);

	if (cc_id == 0) {
		zlog_err("Cannot declare control channel with id %d",
			 cc_id);
		return LRM_ERROR;
	}

	/* At first check if this cc exists */
	for (ALL_LIST_ELEMENTS(lrm->cc_list, node, nnode, cc)) {
		ASSERT(cc);
		if (cc->cc_id == cc_id) {
			zlog_err("A Control channel already exist with ID %d",
				 cc_id);
			return LRM_ERROR;
		}
	}

	/* Then check if the lcl_ctrl_if exists */
	for (ALL_LIST_ELEMENTS_RO(lrm->scn_if_list, node, ctrl_if)) {
		if (addr_equal(ctrl_if->addr, lcl_scn_addr) == 1) {
			found_ctrl_if = 1;
			break;
		}
	}

	if (!found_ctrl_if) {
		zlog_err("SCN Control interface %s does not exist in LRM",
			 addr_ntoa(lcl_scn_addr));
		zlog_err("Control channel %d cannot be bound",
			 cc_id);
		return LRM_ERROR;
	}


	/* if you are here you need to really add */
	cc = XCALLOC(MTYPE_LRM_CC, sizeof(*cc));
	if (cc == NULL) {
		zlog_err("Cannot allocate a control channel data structure");
		return LRM_ERROR;
	}

	memset(cc, 0, sizeof(cc_t));

	cc->cc_id     = cc_id;
	cc->rem_cc_id = 0xFFFFFFFF; /* XXX-FIXME HARD WIRE to remove with LMP*/

	memcpy(&(cc->lcl_scn_addr), &lcl_scn_addr, sizeof(lcl_scn_addr));
	memcpy(&(cc->rem_scn_addr), &rem_scn_addr, sizeof(rem_scn_addr));

	cc->op_state     = DOWN;
	cc->adm_state    = DISABLED;
	cc->te_link_list = list_new();
	cc->sync_status = NOT_SENT;

	listnode_add(lrm->cc_list, cc);

	return LRM_OK;
}

int
control_channel_del(lrm_t * lrm, u_int32_t cc_id)
{
	struct zlistnode * node, *node2;
	struct zlistnode *nnode, *nnode2;
	cc_t            *cc, *cc2;
	te_link_t	*tel;
	int cc_found = 0;

	node    = 0;
	nnode   = 0;
	cc      = 0;

	ASSERT(lrm);

	/*if (cc_id == 0) {
	  zlog_err("Cannot delete any control channel with id %d",
	  cc_id);
	  return LRM_ERROR;
	  }*/

	for (ALL_LIST_ELEMENTS(lrm->cc_list, node, nnode,cc)) {
		ASSERT(cc);
		if (cc->cc_id == cc_id) {
			cc_found = 1;
			break;
		}
	}

	if (cc_found) {
		if(cc){
			if (cc->adm_state == ENABLED) {
				zlog_err("Cannot delete control channel %d:"
					 "admin state is UP",
					 cc_id);
				return  LRM_ERROR;
			}
			lrm_cc_del_notify(cc);

			for (ALL_LIST_ELEMENTS(cc->te_link_list, node, nnode,tel)) {
				for (ALL_LIST_ELEMENTS(tel->cc_list, node2, nnode2, cc2)){
					if(cc2->cc_id == cc->cc_id)
						listnode_delete(tel->cc_list, cc2);

				}
				if(tel->assoc_cc->cc_id == cc->cc_id){
					zlog_debug("deleting the associated cc");
					tel->assoc_cc = NULL;
					for (ALL_LIST_ELEMENTS(tel->cc_list, node2, nnode2, cc2)){
						if((cc2->op_state == UP) && (cc2->adm_state == ENABLED)){
							tel->assoc_cc = cc2;
							break;
						}
					}

				}
			}
			listnode_delete(lrm->cc_list, cc);
			free(cc);
		}
		return LRM_OK;
	}
	else{ /*cc_found == 0 the cc was not found*/
		zlog_err("Control channel %d not found", cc_id);
		return LRM_ERROR;
	}
}


int
control_channel_ena(lrm_t * lrm, u_int32_t cc_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	cc_t            * cc;

	node    = 0;
	nnode   = 0;
	cc      = 0;

	ASSERT(lrm);

	if (cc_id == 0) {
		zlog_err("Cannot handle any control channel with id %d",
			 cc_id);
		return LRM_ERROR;
	}

	for (ALL_LIST_ELEMENTS(lrm->cc_list, node, nnode, cc)) {
		if (cc->cc_id == cc_id) {
			cc->adm_state = ENABLED;
			return  LRM_OK;
		}
	}

	zlog_err("Control channel %d not found", cc_id);
	return LRM_ERROR;
}


int
control_channel_dis(lrm_t * lrm, u_int32_t cc_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	cc_t            * cc;

	node    = 0;
	nnode   = 0;
	cc      = 0;

	ASSERT(lrm);

	if (cc_id == 0) {
		zlog_err("Cannot handle any control channel with id %d",
			 cc_id);
		return LRM_ERROR;
	}

	for (ALL_LIST_ELEMENTS(lrm->cc_list, node, nnode, cc)) {
		if (cc->cc_id == cc_id) {
			cc->adm_state = DISABLED;
			return  LRM_OK;
		}
	}

	zlog_err("Control channel %d not found", cc_id);
	return LRM_ERROR;
}

int
control_channel_up(lrm_t * lrm, u_int32_t cc_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	cc_t            * cc;

	ASSERT(lrm);

	node    = 0;
	nnode   = 0;
	cc      = 0;

	if (cc_id == 0) {
		zlog_err("Cannot handle any control channel with id %d",
			 cc_id);
		return LRM_ERROR;
	}

	for (ALL_LIST_ELEMENTS(lrm->cc_list, node, nnode, cc)) {
		if (cc->cc_id == cc_id) {
			cc->op_state = UP;
			/* and now  notify to all the interested listeners */
			lrm_cc_add_notify(cc);
			return  LRM_OK;
		}
	}

	zlog_err("Control channel %d not found", cc_id);
	return LRM_ERROR;
}


int
control_channel_down(lrm_t * lrm, u_int32_t cc_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	cc_t            * cc;

	node    = NULL;
	nnode   = NULL;
	cc      = NULL;

	ASSERT(lrm);

	if (cc_id == 0) {
		zlog_err("Cannot handle any control channel with id %d",
			 cc_id);
		return LRM_ERROR;
	}

	for (ALL_LIST_ELEMENTS(lrm->cc_list, node, nnode, cc)) {
		if (cc->cc_id == cc_id) {
			cc->op_state = DOWN;
			/* and now  notify to all the interested listeners */
			lrm_cc_del_notify(cc);

			return  LRM_OK;
		}

	}

	zlog_err("Control channel %d not found", cc_id);
	return LRM_ERROR;
}

int
data_link_retrieve_info_from_tnrc(datalink_t * dl)
{
	ASSERT(dl);

#if HAVE_OMNIORB
	if (!tnrc_get_dlink_details(dl)) {
		return LRM_ERROR;
	}
#else
	/* XXX-FIXME: hard wire to be removed once TNRC is ready */
	dl->op_state   = UP;
	dl->swcap      = SWCAP_LSC;
	dl->enctype    = ENCT_LAMBDA;
	dl->max_bw     = BwEnc_10GigELAN;
	dl->max_res_bw = BwEnc_10GigELAN;
	dl->min_LSP_bw = BwEnc_Fast_Ethernet;

	for (int i = 0; i < MAX_BW_PRIORITIES; i++) {
		dl->avail_bw_per_prio[i] = BwEnc_GigE;
		dl->max_LSP_bw[i]        = BwEnc_GigE;
	}
#endif /* HAVE_OMNIORB */

	return LRM_OK;
}


int
data_link_add(lrm_t * lrm, g2mpls_addr_t dl_id, g2mpls_addr_t rem_dl_id)
{
	struct zlistnode * node;
	datalink_t      * dl;

	node    = 0;
	dl      = 0;

	ASSERT(lrm);

	/* At first check if this dl exists */
	for (ALL_LIST_ELEMENTS_RO(lrm->datalink_list, node, dl)) {
		ASSERT(dl);
		if (addr_equal(dl->lcl_id, dl_id)) {
			zlog_err("A data-link just exist with ID %s",
				 addr_ntoa(dl_id));
			return LRM_ERROR;
		}
	}

	/* if you are here you need to add */
	dl = XCALLOC(MTYPE_LRM_DL, sizeof(*dl));
	if (dl == NULL) {
		zlog_err("Cannot allocate a data-link data structure");
		return LRM_ERROR;
	}

	memset(dl, 0, sizeof(datalink_t));
	memcpy(&(dl->lcl_id), &dl_id,     sizeof(dl_id));
	memcpy(&(dl->rem_id), &rem_dl_id, sizeof(rem_dl_id));

	dl->op_state     = DOWN;
	dl->adm_state    = DISABLED;

	dl->calendar      = list_new();
	dl->calendar->del = delete_calendar_event;
	dl->calendar->cmp = compare_calendar_event;

	if (data_link_retrieve_info_from_tnrc(dl) != LRM_OK) {
		zlog_warn("Cannot retrieve hardware specific info "
			 "for data-link %s...",
			 addr_ntoa(dl_id));
		dl->swcap   = SWCAP_UNKNOWN;
		dl->enctype = ENCT_UNKNOWN;

		//return LRM_ERROR;
	}

	/* finally add the dl to the list of data links */
	listnode_add(lrm->datalink_list, dl);

	return LRM_OK;
}

int
data_link_del(lrm_t * lrm, g2mpls_addr_t dl_id)
{
	struct zlistnode * node, *node2;
	struct zlistnode * nnode, *nnode2;
	datalink_t      * dl, *dl2;
	te_link_t *tel;
	int dl_found = 0;

	node  = 0;
	nnode = 0;
	dl    = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->datalink_list, node, nnode, dl)) {
		ASSERT(dl);
		if (addr_equal(dl->lcl_id, dl_id) == 1) {
			if (dl->adm_state == ENABLED) {
				zlog_err("Cannot delete data-link %s:"
					 "administrative state is "
					 "ENABLED",
					 addr_ntoa(dl_id));
				return  LRM_ERROR;
			}
			dl_found = 1;
			break;
		}
	}

	if (dl_found) {
		//delete the datalink from all the te-links containing it
		for(ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)){
			for(ALL_LIST_ELEMENTS(tel->dl_list, node2, nnode2, dl2)){
				if(addr_equal(dl->lcl_id, dl2->lcl_id)){
					zlog_err("cannot delete datalink: used in te-link %s", tel->tel_name);
					return LRM_ERROR;
				}

			}
		}
		listnode_delete(lrm->datalink_list, dl);
		free(dl);
		return LRM_OK;
	}
	zlog_err("Data link %s not found", addr_ntoa(dl_id));
	return LRM_ERROR;

}

int
data_link_ena(lrm_t * lrm, g2mpls_addr_t dl_id)
{
	struct zlistnode * node;
	datalink_t      * dl;

	node = 0;
	dl   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS_RO(lrm->datalink_list, node, dl)) {
		ASSERT(dl);
		if (addr_equal(dl->lcl_id, dl_id) == 1) {
			dl->adm_state = ENABLED;
			return LRM_OK;
		}
	}

	zlog_err("Data link %s not found", addr_ntoa(dl_id));
	return LRM_ERROR;
}

int
data_link_up(lrm_t * lrm, g2mpls_addr_t dl_id)
{
	struct zlistnode *node;
	datalink_t       *dl;

	node = 0;
	dl   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS_RO(lrm->datalink_list, node, dl)) {
		ASSERT(dl);
		if (addr_equal(dl->lcl_id, dl_id) == 1) {
			dl->op_state = UP;
			return LRM_OK;
		}
	}
	zlog_err("Data link %s not found", addr_ntoa(dl_id));
	return LRM_ERROR;
}

int
data_link_dis(lrm_t * lrm, g2mpls_addr_t dl_id)
{
	struct zlistnode *node, *node1, *nnode1, *node2, *nnode2;
	datalink_t      *dl, *dl2;
	te_link_t *tel;

	node = 0;
	dl   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS_RO(lrm->datalink_list, node, dl)) {
		ASSERT(dl);
		if (addr_equal(dl->lcl_id, dl_id) == 1) {
			for(ALL_LIST_ELEMENTS(lrm->telink_list, node1, nnode1, tel)){
				for(ALL_LIST_ELEMENTS(tel->dl_list, node2, nnode2, dl2)){
					if(addr_equal(dl->lcl_id, dl2->lcl_id)){
						zlog_err("cannot disable datalink: used in te-link %s", tel->tel_name);
						return LRM_ERROR;
					}
				}
			}
			dl->adm_state = DISABLED;
			return LRM_OK;
		}
	}
	zlog_err("Data link %s not found", addr_ntoa(dl_id));
	return LRM_ERROR;
}

int
data_link_down(lrm_t * lrm, g2mpls_addr_t dl_id)
{
	struct zlistnode *node;
	datalink_t       *dl;

	node = 0;
	dl   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS_RO(lrm->datalink_list, node, dl)) {
		ASSERT(dl);
		if (addr_equal(dl->lcl_id, dl_id) == 1) {
			dl->op_state = DOWN;
			return LRM_OK;
		}
	}
	zlog_err("Data link %s not found", addr_ntoa(dl_id));
	return LRM_ERROR;
}

int
te_link_add(lrm_t *       lrm,
	    g2mpls_addr_t tel_id,
	    g2mpls_addr_t rem_tel_id,
	    u_int32_t     adj_rid,
	    adj_type_t    adj_type)
{
	struct zlistnode * node, *nnode;
	te_link_t       * tel;
	char              t_name[7];
	adj_t           * adj;
	adj_t           * adj_ptr = NULL;

	int adj_found = 0;

	ASSERT(lrm);

	/* check if the adjacency exists, if not add it to the adj list
	 * */
	for(ALL_LIST_ELEMENTS(lrm->adj_list, node, nnode, adj)){

		if(adj_rid == adj->adj_addr){
			adj_ptr = adj;
			adj_found = 1;
		}
	}

	if(!adj_found){
		zlog_debug("the adjacency %s does not exist already", numb_htoa(adj_rid));
		adj_ptr = XCALLOC(MTYPE_LRM_ADJ, sizeof(adj_t));
		adj_ptr->tel_list = list_new();
		adj_ptr->adj_addr = adj_rid;
		adj_ptr->link_type = adj_type;
		listnode_add(lrm->adj_list, adj_ptr);
	}
	else{
		/*the adjacency was already exists we need to check if the te-link type is
		  of the same type of the adjacency
		*/
		if(adj_type != adj_ptr->link_type){
			zlog_err("cannot add te-link to adjacency %s: incompatible types", numb_htoa(adj_rid));
			return LRM_ERROR;
		}
	}

	/* At first check if this tel exists */
	for (ALL_LIST_ELEMENTS_RO(lrm->telink_list, node, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id)) {
			zlog_err("A TE-link just exist with ID %s",
				 addr_ntoa(tel_id));
			return LRM_ERROR;
		}
	}

	/* if you are here you need to really add */
	tel = XCALLOC(MTYPE_LRM_TEL, sizeof(*tel));
	if (tel == NULL) {
		zlog_err("Cannot allocate a TE-link data structure");
		return LRM_ERROR;
	}

	memset(tel, 0, sizeof(te_link_t));

	memcpy(&(tel->lcl_id), &tel_id,     sizeof(tel_id));
	memcpy(&(tel->rem_id), &rem_tel_id, sizeof(rem_tel_id));
	/* XXX-FIXME: remeber to add adj and check rem_rid + rem_tel_id */
	tel->rem_node_id = adj_rid;

	/*set the name of the TE Link*/
	//sprintf(t_name, "tel%d", lrm->telink_list->count);
	sprintf(t_name, "tel%d", lrm->telink_count + 1);
	strncpy(tel->tel_name, t_name, strlen(t_name));
        tel->tel_name[strlen(t_name)] = '\0';

	tel->adj_type   = adj_type;
	//tel->tel_key    = lrm->telink_list->count;
	tel->tel_key    = lrm->telink_count + 1;
	tel->op_state   = DOWN;
	tel->adm_state  = DISABLED;

	tel->te_metric  = 0;
	tel->link_color = 0;
	tel->SRLG_ids   = list_new();

	/* No need to set up BW TE info, they will be retrieved upon enable */

	tel->calendar      = list_new();
	tel->calendar->del = delete_calendar_event;
	tel->calendar->cmp = compare_calendar_event;

	tel->dl_list  = list_new();

	tel->assoc_cc   = NULL;
	tel->cc_list    = list_new();
	tel->num_cc_up  = 0;

	tel->max_bw = 0;
	tel->max_res_bw = 0;
	tel->min_LSP_bw = 0;
	tel->protection_type = 0;
	tel->sync_status = NOT_SENT;
	tel->adj = adj_ptr;

	listnode_add(lrm->telink_list, tel);
	listnode_add(adj_ptr->tel_list, tel);

	lrm->telink_count++;

	return LRM_OK;
}

int
te_link_del(lrm_t * lrm, g2mpls_addr_t tel_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;
	int tel_found = 0;

	node  = 0;
	nnode = 0;
	tel   = 0;
	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			tel_found  = 1;
			break;
		}
	}

	if (tel_found) {
		ASSERT(tel);
		if (tel->adm_state == ENABLED) {
			zlog_err("Cannot delete TE-link %s, "
				 "administrative state is "
				 "still ENABLED",
				 addr_ntoa(tel_id));
			return  LRM_ERROR;
		}
		/* First notify clients*/
		struct interface *ifp = TEL2interface(tel);
		/**
		 * this notifies ospf
		 */
		lrm_interface_delete_notify(ifp);
		/*
		 * this notifies SCNGWS
		 * */
		lrm_tel_del_notify(tel);

		/* then free the list node*/
		listnode_delete(lrm->telink_list, tel);
		listnode_delete(tel->adj->tel_list, tel);
		if(tel->adj->tel_list->count == 0){
			listnode_delete(lrm->adj_list, tel->adj);
			free(tel->adj);
		}
		free(tel);
		return LRM_OK;
	}

	zlog_err("TE link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

int
te_link_ena(lrm_t * lrm, g2mpls_addr_t tel_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			tel->adm_state = ENABLED;
			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

int
te_link_up(lrm_t * lrm, g2mpls_addr_t tel_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			tel->op_state = UP;
			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

int
te_link_dis(lrm_t * lrm, g2mpls_addr_t tel_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			tel->adm_state = DISABLED;
			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

int
te_link_down(lrm_t * lrm, g2mpls_addr_t tel_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			tel->op_state = DOWN;
			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

static int
te_link_bundle_info(te_link_t * tel)
{
	ASSERT(tel);

	/* XXX-FIXME: to be implemeted just for */

	tel->op_state = UP;

	return 1;
}

/* Once the te-link run command executed, the updates made on a te-link such as addition or
 * removal of a srlg id or te metric will be immedialtly sent to osfd
 */
int te_link_run(lrm_t * lrm, g2mpls_addr_t tel_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			if ((tel->adm_state == ENABLED)    &&
			    (tel->num_cc_up != 0) && (te_link_bundle_info(tel) == 1)) {

				/*aggregate te-link information*/
				te_bundle(tel);
				/*put the te-link admin state to UP*/
				tel->op_state = UP;
				struct interface * ifp;
				/* First notify listeners */
				/*notify SCNGW of the presence of the te-link*/
				lrm_tel_add_notify(tel);
				ifp = TEL2interface(tel);
				/*notify ospf of the creation of the te-link and the corresponding interface*/
				lrm_interface_add_notify(ifp);
				return LRM_OK;
			}
			zlog_err("TE-link %s cannot be run... ",
				 addr_ntoa(tel->lcl_id));
			zlog_err("\t- op state             : %s",
				 SHOW_OPSTATE(tel->op_state));
			zlog_err("\t- adm state            : %s",
				 SHOW_ADMSTATE(tel->adm_state));
			return LRM_ERROR;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}


/*
 * This function binds a control channel specified by cc_id to a te-link specified by its local address
 */
int te_link_bind_cc(lrm_t * lrm, g2mpls_addr_t tel_id, u_int32_t cc_id)
{
	struct zlistnode * t_node;
	struct zlistnode * t_nnode;
	struct zlistnode * c_node;
	struct zlistnode * c_nnode;
	te_link_t       * tel, *tel_ptr;
	cc_t            * cc, *cc_ptr;
	int cc_found = 0;
	int tel_found = 0;


	t_node  = 0;
	t_nnode = 0;
	c_node  = 0;
	c_nnode = 0;
	tel     = 0;
	cc      = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->cc_list, c_node, c_nnode, cc)) {
		ASSERT(cc);
		if (cc->cc_id == cc_id) {
			cc_ptr = cc;
			cc_found = 1;
			break;
		}
	}

	if (!cc_found) {
		zlog_err("Control Channel %d  doesn't exist in LRM", cc_id);
		return LRM_ERROR;
	}

	for (ALL_LIST_ELEMENTS(lrm->telink_list, t_node, t_nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			tel_found = 1;
			tel_ptr = tel;
			break;
		}
	}

	if (!tel_found) {
		zlog_err("TE-link %s not found", addr_ntoa(tel_id));
		return LRM_ERROR;
	}


	for (ALL_LIST_ELEMENTS(tel->cc_list, c_node, c_nnode, cc)) {
		ASSERT(cc);
		if (cc->cc_id == cc_id) {
			zlog_err("Control Channel %d already bound to "
				 "TE-link %s",
				 cc_id, addr_ntoa(tel_id));
			return LRM_ERROR;
		}
	}





	//a cc must enabled and up to be bound to a te-link
	if ( (cc_ptr->adm_state == ENABLED) && (cc_ptr->op_state == UP) ) {
		tel_ptr->assoc_cc = cc_ptr;
		tel_ptr->num_cc_up++;
		listnode_add(cc_ptr->te_link_list, tel_ptr);
		listnode_add(tel_ptr->cc_list, cc_ptr);
	}
	else
		return LRM_ERROR;

	/*if ((tel->adm_state == ENABLED)    &&
	  (tel->num_cc_up != 0)          &&
	  (te_link_bundle_info(tel) == 1)) {

	  lrm_tel_update_notify(tel, ADD_OP);
	  }*/

	if ((tel->adm_state == ENABLED) && (tel->op_state == UP)) {
		lrm_tel_update_notify(tel, ADD_OP, cc_id);
	}
	return LRM_OK;
}

/*
 * This function unbinds a control channel specified by cc_id to a te-link specified
 * by its ip address
 */
int te_link_unbind_cc(lrm_t * lrm, g2mpls_addr_t tel_id, u_int32_t cc_id)
{
	struct zlistnode * t_node;
	struct zlistnode * t_nnode;
	struct zlistnode * c_node;
	struct zlistnode * c_nnode;
	te_link_t       * tel;
	cc_t            * cc;
	int cc_found = 0;
	int tel_found = 0;


	t_node  = 0;
	t_nnode = 0;
	c_node  = 0;
	c_nnode = 0;
	tel     = 0;
	cc      = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->cc_list, c_node, c_nnode, cc)) {
		ASSERT(cc);
		if (cc->cc_id == cc_id) {
			cc_found = 1;
			break;
		}
	}

	if (!cc_found) {
		zlog_err("Control Channel %d  doesn't exist in LRM", cc_id);
		return LRM_ERROR;
	}

	for (ALL_LIST_ELEMENTS(lrm->telink_list, t_node, t_nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			tel_found = 1;
			break;
		}
	}

	if (!tel_found) {
		zlog_err("TE-link %s not found", addr_ntoa(tel_id));
		return LRM_ERROR;
	}

	cc_found = 0;
	cc = NULL;
	for (ALL_LIST_ELEMENTS(tel->cc_list, c_node, c_nnode, cc)) {
		ASSERT(cc);
		if (cc->cc_id == cc_id) {
			cc_found = 1;
			break;
		}
	}

	if (!cc_found) {
		zlog_err("Control Channel %d not bound to "
			 "TE-link %s",
			 cc_id, addr_ntoa(tel_id)
			 );
		return LRM_ERROR;
	}

	listnode_delete(cc->te_link_list, tel);
	listnode_delete(tel->cc_list, cc);

	if ((cc->adm_state == ENABLED) && (cc->op_state == UP)) {
		if (tel->assoc_cc == cc) {

			tel->assoc_cc = NULL;
			cc = NULL;
			/* select another cc UP and ENABLED if possible*/
			for (ALL_LIST_ELEMENTS(tel->cc_list, c_node, c_nnode, cc)) {
				ASSERT(cc);
				if ((cc->adm_state == ENABLED) && (cc->op_state == UP)) {
					tel->assoc_cc = cc;
					break;
				}
			}
		}
		tel->num_cc_up--;
	}

	if ((tel->adm_state == ENABLED) &&
	    (tel->op_state == UP) /*&&
				    (te_link_bundle_info(tel) == 1) */) {
		lrm_tel_update_notify(tel, DEL_OP, cc_id);
	}
	/*if the number of cc up associated to the te-link is 0 the we put the te-link to down
	  thus next time we add a new cc we need to execute te-link run command to bring the te-link up
	  again
	  if(tel->num_cc_up == 0){
	  tel->op_state = DOWN;
	  }
	*/

	return LRM_OK;
}


/*
 * This function adds a data-link to the te-link specified by its local address
 */
int te_link_push_dl(lrm_t * lrm, g2mpls_addr_t tel_id, g2mpls_addr_t dl_id)
{
	struct zlistnode * t_node;
	struct zlistnode * t_nnode;
	struct zlistnode * d_node;
	struct zlistnode * d_nnode;
	te_link_t       * tel;
	datalink_t      * dl, *dl_ptr;
	int dl_found = 0;

	t_node  = 0;
	t_nnode = 0;
	d_node  = 0;
	d_nnode = 0;
	tel     = 0;
	dl      = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->datalink_list, d_node, d_nnode, dl)) {
		ASSERT(dl);

		if (addr_equal(dl->lcl_id, dl_id) == 1) {

			if(dl->adm_state == ENABLED){
				dl_found = 1;
				dl_ptr = dl;
				break;
			}
		}
	}

	if (!dl_found) {
		zlog_err("Data-link %s not ENABLED or doesn't exist in LRM",
			 addr_ntoa(dl_id));
		return LRM_ERROR;
	}

	for (ALL_LIST_ELEMENTS(lrm->telink_list, t_node, t_nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) != 1){
			continue;
		}
		if ((tel->dl_list->count > 0) &&
		    (tel->swcap == SWCAP_LSC)) {
			zlog_err("Trying to insert more than one data link in LSC te link: "
				 "not allowed");
			return LRM_ERROR;
		}
		for (ALL_LIST_ELEMENTS(tel->dl_list, d_node, d_nnode, dl)) {
			ASSERT(dl);
			if (addr_equal(dl->lcl_id, dl_id) == 1) {

				char *id_dl = (char *)strdup(addr_ntoa(dl_id));
				char *id_tel = (char *)strdup(addr_ntoa(tel_id));
				if( id_dl && id_tel){
					zlog_err("Data link %s is already in "
						 "TE-link %s",
						 id_dl, id_tel);
					free(id_dl);
					free(id_tel);
				}
				return LRM_ERROR;
			}
		}
		if(tel->dl_list->count == 0){
			tel->swcap = dl_ptr->swcap;
			tel->enctype = dl_ptr->enctype;

		}
		else {
			if( (dl_ptr->enctype != tel->enctype) || (dl_ptr->swcap != tel->swcap)){
				zlog_err("Trying to insert data link in te link of different "
					 "switching capability or encoding type");
				return LRM_ERROR;
			}

		}

		/*if (dl_ptr->rem_id.type == UNNUMBERED) {*/
		/* Here we correlate rem info on this DL */
		/*dl_ptr->rem_id.value.unnum.node_id = tel->rem_node_id;*/
		/*}*/

		listnode_add(tel->dl_list, dl_ptr);

		/*if (  (tel->adm_state == ENABLED)    &&
		  (tel->num_cc_up != 0)          &&
		  (te_link_bundle_info(tel) == 1)
		  )
		  {

		  struct interface * ifp;
		  lrm_tel_add_notify(tel);
		  ifp = TEL2interface(tel);
		  lrm_interface_add_notify(ifp);
		  }*/

		if (  (tel->adm_state == ENABLED)    &&
		      (tel->op_state == UP)) {

			struct interface * ifp;
			/*when a data-link is added or removed update the bandwidth parameters*/
			te_bundle(tel);
			ifp = TEL2interface(tel);
			//lrm_interface_add_notify(ifp);
			lrm_interface_update_notify(ifp, BW_UPDATE);
		}


		zlog_debug("Got TE-link %s", addr_ntoa(tel->lcl_id));
		return LRM_OK;
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}


/*
 * This function removes a data-link from a te-link specified by its local address
 */
int te_link_pop_dl(lrm_t * lrm, g2mpls_addr_t tel_id, g2mpls_addr_t dl_id)
{
	struct zlistnode * t_node;
	struct zlistnode * t_nnode;
	struct zlistnode * d_node;
	struct zlistnode * d_nnode;
	te_link_t       * tel;
	datalink_t      * dl;


	t_node  = 0;
	t_nnode = 0;
	d_node  = 0;
	d_nnode = 0;
	tel     = 0;
	dl      = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, t_node, t_nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) != 1) {
			continue;
		}

		for (ALL_LIST_ELEMENTS(tel->dl_list, d_node, d_nnode, dl)) {
			ASSERT(dl);
			if (addr_equal(dl->lcl_id, dl_id) == 1) {
				listnode_delete(tel->dl_list, dl);

				if((tel->adm_state == ENABLED)    &&
				   (tel->op_state == UP) /*&&
							   (te_link_bundle_info(tel) == 1)*/) {
					struct interface * ifp;
					te_bundle(tel);
					ifp = TEL2interface(tel);
					//lrm_interface_add_notify(ifp);
					lrm_interface_update_notify(ifp, BW_UPDATE);
				}
				return LRM_OK;
			}
		}
		return LRM_ERROR;
	}
	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

/*
 * This function sets the te-link metric
 */
int te_link_set_te_metric(lrm_t *       lrm,
			  g2mpls_addr_t tel_id,
			  u_int32_t     te_metric)
{	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {

			tel->te_metric = te_metric;

			/*if ((tel->adm_state == ENABLED)    &&
			  (tel->num_cc_up != 0)          &&
			  (te_link_bundle_info(tel) == 1)) {
			  struct interface * ifp;
			  ifp = TEL2interface(tel);
			  lrm_interface_add_notify(ifp);
			  }*/

			if ((tel->adm_state == ENABLED)    &&
			    (tel->op_state == UP) /*&&
						    (te_link_bundle_info(tel) == 1)*/) {
				struct interface * ifp;
				ifp = TEL2interface(tel);
				//lrm_interface_add_notify(ifp);
				lrm_interface_update_notify(ifp, METRIC_UPDATE);
			}

			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

/*
 * This function sets the te-link protection type
 */
int   te_link_set_link_protection(lrm_t *       lrm,
			     g2mpls_addr_t tel_id,
			     u_int8_t     protection_type)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {

			tel->protection_type = protection_type;

			/*if ((tel->adm_state == ENABLED)    &&
			  (tel->num_cc_up != 0)          &&
			  (te_link_bundle_info(tel) == 1)) {
			  struct interface * ifp;

			  lrm_tel_add_notify(tel);

			  ifp = TEL2interface(tel);
			  lrm_interface_add_notify(ifp);
			  }*/

			if ((tel->adm_state == ENABLED)    &&
			    (tel->op_state == UP)  /*        &&
							     (te_link_bundle_info(tel) == 1)*/ ) {
				struct interface * ifp;
				lrm_tel_add_notify(tel);
				ifp = TEL2interface(tel);
				//lrm_interface_add_notify(ifp);
				lrm_interface_update_notify(ifp, PROTECTION_UPDATE);
			}

			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}


/*
 * This function sets the te-link link color
 */
int te_link_set_link_color(lrm_t *        lrm,
			   g2mpls_addr_t  tel_id,
			   u_int32_t      colormask)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {

			tel->link_color = colormask;

			/*if ((tel->adm_state == ENABLED)    &&
			  (tel->num_cc_up != 0)          &&
			  (te_link_bundle_info(tel) == 1)) {
			  struct interface * ifp;

			  lrm_tel_add_notify(tel);

			  ifp = TEL2interface(tel);
			  lrm_interface_add_notify(ifp);
			  }*/

			if ((tel->adm_state == ENABLED)    &&
			    (tel->op_state == UP)  /*        &&
							     (te_link_bundle_info(tel) == 1)*/ ) {
				struct interface * ifp;
				lrm_tel_add_notify(tel);
				ifp = TEL2interface(tel);
				//lrm_interface_add_notify(ifp);
				lrm_interface_update_notify(ifp, LINK_CLR_UPDATE);
			}

			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}


/*
 * This function adds an srlg id from the list of srlg ids of a te-link
 */
int te_link_add_srlg_id(lrm_t * lrm, g2mpls_addr_t tel_id, u_int32_t SRLG_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {
			struct zlistnode * s_node;
			struct zlistnode * s_nnode;
			u_int32_t       * id;

			id = NULL;

			for (ALL_LIST_ELEMENTS(tel->SRLG_ids, s_node,
					       s_nnode, id)) {
				ASSERT(id);

				if (*id == SRLG_id) {
					zlog_err("SRLG_id %d already present "
						 "TE link %s",
						 SRLG_id, addr_ntoa(tel_id));
					return LRM_ERROR;
				}
			}

			id = XCALLOC(MTYPE_LRM_OTHER, sizeof(*id));
			if (id == NULL) {
				zlog_err("Cannot allocate a SRLG id "
					 "data structure");
				return LRM_ERROR;
			}

			*id = SRLG_id;
			listnode_add(tel->SRLG_ids, id);
			/*if ((tel->adm_state == ENABLED)    &&
			  (tel->num_cc_up != 0)          &&
			  (te_link_bundle_info(tel) == 1)) {
			  struct interface * ifp;
			  //lrm_tel_add_notify(tel);
			  ifp = TEL2interface(tel);
			  lrm_interface_add_notify(ifp);
			  }*/

			if ((tel->adm_state == ENABLED)    &&
			    (tel->op_state == UP) /*         &&
							     (te_link_bundle_info(tel) == 1)*/) {
				struct interface * ifp;
				ifp = TEL2interface(tel);
				//lrm_interface_add_notify(ifp);
				lrm_interface_update_notify(ifp, SRLGid_UPDATE);
			}

			return LRM_OK;
		}
	}
	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}


/*
 * This function removes an srlg id from the list of srlg ids of a te-link
 */
int te_link_rem_srlg_id(lrm_t * lrm, g2mpls_addr_t tel_id, u_int32_t SRLG_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t       * tel;
	int found = 0;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id) == 1) {

			struct zlistnode * s_node;
			struct zlistnode * s_nnode;
			u_int32_t       * id;

			id = NULL;

			for (ALL_LIST_ELEMENTS(tel->SRLG_ids, s_node,
					       s_nnode, id)) {
				ASSERT(id);

				if (*id == SRLG_id) {
					found = 1;
					break;
				}
			}

			if (!found) {
				zlog_err("SRLG id %d not found in TE link %s",
					 SRLG_id, addr_ntoa(tel_id) );
				return LRM_ERROR;
			}
			else{
				listnode_delete(tel->SRLG_ids, id);
				free(id);
			}

			/*if ((tel->adm_state == ENABLED)    &&
			  (tel->num_cc_up != 0)          &&
			  (te_link_bundle_info(tel) == 1)) {
			  struct interface * ifp;
			  //lrm_tel_add_notify(tel);
			  ifp = TEL2interface(tel);
			  lrm_interface_add_notify(ifp);
			  }*/

			if ((tel->adm_state == ENABLED)    &&
			    (tel->op_state == UP)  /*        &&
							     (te_link_bundle_info(tel) == 1)*/) {
				struct interface * ifp;
				ifp = TEL2interface(tel);
				//lrm_interface_add_notify(ifp);
				lrm_interface_update_notify(ifp, SRLGid_UPDATE);
			}

			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

/*
 * This function sets the TNA for a UNI/ENNI te-link
 */
int te_link_set_tna(lrm_t *       lrm,
		    g2mpls_addr_t tel_id,
		    g2mpls_addr_t tna_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t *        tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);

		if (tel->adj && tel->adj->link_type != UNI &&
		    tel->adj->link_type != ENNI) {
			continue;
		}

		if (addr_equal(tel->lcl_id, tel_id) == 1) {

			if (tel->adj->link_type == ENNI &&
			    addr_equal(tel_id, tna_id) == 0) {
				zlog_err("ENNI TE-link %s and",
					 addr_ntoa(tel_id));
				zlog_err("TNA %s mismatch: TNA not set",
					 addr_ntoa(tna_id));
				return LRM_ERROR;
			}

			memcpy(&(tel->tna_id), &tna_id, sizeof(tna_id));

			if ((tel->adm_state == ENABLED)    &&
			    (tel->op_state == UP)) {
				struct interface * ifp;
				ifp = TEL2interface(tel);
				lrm_interface_add_notify(ifp);
			}

			return LRM_OK;
		}
	}

	zlog_err("TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

/*
 * This function sets the remote RC ID for a E-NNI te-link
 */
int te_link_set_remrcid(lrm_t *       lrm,
			g2mpls_addr_t tel_id,
			g2mpls_addr_t rem_rc_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	te_link_t *        tel;

	node  = 0;
	nnode = 0;
	tel   = 0;

	ASSERT(lrm);

	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);

		if (tel->adj && tel->adj->link_type != ENNI) {
			continue;
		}

		if (addr_equal(tel->lcl_id, tel_id) == 1) {

			memcpy(&(tel->rem_rc_id),
			       &rem_rc_id,
			       sizeof(rem_rc_id));

			if ((tel->adm_state == ENABLED)    &&
			    (tel->op_state == UP)) {
				struct interface * ifp;
				ifp = TEL2interface(tel);
				lrm_interface_add_notify(ifp);
			}

			return LRM_OK;
		}
	}

	zlog_err("E-NNI TE-link %s not found", addr_ntoa(tel_id));
	return LRM_ERROR;
}

/*this function copies two advance reservation calendars*/
int adv_rsrv_calendar_copy(struct zlist * dst, struct zlist * src)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	calendar_event_t * event;
	calendar_event_t * new_event;

	list_delete_all_node(dst);

	for(ALL_LIST_ELEMENTS(src, node, nnode, event)) {
		new_event = (calendar_event_t *) malloc(sizeof(*new_event));
		if (new_event == NULL) {
			zlog_err("Cannot allocate a new calendar "
				 "event data structure");
			return 0;
		}
		new_event->time_stamp  = event->time_stamp;
		for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
			new_event->avail_bw[j] = event->avail_bw[j];
		}
		listnode_add(dst, new_event);
	}

	return 1;
}

int filter_bundle_calendar(struct zlist * dst, struct zlist * src)
{
	struct zlistnode * src_node;
	struct zlistnode * begin_node;
	calendar_event_t * src_event;
	calendar_event_t * new_event;
	uint32_t           val[MAX_BW_PRIORITIES]; //current value
	uint32_t           min[MAX_BW_PRIORITIES]; //min value
	uint32_t           pre[MAX_BW_PRIORITIES]; //previous min
	uint32_t           start_time;
	uint32_t           low_thr;                //low threshold
	uint32_t           high_thr;               //high threshold
	uint32_t           n_quant;
	uint32_t           time_quantum = LRM_DEFAULT_QUANTUM;
	uint32_t           size = LRM_DEFAULT_NUM_EVENTS;

	//initialize
	src_node  = listhead(src);
	src_event = (calendar_event_t *) listgetdata(src_node);

	start_time = src_event->time_stamp;
	low_thr    = src_event->time_stamp;
	high_thr   = start_time + time_quantum;
	for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
		pre[j] = 0xFFFFFFFF; //dummy value
		min[j] = src_event->avail_bw[j];
		val[j] = 0x0;
	}

	begin_node = listhead(src);

	for (src_node = begin_node; src_node; src_node = listnextnode(src_node)) {
		src_event = (calendar_event_t *) listgetdata(src_node);
		if (src_event->time_stamp >= high_thr) {
			n_quant = (src_event->time_stamp - start_time)/
				time_quantum;
			if (min[0] != pre[0]) {
				for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
					pre[j] = min[j];
				}
				new_event = (calendar_event_t *) malloc(sizeof(*new_event));
				if (new_event == NULL) {
					zlog_err("Cannot allocate a new calendar "
						 "event data structure");
					return 0;
				}
				new_event->time_stamp  = low_thr;
				for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
					new_event->avail_bw[j] =  min[j];
				}
				listnode_add(dst, new_event);
			}
			if (((src_event->time_stamp - high_thr) >= time_quantum) &&
			    (val[0] != pre[0])) {
				low_thr = high_thr;
				for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
					pre[j] = val[j];
				}
				new_event = (calendar_event_t *) malloc(sizeof(*new_event));
				if (new_event == NULL) {
					zlog_err("Cannot allocate a new calendar "
						 "event data structure");
					return 0;
				}
				new_event->time_stamp  = low_thr;
				for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
					new_event->avail_bw[j] =  val[j];
				}
				listnode_add(dst, new_event);
			}
			for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
				min[j] = val[j];
			}
			low_thr = start_time + n_quant*time_quantum;
			high_thr = start_time + (n_quant + 1)*time_quantum;
		}
		for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
			val[j] = src_event->avail_bw[j];
		}
		if (val[0] < min[0]) {
			for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
				min[j] = val[j];
			}
		}
		if (dst->count >= size) {
			break;
		}
	}
	//insert last event
	if (dst->count < size) {
		new_event = (calendar_event_t *) malloc(sizeof(*new_event));
		if (new_event == NULL) {
			zlog_err("Cannot allocate a new calendar "
				 "event data structure");
			return 0;
		}
		new_event->time_stamp  = low_thr;
		for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
			new_event->avail_bw[j] =  val[j];
		}
		listnode_add(dst, new_event);
	}
	return 1;
}

/*this function updates the bundle advance reservation calendar*/
int
update_bundle_calendar(struct zlist * bundle_calendar, struct zlist * dl_calendar)
{
	struct zlistnode * bundle_node;
	struct zlistnode * dl_node;
	struct zlistnode * dl_nnode;
	struct zlistnode * dl_start_node = NULL;
	struct zlistnode * bundle_pnode = NULL;
	struct zlistnode * current_bundle_node;
	struct zlistnode * next_node;
	calendar_event_t * dl_event;
	calendar_event_t * dl_nevent;
	calendar_event_t * bundle_event;
	calendar_event_t * bundle_pevent = NULL;
	calendar_event_t * new_event;
	uint32_t           prev_bw[MAX_BW_PRIORITIES];
	uint32_t           pavail_bw_0 = 0;
	float              bw;
	float              bw_0 = 0;
	int                found;

	for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
		prev_bw[j] = 0;
	}

	current_bundle_node = listhead(bundle_calendar);

	for (ALL_LIST_ELEMENTS(dl_calendar, dl_node, dl_nnode, dl_event)) {
		found = 0;
		for (bundle_node = current_bundle_node; bundle_node; bundle_node = listnextnode(bundle_node)) {
			bundle_event = (calendar_event_t *) listgetdata(bundle_node);
			bundle_pnode = bundle_node->prev;
			if (bundle_pnode) {
				bundle_pevent = listgetdata(bundle_pnode);
			}
			if (bundle_event->time_stamp > dl_event->time_stamp) {
				found = 1;
				/* process this dl_event */
				if (bundle_pnode) {
					/* is not head */
					for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
						prev_bw[j] = bundle_pevent->avail_bw[j];
					}
				}
				if (dl_node == listhead(dl_calendar)) {
					new_event = (calendar_event_t *) malloc(sizeof(*new_event));
					if (new_event == NULL) {
						zlog_err("Cannot allocate a new calendar "
							 "event data structure");
						return 0;
					}
					new_event->time_stamp  = dl_event->time_stamp;
					for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
						bw = (BW_HEX2BPS(dl_event->avail_bw[j]) +
						      BW_HEX2BPS(prev_bw[j]));
						new_event->avail_bw[j] = BW_BPS2HEX(bw);
					}
					listnode_add_after(bundle_calendar, bundle_pnode, new_event);
				} else {
					for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
						bw = (BW_HEX2BPS(dl_event->avail_bw[j]) +
						      BW_HEX2BPS(bundle_pevent->avail_bw[j]));
						bundle_pevent->avail_bw[j] = BW_BPS2HEX(bw);
					}
				}
				/* process next dl_event */
				if (dl_nnode == NULL) {
					break;
				}
				dl_nevent = (calendar_event_t *) listgetdata(dl_nnode);
				if (dl_nevent->time_stamp < bundle_event->time_stamp) {
					new_event = (calendar_event_t *) malloc(sizeof(*new_event));
					if (new_event == NULL) {
						zlog_err("Cannot allocate a new calendar "
							 "event data structure");
						return 0;
					}
					new_event->time_stamp  = dl_nevent->time_stamp;
					for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
						new_event->avail_bw[j] = prev_bw[j];
					}
					listnode_add_sort(bundle_calendar, new_event);
					current_bundle_node = bundle_node;
					break;
				} else if (dl_nevent->time_stamp != bundle_event->time_stamp) {
					next_node = bundle_node;
					bundle_pevent = bundle_event;
					while (dl_nevent->time_stamp > bundle_event->time_stamp) {
						bundle_pevent = bundle_event;
						for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
							bw = (BW_HEX2BPS(dl_event->avail_bw[j]) +
							      BW_HEX2BPS(bundle_event->avail_bw[j]));
							prev_bw[j] = bundle_event->avail_bw[j];
							bundle_event->avail_bw[j] = BW_BPS2HEX(bw);
						}
						next_node = listnextnode(next_node);
						if (!next_node) {
							break;
						}
						bundle_event = (calendar_event_t *) listgetdata(next_node);
					}
					if (dl_nevent->time_stamp != bundle_pevent->time_stamp) {
						new_event = (calendar_event_t *) malloc(sizeof(*new_event));
						if (new_event == NULL) {
							zlog_err("Cannot allocate a new calendar "
								 "event data structure");
							return 0;
						}
						new_event->time_stamp  = dl_nevent->time_stamp;
						for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
							new_event->avail_bw[j] = prev_bw[j];
						}
						if (next_node) {
							listnode_add_after(bundle_calendar, next_node->prev, new_event);
						} else {
							listnode_add(bundle_calendar, new_event);
						}
					}
					if (next_node) {
						current_bundle_node = next_node;
					} else {
						current_bundle_node = listtail(bundle_calendar);
					}
					break;
				}
				current_bundle_node = bundle_node;
				break;
			} else if (bundle_event->time_stamp == dl_event->time_stamp) {
				found = 1;
				if (bundle_node != listtail(bundle_calendar)) {
					current_bundle_node = listnextnode(bundle_node);
				} else {
					current_bundle_node = bundle_node;
				}
				for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
					bw = (BW_HEX2BPS(dl_event->avail_bw[j]) +
					      BW_HEX2BPS(bundle_event->avail_bw[j]));
					if (j == 0) {
						bw_0       = bw;
						pavail_bw_0 = bundle_pevent->avail_bw[j];
					}
					if (!bundle_pnode ||
					    (bundle_pnode && pavail_bw_0 != BW_BPS2HEX(bw_0))) {
						bundle_event->avail_bw[j] = BW_BPS2HEX(bw);
					} else {
						listnode_delete(bundle_calendar, bundle_node);
						(*bundle_calendar->del) (bundle_event);
						break;
					}
				}
				break;
			}
		}
		if (!found) {
			//no bundle_event->time_stamp >= dl_event->time_stamp
			if (bundle_pnode) {
				/* get the offset */
				for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
					prev_bw[j] = bundle_pevent->avail_bw[j];
				}
			}
			dl_start_node = dl_node;
			break;
		}
	}
	if (!found) {
		//this point is reached when dl events are after bundle events
		for (dl_node = dl_start_node; dl_node; dl_node = listnextnode(dl_node)) {
			dl_event = (calendar_event_t *) listgetdata(dl_node);
			new_event = (calendar_event_t *) malloc(sizeof(*new_event));
			if (new_event == NULL) {
				zlog_err("Cannot allocate a new calendar "
					 "event data structure");
				return 0;
			}
			new_event->time_stamp  = dl_event->time_stamp;
			for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
				bw = (BW_HEX2BPS(dl_event->avail_bw[j]) +
				      BW_HEX2BPS(prev_bw[j]));
				new_event->avail_bw[j] = BW_BPS2HEX(bw);
			}
			listnode_add(bundle_calendar, new_event);
		}
	}
	return 1;
}

/*this function copies two lambdas bitmap*/
int lambdas_bitmap_copy(wdm_link_lambdas_bitmap_t * dst, wdm_link_lambdas_bitmap_t * src)
{
	size_t size;

	if (dst->bitmap_word) {
		free(dst->bitmap_word);
	}

	memset(dst, 0, sizeof(wdm_link_lambdas_bitmap_t));

	dst->base_lambda_label = src->base_lambda_label;
	dst->num_wavelengths   = src->num_wavelengths;
	dst->bitmap_size       = src->bitmap_size;

	size = (size_t) (dst->bitmap_size * 4);
	dst->bitmap_word = (uint32_t *) malloc(size);
	memcpy(dst->bitmap_word, src->bitmap_word, size);

	return 1;
}

/*this function aggregates the traffic engineering params of the data links into the TE link*/
int te_bundle(te_link_t *tel)
{
	int i     = 0;
	int count = 1;
	int initialized = 0;
	struct zlistnode *node, *nnode;
	datalink_t *data = NULL;
	float tel_bw;
	lrm_t * lrm;
	struct zlist * bundle_calendar = NULL;
	struct zlist * tmp_calendar = NULL;

	lrm = lrm_lookup();
	if (lrm == NULL) {
		zlog_err ("There isn't an active LRM instance.");
		return 0;
	}

	//INITIALIZATION
	for(i = 0; i < MAX_BW_PRIORITIES; i++){
		tel->avail_bw_per_prio[i] = 0x0;
	}
	/*tel max_bw*/
	tel->max_bw = 0x0;
	/*tel max_res_bw*/
	tel->max_res_bw = 0x0;
	/*tel min_LSP_bw*/
	tel->min_LSP_bw = 0x0;

	for(i = 0; i < MAX_BW_PRIORITIES; i++){
		tel->max_LSP_bw[i] = 0x0;
	}
	//END INITIALIZATION

	//if no data-links associated return
	if(tel->dl_list->count == 0){
		return 0;
	}

	list_delete_all_node(tel->calendar);

	for(ALL_LIST_ELEMENTS(tel->dl_list, node, nnode, data)){
		if (data->op_state == UP)  {
			if (tel->enctype == SWCAP_UNKNOWN ||
			    tel->swcap   == SWCAP_UNKNOWN) {
				if (count == 1) {
					tel->enctype = data->enctype;
					tel->swcap   = data->swcap;
				}
			}
			if ((data->enctype != tel->enctype) ||
			    (data->swcap   != tel->swcap  )) {
				te_link_pop_dl(lrm, tel->lcl_id, data->lcl_id);
				zlog_debug("Found a data link in te link with "
					   "different switching capability or "
					   "encoding type and removed "
					   "from associated list");
				continue;
			}

			/* min_LSP_bw: max between all data links */
			if (data->min_LSP_bw > tel->min_LSP_bw) {
				tel->min_LSP_bw = data->min_LSP_bw;
			}

			/* max_LSP_bw by prio: max between all data links */
			for (i = 0; i < MAX_BW_PRIORITIES; i++) {
				if (data->max_LSP_bw[i] > tel->max_LSP_bw[i]) {
					tel->max_LSP_bw[i] =
						data->max_LSP_bw[i];
				}
			}

			/* max_bw: max_LSP_bw[7] */
			tel->max_bw = tel->max_LSP_bw[7];

			/* max_res_bw: sum of components' max_res_bws */
			tel_bw = (BW_HEX2BPS(tel->max_res_bw) +
				  BW_HEX2BPS(data->max_res_bw));

			tel->max_res_bw = BW_BPS2HEX(tel_bw);

			/* avail_bw_per_prio: sum of components' avail_bws */
			for (i = 0; i < MAX_BW_PRIORITIES; i++) {

				tel_bw = (BW_HEX2BPS(tel->avail_bw_per_prio[i]) +
					 BW_HEX2BPS(data->avail_bw_per_prio[i]));

				tel->avail_bw_per_prio[i] = BW_BPS2HEX(tel_bw);

			}

			if (tel->swcap != SWCAP_LSC) {
				if (!initialized) {
					/* create bundle tel advance
					   reservation calendar */
					bundle_calendar      = list_new();
					bundle_calendar->del = delete_calendar_event;
					bundle_calendar->cmp = compare_calendar_event;
					initialized = 1;
				}
				if (data->calendar->count) {
					if (bundle_calendar->count) {
						/* update bundle calendar */
						if (!update_bundle_calendar(bundle_calendar, data->calendar)) {
							zlog_err("Cannot update bundle advance "
								 "reservation calendar");
							list_free(bundle_calendar);
							return 0;
						}
					} else {
						/* initailize bundle calendar */
						if (!adv_rsrv_calendar_copy(bundle_calendar, data->calendar)) {
							zlog_debug("Cannot copy advance reservation "
								   "calendar in LSC te link bundle");
							list_free(bundle_calendar);
							return 0;
						}
					}
				} else {
					struct zlistnode * bundle_node;
					struct zlistnode * bundle_nnode;
					calendar_event_t * bundle_event;
					float              bw;
					for (ALL_LIST_ELEMENTS(bundle_calendar,
							       bundle_node,
							       bundle_nnode,
							       bundle_event)) {
						for (size_t j = 0; j < MAX_BW_PRIORITIES; j++) {
							bw = (BW_HEX2BPS(bundle_event->avail_bw[j]) +
							      BW_HEX2BPS(data->max_res_bw));
							bundle_event->avail_bw[j] = BW_BPS2HEX(bw);
						}
					}
				}
			}
		}
		count++;
	}
	if ((tel->swcap == SWCAP_LSC) && (data != NULL)) {
		/*data is the pointer to the unique data link in te link*/
		if (!lambdas_bitmap_copy(&(tel->lambdas_bitmap), &(data->lambdas_bitmap))) {
			zlog_err("Cannot copy lambdas bitmap in LSC te link bundle");
			return 0;
		}
		if (!adv_rsrv_calendar_copy(tel->calendar, data->calendar)) {
			zlog_err("Cannot copy advance reservation calendar in LSC te link bundle");
			return 0;
		}
	} else if (tel->swcap != SWCAP_LSC) {
		if (bundle_calendar == NULL) {
			return 1;
		} else if (!bundle_calendar->count) {
			list_free(bundle_calendar);
			return 1;
		}
		tmp_calendar      = list_new();
		tmp_calendar->del = delete_calendar_event;
		tmp_calendar->cmp = compare_calendar_event;
		if (!filter_bundle_calendar(tmp_calendar, bundle_calendar)) {
			zlog_err("Cannot copy advance reservation calendar in te link bundle");
			list_free(tmp_calendar);
			list_free(bundle_calendar);
			return 0;
		}
		list_free(bundle_calendar);
		if (!adv_rsrv_calendar_copy(tel->calendar, tmp_calendar)) {
			zlog_err("Cannot copy advance reservation calendar in te link bundle");
			list_free(tmp_calendar);
			return 0;
		}
		list_free(tmp_calendar);
	}

	return 1;
}

int adj_add(lrm_t *lrm, u_int32_t adj_id){


	struct zlistnode * node;
	struct zlistnode * nnode;
	adj_t *adj;

	node    = 0;
	nnode   = 0;

	ASSERT(lrm);

	/* At first check if this adjacency exists */
	for (ALL_LIST_ELEMENTS(lrm->adj_list, node, nnode, adj)) {
		ASSERT(adj);
		if (adj->adj_id == adj_id) {
			zlog_err("An adjacency already exist with ID %d",
				 adj_id);
			return LRM_ERROR;
		}
	}

	adj = XCALLOC(MTYPE_LRM_ADJ, sizeof(adj_t));
	if (adj == NULL) {
		zlog_err("Cannot allocate a new adjacency data structure");
		return LRM_ERROR;
	}

	memset(adj, 0, sizeof(adj_t));

	adj->adj_id     = adj_id;

	adj->tel_list = list_new();

	listnode_add(lrm->adj_list, adj);

	return LRM_OK;
}

int adj_delete(lrm_t *lrm, u_int32_t adj_id){

	struct zlistnode * node;
	struct zlistnode * nnode;
	adj_t *adj;
	int found_adj = 0;

	node    = 0;
	nnode   = 0;

	ASSERT(lrm);

	/* At first check if this adjacency exists */
	for (ALL_LIST_ELEMENTS(lrm->adj_list, node, nnode, adj)) {
		ASSERT(adj);
		if (adj->adj_id == adj_id) {
			found_adj = 1;
			break;
		}
	}

	if(found_adj){

		listnode_delete(lrm->adj_list, adj);
		free(adj);
		return LRM_OK;
	}
	else
		return LRM_ERROR;
}

int
adj_add_tel(lrm_t *lrm, u_int32_t adj_id, g2mpls_addr_t tel_id)
{
	struct zlistnode * node;
	struct zlistnode * nnode;
	adj_t* adj = NULL;
	adj_t* adj_found;
	te_link_t* tel;
	te_link_t *tel_found;
	int found_adj = 0;
	int found_tel = 0;

	node    = 0;
	nnode   = 0;

	ASSERT(lrm);




	for (ALL_LIST_ELEMENTS(lrm->adj_list, node, nnode, adj)) {
		ASSERT(adj);
		if (adj->adj_id == adj_id) {
			adj_found = adj;
			found_adj = 1;
			break;
		}
	}

	if(!found_adj){
		zlog_err("error while adding te-link %s to adj %d: no such adjacency",
			 addr_ntoa(tel_id),
			 adj->adj_id);
		return LRM_ERROR;
	}


	/*check if the te-link exists in telink_list */
	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);

		if (addr_equal(tel->lcl_id, tel_id)) {
			found_tel = 1;
			tel_found = tel;
			break;
		}

	}

	if(!found_tel){
		zlog_err("error while adding te-link %s to adj %d: no such te-link",
			 addr_ntoa(tel_id),
			 adj->adj_id);
		return LRM_ERROR;
	}

	/*check if the te-link is already in the adj list of te-links*/
	for (ALL_LIST_ELEMENTS(adj_found->tel_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id)) {
			zlog_err("cannot add te-link %s to adj %d: te-link already exists",
				 addr_ntoa(tel_id),
				 adj->adj_id);
			return LRM_ERROR;
		}
	}

	listnode_add(adj_found->tel_list, tel_found);
	return LRM_OK;

}

int
adj_rem_tel(lrm_t * lrm, u_int32_t adj_id, g2mpls_addr_t tel_id)
{



	struct zlistnode * node;
	struct zlistnode * nnode;
	adj_t* adj = NULL;
	adj_t* adj_found;
	te_link_t* tel;
	te_link_t *tel_found;
	int found_adj = 0;
	int found_tel = 0;

	node    = 0;
	nnode   = 0;

	ASSERT(lrm);




	for (ALL_LIST_ELEMENTS(lrm->adj_list, node, nnode, adj)) {
		ASSERT(adj);
		if (adj->adj_id == adj_id) {
			adj_found = adj;
			found_adj = 1;
			break;
		}
	}

	if(!found_adj){
		zlog_err("error while removing te-link %s to adj %d: no such adjacency",
			 addr_ntoa(tel_id),
			 adj->adj_id);
		return LRM_ERROR;
	}


	/*check if the te-link exists in telink_list */
	for (ALL_LIST_ELEMENTS(lrm->telink_list, node, nnode, tel)) {
		ASSERT(tel);

		if (addr_equal(tel->lcl_id, tel_id)) {
			found_tel = 1;
			tel_found = tel;
			break;
		}

	}

	if(!found_tel){
		zlog_err("error while removing te-link %s to adj %d: no such te-link",
			 addr_ntoa(tel_id),
			 adj->adj_id);
		return LRM_ERROR;
	}

	/*check if the te-link is already in the adj list of te-links*/
	for (ALL_LIST_ELEMENTS(adj_found->tel_list, node, nnode, tel)) {
		ASSERT(tel);
		if (addr_equal(tel->lcl_id, tel_id)) {
			listnode_delete(adj_found->tel_list, tel);
			zlog_err("te-link %s deleted from adjacency %d", addr_ntoa(tel_id),
				 adj->adj_id);
			return LRM_OK;
		}
	}

	return LRM_ERROR;
}
