/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#ifndef LRM_CORE_H
#define LRM_CORE_H

#include <sys/types.h>
#include <netinet/in.h>
#include "linklist.h"
#include "g2mpls_addr.h"
#include "g2mpls_types.h"

/* general purpose defines */
#define LRM_ERROR       0
#define LRM_OK          1
#define LRM_WARN        2
#define MAX_INT		4294967296
#define SENT		1
#define NOT_SENT		2
//advance reservation calendar utilities
#define LRM_DEFAULT_QUANTUM    60 //SECONDS
#define LRM_DEFAULT_NUM_EVENTS 10

#define ASSERT(X) {							 \
	if (!(X)) {							 \
	      zlog_err("ASSERTION failed in %s:%d", __FILE__, __LINE__); \
	}								 \
	assert(X);							 \
}


#define BW_UPDATE         0
#define SRLGid_UPDATE     1
#define METRIC_UPDATE     2
#define LINK_CLR_UPDATE   3
#define TNA_UPDATE        4
#define PROTECTION_UPDATE 5

/*
 * LRM instance structure
*/
typedef struct lrm {
	u_int32_t      router_id;
	u_int32_t      router_enni_id;
	u_int32_t      router_uni_id;
	struct zlist * scn_if_list;
	struct zlist * cc_list;
	struct zlist * datalink_list;
	struct zlist * telink_list;
	struct zlist * adj_list;
	uint32_t       telink_count;
} lrm_t;

/*
 * CONTROL INTERFACE
 */
typedef struct ctrl_interface {
        g2mpls_addr_t     addr;
        if_type_t         type;
        opstate_t         op_state;
        admstate_t        adm_state;
	int               sync_status;
} ctrl_intf_t;

/*
 * CONTROL CHANNEL
 */
typedef struct control_channel {
        u_int32_t        cc_id;              /* local  and node-unique CC id */
        u_int32_t        rem_cc_id;          /* remote and node-unique CC id */
        g2mpls_addr_t    lcl_scn_addr;
        g2mpls_addr_t    rem_scn_addr;
        opstate_t        op_state;
        admstate_t       adm_state;
        struct zlist   * te_link_list;       /*  te-links managed by this CC */
	int              sync_status;
} cc_t;

/*
 * DATA LINK
 */
typedef struct datalink {
        g2mpls_addr_t             lcl_id;
        g2mpls_addr_t             rem_id;
        opstate_t                 op_state;
        admstate_t                adm_state;
        sw_cap_t                  swcap;  /* switching capability */
        enc_type_t                enctype;/* encoding type of this data link */
        u_int32_t                 max_bw;
        u_int32_t                 max_res_bw;
        u_int32_t                 avail_bw_per_prio[MAX_BW_PRIORITIES];
        u_int32_t                 max_LSP_bw[MAX_BW_PRIORITIES];
        u_int32_t                 min_LSP_bw;
	wdm_link_lambdas_bitmap_t lambdas_bitmap;
	struct zlist *            calendar;/* advance reservation calendar */
} datalink_t;

typedef struct adj {
	u_int32_t       adj_id;
	u_int32_t       adj_addr;
	adj_type_t      link_type;
	struct zlist  * tel_list;
} adj_t;


/*
 * TE LINK
 */
typedef struct _te_link {
        g2mpls_addr_t             lcl_id;
        g2mpls_addr_t             rem_id;
        u_int32_t                 rem_node_id; /* rem_id MUST be contained */
        char                      tel_name[20 + 1]; /* name of TEL */
        int                       tel_key;    /* used for internal purposes */
        opstate_t                 op_state;
        admstate_t                adm_state;

        g2mpls_addr_t             tna_id;    /* TNA, meaningfull just for UNI TElinks */


        g2mpls_addr_t             rem_rc_id; /* remote RC ID, i.e. rem domain ctrl,
				                meaningfull just for E-NNI TE links */

        /* Summary (after bundling) or configured TE info */
        adj_type_t                adj_type;
        u_int32_t                 te_metric;
        u_int32_t                 link_color;
        struct zlist *            SRLG_ids;

        sw_cap_t                  swcap;      /* switching capability */
        enc_type_t                enctype;    /* the encoding type of this data link */
        u_int32_t                 max_bw;
        u_int32_t                 max_res_bw;
        u_int32_t                 avail_bw_per_prio[MAX_BW_PRIORITIES];
	/* unreserved bw per priority (sum of unreserved bw per priority
	 * of componenet links
	 */
        u_int32_t                 max_LSP_bw[MAX_BW_PRIORITIES];
	/*max of max LSP per priority p bw of component links*/
        u_int32_t                 min_LSP_bw;
        u_int8_t                  protection_type;


	wdm_link_lambdas_bitmap_t lambdas_bitmap;

	struct zlist *            calendar;        /* advance reservation calendar */

        struct zlist *            dl_list;         /* list of data links into te-link */
	adj_t *                   adj;
        cc_t *                    assoc_cc;        /* cc associated whith this te-link */
        struct zlist *            cc_list;         /* list of CCs for this TEL */
        u_int32_t                 num_cc_up;       /* number of available CCs in up */
	int                       sync_status;
} te_link_t;

/*
 * RELATED FUNCTIONS
 */

/* LRM instance related functions */
#ifdef __cplusplus
extern "C" {
#endif
lrm_t *  lrm_lookup(void);
#ifdef __cplusplus
}
#endif
lrm_t *  lrm_get(void);
void     lrm_finish(lrm_t * lrm);
int      lrm_set_rid(lrm_t * lrm, adj_type_t adjacency, u_int32_t rid);
int      lambdas_bitmap_copy(wdm_link_lambdas_bitmap_t * dst, wdm_link_lambdas_bitmap_t * src);
int      adv_rsrv_calendar_copy(struct zlist * dst, struct zlist * src);

/* CTRL IF related functions */
int   scn_if_add(lrm_t * lrm, g2mpls_addr_t addr, if_type_t intf_type);
int   scn_if_del(lrm_t * lrm, g2mpls_addr_t addr);
int   scn_if_ena(lrm_t * lrm, g2mpls_addr_t addr);
int   scn_if_dis(lrm_t * lrm, g2mpls_addr_t addr);


/* CC related functions */
int   control_channel_add(lrm_t *            lrm,
			  u_int32_t          cc_id,
			  g2mpls_addr_t      lcl_scn,
			  g2mpls_addr_t      rem_scn);
int   control_channel_del(lrm_t * lrm, u_int32_t cc_id);
int   control_channel_ena(lrm_t * lrm, u_int32_t cc_id);
int   control_channel_dis(lrm_t * lrm, u_int32_t cc_id);
int   control_channel_up(lrm_t * lrm, u_int32_t cc_id);   /* static-LMP */
int   control_channel_down(lrm_t * lrm, u_int32_t cc_id); /* static-LMP */


/* DATA LINK related functions */
int   data_link_add (lrm_t * lrm, g2mpls_addr_t dl_id, g2mpls_addr_t rem_dl_id);
int   data_link_del (lrm_t * lrm, g2mpls_addr_t dl_id);
int   data_link_ena (lrm_t * lrm, g2mpls_addr_t dl_id);
int   data_link_up  (lrm_t * lrm, g2mpls_addr_t dl_id);
int   data_link_dis (lrm_t * lrm, g2mpls_addr_t dl_id);
int   data_link_down(lrm_t * lrm, g2mpls_addr_t dl_id);


/* TE-LINK related functions */
int   te_link_add(lrm_t *       lrm,
		  g2mpls_addr_t tel_id,
		  g2mpls_addr_t r_tel_id,
		  u_int32_t     adj_rid,
		  adj_type_t    adj_type);
int   te_link_del (lrm_t * lrm, g2mpls_addr_t tel_id);
int   te_link_ena (lrm_t * lrm, g2mpls_addr_t tel_id);
int   te_link_up  (lrm_t * lrm, g2mpls_addr_t tel_id);
int   te_link_dis (lrm_t * lrm, g2mpls_addr_t tel_id);
int   te_link_down(lrm_t * lrm, g2mpls_addr_t tel_id);
int   te_link_run (lrm_t * lrm, g2mpls_addr_t tel_id);

int   te_link_bind_cc(lrm_t * lrm, g2mpls_addr_t tel_id, u_int32_t cc_id);
int   te_link_unbind_cc(lrm_t * lrm, g2mpls_addr_t tel_id, u_int32_t cc_id);

int   te_link_push_dl(lrm_t * lrm, g2mpls_addr_t tel_id, g2mpls_addr_t dl_id);
int   te_link_pop_dl(lrm_t * lrm, g2mpls_addr_t tel_id, g2mpls_addr_t dl_id);

int   te_link_set_te_metric(lrm_t *       lrm,
			    g2mpls_addr_t tel_id,
			    u_int32_t     te_metric);
int   te_link_set_link_color(lrm_t *       lrm,
			     g2mpls_addr_t tel_id,
			     u_int32_t     colotmask);

int   te_link_set_link_protection(lrm_t *       lrm,
			     g2mpls_addr_t tel_id,
			     u_int8_t     protection_type);

int   te_link_add_srlg_id(lrm_t *        lrm,
			  g2mpls_addr_t  tel_id,
			  u_int32_t      SRLG_id);
int   te_link_rem_srlg_id(lrm_t *        lrm,
			  g2mpls_addr_t  tel_id,
			  u_int32_t      SRLG_id);
int   te_link_set_tna(lrm_t *        lrm,
		      g2mpls_addr_t  tel_id,
		      g2mpls_addr_t  tna_id);
int   te_link_set_remrcid(lrm_t *        lrm,
			  g2mpls_addr_t  tel_id,
			  g2mpls_addr_t  rem_rc_id);

int   data_link_retrieve_info_from_tnrc(datalink_t * dl);

int   te_bundle(te_link_t *tel);

int   adj_add(lrm_t* lrm, u_int32_t adj_id);
int   adj_delete(lrm_t* lrm, u_int32_t adj_id);
int   adj_add_tel(lrm_t* lrm, u_int32_t adj, g2mpls_addr_t tel_id);
int   adj_rem_tel(lrm_t* lrm, u_int32_t adj, g2mpls_addr_t tel_id);



#endif /* LRM_CORE_H */
