/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "log.h"
#include "version.h"
#include <getopt.h>
#include "command.h"
#include "thread.h"
#include "lrmd.h"
#include "lrm_zebra.h"
#include "sigevent.h"
#include "privs.h"
#include "memory.h"
#include "filter.h"
#include "prefix.h"
#include "plist.h"
#include "lrm_vty.h"
#include "lrm_core.h"
#include "zserv.h"
#include "lrm_corba.h"
#include "lib/corba.h"

/*
 * GLOBALS
 */

/* lrmd privileges */
zebra_capabilities_t _caps_p [] = {
	ZCAP_NET_RAW,
	ZCAP_BIND,
	ZCAP_NET_ADMIN,
};

struct zebra_privs_t lrmd_privs = {
#if defined(QUAGGA_USER) && defined(QUAGGA_GROUP)
	.user      = QUAGGA_USER,
	.group     = QUAGGA_GROUP,
#endif
#if defined(VTY_GROUP)
	.vty_group = VTY_GROUP,
#endif
	.caps_p    = _caps_p,
	.cap_num_p = sizeof(_caps_p)/sizeof(_caps_p[0]),
	.cap_num_i = 0
};

char config_default[] = LRMD_DEFAULT_CONFIG;

struct option longopts[] = {
	{ "daemon",      no_argument,       NULL, 'd'},
	{ "config_file", required_argument, NULL, 'f'},
	{ "pid_file",    required_argument, NULL, 'i'},
	{ "iors_dir",    required_argument, NULL, 'o'},
	{ "dryrun",      no_argument,       NULL, 'C'},
	{ "vty_port",    required_argument, NULL, 'P'},
	{ "user",        required_argument, NULL, 'u'},
	{ "group",       required_argument, NULL, 'g'},
	{ "version",     no_argument,       NULL, 'v'},
	{ "help",        no_argument,       NULL, 'h'},
	{ 0 }
};


char * progname;

/* Process ID saved for use by init system */
const char * pid_file = PATH_LRMD_PID;

struct thread_master * master;
/* needed by the thread implementation */

struct zebra_t zebrad;

static void
sighup (void)
{
	zlog(NULL, LOG_INFO, "SIGHUP received");
}

static int terminate_execution = 0;

/* SIGINT / SIGTERM handler. */
static void
sigint (void)
{
	zlog_notice("Terminating on signal");
	/* lrm_terminate();*/

	terminate_execution = 1;
}

/* SIGUSR1 handler. */
static void
sigusr1 (void)
{
	zlog_rotate (NULL);
}

struct quagga_signal_t lrm_signals[] = {
	{
		.signal = SIGHUP,
		.handler = &sighup,
	},
	{
		.signal = SIGUSR1,
		.handler = &sigusr1,
	},
	{
		.signal = SIGINT,
		.handler = &sigint,
	},
	{
		.signal = SIGTERM,
		.handler = &sigint,
	},
};

/* Help information display. */
static void __attribute__ ((noreturn))
     usage (char *progname, int status)
{
	if (status != 0) {
		fprintf (stderr,
			 "Try `%s --help' for more information.\n",
			 progname);
	} else {
		printf ("Usage : %s [OPTION...]\n"
			"Daemon which manages LRM.\n\n"
			"-d, --daemon       Runs in daemon mode\n"
			"-f, --config_file  Set configuration file name\n"
			"-i, --pid_file     Set process identifier file name\n"
			"-o, --iors_dir     Set IORs directory\n"
			"-C, --dryrun       Check configuration for validity and exit\n"
			"-P, --vty_port     Set vty's port number\n"
			"-u, --user         User to run as\n"
			"-g, --group        Group to run as\n"
			"-v, --version      Print program version\n"
			"-h, --help         Display this help and exit\n"
			"\n"
			"Report bugs to %s\n",
			progname, ZEBRA_BUG_ADDRESS);
	}

	exit(status);
}

int main(int argc, char ** argv)
{
	char          * p;
	int             vty_port;
	char          * vty_addr;
	int             daemon_mode;
	char          * config_file;
	char          * iors_dir;
	struct thread   thread;
	int             dryrun;

	vty_port    = LRMD_VTY_PORT;
	vty_addr    = NULL;
	daemon_mode = 0;
	config_file = NULL;
	iors_dir    = NULL;
	dryrun      = 0;

	umask(0027);

	progname = ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);

	/*
	 * initialize the log subsystem, you will have to include
	 * ZLOG_ZAP in the zlog_proto_t enum type definition in
	 * lib/log.h
	 */
	zlog_default = openzlog(progname,
				ZLOG_LRM,
				LOG_CONS | LOG_NDELAY | LOG_PID,
				LOG_DAEMON);


     	/* this while just reads the options */
	while (1) {
		int    opt;

		opt = getopt_long(argc, argv, "dChvf:P:u:g:i:o:", longopts, 0);

		if (opt == EOF) {
			break;
		}

		switch (opt) {
			case 0:
				break;
			case 'd':
				daemon_mode = 1;
				break;
			case 'f':
				config_file = optarg;
				break;
			case 'i':
				pid_file = optarg;
				break;
			case 'o':
				iors_dir = optarg;
				break;
			case 'C':
				dryrun = 1;
				break;
			case 'P':
				vty_port = atoi(optarg);
				break;
			case 'u':
				lrmd_privs.user = optarg;
				break;
			case 'g':
				lrmd_privs.group = optarg;
				break;
			case 'v':
				printf("LRM daemon v.%s: vty@%d\n",
                                       LRM_CODE_VERSION, vty_port);
                                printf("This program is part of "
                                       "phosphorus-g2mpls\n");
				//print_version(progname);
				exit (0);
				break;
			case 'h':
				usage(progname, 0);
				break;
			default:
				usage(progname, 1);
				break;
		}
	}

	/*initialization of the LRM data structure*/
	lrm_master_init();

	/*this the main thread controlling structure*/
	master = lrm_m->master;

	zebra_init();

        zprivs_init(&lrmd_privs);
	signal_init(master, Q_SIGC(lrm_signals), lrm_signals);

	/*
	 * Library inits.
	 */


	/* initializes the command sub-system */
	cmd_init(1);

	/* these are all from libzebra */
	vty_init(master);
	memory_init();
	access_list_init();
	prefix_list_init();

	/*
	 * LRM inits
	 */

	//lrm_serv_init();
	lrm_init();
	lrm_if_init();
	/* lrm_zebra_init (); */

	/* LRM vty inits. */
	lrm_vty_init();

	/* sorting is needed by the command subsystem to finalize init */
	sort_node();

	/* Don't start execution if we are in dry-run mode */
	if (dryrun)
		exit(0);

	/* Change to the daemon program. */
	if (daemon_mode) {
		daemon(0, 0);
	}

#if HAVE_OMNIORB
	if (!corba_init(iors_dir, CORBA_SERVANT_LRM, master)) {
		fprintf(stderr, "Cannot initialize CORBA subsystem\n");
		exit(1);
	}
	if (!corba_server_setup()) {
		fprintf(stderr, "Cannot setup CORBA subsystem\n");
		exit(1);
	}
	if (!corba_tnrc_client_setup()) {
		fprintf(stderr, "Cannot setup TNRC client object\n");
		exit(-1);
	}

	/*
	 * SCNGW client setup if and only if SCNGW is up
	 *   -> moved into scngw_isup CORBA method
	 */
	/*if (!corba_scngw_client_setup()) {
		zlog_err("Cannot setup client object");
		exit(1);
	}*/
#endif

	/* Get configuration file. */
	vty_read_config(config_file, config_default);

#if HAVE_OMNIORB
	if (!corba_lrm_isup()) {
		fprintf(stderr, "Cannot setup communication with TNRC\n");
		exit(-1);
	}
#endif
	/* Create VTY socket
	 * start the TCP and unix socket listeners
	 */
	vty_serv_sock(vty_addr,vty_port,LRM_VTYSH_PATH);

	/* Process id file create. */
	pid_output(pid_file);

	/* Print banner. */
	zlog_notice("LRMd %s starting: vty@%d", QUAGGA_VERSION, vty_port);

	while (!terminate_execution) {
		if (thread_fetch(master, &thread)) {
			thread_call(&thread);
		}
	}

	zlog_debug("Termination in progress ...");

#if HAVE_OMNIORB
	(void) corba_scngw_client_shutdown();
	(void) corba_tnrc_client_shutdown();
	(void) corba_server_shutdown();
	(void) corba_fini();
#endif

	exit(0);
}
