/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "network.h"
#include "prefix.h"
#include "stream.h"
#include "zclient.h"
#include "log.h"
#include "linklist.h"
#include "lrmd.h"
#include "lrm_zebra.h"

/* Zebra structure to hold current status. */
struct zclient * zclient = NULL;

/* Inteface addition message from zebra. */
static int
lrm_interface_add (int              command, 
		   struct zclient * zclient, 
		   zebra_size_t     length)
{
	struct interface * ifp;
  
	ifp = zebra_interface_add_read(zclient->ibuf);

	listnode_add(lrmiflist, ifp);

	zlog_debug ("Zebra interface add %s - "
		    "index %d flags %llu metric %d mtu %d",
		    ifp->name, ifp->ifindex, ifp->flags, 
		    ifp->metric, ifp->mtu);

	return 0;
}

static int
lrm_interface_delete (int              command, 
		      struct zclient * zclient, 
		      zebra_size_t     length)
{
	struct interface * ifp;
	struct stream    * s;

	s = zclient->ibuf;
	
	/* update interface structure in iflist */
	ifp = zebra_interface_state_read(s);
	
	if (ifp == NULL) {
		return 0;
	}

	if (if_is_up(ifp)) {
		zlog_warn ("Zebra: got delete of %s, "
			   "but interface is still UP",
			   ifp->name);
	}
		
	zlog_debug ("Zebra interface delete %s - "
		    "index %d flags %llu metric %d mtu %d",
		    ifp->name, ifp->ifindex, ifp->flags, 
		    ifp->metric, ifp->mtu);
	
	ifp->ifindex = IFINDEX_INTERNAL;
	
	listnode_delete(lrmiflist, ifp);
	
	return 0;
}

static struct interface *
zebra_interface_if_lookup(struct stream * s)
{
  char ifname_tmp[INTERFACE_NAMSIZ];

  /* Read interface name. */
  stream_get(ifname_tmp, s, INTERFACE_NAMSIZ);

  /* And look it up. */
  return if_lookup_by_name_len(ifname_tmp,
			       strnlen(ifname_tmp, INTERFACE_NAMSIZ));
}

static int
lrm_interface_state_up(int              command, 
		       struct zclient * zclient,
			zebra_size_t    length)
{
	struct interface * ifp;
	
	ifp = zebra_interface_if_lookup(zclient->ibuf);
	
	if (ifp == NULL) {
		return 0;
	}

       	/* Interface is already up. */
	if (if_is_operative(ifp)) {

		/* Temporarily keep ifp values. */
		struct interface if_tmp;
		memcpy (&if_tmp, ifp, sizeof (struct interface));

		zebra_interface_if_set_value(zclient->ibuf, ifp);

		zlog_debug ("Zebra: Interface[%s] state update.", 
			    ifp->name);
		
		if (if_tmp.bandwidth != ifp->bandwidth) { 
			zlog_debug ("Zebra: Interface[%s] "
				    "bandwidth change %d -> %d.",
				    ifp->name, if_tmp.bandwidth, 
				    ifp->bandwidth);		
       		}	
      
		if (if_tmp.mtu != ifp->mtu) {
			zlog_debug ("Zebra: Interface[%s] "
				    "MTU change %u -> %u.",
				    ifp->name, if_tmp.mtu, ifp->mtu);
		}

		return 0;
	}

	zebra_interface_if_set_value(zclient->ibuf, ifp);
	
	zlog_debug ("Zebra: Interface[%s] state change to up.", 
		    ifp->name);

	return 0;
}

static int
lrm_interface_state_down(int              command, 
			 struct zclient * zclient,
			 zebra_size_t     length)
{
	struct interface * ifp;

	ifp = zebra_interface_state_read(zclient->ibuf);
	
        if (ifp == NULL) {
		return 0;
	}

	zlog_debug ("Zebra: Interface[%s] state change to down.", 
		    ifp->name);

	return 0;
}

void
lrm_zebra_init ()
{
	/* Allocate zebra structure. */
	zclient = zclient_new ();

	zclient_init(zclient, ZEBRA_ROUTE_LRM, 0);

	zclient->interface_add    = lrm_interface_add;
	zclient->interface_delete = lrm_interface_delete;
	zclient->interface_up     = lrm_interface_state_up;
	zclient->interface_down   = lrm_interface_state_down;
}
