//
//  This file is part of phosphorus-g2mpls.
//
//  Copyright (C) 2006, 2007, 2008, 2009 PSNC
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along
//  with this program; if not, write to the Free Software Foundation, Inc.,
//  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//  Authors:
//
//  Adam Kaliszan         (PSNC)             <kaliszan_at_man.poznan.pl>
//  Damian Parniewicz     (PSNC)             <damianp_at_man.poznan.pl>
//  Lukasz Lopatowski     (PSNC)             <llopat_at_man.poznan.pl>
//  Jakub Gutkowski       (PSNC)             <jgutkow_at_man.poznan.pl>
//

#if HAVE_OMNIORB

#include <zebra.h>
#include "log.h"
#include "g2mpls_addr.h"
#include "g2mpls_types.h"
#include "ospfd/ospf_corba.h"
#include "idl/g2mplsTypes.hh"


typedef enum {
  OCD_ACTION_SENT,
  OCD_ACTION_ABORTED
} ocd_action_t;

typedef enum {
  OCD_SERV_G2PCERA,
  OCD_SERV_GUNIGW
} ocd_server_t;

typedef enum {
  OCD_TYPE_GRID,
  OCD_TYPE_TNA,
  OCD_TYPE_TELINK
} ocd_type_t;

void OSPF_CORBA_DEBUG(ocd_type_t type, ocd_action_t action, ocd_server_t server, string msgStr,
                      g2mplsTypes::tnaIdent tnaIdent, g2mplsTypes::teLinkIdent linkIdent);

g2mplsTypes::nodeType                   nodeType_from_node_type_t               (node_type_t type);
g2mplsTypes::gridSubNodeType            gridSubNodeType_from_grid_s_node_type_t (grid_s_node_type_t type);
g2mplsTypes::gridServiceType            service_type_from_uint                  (uint16_t value);
g2mplsTypes::gridServiceState           service_state_from_uint                 (uint8_t value);
g2mplsTypes::gridLrmsType               lrms_type_from_uint                     (uint16_t value);
g2mplsTypes::gridCeSeState              cese_state_from_uint                    (uint8_t value);
g2mplsTypes::gridCpuArch                cpu_arch_from_uint                      (uint8_t value);
g2mplsTypes::gridOsType                 os_type_from_uint                       (uint16_t value);
g2mplsTypes::gridApplicationType        appl_type_from_uint                     (uint16_t value);
g2mplsTypes::gridStorageArch            storage_arch_from_uint                  (uint8_t value);
g2mplsTypes::gridStorageRetentionPolicy storage_retpol_from_uint                (uint8_t value);
g2mplsTypes::gridStorageAccessLatency   storage_acclat_from_uint                (uint8_t value);
g2mplsTypes::gridStorageExpirationMode  storage_expmode_from_uint               (uint8_t value);
g2mplsTypes::linkMode                   linkMode_from_link_mode_t               (link_mode_t mode);

g2mplsTypes::switchingCap               switching_cap_from_uchar                (uint8_t value);
uint8_t                                 switching_cap_from_enum                 (g2mplsTypes::switchingCap swCap);
g2mplsTypes::encodingType               encoding_type_from_uchar                (uint8_t value);
uint8_t                                 encoding_type_from_enum                 (g2mplsTypes::encodingType encType);

g2mplsTypes::srlgSeq                    diffSrlgs                               (g2mplsTypes::srlgSeq srlgs1, g2mplsTypes::srlgSeq srlgs2);
g2mplsTypes::teLinkCalendarSeq          diffCalendars                           (g2mplsTypes::teLinkCalendarSeq cal1, g2mplsTypes::teLinkCalendarSeq cal2);
g2mplsTypes::iscSeq                     diffIscs                                (g2mplsTypes::iscSeq iscs1, g2mplsTypes::iscSeq iscs2);

bool  isNodeInSeq                       (g2mplsTypes::nodeIdent nodeIdent, g2mplsTypes::nodeIdentSeq* nodeSeq);
bool  isSubNodeInSeq                    (g2mplsTypes::gridSubNodeIdent subNodeIdent, g2mplsTypes::gridSubNodeIdentSeq* snodes);
bool  isTnaIdinSeq                      (g2mplsTypes::tnaIdent tna, g2mplsTypes::tnaIdentSeq* tnaSeq);
bool  isTELinkInSeq                     (g2mplsTypes::teLinkIdent ident, g2mplsTypes::teLinkIdentSeq* linkSeq);

bool  equalNetNodes                     (g2mplsTypes::netNodeParams* nodeParams1, g2mplsTypes::netNodeParams* nodeParams2);
bool  equalTeLinks                      (const g2mplsTypes::teLinkIdent* linkParams1, g2mplsTypes::teLinkIdent* linkParams2);
bool  equalGridSites                    (g2mplsTypes::gridSiteParams* siteInfo1, g2mplsTypes::gridSiteParams* siteInfo2);
bool  equalGridServices                 (g2mplsTypes::gridServiceParams* serviceInfo1, g2mplsTypes::gridServiceParams* serviceInfo2);
bool  equalGridCompElems                (g2mplsTypes::gridCEParams* compInfo1, g2mplsTypes::gridCEParams* compInfo2);
bool  equalGridSubClusters              (g2mplsTypes::gridSubClusterParams* subClInfo1, g2mplsTypes::gridSubClusterParams* subClInfo2);
bool  equalGridStorages                 (g2mplsTypes::gridSEParams* storageInfo1, g2mplsTypes::gridSEParams* storageInfo2);
bool  equalComParams                    (g2mplsTypes::teLinkComParams* comParam1, g2mplsTypes::teLinkComParams* comParam2);
bool  equalTdmParams                    (g2mplsTypes::teLinkTdmParams* tdmParam1, g2mplsTypes::teLinkTdmParams* tdmParam2);
bool  equalLscG709Params                (g2mplsTypes::teLinkLscG709Params* lscG709Params1, g2mplsTypes::teLinkLscG709Params* lscG709Params2);
bool  equalLscWdmParams                 (g2mplsTypes::teLinkLscWdmParams lscWdmParam1, g2mplsTypes::teLinkLscWdmParams lscWdmParam2);
bool  equalLinkStates                   (g2mplsTypes::statesBundle* states1, g2mplsTypes::statesBundle* states2);
bool  equalGenBw                        (g2mplsTypes::availBwPerPrio bw1, g2mplsTypes::availBwPerPrio bw2);
bool  equalTdmBw                        (g2mplsTypes::freeCTPSeq freeCTP1, g2mplsTypes::freeCTPSeq freeCTP2);
bool  equalLscG709Bw                    (g2mplsTypes::freeCTPSeq freeODUk, g2mplsTypes::freeCTPSeq freeOCh, g2mplsTypes::freeCTPSeq serverFreeODUk, g2mplsTypes::freeCTPSeq serverFreeOCh);
bool  equalLscWdmBw                     (g2mplsTypes::teLinkWdmLambdasBitmap bitmap1, g2mplsTypes::teLinkWdmLambdasBitmap bitmap2);

typedef enum {
  ISCS_GEN  ,
  ISCS_TDM  ,
  ISCS_PSC
} iscs_type_t;


#endif // HAVE_OMNIORB
