/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "scngwc.h"
#include "scngws.h"
#include "log.h"
#include "thread.h"
#include "sockopt.h"
#include "network.h"
#include "memory.h"
#include "privs.h"

/* SCNGWs process wide configuration. */
static struct scn_master scn_master;

/* SCNGWs process wide configuration pointer to export. */
struct scn_master *scn_m;

/* SCNGWs master of the threads. */
extern struct thread_master *master;

extern struct zebra_privs_t scngwsd_privs;

/* This function initialize the structure containing the process wide configuration and variables*/
void
scngws_master_init ()
{
	memset (&scn_master, 0, sizeof (struct scn_master));

	scn_m = &scn_master;
	scn_m->master = thread_master_create ();
	scn_m->client_list = list_new();
	scn_m->start_time = quagga_time (NULL);
	scn_m->lrm_conn_status = STATUS_NEVER_CONN;
	scn_m->timeout_lrmsync = TIMEOUT_LRMSYNC_DEFUALT;
	scn_m->timer_thread = NULL;

}

/* This function open the net socket for the clients */
static int
net_sock_open (int proto, struct thread ** thread)
{
	int fd, ret, hincl = 1;
	assert(thread);
	*thread = 0;

	if ( scngwsd_privs.change (ZPRIVS_RAISE) )
		zlog_err ("net_sock_open: could not raise privs, %s", safe_strerror (errno));

	fd = socket (AF_INET, SOCK_RAW, proto);
	if (fd < 0) {
		int save_errno = errno;
		if ( scngwsd_privs.change (ZPRIVS_LOWER) )
			zlog_err ("net_sock_open: could not lower privs, %s", safe_strerror (errno));
		zlog_err ("net_sock_open: socket: %s", safe_strerror (save_errno));
		exit(1);
	}
#ifdef IP_HDRINCL
	/* we will include IP header with packet */
	ret = setsockopt (fd, IPPROTO_IP, IP_HDRINCL, &hincl, sizeof (hincl));
	if (ret < 0) {
		int save_errno = errno;
		if ( scngwsd_privs.change (ZPRIVS_LOWER) )
			zlog_err ("net_sock_open: could not lower privs, %s", safe_strerror (errno));
		zlog_warn ("Can't set IP_HDRINCL option for fd %d: %s", fd, safe_strerror(save_errno));
	}
#elif defined (IPTOS_PREC_INTERNETCONTROL)
#warning "IP_HDRINCL not available on this system"
#warning "using IPTOS_PREC_INTERNETCONTROL"
	/* Set precedence field. */
	tos = IPTOS_PREC_INTERNETCONTROL;

	ret = setsockopt (fd, IPPROTO_IP, IP_TOS, (char *) &tos, sizeof (int));

	if (ret < 0) {
		int save_errno = errno;
		if ( scngwsd_privs.change (ZPRIVS_LOWER) )
			zlog_err ("net_sock_open: could not lower privs, %s", safe_strerror (errno));
		zlog_warn ("can't set sockopt IP_TOS %d to socket %d: %s", tos, fd, safe_strerror(save_errno));

		close (fd);       /* Prevent sd leak. */

		return ret;
	}
#else /* !IPTOS_PREC_INTERNETCONTROL */
#warning "IP_HDRINCL not available, nor is IPTOS_PREC_INTERNETCONTROL"
	zlog_warn ("IP_HDRINCL option not available");
#endif /* IP_HDRINCL */

	if ( scngwsd_privs.change (ZPRIVS_LOWER) )
		zlog_err ("net_sock_open: could not lower privs, %s", safe_strerror (errno));

	*thread = thread_add_read (master, net_packet_read, NULL, fd);

	return fd;
}

static int
net_sock_init (int encap, int proto, struct thread ** thread)
{
	switch (encap) {
		case TUNNEL:
			return net_sock_open (IPPROTO_IPIP, thread);
			break;
		case NO_TUNNEL:
			return net_sock_open (proto, thread);
			break;
		default:
			return 0;
			break;
	}
}

static struct scn_client *
client_fetch_tun ()
{
	struct zlistnode *client_node;
	struct zlistnode *client_nnode;
	struct scn_client *client;

	for (ALL_LIST_ELEMENTS (scn_m->client_list, client_node, client_nnode, client)) {
		if (client->encap == TUNNEL) {
			return client;
		}
	}
	return NULL;
}

static struct scn_client *
client_fetch_notun (int proto)
{
	struct zlistnode *client_node;
	struct zlistnode *client_nnode;
	struct scn_client *client;

	for (ALL_LIST_ELEMENTS (scn_m->client_list, client_node, client_nnode, client)) {
		if ((client->encap == NO_TUNNEL) && (client->proto == proto)) {
			return client;
		}
	}
	return NULL;
}

static struct scn_client *
client_fetch (int encap, int proto)
{
	switch (encap) {
		case TUNNEL:
			return client_fetch_tun ();
			break;
		case NO_TUNNEL:
			return client_fetch_notun (proto);
			break;
		default:
			return NULL;
			break;
	}
}

int
client_set_thread_read_net(int encap, int proto, struct thread * th)
{
	struct scn_client * client;

	switch (encap) {
		case TUNNEL:
			client = client_fetch_tun ();
			break;
		case NO_TUNNEL:
			client = client_fetch_notun (proto);
			break;
		default:
			return -1;
	}

	if (!client) {
		return -1;
	}

	client->t_read_net = th;
	return 1;
}

/* This function creates a new client structure. */
static int
scngws_client_create (int fd_cl, int proto, int interface_type, int encap, int want_ack)
{
	struct scn_client *client;
	struct scn_client *tmp_client;
	int fd_net;

	client = XCALLOC (0, sizeof (struct scn_client));

	tmp_client = client_fetch (encap, proto);
	if (!tmp_client) {
		fd_net = net_sock_init (encap, proto, &client->t_read_net);
		if (fd_net < 0) {
			XFREE(0, client);
			return fd_net;
		}
	}
	else {
		fd_net = tmp_client->fd_net;
		client->t_read_net = tmp_client->t_read_net;
	}

	client->fd_cl = fd_cl;
	client->fd_net = fd_net;
	client->proto = proto;
	client->interface_type = interface_type;
	client->encap = encap;
	client->want_ack = want_ack;
	client->pckts_sent = 0;
	client->pckts_rcvd = 0;

	client->ibuf_cl = stream_new (MAX_PACKET_SIZE);
	stream_reset (client->ibuf_cl);
	memset (client->ibuf_cl->data, 0, client->ibuf_cl->size);

	client->obufq_cl = scngws_fifo_new ();
	client->obufq_net = scngws_fifo_new ();

	/* Add this client to linked list. */
	listnode_add (scn_m->client_list, client);

	client->t_read_cl = thread_add_read(master, scngwc_packet_read, client, fd_cl);

	return 1;
}

/* This function delete an existent client structure. */
void
scngws_client_close (struct scn_client *client)
{
	struct linger ling;

	ling.l_onoff = 1;
	ling.l_linger = 0.1;

	/* Close file descriptors. */
	if (client->fd_cl) {
		setsockopt (client->fd_cl, SOL_SOCKET, SO_LINGER, &ling, sizeof (ling));
		close (client->fd_cl);
		client->fd_cl = -1;
	}
	/* Free stream buffers. */
	if (client->ibuf_cl)
		stream_free (client->ibuf_cl);
	/* Free queues. */
	if (client->obufq_cl)
		scngws_fifo_flush (client->obufq_cl);
	if (client->obufq_net)
		scngws_fifo_flush (client->obufq_net);

	/* Free client structure. */
	listnode_delete (scn_m->client_list, client);

	/* Release threads. */
	if (client->t_read_cl) {
		thread_cancel (client->t_read_cl);
	}
	if (client->t_read_net && !client_fetch(client->encap, client->proto)) {
		thread_cancel (client->t_read_net);
	}
	if (client->t_write_cl) {
		thread_cancel (client->t_write_cl);
	}
	if (client->t_write_net) {
		thread_cancel (client->t_write_net);
	}

	/* Free client struct */
	XFREE (0, client);

	zlog_debug ("Listening for new client connections");
}

/* This function reads a registration request message from client */
static int
read_registration (struct stream *buff, int sock)
{
	u_int32_t msg_type;
	u_int32_t interface_type;
	u_int32_t proto;
	u_int32_t encap;
	u_int32_t want_ack;
	int ret;

	msg_type = stream_getl (buff);
	if (msg_type != REGISTRATION_MESSAGE) {
		return -1;
	}
	interface_type = stream_getl (buff);
	if (((int)interface_type != SCN_INNI) &&
	    ((int)interface_type != SCN_ENNI) &&
	    ((int)interface_type != SCN_UNI)) {
		return -1;
	}
	proto = stream_getl (buff);
	if (((int)proto != IPPROTO_OSPFIGP) &&
	    ((int)proto != IPPROTO_RSVP)) {
		return -1;
	}
	encap = stream_getl (buff);
	if (((int)encap != TUNNEL) &&
	    ((int)encap != NO_TUNNEL)) {
		return -1;
	}
	want_ack = stream_getl (buff);
	if (((int)want_ack != WANT_ACK) &&
	    ((int)want_ack!= WANT_NO_ACK)) {
		return -1;
	}

	ret = scngws_client_create (sock, (int)proto, (int)interface_type, (int)encap, (int)want_ack);

	return ret;

}

/* This function manage the arrival of a registration mesasge from client */
static int
accept_registration (int sock)
{
	struct stream *buff;
        int ret;
	int resp;
	u_int32_t size = REGISTRATION_MSG_SIZE;

	buff = stream_new (size);
	memset (buff->data, 0, size);
	stream_reset (buff);

	/* Read registration message from client */
	ret = scngws_recv_hdr (sock, buff, size, MSG_WAITALL);
	if (ret < 0) {
		stream_free (buff);
		return ret;
	}
	buff->endp += ret;
	resp = read_registration (buff, sock);

	/* send response to client */
	stream_reset (buff);
	if (resp < 0) {
		stream_putl (buff, NACK_MESSAGE);
	}
	else {
		stream_putl (buff, ACK_MESSAGE);
	}
	scngws_send_hdr (sock, buff, 4U, 0);
	stream_free (buff);

	return ret;

}

/*This function is called when a client tries to connect with the SCNGWs. Here we accept the connection */
int
accept_sock (struct thread *thread)
{
	int sock, fd_cl, ret;
	int buf_size = MAX_PACKET_SIZE;
	size_t b_len = sizeof (int);
	struct sockaddr_in client;
	socklen_t len = sizeof (struct sockaddr_in);

	sock = THREAD_FD(thread);

	zlog_debug ("Establishment of a new connection with a new client...");
	/* Reschedule for other clients connections */
	thread_add_read (master, accept_sock, NULL, sock);

	fd_cl = accept (sock, (struct sockaddr *) &client, &len);
	if (fd_cl < 0) {
		zlog_warn ("accept_sock: Can't accept client connection: %s", safe_strerror (errno));
		return -1;
	}
	ret = setsockopt (fd_cl, SOL_SOCKET, SO_RCVBUF, (char *)&buf_size, b_len);
	if (ret < 0) {
		zlog_debug ("accept_sock: cannot set SO_RCVBUF %s", strerror (errno));
	}
	ret = setsockopt (fd_cl, SOL_SOCKET, SO_SNDBUF, (char *)&buf_size, b_len);
	if (ret < 0) {
		zlog_debug ("accept_sock: cannot set SO_SNDBUF %s", strerror (errno));
	}

	ret = accept_registration (fd_cl);
	if (ret < 0) {
		zlog_debug ("Denied a connection from client on socket %d. Something is wrong in his registration message", fd_cl);
		return -1;
	}
	zlog_debug ("Accepted a connection from client on socket %d", fd_cl);

	return 0;
}

/* This function opens the socket where SCNGW server listen for clients connections */
static int
listening_sock_init (int port)
{
	int sock, ret;
	int on = 1;
	struct sockaddr_in addr;

	sock = socket (AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		zlog_err("Can't open SCNGW server socket");
		exit (1);
	}

	ret = setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, (void *) &on, sizeof (on));
	if (ret < 0) {
		zlog_warn ("Can't set sockopt SO_REUSEADDR to socket %d", sock);
	}

	memset (&addr, 0, sizeof (struct sockaddr_in));
	addr.sin_family = AF_INET;
	addr.sin_port = htons (port);
#ifdef HAVE_SIN_LEN
	addr.sin_len = sizeof (struct sockaddr_in);
#endif /* HAVE_SIN_LEN */
	addr.sin_addr.s_addr = htonl (INADDR_LOOPBACK);

	ret  = bind (sock, (struct sockaddr *)&addr, sizeof (struct sockaddr_in));
	if (ret < 0) {
		zlog_err ("listening_sock_init: Can't bind to port %d: %s.\n", port, safe_strerror(errno));
		exit(1);
	}

	ret = listen (sock, 5);
	if (ret < 0) {
		zlog_err ("listening_sock_init: Can't listen to stream socket: %s", safe_strerror (errno));
		zlog_err ("SCNGW server can't provice full functionality due to above error");
		close (sock);
		exit(1);
	}
	zlog_debug ("Listening on socket %d for a new client", sock);

	thread_add_read (master, accept_sock, NULL, sock);

	return 0;
}

/* Initialization of server beahavior */
int
server_sock_init ()
{
	/* Open socket for clients connections */
	listening_sock_init (SCNGWS_PORT);

	return 0;
}
