/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "scngwc.h"
#include "scngws.h"
#include "scngws_packet.h"
#include "scngws_lrm.h"
#include "log.h"
#include "thread.h"
#include "sockopt.h"
#include "memory.h"
#include "linklist.h"
#include "zclient.h"
#include <sys/ioctl.h>
#include <net/if.h>

/* SCNGWs master of the threads. */
extern struct scn_master *scn_m;

/* SCNGWs process wide configuration pointer. */
extern struct thread_master *master;

/* SCNGWs tel_cc and cc_scnif associations lists*/
/*extern struct zlist *tel_cc_list;*/
/*extern struct zlist *cc_scnif_list;*/

extern struct zclient *zclient;

/* This function returns the first packet in the queue */
static struct scngws_packet *
scngws_fifo_head (struct scngws_fifo *fifo)
{
  return fifo->head;
}

/* This function creates a new packet */
static struct scngws_packet *
scngws_packet_new (size_t size)
{
	struct scngws_packet *new;

	new = XCALLOC (MTYPE_SCNGWS_PACKET, sizeof (struct scngws_packet));
	new->s = stream_new (size);
	stream_reset (new->s);
        memset (new->s->data, 0, new->s->size);

	return new;
}

/* This function creates a new FIFO queue */
struct scngws_fifo *
scngws_fifo_new ()
{
	struct scngws_fifo *new;

	new = XCALLOC (MTYPE_SCNGWS_FIFO, sizeof (struct scngws_fifo));
	return new;
}

/* This function puts a packet at the end of the queue */
static void
scngws_fifo_push (struct scngws_fifo *fifo, struct scngws_packet *sp)
{
	if (fifo->tail)
		fifo->tail->next = sp;
	else
		fifo->head = sp;

	fifo->tail = sp;

	fifo->count++;
}

/* This function extract the first packet in the queue and updates the queue */
static struct scngws_packet *
scngws_fifo_pop (struct scngws_fifo *fifo)
{
	struct scngws_packet *sp;

	sp = fifo->head;

	if (sp) {
		fifo->head = sp->next;

		if (fifo->head == NULL)
			fifo->tail = NULL;

		fifo->count--;
	}

	return sp;
}

/* This funtion frees the memory space for a packet */
static void
scngws_packet_free (struct scngws_packet *sp)
{
	if (sp->s)
		stream_free (sp->s);

	XFREE (MTYPE_SCNGWS_PACKET, sp);

	sp = NULL;
}

/*This function deletes all packets in the queue */
void
scngws_fifo_flush (struct scngws_fifo *fifo)
{
	struct scngws_packet *sp;
	struct scngws_packet *next;

	for (sp = fifo->head; sp; sp = next) {
			next = sp->next;
			scngws_packet_free (sp);
	}
	fifo->head = fifo->tail = NULL;
	fifo->count = 0;
}

/* This function deletes first packet from the queue */
static void
scngws_packet_delete (struct scngws_fifo *fifo)
{
	struct scngws_packet *sp;

	sp = scngws_fifo_pop (fifo);

	if (sp)
		scngws_packet_free (sp);
}

/* This function sends the SCNGW header in a message to client */
int
scngws_send_hdr (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_hdr;

	nbytes_hdr = send (sock, buf->data, size, flags);
	if (nbytes_hdr < 0) {
		zlog_warn ("[WRN] scngws_send_hdr: send failed, %s", safe_strerror(errno));
	}
	else if (nbytes_hdr != (int)size) {
		zlog_debug ("scngws_send_hdr: sent wrong number of bytes, %d instead of %d", nbytes_hdr, (int)size);
	}
	return nbytes_hdr;
}

/* This function sends the body of a message to client */
static int
scngws_send_msg (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_msg;

	nbytes_msg = send (sock, buf->data + SCNGW_HDR_SIZE, size, flags);

	return nbytes_msg;
}

/* This function adds the SCNGW header in the message to send to SCNGWc */
static int
scngws_hdr_add (struct stream *buff, struct scngws_packet *sp, u_int32_t msg_type)
{
	stream_putl (buff, msg_type);
	stream_putl (buff, SCNGW_HDR_SIZE);
	stream_putl (buff, sp->length);
	stream_putl (buff, sp->msg_id);
	stream_put_ipv4 (buff, sp->src_addr.s_addr);
	stream_put_ipv4 (buff, sp->dst_addr.s_addr);
	stream_putl (buff, sp->flags);
	stream_putl (buff, sp->cc);

	memcpy (buff->data + buff->endp, sp->s->data, sp->length);

	return (sp->length + SCNGW_HDR_SIZE);
}

/* This function sends the packet received from the network to the client */
int
scngwc_packet_send (struct thread *thread)
{
	int ret, nbytes;
	struct scn_client *client;
	struct stream *buff;
	struct scngws_packet *sp;

	client = THREAD_ARG (thread);
	sp = scngws_fifo_head (client->obufq_cl);

	if (sp == NULL) {
		zlog_debug ("scngwc_packet_send: there are no packets queued to send");
		return 0;
	}
	stream_set_getp (sp->s, 0);

	buff = stream_new (sp->length + SCNGW_HDR_SIZE +1);
	stream_reset (buff);
        memset (buff->data, 0, buff->size);

	nbytes = scngws_hdr_add (buff, sp, PACKET_MESSAGE);

	ret = scngws_send_hdr (client->fd_cl, buff, SCNGW_HDR_SIZE, 0);

	ret = scngws_send_msg (client->fd_cl, buff, sp->length, 0);

	scngws_packet_delete (client->obufq_cl);

	if (ret > 0) {
		switch (client->proto) {
			case IPPROTO_OSPFIGP:
				switch (client->interface_type) {
					case SCN_INNI:
						zlog_debug ("              Sent %d bytes to OSPF-INNI module", sp->length + SCNGW_HDR_SIZE);
						break;
					case SCN_ENNI:
						zlog_debug ("              Sent %d bytes to OSPF-ENNI module", sp->length + SCNGW_HDR_SIZE);
						break;
					case SCN_UNI:
						zlog_debug ("              Sent %d bytes to OSPF-UNI module", sp->length + SCNGW_HDR_SIZE);
						break;
				}
				break;
			case IPPROTO_RSVP:
				switch (client->interface_type) {
					case SCN_INNI:
						zlog_debug ("              Sent %d bytes to RSVP-INNI module", sp->length + SCNGW_HDR_SIZE);
						break;
					case SCN_ENNI:
						zlog_debug ("              Sent %d bytes to RSVP-ENNI module", sp->length + SCNGW_HDR_SIZE);
						break;
					case SCN_UNI:
						zlog_debug ("              Sent %d bytes to RSVP-UNI module", sp->length + SCNGW_HDR_SIZE);
						break;
				}
				break;
		}
	}
	stream_free (buff);
	/* If packets still remain in queue, call write thread. */
	if (!list_isempty (client->obufq_cl))
		client->t_write_cl = thread_add_write (master, scngwc_packet_send, client, client->fd_cl);

	return 0;
}

/* This function reads SCNGW header in the mesasge from SCNGWc */
static int
scngws_hdr_read (struct stream *buf, struct scngw_hdr *hdr)
{
	hdr->msg_type = stream_getl (buf);
	hdr->hdr_len = stream_getl (buf);
	hdr->sdu_len = stream_getl (buf);
	hdr->msg_id = stream_getl (buf);

	hdr->flags = stream_getl (buf);
	hdr->src_addr = stream_get_ipv4 (buf);
	hdr->dst_addr = stream_get_ipv4 (buf);

	hdr->cc = stream_getl (buf);

	return 0;
}

/* This function receives the SCNGW header in a message from client */
int
scngws_recv_hdr (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_hdr;

	nbytes_hdr = recv (sock, buf->data, size, flags);
	if (nbytes_hdr < 0) {
		zlog_debug("scngws_recv_hdr: recv failed %s", strerror (errno));
	}
	return nbytes_hdr;
}

/* This function receives the body of a message from client */
static int
scngws_recv_msg (int sock, struct stream *buf, u_int32_t size, int flags)
{
	int nbytes_msg;

	nbytes_msg = recv (sock, buf->data + SCNGW_HDR_SIZE, size, flags);
	if (nbytes_msg < 0) {
		zlog_warn ("[WRN] scngws_recv_msg: recv failed: %s", strerror (errno));
	}
	return nbytes_msg;
}


/* This function receives the packet from SCNGWc (containing the client packet) */
static int
scngwc_recv_packet (struct scn_client *client)
{
	int nbytes_hdr;
	int nbytes_msg;
	int nbytes;
	struct scngw_hdr hdr;

	nbytes_hdr = scngws_recv_hdr (client->fd_cl, client->ibuf_cl, SCNGW_HDR_SIZE, MSG_WAITALL);
	if (nbytes_hdr < 0)  {
		zlog_warn("[WRN] scngwc_recv_packetL nbytes_hdr(%d) < 0", nbytes_hdr);
		return nbytes_hdr;
	}
	else if (nbytes_hdr == 0) {
		switch (client->proto) {
			case IPPROTO_OSPFIGP:
				switch (client->interface_type) {
					case SCN_INNI:
						zlog_warn ("[WRN] scngwc_recv_packet: OSPF-INNI shutdown");
						break;
					case SCN_ENNI:
						zlog_warn ("[WRN] scngwc_recv_packet: OSPF-ENNI shutdown");
						break;
					case SCN_UNI:
						zlog_warn ("[WRN] scngwc_recv_packet: OSPF-UNI shutdown");
						break;
				}
				break;
			case IPPROTO_RSVP:
				switch (client->interface_type) {
					case SCN_INNI:
						zlog_warn ("[WRN] scngwc_recv_packet: RSVP-INNI shutdown");
						break;
					case SCN_ENNI:
						zlog_warn ("[WRN] scngwc_recv_packet: RSVP-ENNI shutdown");
						break;
					case SCN_UNI:
						zlog_warn ("[WRN] scngwc_recv_packet: RSVP-UNI shutdown");
						break;
				}
				break;
		}
		scngws_client_close (client);
		return 0;
	}
	client->ibuf_cl->endp += (size_t)nbytes_hdr;
	memset (&hdr, 0, sizeof (struct scngw_hdr));
	scngws_hdr_read (client->ibuf_cl, &hdr);

	nbytes_msg = scngws_recv_msg (client->fd_cl, client->ibuf_cl, hdr.sdu_len, MSG_WAITALL);
	if (nbytes_msg < 0)  {
		return nbytes_msg;
	}
	else if (nbytes_msg == 0) {
		switch (client->proto) {
			case IPPROTO_OSPFIGP:
				switch (client->interface_type) {
					case SCN_INNI:
						zlog_warn ("[WRN] scngwc_recv_packet: OSPF-INNI shutdown");
						break;
					case SCN_ENNI:
						zlog_warn ("[WRN] scngwc_recv_packet: OSPF-ENNI shutdown");
						break;
					case SCN_UNI:
						zlog_warn ("[WRN] scngwc_recv_packet: OSPF-UNI shutdown");
						break;
				}
				break;
			case IPPROTO_RSVP:
				switch (client->interface_type) {
					case SCN_INNI:
						zlog_warn ("[WRN] scngwc_recv_packet: RSVP-INNI shutdown");
						break;
					case SCN_ENNI:
						zlog_warn ("[WRN] scngwc_recv_packet: RSVP-ENNI shutdown");
						break;
					case SCN_UNI:
						zlog_warn ("[WRN] scngwc_recv_packet: RSVP-UNI shutdown");
						break;
				}
				break;
		}
		scngws_client_close (client);
		return 0;
	}
	nbytes = nbytes_hdr + nbytes_msg;
	client->ibuf_cl->endp += (size_t)nbytes_msg;
	return nbytes;
}

/* This function is called when a new message (containing a client packet) arrives */
int
scngwc_packet_read (struct thread *thread)
{
	int sdu_length;
	struct scn_client *client;
	struct scngw_hdr scnhdr;
	struct scngws_packet *sp;

	client = THREAD_ARG (thread);

	stream_reset (client->ibuf_cl);
        memset (client->ibuf_cl->data, 0, client->ibuf_cl->size);

	/* reschedule for next packet. */
	client->t_read_cl = thread_add_read (master, scngwc_packet_read, client, client->fd_cl);

	switch (client->proto) {
		case IPPROTO_OSPFIGP:
			switch (client->interface_type) {
				case SCN_INNI:
					zlog_debug ("Received message from OSPF-INNI module:");
					break;
				case SCN_ENNI:
					zlog_debug ("Received message from OSPF-ENNI module:");
					break;
				case SCN_UNI:
					zlog_debug ("Received message from OSPF-UNI module:");
					break;
			}
			break;
		case IPPROTO_RSVP:
			switch (client->interface_type) {
				case SCN_INNI:
					zlog_debug ("Received message from RSVP-INNI module:");
					break;
				case SCN_ENNI:
					zlog_debug ("Received message from RSVP-ENNI module:");
					break;
				case SCN_UNI:
					zlog_debug ("Received message from RSVP-UNI module:");
					break;
			}
			break;
	}
	sdu_length = scngwc_recv_packet (client);
	if (sdu_length > 0) {
		if (client->ibuf_cl == NULL) {
		return -1;
		}
		/* reset get pointer */
		stream_set_getp (client->ibuf_cl, 0);
		memset (&scnhdr, 0, sizeof (struct scngw_hdr));
		/* read scngw header */
		scngws_hdr_read (client->ibuf_cl, &scnhdr);

//		zlog_debug ("scngwc_packet_read: message received.");
		zlog_debug ("              SCN header length : %u", scnhdr.hdr_len);
		zlog_debug ("              SDU length        : %u", scnhdr.sdu_len);

		sp = scngws_packet_new (scnhdr.sdu_len);

		memcpy  (sp->s->data, client->ibuf_cl->data + client->ibuf_cl->getp, scnhdr.sdu_len);

		sp->s->endp += scnhdr.sdu_len;
		sp->length = scnhdr.sdu_len;
		sp->flags = scnhdr.flags;
		sp->src_addr.s_addr = scnhdr.src_addr;
		sp->dst_addr.s_addr = scnhdr.dst_addr;
		sp->cc = scnhdr.cc;
		sp->client_type = client->proto;
		sp->msg_id = scnhdr.msg_id;

		scngws_fifo_push (client->obufq_net, sp);

		zlog_debug ("              SRC addr          : %s", inet_ntoa(sp->src_addr));
		zlog_debug ("              DST addr          : %s", inet_ntoa(sp->dst_addr));

		zlog_debug ("              Packet from client received and queued.");
		zlog_debug ("              There are %u packets waiting to be transmitted to network", client->obufq_net->count);

		client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);

		return 0;
	}
	else
		return sdu_length;
}

/* This function is used to retrieve the control channel associated with the TE-link speciifed by client */
static int
scnif_fetch (struct in_addr src_addr, struct in_addr dst_addr, struct scn_if_addrs *scnif, int encap)
{
	int sock, n, numreqs = 30;
	struct ifconf ifc;
	struct ifreq *ifr;
	struct sockaddr_in *ifaddr;
	struct zlistnode *telcc_node, *telcc_nnode;
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct tel_cc_assoc *tel_cc;
	struct cc_scnif_assoc *cc_scnif;
	struct zlist *tel_cc_list;
	struct zlist *cc_scnif_list;
	u_int32_t cc_id = 0, found_cc = 0, found_if = 0;

	tel_cc_list = get_tel_cc_list();
	cc_scnif_list = get_cc_scnif_list();


	if (tel_cc_list != NULL)
	{
		for (ALL_LIST_ELEMENTS (tel_cc_list, telcc_node, telcc_nnode, tel_cc)) {
			if (src_addr.s_addr == tel_cc->tel_loc.s_addr) {
				/* we select the first CC in the list waiting
				* to define an algorithm
				*/
	
				u_int32_t * tmp;
	
				tmp = listnode_head (tel_cc->cclist);
	
				cc_id = tmp ? *(tmp) : 0;
	
				break;
			}
		}
	}
	else
	{	
		zlog_err("scnif_fetch: tel_cc_list = NULL");
		return -1;
	}
	if (cc_id == 0) {
		if (encap == TUNNEL) {
			/* tunnelled clients must have a cc binding and
			 * then you can fetch related scn i/f
			 */
			return -1;
		} else if (encap == NO_TUNNEL) {
			/* non-tunnelled clients can specify scn i/f directly,
			 * if a cc-tel binding has not been found
			 */
			struct zlistnode *    scnif_node;
			struct zlistnode *    scnif_nnode;
			struct zlist *        scnif_list;
			struct scnif *        tmp;

			scnif_list = get_scnif_list();

			for (ALL_LIST_ELEMENTS(scnif_list,
					       scnif_node,
					       scnif_nnode,
					       tmp)) {
				if (src_addr.s_addr == (tmp->loc_addr).s_addr){
					memcpy (&(scnif->loc_addr),
						&(tmp->loc_addr),
						sizeof (struct in_addr));
					memcpy (&(scnif->rem_addr),
						&dst_addr,
						sizeof (struct in_addr));
					found_cc = 1;
					cc_id    = 0;
				}
				break;
			}
		} else {
			zlog_err("Bad encapsulation type %d", encap);
			return -4;
		}
	} else {
		if (cc_scnif_list != NULL)
		{
			for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, cc_scnif)) {
				if (cc_id == cc_scnif->cc_id) {
					memcpy (&(scnif->loc_addr), &(cc_scnif->scnif_loc), sizeof (struct in_addr));
					memcpy (&(scnif->rem_addr), &(cc_scnif->scnif_rem), sizeof (struct in_addr));
					found_cc = 1;
					break;
				}
			}
		}
		else
		{
			zlog_err("scnif_fetch: cc_scnif_list = NULL");
			return -2;
		}
		if (found_cc == 0)
			return -2;
	}

	/* Open a socket for ioctl() */
	sock = socket (AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) {
		int save_errno = errno;
		zlog_err("Cannot create UDP socket: %s", safe_strerror(save_errno));
		exit (1);
	}

	ifc.ifc_buf = NULL;
	for (;;) {
		ifc.ifc_len = sizeof(struct ifreq) * numreqs;
		ifc.ifc_buf = realloc(ifc.ifc_buf, ifc.ifc_len);

		if (ioctl(sock, SIOCGIFCONF, &ifc) < 0) {
			zlog_err ("Can't lookup mtu by ioctl(SIOCGIFCONF): %s", safe_strerror (errno));
			close (sock);
			free(ifc.ifc_buf);
			return -3;
		}
		if ((u_int32_t)ifc.ifc_len == sizeof(struct ifreq) * numreqs) {
			/* assume it overflowed and try again */
			numreqs += 10;
			continue;
		}
		break;
	}
	ifr = ifc.ifc_req;
	for (n = 0; n < ifc.ifc_len; n += sizeof(struct ifreq)) {
		ifaddr = (struct sockaddr_in *)(&(ifr->ifr_addr));
		if (ifaddr->sin_addr.s_addr == scnif->loc_addr.s_addr) {
			struct ifreq ifr_match;
			strcpy (ifr_match.ifr_name, ifr->ifr_name);
			if (ioctl(sock, SIOCGIFMTU, &ifr_match) < 0 ) {
				zlog_err ("Can't lookup mtu by ioctl(SIOCGIFCONF): %s", safe_strerror (errno));
				close (sock);
				free(ifc.ifc_buf);
				return -3;
			}
			memcpy (&(scnif->mtu), &(ifr_match.ifr_mtu), sizeof (int));
			found_if = 1;
			break;
		}
		ifr++;
	}
	close (sock);
	free(ifc.ifc_buf);

	if (found_if == 0)
		return -3;

	return cc_id;
}

#ifdef WANT_SCNGW_WRITE_FRAGMENT
/* This function sends a fragment of the packet (containing an client packet) if fragmentation is necessary */
static void
scngws_write_frags (struct scn_client *client, struct scngws_packet *sp, struct ip *iph, struct msghdr *msg, unsigned int maxdatasize, int mtu, int flags)
{
#define SCNGWS_WRITE_FRAG_SHIFT 3
	u_int16_t offset;
	struct iovec *iovp;
	int ret;

	assert ( (size_t)sp->length == sp->s->endp );
	assert (msg->msg_iovlen == 2);

	/* set More Frag */
	iph->ip_off |= IP_MF;

	/* ip frag offset is expressed in units of 8byte words */
	offset = maxdatasize >> SCNGWS_WRITE_FRAG_SHIFT;

	iovp = &msg->msg_iov[1];

	while ( (stream_get_endp(sp->s) - stream_get_getp (sp->s)) > maxdatasize ) {
		/* data length of this frag is to next offset value */
		iovp->iov_len = offset << SCNGWS_WRITE_FRAG_SHIFT;
		iph->ip_len = iovp->iov_len + sizeof (struct ip);

		assert (iph->ip_len <= mtu);

		sockopt_iphdrincl_swab_htosys (iph);

		ret = sendmsg (client->fd_net, msg, flags);

		sockopt_iphdrincl_swab_systoh (iph);

		if (ret < 0)
			zlog_warn ("[WRN] scngws_write_frags: sendmsg failed to %s,"
				   " id %d, off %d, len %d, mtu %u failed with %s",
				   inet_ntoa (iph->ip_dst),
				   iph->ip_id,
				   iph->ip_off,
				   iph->ip_len,
				   mtu,
				   safe_strerror (errno));

		iph->ip_off += offset;
		stream_forward_getp (sp->s, iovp->iov_len);
		iovp->iov_base = STREAM_PNT (sp->s);
	}

	/* setup for final fragment */
	iovp->iov_len = stream_get_endp(sp->s) - stream_get_getp (sp->s);
	iph->ip_len = iovp->iov_len + sizeof (struct ip);
	iph->ip_off &= (~IP_MF);
}
#endif /* WANT_SCNGW_WRITE_FRAGMENT */

/* This function sends ack/nack response to client for a specific packet (if client speciifed WANT_ACK) */
static int
send_response_to_client (struct scn_client *client, u_int32_t msg_id, int resp)
{
	struct stream *buf;
        size_t size = 4U;
	u_int32_t msg_type;
	int ret, nbytes_hdr, nbytes_msg;
	struct scngws_packet *sp;

	nbytes_hdr = 0;
	nbytes_msg = 0;

	buf = stream_new (SCNGW_HDR_SIZE + size);
	stream_reset (buf);
        memset (buf->data, 0, buf->size);

	sp = scngws_packet_new (size);

	if (resp < 0) {
		msg_type = NACK_MESSAGE;
		resp = abs (resp);
	}
	else {
		msg_type = ACK_MESSAGE;
	}

	stream_putl (sp->s, (u_int32_t)resp);

	sp->length = size;
	sp->msg_id = msg_id;
        sp->src_addr.s_addr = 0;
	sp->dst_addr.s_addr = 0;
	sp->flags = 0;
	sp->cc = 0;

	scngws_hdr_add (buf, sp, msg_type);

	ret = scngws_send_hdr (client->fd_cl, buf, SCNGW_HDR_SIZE, 0);
	if (ret < 0) {
		stream_free (buf);
		return ret;
	}
	ret = scngws_send_msg (client->fd_cl, buf, size, 0);
	if (ret < 0) {
		stream_free (buf);
		return ret;
	}
	switch (msg_type) {
		case ACK_MESSAGE:
			zlog_debug("Sent ACK_MESSAGE to client for message ID %d.", msg_id);
			break;
		case NACK_MESSAGE:
			zlog_debug("Sent NACK_MESSAGE to client for message ID %d.", msg_id);
			break;
	}
	stream_free (buf);

	return ret;
}

/* This function creates the external IP header for the packet to be sent on network */
static void
create_iphdr (struct ip *iph, struct scngws_packet *sp, struct scn_if_addrs scnif, int encap, u_int8_t proto)
{
#define SCNGWS_WRITE_IPHL_SHIFT 2

	iph->ip_hl = sizeof (struct ip) >> SCNGWS_WRITE_IPHL_SHIFT;
	iph->ip_v = IPVERSION;
	iph->ip_tos = IPTOS_PREC_INTERNETCONTROL;
	iph->ip_len = (iph->ip_hl << SCNGWS_WRITE_IPHL_SHIFT) + sp->length;
	iph->ip_off = 0;
	iph->ip_ttl = SCNGW_IP_TTL;
	if (encap == TUNNEL) {
		iph->ip_p = IPPROTO_IPIP; /* IPIP encapsulation */
	}
	else {
		iph->ip_p = proto;
	}
	iph->ip_sum = 0;
	iph->ip_src.s_addr = scnif.loc_addr.s_addr;
	iph->ip_dst.s_addr = scnif.rem_addr.s_addr;
}

/* This function sends the packet (containing a client packet) to the peer identified by SCN-if remote address */
int
net_packet_send (struct thread *thread)
{
	int ret, cc_id;
	struct scn_client *client;
	struct scngws_packet *sp;
	struct msghdr msg;
	struct ip iph;
	struct iovec iov[2];
	struct sockaddr_in sa_dst;
	struct scn_if_addrs scnif;
#ifdef WANT_SCNGW_WRITE_FRAGMENT
	static u_int16_t ipid = 0;
#endif /* WANT_SCNGW_WRITE_FRAGMENT */
	u_int16_t maxdatasize;

	client = THREAD_ARG (thread);
	sp = scngws_fifo_head (client->obufq_net);

	if (sp == NULL) {
		zlog_debug ("net_packet_send: there are no packets queued to send");
		return 0;
	}
	switch (scn_m->lrm_conn_status) {
		case STATUS_UP:
			break;
		case STATUS_DOWN:
			if (client->want_ack == WANT_ACK) {
			send_response_to_client (client, sp->msg_id, -1);
			}
			scngws_packet_delete (client->obufq_net);
			zlog_warn ("[WRN] net_packet_send: The connection with LRM is down. No packages sent to the network");
			/* If packets still remain in queue, call write thread. */
			if (!list_isempty (client->obufq_net))
				client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);
			return 0;
		case STATUS_SYNC:
			if (client->want_ack == WANT_ACK) {
			send_response_to_client (client, sp->msg_id, -1);
			}
			scngws_packet_delete (client->obufq_net);
			zlog_warn ("[WRN] net_packet_send: The connection with LRM is in synchronization status. No packages sent to the network");
			/* If packets still remain in queue, call write thread. */
			if (!list_isempty (client->obufq_net))
				client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);
			return 0;
	}
	stream_set_getp (sp->s, 0);

	memset (&sa_dst, 0, sizeof (sa_dst));

	sa_dst.sin_family = AF_INET;
#ifdef HAVE_SIN_LEN
	sa_dst.sin_len = sizeof(sa_dst);
#endif /* HAVE_SIN_LEN */
	sa_dst.sin_port = htons (0);

	memset (&scnif, 0, sizeof (struct scn_if_addrs));

	cc_id = scnif_fetch(sp->src_addr, sp->dst_addr, &scnif, client->encap);

	switch (cc_id) {
		case -1:
			scngws_packet_delete (client->obufq_net);
			zlog_err ("net_packet_send: can't retrieve TE_link local address %s. Packet discarded.", inet_ntoa(sp->src_addr));
			if (client->want_ack == WANT_ACK) {
			send_response_to_client (client, sp->msg_id, -1);
			}
			/* If packets still remain in queue, call write thread. */
			if (!list_isempty (client->obufq_net))
				client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);
			return 0;
		case -2:
			scngws_packet_delete (client->obufq_net);
			zlog_err ("net_packet_send: can't retrieve control channel ID. Packet discarded.");
			if (client->want_ack == WANT_ACK) {
				send_response_to_client (client, sp->msg_id, -1);
			}
			/* If packets still remain in queue, call write thread. */
			if (!list_isempty (client->obufq_net))
				client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);
			return 0;
		case -3:
			scngws_packet_delete (client->obufq_net);
			zlog_err ("net_packet_send: can't retrieve MTU of local SCN-if %s. Packet discarded.", inet_ntoa(scnif.loc_addr));
			if (client->want_ack == WANT_ACK) {
			send_response_to_client (client, sp->msg_id, -1);
			}
			/* If packets still remain in queue, call write thread. */
			if (!list_isempty (client->obufq_net))
				client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);
			return 0;
		case -4:
			scngws_packet_delete (client->obufq_net);
			zlog_err ("net_packet_send: bad encapsulation type for SCNGW client.");
			if (client->want_ack == WANT_ACK) {
				send_response_to_client (client, sp->msg_id, -1);
			}
			/* If packets still remain in queue, call write thread. */
			if (!list_isempty (client->obufq_net))
				client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);
			return 0;
	}

	sa_dst.sin_addr.s_addr = scnif.rem_addr.s_addr;

	maxdatasize = scnif.mtu - sizeof (struct ip);

	memset (&iph, 0, sizeof (struct ip));
	create_iphdr (&iph, sp, scnif, client->encap, (u_int8_t)client->proto);
#ifdef WANT_SCNGW_WRITE_FRAGMENT
	iph.ip_id = ++ipid;
#endif /* WANT_SCNGW_WRITE_FRAGMENT */

	memset (&msg, 0, sizeof (msg));
	msg.msg_name = (caddr_t) &sa_dst;
	msg.msg_namelen = sizeof (sa_dst);
	msg.msg_iov = iov;
	msg.msg_iovlen = 2;
	iov[0].iov_base = (char*)&iph;
	iov[0].iov_len = iph.ip_hl << SCNGWS_WRITE_IPHL_SHIFT;
	iov[1].iov_base = STREAM_PNT (sp->s);
	iov[1].iov_len = sp->length;

#ifdef WANT_SCNGW_WRITE_FRAGMENT
	if (sp->length > maxdatasize)
		scngws_write_frags (client, sp, &iph, &msg, maxdatasize, scnif.mtu, sp->flags);
#endif /* WANT_SCNGW_WRITE_FRAGMENT */
	sockopt_iphdrincl_swab_htosys (&iph);
	ret = sendmsg (client->fd_net, &msg, sp->flags);
	sockopt_iphdrincl_swab_systoh (&iph);

	if (ret < 0)
		zlog_warn ("[WRN] net_packet_send: sendmsg failed (%s, "
			   "id %d, off %d, len %d, mtu %d: %s)",
			   inet_ntoa (iph.ip_dst), iph.ip_id, iph.ip_off, iph.ip_len,
			   scnif.mtu, safe_strerror (errno));

	if (client->want_ack == WANT_ACK) {
		send_response_to_client (client, sp->msg_id, ret);
	}
	client->pckts_sent ++;
	switch (client->proto) {
		case IPPROTO_OSPFIGP:
			switch (client->interface_type) {
				case SCN_INNI:
					zlog_debug ("OSPF-INNI packet sent to network");
					break;
				case SCN_ENNI:
					zlog_debug ("OSPF-ENNI packet sent to network");
					break;
				case SCN_UNI:
					zlog_debug ("OSPF-UNI packet sent to network");
					break;
			}
			break;
		case IPPROTO_RSVP:
			switch (client->interface_type) {
				case SCN_INNI:
					zlog_debug ("RSVP-INNI packet sent to network");
					break;
				case SCN_ENNI:
					zlog_debug ("RSVP-ENNI packet sent to network");
					break;
				case SCN_UNI:
					zlog_debug ("RSVP-UNI packet sent to network");
					break;
			}
			break;
	}
	zlog_debug ("              Local  SCN-if: %s", inet_ntoa(iph.ip_src));
	zlog_debug ("              Remote SCN-if: %s", inet_ntoa(iph.ip_dst));
	zlog_debug ("              IPhdr length : %d", iph.ip_hl << SCNGWS_WRITE_IPHL_SHIFT);
	zlog_debug ("              Total length : %d", iph.ip_len);
	zlog_debug ("              MTU          : %d", scnif.mtu);

	scngws_packet_delete (client->obufq_net);

	/* If packets still remain in queue, call write thread. */
	if (!list_isempty (client->obufq_net))
		client->t_write_net = thread_add_write (master, net_packet_send, client, client->fd_net);

	return 0;
}

/* This function return an scn_client structure if there's only one client with "encap" and "proto" */
static struct scn_client *
is_unique_client (int encap, u_int8_t proto)
{
	struct zlistnode *client_node, *client_nnode;
	struct scn_client *client;
	struct scn_client *client_ok = NULL;
	int found;
	found = 0;
	for (ALL_LIST_ELEMENTS (scn_m->client_list, client_node, client_nnode, client)) {
		if ( (client->encap == encap) && ((u_int8_t)client->proto == proto) ) {
			found ++;
			if (found == 1) {
				client_ok = client;
			}
		}
	}
	if (found == 1) {
		return client_ok;
	}
	return NULL;
}

/* This function retrieve the right client (if present) for the incoming packet, based on the protocol
   specified in the encapsulated IP packet header */
static struct scn_client *
client_retrieve_tun (struct stream *buff, struct ip *iph)
{
	struct ip *iph_enc;
	struct zlistnode *client_node, *client_nnode;
	struct scn_client *client;
	struct zlistnode *tel_cc_node, *tel_cc_nnode;
	struct tel_cc_assoc *tel_cc;
	struct zlist *tel_cc_list;

	tel_cc_list = get_tel_cc_list();

	stream_forward_getp (buff, iph->ip_hl * 4);

	/* Get the encapsulated IP packet header to recognize the client type */
	iph_enc = (struct ip *) STREAM_PNT (buff);
	sockopt_iphdrincl_swab_systoh (iph_enc);
	/* Look if there's only one client "tunnelled" with protocol iph_enc->ip_p */
	client = is_unique_client (TUNNEL, iph_enc->ip_p);
	if (client != NULL) {
		sockopt_iphdrincl_swab_htosys (iph_enc);
		return client;
	}
	/* Fetch client for incoming packet */
	for (ALL_LIST_ELEMENTS (tel_cc_list, tel_cc_node, tel_cc_nnode, tel_cc)) {
		if (tel_cc->tel_rem.s_addr == iph_enc->ip_src.s_addr) {
			for (ALL_LIST_ELEMENTS (scn_m->client_list, client_node, client_nnode, client)) {
				if ((client->interface_type == (int)tel_cc->interface_type) &&
				    ((u_int8_t)client->proto == iph_enc->ip_p) &&
				    (client->encap == TUNNEL)) {
					sockopt_iphdrincl_swab_htosys (iph_enc);
					return client;
				}
			}
		}
	}
	sockopt_iphdrincl_swab_htosys (iph_enc);
	return NULL;
}

/* This function retrieves one TE-link who has associated CC specified in the IP header, to get interface type */
static struct tel_cc_assoc *
fetch_tel_notun (struct ip *iph)
{
	struct zlistnode *ccscnif_node, *ccscnif_nnode;
	struct cc_scnif_assoc *cc_scnif;
	struct zlistnode *tel_cc_node, *tel_cc_nnode;
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *cc_node, *cc_nnode;
	struct zlist *tel_cc_list;
	struct zlist *cc_scnif_list;
        u_int32_t *ccid;

	tel_cc_list = get_tel_cc_list();
	cc_scnif_list = get_cc_scnif_list();

	for (ALL_LIST_ELEMENTS (cc_scnif_list, ccscnif_node, ccscnif_nnode, cc_scnif)) {
		/* retrieve CC */
		if ((cc_scnif->scnif_loc.s_addr == iph->ip_dst.s_addr) &&
		    (cc_scnif->scnif_rem.s_addr == iph->ip_src.s_addr)) {
			/* retrieve TE-link who has associated this CC, to get interface type*/
			for (ALL_LIST_ELEMENTS (tel_cc_list, tel_cc_node, tel_cc_nnode, tel_cc)) {
				for (ALL_LIST_ELEMENTS (tel_cc->cclist, cc_node, cc_nnode, ccid)) {
					if (*((u_int32_t *)ccid) == cc_scnif->cc_id) {
						return tel_cc;
					}
				}
			}
			break;
		}

	}
	return NULL;

}

/* This function retrieve the correct scn_client structure in case of client specified NO_TUNNEL */
static struct scn_client *
client_retrieve_notun (struct stream *buff, struct ip *iph)
{
	struct tel_cc_assoc *tel_cc;
	struct zlistnode *client_node, *client_nnode;
	struct scn_client *client;

	//stream_forward_getp (buff, iph->ip_hl * 4);

	/* Look if there's only one client "no tunnelled" with protocol iph->ip_p */
	client = is_unique_client (NO_TUNNEL, iph->ip_p);
	if (client != NULL) {
		return client;
	}
	/* Fetch TE-link for incoming packet */
	tel_cc = fetch_tel_notun (iph);
	if (tel_cc == NULL) {
		return NULL;
	}
	/* Fetch client for incoming packet */
	for (ALL_LIST_ELEMENTS (scn_m->client_list, client_node, client_nnode, client)) {
		if ((client->interface_type == (int)tel_cc->interface_type) &&
		    ((u_int8_t)client->proto == iph->ip_p) &&
		    (client->encap == NO_TUNNEL)) {
			return client;
		}
	}

	return NULL;
}

static struct scn_client *
client_retrieve (int encap, struct stream *buff, struct ip *iph)
{
	if (encap == TUNNEL) {
		return client_retrieve_tun (buff, iph);
	}
	else {
		return client_retrieve_notun (buff, iph);
	}
}

/* This function receives the packet (containing a client packet) from the network */
static struct stream *
recv_net_packet (int fd, struct stream *ibuf)
{
	int ret;
	struct ip *iph;
	u_int16_t ip_len;
	struct iovec iov;
	struct msghdr msgh;

	memset (&msgh, 0, sizeof (struct msghdr));
	msgh.msg_iov = &iov;
	msgh.msg_iovlen = 1;

	ret = stream_recvmsg (ibuf, fd, &msgh, 0, MAX_PACKET_SIZE);

	if (ret < 0) {
		zlog_warn("[WRN] stream_recvmsg failed: %s", safe_strerror(errno));
		return NULL;
	}
	if ((unsigned int)ret < sizeof(iph)) { /* ret must be > 0 now */
		zlog_warn("[WRN] recv_net_packet: discarding runt packet of length %d "
			  "(ip header size is %u)",
			  ret, (u_int)sizeof(iph));
		return NULL;
	}

	/* Note that there should not be alignment problems with this assignment
	   because this is at the beginning of the stream data buffer. */
	iph = (struct ip *) STREAM_DATA(ibuf);
	sockopt_iphdrincl_swab_systoh (iph);

	ip_len = iph->ip_len;

#if !defined(GNU_LINUX) && (OpenBSD < 200311)
	/*
	 * Kernel network code touches incoming IP header parameters,
	 * before protocol specific processing.
	 *
	 *   1) Convert byteorder to host representation.
	 *      --> ip_len, ip_id, ip_off
	 *
	 *   2) Adjust ip_len to strip IP header size!
	 *      --> If user process receives entire IP packet via RAW
	 *          socket, it must consider adding IP header size to
	 *          the "ip_len" field of "ip" structure.
	 *
	 * For more details, see <netinet/ip_input.c>.
	 */
	ip_len = ip_len + (iph->ip_hl << 2);
#endif
	if (ret != ip_len) {
		zlog_warn ("[WRN] recv_net_packet read length mismatch: ip_len is %d, "
			   "but recvmsg returned %d", ip_len, ret);
		return NULL;
	}

	return ibuf;
}

/* This function reads a packet (containing a client packet) received by SCNGWs */
int
net_packet_read (struct thread *thread)
{
	int fd;
	struct stream *ibuf;
	struct scn_client *client;
	struct ip *iph;
	size_t sdu_size;
	struct scngws_packet *sp;
	struct thread * tmp_thread;

	fd = THREAD_FD (thread);

	ibuf = stream_new(MAX_PACKET_SIZE+1);
	stream_reset(ibuf);
	memset (ibuf->data, 0, ibuf->size);

	/* prepare for next packet. */
	tmp_thread = thread_add_read (master, net_packet_read, NULL, fd);

	/* read packet. */
	if (!(ibuf = recv_net_packet (fd, ibuf)))
		return -1;

	/* Note that there should not be alignment problems with this assignment
	   because this is at the beginning of the stream data buffer. */
	iph = (struct ip *) STREAM_DATA (ibuf);
	/* Note that sockopt_iphdrincl_swab_systoh was called in recv_net_packet. */
	if (iph->ip_p == IPPROTO_IPIP) {
		client = client_retrieve (TUNNEL, ibuf, iph);
	}
	else {
		client = client_retrieve (NO_TUNNEL, ibuf, iph);
	}

	if (client == NULL) {
		zlog_warn ("[WRN] A packet was received from network but no correspondent client is running. Packet discarded.");
		stream_free (ibuf);
		return 0;
	}

	client_set_thread_read_net(client->encap, client->proto, tmp_thread);

	/*sdu_size = ((size_t)iph->ip_len - (iph->ip_hl * 4));*/
  if (client->encap == NO_TUNNEL)
    sdu_size = (size_t)iph->ip_len;
  else
    sdu_size = ((size_t)iph->ip_len - (iph->ip_hl * 4));

  client->pckts_rcvd ++;
//  zlog_debug("net_packet_read:");
  switch (client->proto) {
    case IPPROTO_OSPFIGP:
      switch (client->interface_type) {
        case SCN_INNI:
          zlog_debug ("OSPF-INNI packet received from network and queued:");
          break;
        case SCN_ENNI:
          zlog_debug ("OSPF-ENNI packet received from network and queued:");
          break;
        case SCN_UNI:
          zlog_debug ("OSPF-UNI packet received from network and queued:");
          break;
      }
      break;
    case IPPROTO_RSVP:
      switch (client->interface_type) {
        case SCN_INNI:
          zlog_debug ("RSVP-INNI packet received from network and queued:");
          break;
        case SCN_ENNI:
          zlog_debug ("RSVP-ENNI packet received from network and queued:");
          break;
        case SCN_UNI:
          zlog_debug ("RSVP-UNI packet received from network and queued:");
          break;
      }
      break;
  }
	zlog_debug ("              loc addr: %s", inet_ntoa(iph->ip_src));
	zlog_debug ("              rem addr: %s", inet_ntoa(iph->ip_dst));
	zlog_debug ("              SDU size: %u", sdu_size);
	zlog_debug ("              IP hdr  : %d", iph->ip_hl << SCNGWS_WRITE_IPHL_SHIFT);

	sp = scngws_packet_new (sdu_size);

	sp->length = sdu_size;
	sp->client_type = client->proto;
	sp->msg_id = 0;
	sp->flags = 0;
	sp->src_addr.s_addr = iph->ip_src.s_addr;
	sp->dst_addr.s_addr = iph->ip_dst.s_addr;
	sp->cc = 0;

	sockopt_iphdrincl_swab_htosys (iph);
	memcpy (sp->s->data, ibuf->data + ibuf->getp, sdu_size);

	scngws_fifo_push (client->obufq_cl, sp);

	client->t_write_cl = thread_add_write (master, scngwc_packet_send, client, client->fd_cl);

	stream_free (ibuf);

	return 0;
}
