#!/usr/bin/perl

#
# Parse fsm (.fsm) file and generate the definition (.def) file for the 
# FSM Generation (FsmGen) tool.
#
#   Author: Giodi Giorgi.
#


###############################################################################
#                               Utilities                                     #
###############################################################################

# Checks args
sub check_args {
    my @args = @_;
    
    if ($#args != 0) {
	print "usage: $prog <fsm config file>\n";
	exit;
    }
    return @args;
}

# Checks file
sub check_file {
    my $file = $_[0];

    unless (-f $file) {
	print "File $file does not exist!\n";
	exit;
    }
    
    print "Using FSM config file '$file'\n";
}

# Finds the element inside a list - return the index of element or -1 if none 
sub find_elements {
    my ($elem, $list_ref) = @_;
    my @list = @$list_ref;
    my $res = 0;

    foreach my $item (@list) {
	if ($item eq $elem) {
	    return $count;
	}
	$count++;
    }
    return -1;
}

# Finds out and return the root event of a derived one (returns -1 otherwise)
sub find_root_event {
    my ($evs_ref, $derived_ev) = @_;

    my %evs = %$evs_ref;

    foreach my $root_ev (keys %evs) {
	if (&find_elements($derived_ev, \@{$evs{$root_ev}}) != -1) {
	    #print "Root event of '$derived_ev' is $root_ev!\n";
	    return $root_ev;
	}
    }
    return -1;
}

# Counts equal elements from two list
sub count_equal_elements {
    my ($list1_ref, $list2_ref) = @_;
    my @list1 = @$list1_ref;
    my @list2 = @$list2_ref;
    my $count = 0;

    foreach my $item1 (@list1) {
	foreach my $item2 (@list2) {
	    if ($item1 eq $item2) {
		#print "Equal element ($item1)\n";
		$count++;
	    }
	}
    }
    return $count;
}

# Finds out if inside a list of events, at least 2 have the same root ev
sub same_root_event {
    my ($evs_ref, $list_ref) = @_;
    my $count;

    my %evs = %$evs_ref;
    my @list = @$list_ref;

    foreach my $root_ev (keys %evs) {
	$count = &count_equal_elements(\@{$evs{$root_ev}}, \@list);
	if ($count >= 2) {
	    #print "Found some root event!\n";
	    return 1;
	}
    }
    return 0;
}


################################
#       Print utilities        #
################################
sub print_all_gens {
    my %hs = @_;
    print "GENERAL PARMS\n";
    foreach my $key (keys %hs) {
	print "- $key = $hs{$key}\n";
    }
}

sub print_all_evs {
    my %hs = @_;
    print "EVENTS\n";
    foreach my $key (keys %hs) {
	print "- $key -> ";
	foreach my $der (@{$hs{$key}}) {
	    print "$der ";
	}
	print "\n";
    }
}

sub print_all_transitions {
    my %hs = @_;
    print "TRANSITIONS\n";
    foreach my $key (keys %hs) {
	print "- $key:\n";
	foreach my $key2 ( keys %{ $hs{$key} } ) {
	    print "    $key2 -> $hs{$key}{$key2}\n";
	}
    }
}


###############################################################################
#                                Functions                                    #
###############################################################################
# gets all the generals FSM parameters
sub get_all_gen_parms {
    my $data = $_[0];

    my $target = "FSM";
    my $target_found = 0;
    my %gen_hash;

    foreach my $line (@$data) {
	$line =~ s/\#.*//g;    # remove comments
	$line =~ s/\s*//g;     # remove whitespace
	chomp($line);

	###print "->: $line\n";

	if ($line eq "") {
	    next;
	}

	if ($line =~ /^\{\s*(.+)\s*\}/) {
	    
	    if ($target_found) {
		last;
	    }
	    if ($1 eq $target) {
		###print "==> $1\n";
		$target_found = 1;
		next;
	    }
	}

	# SPECIFIC PART
	if ($target_found) {
	    ###print "LINE==> $line\n";

	    my ($key, $val) = split('=', $line);
	    $gen_hash{$key} = $val;
	} else {
	    #
	}
    }

    # Check -> to be improved
    if ($gen_hash{"definition-file"} eq "") {
	print "You MUST specify a definition-file into FSM config file!\n";
	exit;
    }

    return %gen_hash;
}


# gets all the events
sub get_all_events {
    my $data = $_[0];

    my $target = "Events";
    my $target_found = 0;
    my %ev_hash;

    foreach my $line (@$data) {
	$line =~ s/\#.*//g;    # remove comments
	$line =~ s/\s*//g;     # remove whitespace
	chomp($line);

	###print "->: $line\n";

	if ($line eq "") {
	    next;
	}

	if ($line =~ /^\{\s*(.+)\s*\}/) {
	    
	    if ($target_found) {
		last;
	    }
	    if ($1 eq $target) {
		###print "==> $1\n";
		$target_found = 1;
		next;
	    }
	}

	# SPECIFIC PART
	if ($target_found) {
	    ###print "LINE==> $line\n";

	    my ($root_ev, $derived_ev) = split('=', $line);
	    my @derived_evs = split(',', $derived_ev);
	    $ev_hash{$root_ev} = \@derived_evs;
	} else {
	    #
	}
    }

    return %ev_hash;
}


# gets the transition matrix
sub get_transitions {
    my ($data, $gen_ref) = @_;

    my $target = "States";
    my $target_found = 0;
    my %states_hash;

    my $state = "";
    my $state_found = 0;
    my $start_state = "";

    foreach my $line (@$data) {
	$line =~ s/\#.*//g;    # remove comments
	$line =~ s/\s*//g;     # remove whitespace
	chomp($line);

	###print "->: $line\n";

	if ($line eq "") {
	    next;
	}

	if ($line =~ /^\{\s*(.+)\s*\}/) {
	    
	    if ($target_found) {
		last;
	    }
	    if ($1 eq $target) {
		###print "==> $1\n";
		$target_found = 1;
		next;
	    }
	}
	
	# SPECIFIC PART
	if ($target_found) {
	    ###print "LINE==> $line\n";
	    
	    if ($line =~ /^\s*State=\s*(.+)/i) {
		$state = $1;
		if ($state eq "") {
		    print "Invalid state name!!!";
		    exit;
		}
		if ($start_state eq "") {
		    $start_state = $state;
		}
		$state_found = 1;
		###print "----> State: $state\n";
		next;
	    }
	    
	    if ($state_found) {
		my ($event, $next_state) = split("->", $line);
		$states_hash{$state}{$event} = $next_state;
		###print "---->   Added: $event -> $next_state\n";
	    } else {
		#
	    }
	}
    }

    # Sets start state to first state if not yet specified
    if (${%{$gen_ref}}{"start-state"} eq "") {
	${%{$gen_ref}}{"start-state"} = $start_state;
    }

    return %states_hash;
}


# Gets real FSM states and fills in the reference to virtual_states list
sub get_states {
    my ($evs_ref, $trans_ref, $virt_states_ref) = @_;

    my %trans = %$trans_ref;
    my %evs = %$evs_ref;

    my @real_states;
    my $res;
    my @root_evs = keys %evs;

    foreach my $state (keys %trans) {
	my @evs_list = keys %{ $trans{$state}} ;

	#GG removed check -> now all states are VIRTUAL
	$res = 1;
	#$res = &same_root_event($evs_ref, \@evs_list);
	if (!$res) {
	    #print "State '$state' is a real state\n";
	    push(@real_states, $state);
	} else {
	    #print "State '$state' is a VIRTUAL state\n";
	    push(@$virt_states_ref, $state);
	}
    }

    return @real_states;
}


# Opens file to write
sub open_file_to_write {
    my $file = $_[0];
    my $fd;
    
    (! -e $file) or 
	die "File '$file' already exists, remove it before start again! ";
    open ($fd, ">", $file) or die "Can not open '$file' for writing";
    
    return $fd;
}


# Initializes file to write
sub initialize_file {
    my $fd = $_[0];

    print $fd "Autogen Definitions FSM;\n";
    print $fd "fsm = {\n";
}


# Write general parameters
sub write_generals {
    my ($fd, %parms) = @_;

    foreach my $key (keys %parms) {
	unless ($key eq "definition-file") {
	    print $fd "\t$key = \"$parms{$key}\";\n";
	}
    }
}


# Write all states
sub write_states {
    my ($fd, $states_ref, $virt_states_ref) = @_;
    my @states = @$states_ref;
    my @virt_states = @$virt_states_ref;
    my $virtual = 0;

    print $fd "\n";
    print $fd "\tstates = {\n";
    foreach my $st (@states) {
	print $fd "\t\tstate = \"$st\";\n";
    }
    foreach my $st (@virt_states) {
	print $fd "\t\tstate = \"$st\";\n";
    }
    print $fd "\t};\n\n";

    print $fd "\tvirtual-states = {\n";
    foreach my $st (@virt_states) {
	print $fd "\t\tstate = \"virt_$st\";\n";
    }
    print $fd "\t};\n";
}


# Write all events
sub write_events {
    my ($fd, %evs) = @_;

    print $fd "\n";
    print $fd "\tevents = {\n";
    foreach my $key (keys %evs) {
	foreach my $der (@{$evs{$key}}) {
	    print $fd "\t\tevent = \"$der\";\n";
	}
    }
    print $fd "\t};\n\n";

    print $fd "\troot-events = {\n";
    foreach my $key (keys %evs) {
	print $fd "\t\tevent = \"$key\";\n";
    }
    print $fd "\t};\n\n";

    print $fd "\tevent-hierarchy = {\n";
    foreach my $key (keys %evs) {
	print $fd "\t\trelationship = {\n";
	print $fd "\t\t\troot-event = \"$key\";\n";
	foreach my $der (@{$evs{$key}}) {
	    print $fd "\t\t\tevent = \"$der\";\n";
	}
	print $fd "\t\t};\n";
    }
    print $fd "\t};\n";
}


# Writes all transitions
sub write_transitions {
    my ($fd, $evs_ref, $trans_ref, $virt_states_ref) = @_;

    #my %evs = %$evs_ref;
    my %trans = %$trans_ref;
    my @virt_states = @$virt_states_ref;
    my $root_ev = "";

    print $fd "\n";
    print $fd "\ttransitions = {\n";

    foreach my $state (keys %trans) {
	my $res = &find_elements($state, \@virt_states);
	my %evs = %$evs_ref;

	foreach my $ev (keys %{$trans{$state}}) {
	    my $from; 
	    #print "\tstate $state; res = $res; event = $ev;\n";
	    
	    if ($res != -1) {

		$root_ev = &find_root_event(\%evs, $ev);
		#print "\troot-event = $root_ev;\n";

		if ($root_ev != -1) {
		    print $fd "\t\ttransition = {\n";
		    print $fd "\t\t\tfrom-state = \"$state\"; to-state = " .
			"\"virt_$state\"; on-event = \"$root_ev\";\n";
		    print $fd "\t\t};\n";

		    #&print_all_evs(%evs);
		    # Deleting already written root_event from tmp events hash
		    delete $evs{$root_ev};
		}

		$from = "virt_" . $state;

	    } else {
		$from = $state;
	    }

	    print $fd "\t\ttransition = {\n";
	    print $fd "\t\t\tfrom-state = \"$from\"; to-state = " .
		"\"$trans{$state}{$ev}\"; on-event = \"$ev\";\n";
	    print $fd "\t\t};\n";
	}
    }
    print $fd "\t};\n";
}


# Finalizes file to write
sub finalize_file {
    my $fd = $_[0];

    print $fd "};\n";
}


# Close file
sub close_file {
    my ($fd, $file) = @_;

    close($fd);

    print "- - - - - - - -\n File '$file' created successfully!\n";
}


###############################################################################
#                             GraphViz Functions                              #
###############################################################################

# Initializes graphViz file to write
sub initialize_gfile {
    my $fd = $_[0];
    print $fd "digraph finite_state_machine {\n";
    print $fd "    ordering=in;\n";
    print $fd "    concentrate=true;\n";
    print $fd "    rankdir=LR;\n";
    print $fd "    ranksep=1.25;\n";
    print $fd "    //size=\"8,5\"\n";
    print $fd "    node[height = 1.3];\n";
    print $fd "    node [ fontsize=12 \n".
	"           fixedsize=true  \n".
	"           shape=circle  \n".
	"           color=lightsteelblue3 \n". 
	"           style=filled];\n";
    print $fd "    edge [fontsize=9];\n";
}

# Writes all transitions for graphViz file
sub write_gtrans {
    my ($fd, $evs_ref, $trans_ref, $virt_states_ref) = @_;

    my %trans = %$trans_ref;
    my @virt_states = @$virt_states_ref;
    my $root_ev = "";

    print $fd "\n";
    foreach my $key (keys %trans) {
	foreach my $key2 ( keys %{ $trans{$key} } ) {
	    if ($key eq $trans{$key}{$key2}) {
		print $fd "    $key -> $trans{$key}{$key2} [ label = \"$key2\" ];\n";
	    } else {
#		print $fd "    $key -> $trans{$key}{$key2} [ label = \"$key2->$trans{$key}{$key2}\" ];\n";
		print $fd "    $key -> $trans{$key}{$key2} [ label = \"$key2\" ];\n";
	    }
	}
    }

    #print $fd "\n";
    #print $fd "\ttransitions = {\n";
    #
    #foreach my $state (keys %trans) {
    #	my $res = &find_elements($state, \@virt_states);
    #	my %evs = %$evs_ref;
    #
    #	foreach my $ev (keys %{$trans{$state}}) {
    #	    my $from; 
    #	    #print "\tstate $state; res = $res; event = $ev;\n";
    #	    
    #	    if ($res != -1) {
    #
    #		$root_ev = &find_root_event(\%evs, $ev);
    #		#print "\troot-event = $root_ev;\n";
    #
    #		if ($root_ev != -1) {
    #		    print $fd "\t\ttransition = {\n";
    #		    print $fd "\t\t\tfrom-state = \"$state\"; to-state = " .
    #			"\"virt_$state\"; on-event = \"$root_ev\";\n";
    #		    print $fd "\t\t};\n";
    #
    #		    #&print_all_evs(%evs);
    #		    # Deleting already written root_event from tmp events hash
    #		    delete $evs{$root_ev};
    #		}
    #
    #		$from = "virt_" . $state;
    #
    #	    } else {
    #		$from = $state;
    #	    }
    #
    #	    print $fd "\t\ttransition = {\n";
    #	    print $fd "\t\t\tfrom-state = \"$from\"; to-state = " .
    #		"\"$trans{$state}{$ev}\"; on-event = \"$ev\";\n";
    #	    print $fd "\t\t};\n";
    #	}
    #}
    #print $fd "\t};\n";
}

# Finalizes file to write
sub finalize_gfile {
    my $fd = $_[0];

    print $fd "}\n";
}


###############################################################################
#                                Main Program                                 #
###############################################################################

# Parameters
$prog = "defGen.pl";

# Gets config file
($configfile) = &check_args(@ARGV);
&check_file($configfile);

# Opens config file
@filedata = <>;

# Gets general FSM parameters
%all_generals = &get_all_gen_parms(\@filedata);
&print_all_gens(%all_generals);

# Gets all events (hash with:  key => root_event,  data => derived event list)
%all_events = &get_all_events(\@filedata);
&print_all_evs(%all_events);

# Gets all transitions (hash of hash - key  => state,  
#                                      data => hash with key  => event
#                                                        data => next state )
%all_transitions = &get_transitions(\@filedata, \%all_generals);
&print_all_transitions(%all_transitions);


#
#TODO: check CONSISTENCY, ex:
#        - if all the next-states have the state section into config file
#        - if the derived events have only one root state
#        -  ...


#
# Infers real and virtual FSM states
#
@virtual_states = ();
@states = &get_states(\%all_events, \%all_transitions, \@virtual_states);


#
# Write definition file
#

# Opens write mode definition file
$fd = &open_file_to_write($all_generals{"definition-file"});

# Initializing file to write
&initialize_file($fd);

# Writes all general parameters
&write_generals($fd, %all_generals);

# Writes all states
&write_states($fd, \@states, \@virtual_states);

# Writes all general parameters
&write_events($fd, %all_events);

# Writes all transitions
&write_transitions($fd, \%all_events, \%all_transitions, \@virtual_states);

# Finalizing file to write
&finalize_file($fd);

&close_file($fd, $all_generals{"definition-file"});


#
# Write dot file for graphViz (if required)
#
unless ($all_generals{"graphviz-file"} eq "") {

    # Opens write mode graphViz file
    $fd = &open_file_to_write($all_generals{"graphviz-file"});

    # Initializing graphViz file to write
    &initialize_gfile($fd);

    # Writes all transitions
    &write_gtrans($fd, \%all_events, \%all_transitions, \@virtual_states);

    # Finalizing graphViz file to write
    &finalize_gfile($fd);
    
    &close_file($fd, $all_generals{"graphviz-file"});
}

exit(0);
