#!/usr/bin/python

import os, pickle, socket, sys, time, struct, binascii, select, threading, random, copy, pprint
from struct import calcsize, pack, unpack


########## RECORD FACTORY ########################

class ConcreteRecord:

    def __init__(self, name, fields, data=None):
        self._fields = fields
        self.name = name
        if data == None: return
        for f in fields:
            name,fmt,start,stop = f
            if 'T' in fmt:
                b = unpack('B', data[start:stop-2])[0]
                h = unpack('H', data[start+1:stop])[0]
                val = [(b<<16) + h]
            elif 'A' in fmt:
                val = socket.inet_ntoa(data[start:stop])
            else:
                val = unpack(fmt, data[start:stop])
            if len(val) == 1: val=val[0]
            self.__dict__[name] = val

    def build(self):
        lst = []
        for f in self._fields:
            name,fmt,start,stop = f
            val = self.__dict__.get(name)
            if type(val) in (tuple,list):      # U
                lst += [pack(fmt,*val)]     # G
            else:                                    # L
                lst += [pack(fmt,val)]      # Y
        return ''.join(lst)

    def calcsize(self):
        return self._fields[-1][3]    
    
    def repr(self):
        d = copy.copy(self.__dict__)
        del d['_fields']
        del d['name']
        return d
    
    def __str__(self):
        lst = []
        ret = "%s object:\n" % self.name
        for f in self._fields:
            name,fmt,start,stop = f
            val = self.__dict__.get(name)
            ret += "%20s: %s\n" % (name, str(val))
        return ret


class RecordFactory:

    def __init__(self,record_fmt, name='Unknown'):
        self.fields = []
        self.name = name
        pos = 0
        for field in record_fmt.split():
            if field[0] == "#": continue
            fmt,name = field.split('.')
            if 'T' in fmt:
                size = 3
            elif 'A' in fmt:
                size = 4
            else:
                size = calcsize(fmt)
            self.fields += [(name,fmt,pos,pos+size)]
            pos += size

    def parse(self, data):
        obj = ConcreteRecord(self.name, self.fields, data)
        return obj, data[obj.calcsize():]

#### CONCRETE FACTORY objects  ##########################

ospf_api_header = RecordFactory(
    name       = 'OSPF API header',
    record_fmt = '''
                 !B.version
                 !B.msgtype
                 !H.msglen
                 !I.seqnum
''')

ospf_api_register_opac = RecordFactory(
    name        = 'Register/Unregister opaque type',
    record_fmt  = """
                !B.lsa_type
                !B.opaque_type
                !H.unused
""")

ospf_api_r_event_s_lsdb = RecordFactory(
    name       = 'Register event/sync LSDB',
    record_fmt =  """
                !H.type_mask
                !B.orgin
                !B.num_areas
""")

ospf_api_lsa_originate = RecordFactory(
     name       = 'LSA originate request',
     record_fmt =  """
                !A.ifaddr
                !A.area_id
                !I.lsa_payload
""")

ospf_api_lsa_delete = RecordFactory(
     name       = 'LSA delete reguest',
     record_fmt =  """    
                !A.area_id
                !B.lsa_typ
                !B.opaque_type
                !H.unused
                !I.opaque_id
""")

ospf_api_replay = RecordFactory(
     name       = 'Replay',
     record_fmt =  """                               
                !B.error_code
                !B.unused
                !H.unused
""")

ospf_api_ready_notify = RecordFactory(
     name       = 'Ready notify',
     record_fmt =  """
                !B.lsa_typ
                !B.opaque_type
                !H.unused
                !A.ifaddr_id
""")

ospf_api_lsa_change = RecordFactory(
     name       = 'LSA change notify',
     record_fmt =  """                                   
                !A.ifaddr
                !A.area_id
                !B.self_originate
                !H.unused
                !B.unused
""")

ospf_api_new_interface = RecordFactory(
  name       = 'New Interface',
     record_fmt =  """                                 
                !A.ifaddr
                !A.area_id
""")

ospf_api_delete_interface = RecordFactory(
     name       = 'Delete Interface',
     record_fmt =  """
                !A.ifaddr
""")

ospf_api_ism_change = RecordFactory(
     name       = 'IMS Change',
     record_fmt =  """                              
                !A.ifaddr
                !A.area_id
                !B.ism_status
                !H.unused
                !B.unused
""")

ospf_api_nsm_change = RecordFactory(
     name       = 'NSM Change',
     record_fmt =  """                                 
                !A.ifaddr
                !A.nbraddr
                !A.router_id
                !B.nsm_status
                !B.unused
""")

msg_all_map = {
               1:ospf_api_register_opac,
               2:ospf_api_register_opac,
               3:ospf_api_r_event_s_lsdb,
               4:ospf_api_r_event_s_lsdb,
               5:ospf_api_lsa_originate,
               6:ospf_api_lsa_delete,
               10:ospf_api_replay,
               11:ospf_api_ready_notify,
               12:ospf_api_lsa_change,
               13:ospf_api_lsa_change,
               14:ospf_api_new_interface,
               15:ospf_api_delete_interface,
               16:ospf_api_ism_change,
               17:ospf_api_nsm_change
               }
               
########## OSPF LSAs #######################

########### TE TLVs    ####################

LSA_header = RecordFactory(
     name       = 'LSA header',
     record_fmt =  """                                 
                !H.LS_age
                !B.options
                !B.LS_type
                !B.opaque_type
                !T.opaque_id
                !A.adv_router
                !I.LS_seq
                !H.checksum
                !H.length
""")

TLV_header =  RecordFactory(
     name       = 'TLV header',
     record_fmt =  """                                 
                !H.type
                !H.length
""")

Router_address                  = RecordFactory(name = 'Router address',                      record_fmt = '!A.router_address')
Router_associated_area_ID = RecordFactory(name = 'Router associated area id',        record_fmt = '!I.router_associated_area_id')
Local_TE_router_ID             = RecordFactory(name = 'Local TE router ID',                   record_fmt = '!I.local_te_router_id')
Node_associated_area_ID   = RecordFactory(name = 'Node associated area id',          record_fmt = '!I.node_associated_area_id')


Link_type                            = RecordFactory(name = 'Link type',                               record_fmt = '!B.link_type')
Link_ID                               = RecordFactory(name = 'Link ID',                                   record_fmt = '!A.link_id')
Local_interface_address      = RecordFactory(name = 'Local interface IP address',       record_fmt = '!A.local_interface_address')
Remote_interface_address  = RecordFactory(name = 'Remote interface IP address',   record_fmt = '!A.remote_interface_address')
TE_metric                          = RecordFactory(name = 'Traffic engineering metric',        record_fmt = '!I.te_metric')
Max_bandwidth                  = RecordFactory(name = 'Maximum bandwidth',              record_fmt = '!I.max_bandwidth')
Max_reservable_bandwidth = RecordFactory(name = 'Maximum reservable bandwidth', record_fmt = '!I.max_reservable_bandwidth')
Unreserved_bandwidth        = RecordFactory(name = 'Unreserved bandwidth',            record_fmt = '!I.unreserved_bandwidth') 
Adminstrative_group           = RecordFactory(name = 'Administrative group',               record_fmt = '!I.adminstrative_group')

Local_node_ID                     = RecordFactory(name = 'Local node ID',                         record_fmt = '!A.local_node_id')
Remote_node_ID                 = RecordFactory(name = 'Remote node ID',                    record_fmt = '!A.remote_node_id')
Node_ID                              = RecordFactory(name = 'Node ID',                                record_fmt = '!A.node_id')
Ancestor_RC_ID                   = RecordFactory(name = 'Ancestor routing controller ID', record_fmt = '!A.ancestor_rc_id')
Associated_area_ID             = RecordFactory(name = 'Associated area ID',                 record_fmt = '!I.associated_area_id')
Span_length                        = RecordFactory(name = 'Span length',                          record_fmt = '!I.span_length')
OSNR                                 = RecordFactory(name = 'OSNR',                                   record_fmt = '!I.osnr')
Dpdm                                 = RecordFactory(name = 'Dpdm',                                   record_fmt = '!I.dpdm')

Link_local_remote_identifiers = RecordFactory(name = 'Link local/remote identifiers',
                        record_fmt = '''
                            !I.link_local_id
                            !I.link_remote_id''')

Link_protection_type = RecordFactory(name = 'Link protection type',
                        record_fmt ='''
                            !B.protection_type
                            ''')

Local_and_remote_TE_router_ID = RecordFactory(name = 'Local and remote TE router ID',
                        record_fmt = '''
                            !I.local_router_id
                            !I.remote_router_id''')

TNA_address_IPv4 = RecordFactory(name = 'TNA address IPv4',
                record_fmt = '''
                        !B.tna_ip4_address_length
                        !T.tna_ip4_reserved
                        !A.tna_address_ipv4''')

TNA_address_IPv6 = RecordFactory(name = 'TNA address IPv6',
                record_fmt = '''
                        !B.tna_ip6_address_length
                        !T.tna_ip6_reserved
                        !I.tna_address_ip6_p1
                        !I.tna_address_ip6_p2
                        !I.tna_address_ip6_p3
                        !I.tna_address_ip6_p4''')

TNA_address_nsap = RecordFactory(name = 'TNA address NSAP',
                 record_fmt = '''
                       !B.tna_nsap_address_length
                       !T.tna_nsap_reserved
                       !I.tna_address_nsap_p1
                       !I.tna_address_nsap_p2
                       !I.tna_address_nsap_p3
                       !I.tna_address_nsap_p4
                       !I.tna_address_nsap_p5''')

General_capabilities = RecordFactory(name = 'General capabilities',
                     record_fmt = '''
                           !B.general_cap
                           ''')

BER_estimate = RecordFactory(name = 'BER estimate',
                 record_fmt ='''
                        !B.ber_estimate
                        ''')

#Interface Switching Capability Descriptor TLV parse
def if_sw_cap_desc(data):
    obj = ConcreteRecord('Interface_switching_capability_descriptor', None, None)
    obj.__dict__['if_sw_cap_desc'] = {}
    switching_cap, encoding = struct.unpack("!BB",data[:2])
    obj.__dict__['if_sw_cap_desc']['switching_cap'] = switching_cap
    obj.__dict__['if_sw_cap_desc']['encoding'] = encoding
    data = data[4:]
    i = 0
    while i < 8:
        val = struct.unpack("!I",data[:4])[0]
        data = data[4:]
        obj.__dict__['if_sw_cap_desc']['nr_'+str(i)] = val
        i += 1
    if switching_cap in (1,2,3,4):    #PSC switching capability
        min_lsp_bw = struct.unpack("!I",data[:4])[0]
        mtu = struct.unpack("!H",data[4:6])[0]
        obj.__dict__['if_sw_cap_desc']['min_lsp_bw'] = min_lsp_bw
        obj.__dict__['if_sw_cap_desc']['mtu'] = mtu
    if switching_cap == 100:        #TDM switching capability
        min_lsp_bw = struct.unpack("!I",data[:4])[0]
        indication = struct.unpack("!B",data[4:5])[0]
        obj.__dict__['if_sw_cap_desc']['min_lsp_bw'] = min_lsp_bw
        obj.__dict__['if_sw_cap_desc']['indication'] = indication
    return obj

#Shared Risk Link Group TLV parse
def srlg(data):
    obj = ConcreteRecord('Shared_risk_link_group', None, None)
    i = 1
    obj.__dict__['srlg'] = {}
    while len(data) > 3:
        val = struct.unpack("!I",data[:4])[0]
        data = data[4:]
        obj.__dict__['srlg']['nr_'+str(i)] = val
        i += 1
    return obj

#Sonet and SDH Interface Switching Capability Descriptor TLV parse
def ssdh_if_sw_cap_desc(data):
    obj = ConcreteRecord('Sonet_SDH_interface_switching_capability_descriptor', None, None)
    i = 1
    obj.__dict__['ssdh_if_sw_cap_desc'] = {}
    switching_cap, encoding = struct.unpack("!BB",data[:2])
    obj.__dict__['ssdh_if_sw_cap_desc']['switching_cap'] = switching_cap
    obj.__dict__['ssdh_if_sw_cap_desc']['encoding'] = encoding
    data = data[4:]
    while len(data) > 3:
        type = struct.unpack("B",data[:1])[0]
        data = data[1:]
        b = struct.unpack('B', data[:1])[0]
        h = struct.unpack('!H', data[1:3])[0]
        slots = (b<<16) + h
        val = (type, slots)
        data = data[3:]
        obj.__dict__['ssdh_if_sw_cap_desc']['nr_'+str(i)] = val
        i += 1
    return obj

#Hierarchy List TLV parse
def hierarchy_list(data):
    obj = ConcreteRecord('Hierarchy_list', None, None)
    i = 1
    obj.__dict__['hierarchy_list'] = {}
    while len(data) > 3:
        val = socket.inet_ntoa(data[:4])
        data = data[4:]
        obj.__dict__['hierarchy_list']['nr_'+str(i)] = val
        i += 1
    return obj

#OSPF Downstream Associated Area ID TLV parse
def ospf_down_aa_id(data):
    obj = ConcreteRecord('OSPF_downstream_associated_area_ID', None, None)
    i = 1
    obj.__dict__['ospf_down_aa_id'] = {}
    while len(data) > 3:
        val = struct.unpack("!I",data[:4])[0]
        data = data[4:]
        obj.__dict__['ospf_down_aa_id']['nr_'+str(i)] = val
        i += 1
    return obj

#Technology Specific Bandwidth Accounting TLV parse
def band_account(data):
    obj = ConcreteRecord('Technology_specific_bandwidth_accounting', None, None)
    i = 1
    obj.__dict__['band_account'] = {}
    while len(data) > 3:
        type = struct.unpack("B",data[:1])[0]
        data = data[1:]
        b = struct.unpack('B', data[:1])[0]
        h = struct.unpack('H', data[1:3])[0]
        slots = (h<<8) + b
        val = (type, slots)
        data = data[3:]
        obj.__dict__['band_account']['nr_'+str(i)] = val
        i += 1
    return obj

#Amplifiers List TLV parse
def amplifiers_list(data):
    obj = ConcreteRecord('Amplifiers_list', None, None)
    i = 1
    obj.__dict__['amplifiers_list'] = {}
    while len(data) > 7:
        gain  = struct.unpack("!I",data[:4])[0]
        noise = struct.unpack("!I",data[4:8])[0]
        val = (gain, noise)
        data = data[8:]
        obj.__dict__['amplifiers_list']['nr_'+str(i)] = val
        i += 1
    return obj

#Available Wavelength Mask TLV parse
def av_wavelength_mask(data):
    obj = ConcreteRecord('Available_wavelength_mask', None, None)
    i = 1
    obj.__dict__['av_wavelength_mask'] = {}
    while len(data) > 3:
        val = struct.unpack("!I",data[:4])[0]
        data = data[4:]
        obj.__dict__['av_wavelength_mask']['nr_'+str(i)] = val
        i += 1
    return obj

#TE-link Calendar TLV parse
def te_link_calendar(data):
    obj = ConcreteRecord('TE_link_calendar', None, None)
    obj.__dict__['te_link_calendar'] = []
    while len(data) > 35:
        time = struct.unpack("!I",data[:4])[0]
        band0 = struct.unpack("!I",data[4:8])[0]
        band1 = struct.unpack("!I",data[8:12])[0]
        band2 = struct.unpack("!I",data[12:16])[0]
        band3 = struct.unpack("!I",data[16:20])[0]
        band4 = struct.unpack("!I",data[20:24])[0]
        band5 = struct.unpack("!I",data[24:28])[0]
        band6 = struct.unpack("!I",data[28:32])[0]
        band7 = struct.unpack("!I",data[32:36])[0]
        val = {'time':time, 'band0':band0, 'band1':band1, 'band2':band2, 'band3':band3, 'band4':band4, 'band5':band5, 'band6':band6, 'band7':band7}
        val = {'te_item' : val}
        data = data[36:]
        obj.__dict__['te_link_calendar'].append(val)
    return obj

#Node IPv4 Local Prefix TLV parse
def node_ip4_lcl_prefix(data):
    obj = ConcreteRecord('Node_ipv4_local_prefix', None, None)
    obj.__dict__['node_ip4_lcl_prefix'] = []
    while len(data) > 7:
        mask    = socket.inet_ntoa(data[:4])
        address = socket.inet_ntoa(data[4:8])
        val = {'mask':mask, 'address':address}
        val = {'ip4_prefix' : val}
        data = data[8:]
        obj.__dict__['node_ip4_lcl_prefix'].append(val)
    return obj

#Node IPv6 Local Prefix TLV parse
def node_ip6_lcl_prefix(data):
    obj = ConcreteRecord('Node_ipv6_local_prefix', None, None)
    obj.__dict__['node_ip6_lcl_prefix'] = []
    while len(data) > 19:
        prefixlen, prefixopt, reserved = struct.unpack("!BBH",data[:4])
        address_ip6_p1 = struct.unpack("!I",data[4:8])[0]
        address_ip6_p2 = struct.unpack("!I",data[8:12])[0]
        address_ip6_p3 = struct.unpack("!I",data[12:16])[0]
        address_ip6_p4 = struct.unpack("!I",data[16:20])[0]
        val = {'prefixlen':prefixlen, 'prefixopt':prefixopt, 'address_ip6_p1':address_ip6_p1,
                'address_ip6_p2':address_ip6_p2, 'address_ip6_p3':address_ip6_p3, 'address_ip6_p4':address_ip6_p4}
        val = {'ip6_prefix' : val}
        data = data[20:]
        obj.__dict__['node_ip6_lcl_prefix'].append(val)
    return obj

TE_ROUTER_TLVs = {
     1: Router_address,
     100: Router_associated_area_ID             #temporary TLV number
}

TE_LINK_TLVs = {
      1 : Link_type,
      2 : Link_ID,
      3 : Local_interface_address,
      4 : Remote_interface_address,
      5 : TE_metric,
      6 : Max_bandwidth,
      7 : Max_reservable_bandwidth,
      8 : Unreserved_bandwidth,
      9 : Adminstrative_group,
      11: Link_local_remote_identifiers,
      14: Link_protection_type,
      17: Local_and_remote_TE_router_ID,
      32773: Local_node_ID,
      32774: Remote_node_ID,
      32790: General_capabilities,
      32792: Ancestor_RC_ID,
      32795: Associated_area_ID,
      32796: BER_estimate,
      32781: Span_length,
      32798: OSNR,
      32780: Dpdm,
}

TE_TNA_TLVs = {
      32776: TNA_address_IPv4,
      32778: TNA_address_IPv6,
      32779: TNA_address_nsap,
      32777: Node_ID,
}

TE_NODE_ATTR_TLVs = {
     5: Local_TE_router_ID,
     40005: Node_associated_area_ID
}

TE_TLVs = {
     1: ('Router Address', TE_ROUTER_TLVs),
     2: ('Link',                   TE_LINK_TLVs),
     3: ('TNA',                  TE_TNA_TLVs),
     4: ('Node Attribute',  TE_NODE_ATTR_TLVs)
}


########### Grid TLVs    ####################

ID = RecordFactory(
    name       = 'ID',
    record_fmt = '''
                !I.Id
''')

#Grid Site Latitude TLV parse
def grid_site_latitude(data):
    obj = ConcreteRecord('Grid_site_latitude', None, None)
    obj.__dict__['Latitude'] = {}
    val = struct.unpack("B",data[:1])[0]
    val = val>>2
    val = val & 0x3f
    obj.__dict__['Latitude']['resolution'] = val
    val = struct.unpack("!H",data[:2])[0]
    val = val>>1
    val = val & 0x1ff
    obj.__dict__['Latitude']['integer_part'] = val
    val = struct.unpack("!I",data[1:5])[0]
    val = val & 0x1ffffff
    obj.__dict__['Latitude']['fractional_part'] = val
    return obj

#Grid Site Longitude TLV parse
def grid_site_longitude(data):
    obj = ConcreteRecord('Grid_site_longitude', None, None)
    obj.__dict__['Longitude'] = {}
    val = struct.unpack("B",data[:1])[0]
    val = val>>2
    val = val & 0x3f
    obj.__dict__['Longitude']['resolution'] = val
    val = struct.unpack("!H",data[:2])[0]
    val = val>>1
    val = val & 0x1ff
    obj.__dict__['Longitude']['integer_part'] = val
    val = struct.unpack("!I",data[1:5])[0]
    val = val & 0x1ffffff
    obj.__dict__['Longitude']['fractional_part'] = val
    return obj
    
#Grid Site Name TLV parse
def grid_site_name(data):
    obj = ConcreteRecord('Grid_site_name', None, None)
    obj.__dict__['Name'] = {}
    name = ""
    while len(data) > 0:
        val = struct.unpack("B",data[:1])[0]
        name = name+"%c" % val         
        data = data[1:]
    obj.__dict__['Name']['name'] = name
    return obj
    
Grid_Site_sTLV = {
    1:ID
    }

Parent_Site_ID = RecordFactory(
    name       = 'Parent_Site_ID',
    record_fmt = '''
                !I.Parent_Site_Id
''')

Service_Info = RecordFactory(
    name       = 'Service_Info',
    record_fmt = '''
                !H.Service_type
                !H.Service_version
''')

Status = RecordFactory(
    name       = 'Status',
    record_fmt = '''
                !B.Status
''')

Address_Length = RecordFactory(
    name       = 'Address_Length',
    record_fmt = '''
                      !B.Address_length
''')

IPv4_Endpoint = RecordFactory(name = 'IPv4_Endpoint',
                record_fmt = '''
                        !A.ipv4_endpoint''')

IPv6_Endpoint = RecordFactory(name = 'IPv6_Endpoint',
                record_fmt = '''
                        !I.ip6_endpoint_p1
                        !I.ip6_endpoint_p2
                        !I.ip6_endpoint_p3
                        !I.ip6_endpoint_p4''')

NSAP_Endpoint = RecordFactory(name = 'NSAP_Endpoint',
                 record_fmt = '''
                       !I.nsap_endpoint_p1
                       !I.nsap_endpoint_p2
                       !I.nsap_endpoint_p3
                       !I.nsap_endpoint_p4
                       !I.nsap_endpoint_p5''')
                    
Grid_Service_sTLV = {
    1:ID,
    2:Parent_Site_ID,
    3:Service_Info,
    4:Status,
    5:Address_Length,
    6:IPv4_Endpoint,
    7:IPv6_Endpoint,
    8:NSAP_Endpoint,
}

LRMS_Info = RecordFactory(
    name       = 'LRMS_Info',
    record_fmt = '''
                !H.LRMS_type
                !H.LRMS_version
''')

Gatekeeper_Port = RecordFactory(
    name       = 'Gatekeeper_port',
    record_fmt = '''
                !I.gatekeeper
''')

Default_Storage_Element = RecordFactory(
    name       = 'Default_Storage_Element',
    record_fmt = '''
                !I.Default_SE
''')

Jobs_States = RecordFactory(
    name       = 'Jobs_States',
    record_fmt = '''
                !H.FreeJobSlots
                !B.Status
''')

Jobs_Stat= RecordFactory(
    name       = 'Jobs_Stat',
    record_fmt = '''
                !I.Running_Jobs
                !I.Waiting_Jobs
                !I.Total_Jobs
''')

Jobs_Time_Policy = RecordFactory(
    name       = 'Jobs_Time_Policy',
    record_fmt = '''
                !I.Max_WallClock_Time
                !I.Max_Obtainable_WallClock_Time
                !I.Max_CPU_Time
                !I.Max_Obtainable_CPU_Time
''')

Jobs_Time_Performance = RecordFactory(
    name       = 'Jobs_Time_Performance',
    record_fmt = '''
                !I.Estimated_Respose_Time
                !I.Worst_Response_Time
''')

Jobs_Load_Policy = RecordFactory(
    name       = 'Jobs_Load_Policy',
    record_fmt = '''
                !I.Max_Total_Jobs
                !I.Max_Running_Jobs
                !I.Max_Waiting_Jobs
                !H.Assined_Job_Slots
                !H.Max_Slots_Per_Jobs
                !B.Priority
''')

IPv4_HostName = RecordFactory(name = 'IPv4_Hostname',
                record_fmt = '''
                        !A.ipv4_hostname''')

IPv6_HostName = RecordFactory(name = 'IPv6_Hostname',
                record_fmt = '''
                        !I.ip6_hostname_p1
                        !I.ip6_hostname_p2
                        !I.ip6_hostname_p3
                        !I.ip6_hostname_p4''')

NSAP_HostName = RecordFactory(name = 'NSAP_Hostname',
                 record_fmt = '''
                       !I.nsap_hostname_p1
                       !I.nsap_hostname_p2
                       !I.nsap_hostname_p3
                       !I.nsap_hostname_p4
                       !I.nsap_hostname_p5''')

#Grid CE Jobs Load Policy TLV parse
def grid_ce_jobs_load_policy(data):
    obj = ConcreteRecord('Grid_ce_jobs_load_policy', None, None)
    obj.__dict__['Jobs_load_policy'] = {}
    val = struct.unpack("!I",data[:4])[0]
    obj.__dict__['Jobs_load_policy']['max_total_jobs'] = val
    val = struct.unpack("!I",data[4:8])[0]
    obj.__dict__['Jobs_load_policy']['max_running_jobs'] = val
    val = struct.unpack("!I",data[8:12])[0]
    obj.__dict__['Jobs_load_policy']['max_waiting_jobs'] = val
    val = struct.unpack("!H",data[12:14])[0]
    obj.__dict__['Jobs_load_policy']['assigned_job_slots'] = val
    val = struct.unpack("!H",data[14:16])[0]
    obj.__dict__['Jobs_load_policy']['max_slots_per_jobs'] = val
    val = struct.unpack("B",data[16:17])[0]
    temp = val & 0x1
    obj.__dict__['Jobs_load_policy']['preemption_flag'] = temp
    val = val>>1
    val = val & 0x7f
    obj.__dict__['Jobs_load_policy']['priority'] = val
    return obj

#Grid Computing Element Job Manager TLV parse
def grid_ce_job_manager(data):
    obj = ConcreteRecord('Grid_ce_job_manager', None, None)
    obj.__dict__['Job_manager'] = {}
    jm = ""
    while len(data) > 0:
        val = struct.unpack("B",data[:1])[0]
        jm = jm+"%c" % val         
        data = data[1:]
    obj.__dict__['Job_manager']['job_manager'] = jm
    return obj

#Grid Computing Element Data Dir TLV parse
def grid_ce_data_dir(data):
    obj = ConcreteRecord('Grid_ce_data_dir', None, None)
    obj.__dict__['Data_dir'] = {}
    dd = ""
    while len(data) > 0:
        val = struct.unpack("B",data[:1])[0]
        dd = dd+"%c" % val         
        data = data[1:]
    obj.__dict__['Data_dir']['data_dir'] = dd
    return obj

Grid_ComputingE_sTLV = {
    1:ID,
    2:Parent_Site_ID,
    3:LRMS_Info,
    4:Address_Length,
    5:IPv4_HostName,
    6:IPv6_HostName,
    7:NSAP_HostName,
    8:Gatekeeper_Port,
    11:Default_Storage_Element,
    12:Jobs_States,
    13:Jobs_Stat,
    14:Jobs_Time_Performance,
    15:Jobs_Time_Policy,
}


CPU_Info = RecordFactory(
    name       = 'CPU_Info ',
    record_fmt = '''
                !I.Physical_CPUs
                !I.Logical_CPUs
                !B.CPU_Arch
''')

OS_Info = RecordFactory(
    name       = 'OS_Info ',
    record_fmt = '''
                !H.OS_Type
                !H.Os_info
''')

Memory_Info = RecordFactory(
    name       = 'Memory_Info ',
    record_fmt = '''
                !I.Ram_Size
                !I.Virtual_Memory_Size
''')

#Grid Computing Element Calendar TLV parse
def grid_ce_calendar(data):
    obj = ConcreteRecord('CE_calendar', None, None)
    obj.__dict__['CE_Calendar'] = []
    while len(data) > 5:
        time = struct.unpack("!I",data[:4])[0]
        fjs = struct.unpack("!H",data[4:6])[0]
        val = {'time':time, 'free_job_slots':fjs}
        val = {'ce_item' : val}
        data = data[6:]
        obj.__dict__['CE_Calendar'].append(val)
    return obj

Grid_Sub_Cluster_sTLV = {
    1:ID,
    2:Parent_Site_ID,
    3:CPU_Info,
    4:OS_Info,
    5:Memory_Info,
}

#Grid SubCluster Calendar TLV parse
def grid_sc_calendar(data):
    obj = ConcreteRecord('SubCluster_calendar', None, None)
    obj.__dict__['SC_Calendar'] = []
    while len(data) > 7:
        time = struct.unpack("!I",data[:4])[0]
        p_cpus = struct.unpack("!H",data[4:6])[0]
        l_cpus = struct.unpack("!H",data[6:8])[0]
        val = {'time':time, 'physical_cpus':p_cpus, 'logical_cpus':l_cpus}
        val = {'sc_item' : val}
        data = data[8:]
        obj.__dict__['SC_Calendar'].append(val)
    return obj

#Grid SubCluster Software Package TLV parse
def grid_soft_pack(data):
    obj = ConcreteRecord('Package', None, None)
    obj.__dict__['Package'] = {}
    soft_type = struct.unpack("!H",data[:2])[0]
    soft_version = struct.unpack("!H",data[2:4])[0]
    obj.__dict__['Package']['soft_type'] = soft_type
    obj.__dict__['Package']['soft_version'] = soft_version
    data = data[4:]
    env_set = ""
    while len(data) > 0:
        val = struct.unpack("B",data[:1])[0]
        env_set = env_set+"%c" % val    
        data = data[1:]
    obj.__dict__['Package']['environment_setup'] = env_set
    return obj

Storage_Info = RecordFactory(
    name       = 'Storage_Info ',
    record_fmt = '''
                !B.Arch_and_Status
                !T.Access_p_and_Control_p
''')

#Grid Storage Element Info TLV parse
def grid_se_info(data):
    obj = ConcreteRecord('Storage_Info', None, None)
    obj.__dict__['Storage_Info'] = {}
    val = struct.unpack("B",data[:1])[0]
    temp = val>>4
    temp = temp & 0xf
    obj.__dict__['Storage_Info']['architecture'] = temp
    val = val & 0xf
    obj.__dict__['Storage_Info']['status'] = val
    val = struct.unpack('!H',data[1:3])[0]
    val = val>>4
    val = val & 0xfff
    obj.__dict__['Storage_Info']['access_protocols'] = val
    val = struct.unpack('!H',data[2:4])[0]
    val = val & 0xfff
    obj.__dict__['Storage_Info']['control_protocols'] = val
    return obj

Online_Size = RecordFactory(
    name       = 'Online_Size ',
    record_fmt = '''
                !I.Online_Total_Size
                !I.Online_Used_Size
''')

Nearline_size = RecordFactory(
    name       = 'Nearline_Size ',
    record_fmt = '''
                !I.Nearline_Total_Size
                !I.Nearline_Used_Size
''')

#Grid Storage Element Calendar TLV parse
def grid_se_calendar(data):
    obj = ConcreteRecord('SE_calendar', None, None)
    obj.__dict__['SE_Calendar'] = []
    while len(data) > 11:
        time = struct.unpack("!I",data[:4])[0]
        fos = struct.unpack("!I",data[4:8])[0]
        fns = struct.unpack("!I",data[8:12])[0]
        val = {'time':time, 'free_online_size':fos, 'free_nearline_size':fns}
        val = {'se_item' : val}
        data = data[12:]
        obj.__dict__['SE_Calendar'].append(val)
    return obj

#Grid Storage Element Storage Area TLV parse
def grid_storage_area(data):
    obj = ConcreteRecord('Storage', None, None)
    obj.__dict__['Storage'] = {}
    i = 0
    name = ""
    path = ""
    val = struct.unpack("B",data[:1])[0]
    while val != 0:
        data = data[1:]
        i = i + 1
        name = name+"%c" % val
        val = struct.unpack("B",data[:1])[0]
    while val == 0:
        data = data[1:]
        i = i + 1
        val = struct.unpack("B",data[:1])[0]
    while val != 0:
        data = data[1:]
        i = i + 1
        path = path+"%c" % val
        val = struct.unpack("B",data[:1])[0]
    while i%4 != 0:
        data = data[1:]
        i = i + 1
    obj.__dict__['Storage']['name'] = name
    obj.__dict__['Storage']['path'] = path
    totalOnlineSize = struct.unpack("!I",data[:4])[0]
    obj.__dict__['Storage']['totalOnlineSize'] = totalOnlineSize
    freeOnlineSize = struct.unpack("!I",data[4:8])[0]
    obj.__dict__['Storage']['freeOnlineSize'] = freeOnlineSize
    resTotalOnlineSize = struct.unpack("!I",data[8:12])[0]
    obj.__dict__['Storage']['resTotalOnlineSize'] = resTotalOnlineSize
    totalNearlineSize = struct.unpack("!I",data[12:16])[0]
    obj.__dict__['Storage']['totalNearlineSize'] = totalNearlineSize
    freeNearlineSize = struct.unpack("!I",data[16:20])[0]
    obj.__dict__['Storage']['freeNearlineSize'] = freeNearlineSize
    resNearlineSize = struct.unpack("!I",data[20:24])[0]
    obj.__dict__['Storage']['resNearlineSize'] = resNearlineSize
    val = struct.unpack("B",data[24:25])[0]
    retention_policy = val>>4
    retention_policy = retention_policy & 0xf
    obj.__dict__['Storage']['retention_policy'] = retention_policy
    access_latency = val & 0xf
    obj.__dict__['Storage']['access_latency'] = access_latency
    expiration_mode = struct.unpack("B",data[25:26])[0]
    expiration_mode = expiration_mode>>4
    expiration_mode = expiration_mode & 0xf
    obj.__dict__['Storage']['expiration_mode'] = expiration_mode
    return obj

Grid_StorageE_sTLV = {
    1:ID,
    2:Parent_Site_ID,
    4:Online_Size,
    5:Nearline_size,
}

Grid_TLVs = {
    1: ('Grid Site', Grid_Site_sTLV),
    2: ('Grid Service',Grid_Service_sTLV),
    3: ('Grid Computing Element',Grid_ComputingE_sTLV),
    4: ('Grid Sub-Cluster',Grid_Sub_Cluster_sTLV),
    5: ('Grid Storage Element',Grid_StorageE_sTLV)
}

########## utility functions ####################

def html_generator(bb):

    def aaa(d):
        #xml_string="<data>"
        xml_string=""
        def generate_xml_object(obj, index=1):
            xml_string = ""
            ind = 2*" "*index
            for key, value in obj.items():
                #try:
                key = key.replace(' ', '_')
                #except :
                    #pass
                if type(value) != type({}): 
                    if type(value)==type([]):
                        xml_string+= "%s<%s>\n" % (ind,key)
                        for record in value:
                            xml_string+= generate_xml_object(record, index+1)
                        xml_string+= "%s</%s>\n" % (ind,key)
                    else:
                        xml_string+= "%s<entry><name>%s</name><value>%s</value></entry>\n" % (ind, key, str(value))
                elif value == {}:
                    xml_string+= "%s<%s></%s>\n" % (ind, key, key)
                else:
                    xml_string+= "%s<%s>\n" % (ind,key)
                    xml_string+= generate_xml_object(value, index+1)
                    xml_string+= "%s</%s>\n" % (ind,key)
            return xml_string
        
        xml_string += generate_xml_object(d)
        #x=xml_string+'</data>'
        x=xml_string
        return x
    w= ""
    for key, value in bb.items():

        if key == 'Grid':
            #print key
            local_obj ={"Grid":bb["Grid"]}
            w="<data>"
            w+=aaa(local_obj)
            w+="</data>"
           #print w
        if key == 'Traffic Engineering':
            #print key
            local_obj ={"Traffic Engineering":bb["Traffic Engineering"]}   
            w+="<data>"
            w+=aaa(local_obj)
            w+="</data>"
            #print w
    return w

#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

def pars_xml(xml_s):
    if not os.path.exists("data5.xml"):
        xml_string2="""<?xml version="1.0" encoding="UTF-8"?>\n"""
        xml_string2+="""<?xml-stylesheet type="text/xsl" href="./xslt/transform.xsl"?>"""
        xml_string2+="""<opaque>"""
        xml_string2+="""</opaque>"""
        f = file("data5.xml", 'w')
        f.write(xml_string2)
        f.close()

    f = open("data5.xml", 'r')
    buff = f.read()
    f.close()
    f = open("data5.xml", 'w')
    opaque_shift = len('</opaque>')
    buff = buff[:-opaque_shift] + xml_s +  buff[-opaque_shift:]
    f.write(buff)
    f.close()
    
#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

def tlvs(data):
    '''TLVs parser generator'''
    while data:
        try:
            type, length = struct.unpack('!HH', data[:4])
            value = data[4:4+length]
        except: 
            print "Unproper TLV structure found: ", (data,)
            break
        yield type, value
        rest = length % 4
        if rest != 0:
            length = length + 4 - rest
        data = data[4+length:]
 
#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

def konwersja(id_opa,type_opa):
    def numToAdres(n):
        "convert long int to dotted quad string"
        return socket.inet_ntoa(struct.pack('!L',n))
    adres = type_opa << 24
    id_opa= id_opa>>8
    adres = adres ^ id_opa
    adress = numToAdres(adres)
    return adress

#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

def TE_LSA_parse(data):
    '''parse Traffic Engineering TLVs'''
    print '  Traffic Engineering TLVs:'
    obj = {}
    for type, tlv_data in tlvs(data):
        tlv_type, subtlvs_map = TE_TLVs.get(type, ('Unknown', {}))
        obj[tlv_type] = {}
        tlv_value_obj= {}
        for subtype, subtlv_data in tlvs(tlv_data):     
            subtlv_concrete = subtlvs_map.get(subtype)
            if subtlv_concrete:
                print '      --%s' % subtlv_concrete.name
            if type == 1:           #Router Address
                if subtlv_concrete:  tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data) 
            elif type == 2:        #Link 
                if subtlv_concrete: tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
                elif subtype == 15:          tlv_value_obj = if_sw_cap_desc(subtlv_data)                   
                elif subtype == 16:          tlv_value_obj = srlg(subtlv_data)                                  
                elif subtype == 32775:    tlv_value_obj = ssdh_if_sw_cap_desc(subtlv_data)   
                elif subtype == 32791:    tlv_value_obj = hierarchy_list(subtlv_data)                
                elif subtype == 32793:    tlv_value_obj = band_account(subtlv_data)                
                elif subtype == 32794:    tlv_value_obj = ospf_down_aa_id(subtlv_data)           
                elif subtype == 32782:    tlv_value_obj = amplifiers_list(subtlv_data)           
                elif subtype == 32783:    tlv_value_obj = av_wavelength_mask(subtlv_data)
                elif subtype == 32784:    tlv_value_obj = te_link_calendar(subtlv_data)
            elif type == 3:       #TNA
                tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
            elif type == 4:      #Node Attribute
                if subtlv_concrete:  tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
                elif subtype == 3:    tlv_value_obj = node_ip4_lcl_prefix(subtlv_data)
                elif subtype == 4:    tlv_value_obj = node_ip6_lcl_prefix(subtlv_data)
        
            if tlv_value_obj: obj[tlv_type].update(tlv_value_obj.repr())
    return obj

#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

def GRID_LSA_parse(data):
    '''parse Grid TLVs'''
    print '  Grid TLVs:'
    obj = {}
    sp = 1  #software package tlv index
    sa = 1  #storage area tlv index
    for type, tlv_data in tlvs(data):
        tlv_type, subtlvs_map = Grid_TLVs.get(type, ('Unknown', {}))
        obj[tlv_type] = {}
        tlv_value_obj= {}
        #parse Grid subTLVs
        for subtype, subtlv_data in tlvs(tlv_data): 
            subtlv_concrete = subtlvs_map.get(subtype)
            if subtlv_concrete:
                print '      --%s' % subtlv_concrete.name
            mul = False
            if type == 1: #Grid Site
                if subtlv_concrete:
                    tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
                    tlv_value_obj = {Grid_Site_sTLV[subtype].name : tlv_value_obj.repr()}
                elif subtype == 2:  tlv_value_obj = grid_site_name(subtlv_data)
                elif subtype == 3:  tlv_value_obj = grid_site_latitude(subtlv_data) 
                elif subtype == 4:  tlv_value_obj = grid_site_longitude(subtlv_data) 
            elif type == 2:  #Grid Service
                if subtlv_concrete:
                    tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
                    tlv_value_obj = {Grid_Service_sTLV[subtype].name : tlv_value_obj.repr()}
            elif type == 3:  #Grid Computing Element
                if subtlv_concrete:  
                    tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
                    tlv_value_obj = {Grid_ComputingE_sTLV[subtype].name : tlv_value_obj.repr()}
                elif subtype == 9:  tlv_value_obj = grid_ce_job_manager(subtlv_data)
                elif subtype == 10:  tlv_value_obj = grid_ce_data_dir(subtlv_data)
                elif subtype == 16:  tlv_value_obj = grid_ce_jobs_load_policy(subtlv_data)
                elif subtype == 17:  tlv_value_obj = grid_ce_calendar(subtlv_data)
            elif type == 4:  #Grid Sub-Cluster
                if subtlv_concrete:
                    tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
                    tlv_value_obj = {Grid_Sub_Cluster_sTLV[subtype].name : tlv_value_obj.repr()}
                elif subtype == 6:  
                    mul = True
                    tlv_value_obj = grid_soft_pack(subtlv_data)
                    sp = sp + 1
                    if sp < 3:
                        obj[tlv_type]['Software_Package'] = []
                        obj[tlv_type]['Software_Package'].append(tlv_value_obj.repr())
                    else:
                        obj[tlv_type]['Software_Package'].append(tlv_value_obj.repr())
                elif subtype == 7:  tlv_value_obj = grid_sc_calendar(subtlv_data)
            elif type == 5: #Grid Storage Element
                if subtlv_concrete:
                    tlv_value_obj, _unused = subtlv_concrete.parse(subtlv_data)
                    tlv_value_obj = {Grid_StorageE_sTLV[subtype].name : tlv_value_obj.repr()}
                elif subtype == 3:  tlv_value_obj = grid_se_info(subtlv_data)
                elif subtype == 6:   
                    mul = True
                    tlv_value_obj = grid_storage_area(subtlv_data)       
                    sa = sa + 1                                                                   
                    if sa < 3:
                        obj[tlv_type]['Storage_Area'] = []
                        obj[tlv_type]['Storage_Area'].append(tlv_value_obj.repr())
                    else:
                        obj[tlv_type]['Storage_Area'].append(tlv_value_obj.repr())
                elif subtype == 7:  tlv_value_obj = grid_se_calendar(subtlv_data)     
            
            if tlv_value_obj: 
                if mul: pass
                elif subtlv_concrete: obj[tlv_type].update(tlv_value_obj)
                else: obj[tlv_type].update(tlv_value_obj.repr())
    return obj

#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

def msg_parse(data):
    '''parse OSPF API message'''
    def apend(type, type_obj):
        if type not in obj:
            obj[type] = [type_obj]
        else:
            obj[type].append(type_obj)

    if not data:
        return
    obj = {}
    msg_parse.i  += 1
    #parse OSPF API header
    header, data = ospf_api_header.parse(data)
    factory = msg_all_map.get(header.msgtype)
    if not factory:
        print "Unknown message type: ", header.msgtype
        return
    #parse OSPF API message
    print '-'*60 +'\n%i: Message received: %s' % (msg_parse.i, factory.name)
    obj1,data = factory.parse(data)
    if header.msgtype in (3, 4):
        if obj1.num_areas > 0:
            i = 0
            num_areas = struct.unpack('!'+'I'*obj1.num_areas, data)
            data = data[obj1.num_areas*4:]
            for num_area in num_areas:
                obj1.__dict__['num_area_'+str(i)] = num_area
    #OSPF API message parsed
    if header.msgtype in (5, 12, 13):
        #parse LSA header
        while data:
            lsa_header,data = LSA_header.parse(data)
            if lsa_header.LS_type in (1,2):
                data=data[lsa_header.length:]
            if lsa_header.LS_type in (9, 10, 11):
                if lsa_header.opaque_type == 248: 
                    type, length = struct.unpack('!HH', data[:4])
                    value = data[:4+length] 
                    apend('Grid', GRID_LSA_parse(value))
                    data=data[lsa_header.length:]          
                elif lsa_header.opaque_type == 1 : # TE TLV 
                    type, length = struct.unpack('!HH', data[:4])
                    value = data[:4+length]            
                    adres_opaq= konwersja(lsa_header.opaque_id,lsa_header.opaque_type)
                    te_obj = TE_LSA_parse(value)
                    if 'Link' in te_obj:
                        a=konwersja(lsa_header.opaque_id,lsa_header.opaque_type)
                        te_obj['Link']['Adv_router'] = str(lsa_header.adv_router)
                        te_obj['Link']['opaque_id'] = str(a)
                    apend('Traffic Engineering', te_obj)                       
                    data=data[lsa_header.length:]
    return obj
msg_parse.i = 0

#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

def msg_process(data):
    tree_obj = msg_parse(data)
    pprint.pprint(tree_obj)
    xhtml = html_generator(tree_obj)
    pars_xml(xhtml)
    print 'Messaged processed succesfully'


######### SYNC and ASYNC communication ############

class Server(threading.Thread):
    def __init__(self, server_address): 
        threading.Thread.__init__(self) 
        self.host = '' 
        self.syn_port = random.randint(4000,5000) 
        self.asyn_port = self.syn_port + 1
        self.asyn_conn = None
        self.syn_conn = None 
        self.threads = [] 
        self.dest_addr = server_address
        self.dest_port = 2607
        self.seq_num = 1
        self.active = True

    def open_sockets(self): 
        try: 
            self.asyn_conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
            self.asyn_conn.bind((self.host,self.asyn_port)) 
            self.asyn_conn.settimeout(5)  # timeout for incoming connection accepting
            self.asyn_conn.listen(1) 
        except socket.error, (value,message): 
            if self.asyn_conn: 
                self.asyn_conn.close() 
            print "Could not open asyn socket: " + message 
            sys.exit(1) 
        print "Asyn socket is running"

        try:
            self.syn_conn = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
            self.syn_conn.bind((self.host,self.syn_port)) 
            self.syn_conn.connect((self.dest_addr, self.dest_port)) 
        except socket.error, (value,message): 
            if self.syn_conn: 
                self.syn_conn.close() 
            print "Could not open syn socket: " + message
            print "Please check Quagga is working correctly"
            sys.exit(1) 
        print "Syn socket is running"

    def run(self): 
        self.open_sockets() 
        # handle the asyn server socket 
        c = AsynClient(self.asyn_conn.accept()) 
        self.asyn_conn.settimeout(None) # connection accepted; timeout removing
        c.start()
        self.threads.append(c)

        #self.send_msg(self.register_opaque_msg())
        #self.send_msg(self.sync_lsdb())
        #self.send_msg(self.register_event_msg())
        self.send_msg(self.sync_lsdb())

        while self.active:
            time.sleep(10)

        # close all threads 
        self.syn_conn.close()
        self.asyn_conn.close() 
        for c in self.threads: 
            c.join() 
            
    def stop(self):
        for thread in self.threads:
            thread.stop()
        self.active = False

    def send_msg(self, msg):
        self.syn_conn.send(msg)
        #print "Syn socket send data:", binascii.hexlify(msg)
        
        msg_parse(msg)

        data = self.syn_conn.recv(1024)
        #print "Syn socket recv data:", binascii.hexlify(data)
        msg_parse(data)

        if not data:
            self.syn_conn.close()     

    def register_opaque_msg(self):
        data = struct.pack('!BBHI', 1, 1,  4, self.seq_num)
        data += struct.pack('!BBH', 0xa, 0xf3, 0)
        self.seq_num += 1
        return data
    
    def register_event_msg(self):
        data = struct.pack('!BBHI', 1, 3,  4, self.seq_num)
        data += struct.pack('!HBB', 0xffff, 2, 0)
        self.seq_num += 1
        return data
    
    def sync_lsdb(self):
        data = struct.pack('!BBHI', 1, 4,  4, self.seq_num)
        data += struct.pack('!HBB', 0xffff, 0x02, 0)
        self.seq_num += 1
        return data    

    def ospfapi_header(self):
        data = struct.pack('!BBHI', 1, 1,  4, 1)
        return data

#----------------------------------------------------------------------------------------------------------------------------------------------------------------------

class AsynClient(threading.Thread): 
    def __init__(self,(client,address)): 
        threading.Thread.__init__(self) 
        self.client = client 
        self.address = address 
        self.size = 40000
        self.active = True

    def run(self): 
        while self.active: 
            data = self.client.recv(self.size) 
            #print "Asyn socket received:", binascii.hexlify(data)
            if data:
                msg_process(data)
            else: 
                self.client.close() 
                break 

    def stop(self):
        self.active = False

########### MAIN #################################
   
def test():
    s = Server(server_address = '150.254.171.116') 
    s.start()
    time.sleep(30)
    s.stop()
    
def normal():
    s = Server(server_address = '150.254.171.107') 
    s.start()
    
if __name__ == "__main__": 
    test()
    #normal()




