/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */




#include <zebra.h>

#include "log.h"
#include "version.h"
#include <getopt.h>
#include "command.h"
#include "thread.h"
#include "xxxd.h"
#include "xxx_zebra.h"
#include "sigevent.h"
#include "privs.h"

/*
 * *  GLOBALS
 */

/* xxxd privileges */
zebra_capabilities_t _caps_p [] = {
	ZCAP_NET_RAW,
	ZCAP_BIND,
	ZCAP_NET_ADMIN,
};

struct zebra_privs_t xxxd_privs = {
#if defined(QUAGGA_USER) && defined(QUAGGA_GROUP)
		.user = QUAGGA_USER,
		.group = QUAGGA_GROUP,
#endif
#if defined(VTY_GROUP)
		.vty_group = VTY_GROUP,
#endif
		.caps_p = _caps_p,
		.cap_num_p = sizeof(_caps_p)/sizeof(_caps_p[0]),
		.cap_num_i = 0
};

char config_default[] = XXXD_DEFAULT_CONFIG;

struct option longopts[] = {
		{ "daemon",      no_argument,       NULL, 'd'},
		{ "config_file", required_argument, NULL, 'f'},
		{ "help",        no_argument,       NULL, 'h'},
		{ "vty_port",    required_argument, NULL, 'P'},
		{ "user",        required_argument, NULL, 'u'},
		{ "group",       required_argument, NULL, 'g'},
		{ "version",     no_argument,       NULL, 'v'},
		{ 0 }
};


char* progname;
/* will contain argv[0] */

/* Process ID saved for use by init system */
const char *pid_file = PATH_XXXD_PID;

struct thread_master *master;
/* needed by the thread implementation */


static void 
sighup (void)
{
	zlog (NULL, LOG_INFO, "SIGHUP received");
}

/* SIGINT / SIGTERM handler. */
static void
sigint (void)
{
	zlog_notice ("Terminating on signal");
	/* xxx_terminate ();*/
	exit (0);
}

/* SIGUSR1 handler. */
static void
sigusr1 (void)
{
	zlog_rotate (NULL);
}

struct quagga_signal_t xxx_signals[] = {
		{
			.signal = SIGHUP,
			.handler = &sighup,
		},
		{
			.signal = SIGUSR1,
			.handler = &sigusr1,
		},  
		{
			.signal = SIGINT,
			.handler = &sigint,
		},
		{
			.signal = SIGTERM,
			.handler = &sigint,
		},
};

/* Help information display. */
static void __attribute__ ((noreturn))
     usage (char *progname, int status)
{
	if (status != 0) 
		fprintf (stderr, "Try `%s --help' for more information.\n", progname);
	else {    
		printf ("Usage : %s [OPTION...]\n\
Daemon which manages XXX.\n\n\
-d, --daemon       Runs in daemon mode\n\
-f, --config_file  Set configuration file name\n\
-P, --vty_port     Set vty's port number\n\
-u, --user         User to run as\n\
-g, --group        Group to run as\n\
-v, --version      Print program version\n\
-h, --help         Display this help and exit\n\
\n\
Report bugs to %s\n", progname, ZEBRA_BUG_ADDRESS);
	         
       }
	exit (status);
}


int main(int argc, char** argv) 
{
	char *p;
	int vty_port = XXXD_VTY_PORT;
	int *vty_addr = NULL;
	int daemon_mode = 0;
	char *config_file = NULL;
	struct thread thread;
          
	umask(0027);
 
	progname =  ((p = strrchr (argv[0], '/')) ? ++p : argv[0]);
 
	zlog_default = openzlog (progname, ZLOG_XXX,
				 LOG_CONS|LOG_NDELAY|LOG_PID, LOG_DAEMON);

	/* initialize the log subsystem, you will have to include
	 * ZLOG_ZAP in the zlog_proto_t enum type definition in 
	 * lib/log.h
	 */

     
	/* this while just reads the options */                       
	while (1) {
		int opt;
            
		opt = getopt_long (argc, argv, "df:h:P:u:g:v", longopts, 0);
                      
		if (opt == EOF)
			break;
                                    
		switch (opt) {
			case 0:
			break;
			case 'd':
				daemon_mode = 1;
				break;
			case 'f':
				config_file = optarg;
				break;
			case 'P':
				vty_port = atoi (optarg);
				break;
			case 'u':
				xxxd_privs.user = optarg;
				break;
			case 'g':
				xxxd_privs.group = optarg;
				break;
			case 'v':
				print_version (progname);
				exit (0);
				break;
			case 'h':
				usage (progname,0);
				break;
			default:
				usage (progname,1);
				break;
	               	}
		}

	/*initialization of the XXX data structure*/ 
	xxx_master_init ();

	/*this the main thread controlling structure*/         
	master = xxx_m->master;
  
	zprivs_init (&xxxd_privs);
	signal_init (master, Q_SIGC(xxx_signals), xxx_signals);

	/* 
	 * * Library inits.
	 */
	cmd_init (1);
	/* initializes the command sub-system, if arg, add some commands
	 * which are mostly only useful for humans on the vty
	 */

	/* these are all from libzebra */
	vty_init (master);
	memory_init ();
	access_list_init ();
	prefix_list_init ();
             
	/*
	 * XXX inits
	 */
	xxx_init();
	/* in this function   you could start some threads (the thread subsystem
	 * is not running yet), register some commands, ...
	 */

	xxx_if_init ();
	xxx_zebra_init ();

	/* XXX vty inits. */
	xxx_vty_init ();

	sort_node();
	/* This is needed by the command subsystem to finish initialization.
	 */

	/* Get configuration file. */
	vty_read_config (config_file, config_default);
	/* read the config file, your commands should be defined before this
	 */

	/* Change to the daemon program. */
	if (daemon_mode)
		daemon (0, 0);

	/* Process id file create. */
	pid_output (pid_file);

	/* Create VTY socket */
	vty_serv_sock (vty_addr,vty_port,XXX_VTYSH_PATH);
	/* start the TCP and unix socket listeners
	 */

	/* Print banner. */
	zlog_notice ("XXXd %s starting: vty@%d", QUAGGA_VERSION, vty_port);
    
	/* Fetch next active thread. */
	while (thread_fetch (master, &thread))
		thread_call (&thread);
	/* this is the main event loop */

	/* never reached */
	return 0;
}


/* xxxd_main.c ends */
