/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#include <zebra.h>

#include "thread.h"
#include "vty.h"
#include "command.h"
#include "linklist.h"
#include "prefix.h"
#include "table.h"
#include "if.h"
#include "memory.h"
#include "stream.h"
#include "log.h"
#include "sockunion.h"
#include "zclient.h"
#include "xxxd.h"
#include "xxx_zebra.h"

/* XXX process wide configuration. */
static struct xxx_master xxx_master;

/* XXX process wide configuration pointer to export. */
struct xxx_master *xxx_m;

/* Structure that contains interfaces of XXX */
struct zlist *xxxiflist;

/* Output prefix string to vty. */
static int
prefix_vty_out (struct vty *vty, struct prefix *p)
{
	char str[INET6_ADDRSTRLEN];

	inet_ntop (p->family, &p->u.prefix, str, sizeof (str));
	vty_out (vty, "%s", str);
	return strlen (str);
}

/* Dump if address information to vty. */
static void
connected_dump_vty (struct vty *vty, struct connected *connected)
{
	struct prefix *p;

	/* Print interface address. */
	p = connected->address;
	vty_out (vty, "  %s ", prefix_family_str (p));
	prefix_vty_out (vty, p);
	vty_out (vty, "/%d", p->prefixlen);

	/* If there is destination address, print it. */
	if (connected->destination)
		{
			vty_out (vty, (CONNECTED_PEER(connected) ? " peer " : " broadcast "));
			prefix_vty_out (vty, connected->destination);
		}

	if (CHECK_FLAG (connected->flags, ZEBRA_IFA_SECONDARY))
		vty_out (vty, " secondary");

	if (connected->label)
		vty_out (vty, " %s", connected->label);

	vty_out (vty, "%s", VTY_NEWLINE);
}

/* Interface's information print out to vty interface. */
static void
if_dump_vty (struct vty *vty, struct interface *ifp)
{
#ifdef HAVE_SOCKADDR_DL
	struct sockaddr_dl *sdl;
#endif /* HAVE_SOCKADDR_DL */
	struct connected *connected;
	struct zlistnode *node;
	struct route_node *rn;
	struct zebra_if *zebra_if;

	zebra_if = ifp->info;

	vty_out (vty, "Interface %s is ", ifp->name);
	if (if_is_up(ifp)) {
		vty_out (vty, "up, line protocol ");
    
		if (CHECK_FLAG(ifp->status, ZEBRA_INTERFACE_LINKDETECTION)) {
			if (if_is_running(ifp))
				vty_out (vty, "is up%s", VTY_NEWLINE);
			else
				vty_out (vty, "is down%s", VTY_NEWLINE);
		} else {
			vty_out (vty, "detection is disabled%s", VTY_NEWLINE);
		}
	} else {
		vty_out (vty, "down%s", VTY_NEWLINE);
	}

	if (ifp->desc)
		vty_out (vty, "  Description: %s%s", ifp->desc,
			 VTY_NEWLINE);
	if (ifp->ifindex == IFINDEX_INTERNAL)
		{
			vty_out(vty, "  pseudo interface%s", VTY_NEWLINE);
			return;
		}
	else if (! CHECK_FLAG (ifp->status, ZEBRA_INTERFACE_ACTIVE))
		{
			vty_out(vty, "  index %d inactive interface%s", 
				ifp->ifindex, 
				VTY_NEWLINE);
			return;
		}

	vty_out (vty, "  index %d metric %d mtu %d ",
		 ifp->ifindex, ifp->metric, ifp->mtu);
#ifdef HAVE_IPV6
	if (ifp->mtu6 != ifp->mtu)
		vty_out (vty, "mtu6 %d ", ifp->mtu6);
#endif 
	vty_out (vty, "%s  flags: %s%s", VTY_NEWLINE,
		 if_flag_dump (ifp->flags), VTY_NEWLINE);
  
	/* Hardware address. */
#ifdef HAVE_SOCKADDR_DL
	sdl = &ifp->sdl;
	if (sdl != NULL && sdl->sdl_alen != 0)
		{
			int i;
			u_char *ptr;

			vty_out (vty, "  HWaddr: ");
			for (i = 0, ptr = (u_char *)LLADDR (sdl); i < sdl->sdl_alen; i++, ptr++)
				vty_out (vty, "%s%02x", i == 0 ? "" : ":", *ptr);
			vty_out (vty, "%s", VTY_NEWLINE);
		}
#else
	if (ifp->hw_addr_len != 0)
		{
			int i;

			vty_out (vty, "  HWaddr: ");
			for (i = 0; i < ifp->hw_addr_len; i++)
				vty_out (vty, "%s%02x", i == 0 ? "" : ":", ifp->hw_addr[i]);
			vty_out (vty, "%s", VTY_NEWLINE);
		}
#endif
	  
}


DEFUN (show_xxx_interface, show_xxx_interface_cmd, "show xxx interface", "Show information on interfaces")
{
	struct interface *ifp;
	struct zlistnode *node;

	/* All interface print. */
	for (ALL_LIST_ELEMENTS_RO (xxxiflist, node, ifp))
		if_dump_vty (vty, ifp);

	return CMD_SUCCESS;
}


DEFUN (tel_xxx, tel_xxx_cmd, "tel", "Set TE-Link")
{
	vty_out (vty, "Testing Set TE-Link command %s", VTY_NEWLINE);

	return CMD_SUCCESS;
}

struct cmd_node xxx_node = {
	XXX_NODE,
	"%s(config-xxx)# ",
	1
};

/* XXX configuration write function. */
static int
xxx_config_write (struct vty *vty)
{
	vty_out (vty, "Testing Set TE-Link command %s", VTY_NEWLINE);
  
	return 1;
}


void
xxx_vty_init (void)
{
	/* Install xxx  top node. */
	install_node (&xxx_node, xxx_config_write);

	/* "tel" command . */
	install_element (CONFIG_NODE, &tel_xxx_cmd);

	/* "show ip ospf interface" commands. */
	install_element (VIEW_NODE, &show_xxx_interface_cmd);
	install_element (ENABLE_NODE, &show_xxx_interface_cmd);
}


int
xxx_sleep_thread (int dummy)
{
	sleep (0.1);
	thread_add_read (master, &xxx_sleep_thread, 0, 0);
	
	return 1;
}

void
xxx_init (void)
{
	thread_add_read (master, &xxx_sleep_thread, 0, 0);
}


void
xxx_if_init ()
{
	/* Initialize XXX interface data structure.(similar to Zebra interface structure) */
	xxxiflist = list_new ();
	if_init();

	if (xxxiflist) {
	        xxxiflist->cmp = (int (*)(void *, void *))if_cmp_func;
		return;
	}
	xxx_m->iflist = iflist;
}

void
xxx_master_init ()
{
	memset (&xxx_master, 0, sizeof (struct xxx_master));

	xxx_m = &xxx_master;
	xxx_m->master = thread_master_create ();
	xxx_m->start_time = quagga_time (NULL);
}
