/*
 *  This file is part of phosphorus-g2mpls.
 *
 *  Copyright (C) 2006, 2007, 2008, 2009 Nextworks s.r.l.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Authors:
 *
 *  Giacomo Bernini       (Nextworks s.r.l.) <g.bernini_at_nextworks.it>
 *  Gino Carrozzo         (Nextworks s.r.l.) <g.carrozzo_at_nextworks.it>
 *  Nicola Ciulli         (Nextworks s.r.l.) <n.ciulli_at_nextworks.it>
 *  Giodi Giorgi          (Nextworks s.r.l.) <g.giorgi_at_nextworks.it>
 *  Francesco Salvestrini (Nextworks s.r.l.) <f.salvestrini_at_nextworks.it>
 */



#ifndef G2RSVPTE_MESSAGE_OBJECTS_H
#define G2RSVPTE_MESSAGE_OBJECTS_H

#include <list>

#include <netinet/in.h>
#include <stdint.h>

#include "g2rsvpte_config.h"
#include "g2rsvpte_utils.h"
#include "g2rsvpte_message_constants.h"
#include "g2mpls_addr.h"
#include "g2rsvpte_trace.h"
#include "g2rsvpte_buffer.h"
#include "g2rsvpte_message_objects_utils.h"

#include "g2mpls_types.h"

namespace RSVP {

	typedef uint8_t  bit1_t;
	typedef uint8_t  bit2_t;
	typedef uint8_t  bit3_t;
	typedef uint8_t  bit4_t;
	typedef uint8_t  bit6_t;
	typedef uint8_t	 bit7_t;
	typedef uint8_t  bit8_t;
	typedef uint8_t	 bit10_t;
	typedef uint16_t bit12_t;
	typedef uint16_t bit14_t;
	typedef uint16_t bit16_t;
	typedef uint32_t bit23_t;
	typedef uint32_t bit24_t;
	typedef uint32_t bit25_t;
	typedef uint32_t bit28_t;
	typedef uint32_t bit29_t;
	typedef uint32_t bit30_t;
	typedef uint32_t bit32_t;
	typedef uint64_t bit48_t;
	typedef uint64_t bit64_t;

	typedef struct {
		ipv4_t ipv4_addr;
	} ipv4_addr_t;

	typedef struct {
		ipv6_t ipv6_addr;
	} ipv6_addr_t;

// Object Formats (RFC 2205)
//
//           0             1              2             3
//    +-------------+-------------+-------------+-------------+
//    |       Length (bytes)      |  Class-Num  |   C-Type    |
//    +-------------+-------------+-------------+-------------+
//    |                                                       |
//    //                  (Object contents)                   //
//    |                                                       |
//    +-------------+-------------+-------------+-------------+
//

	class objHeader {
	public:
		objHeader(uint8_t cNum) :
			cNum_(cNum) { }
		virtual ~objHeader(void) { }

		uint8_t cNum(void) const { return cNum_; }

		virtual uint8_t  cType(void) const = 0;
		virtual uint16_t length(void) const = 0;

	protected:
		uint8_t cNum_;

	public:
		static const uint16_t objHdrLength = 4;
	};

#define DEFINE_CLASS_BEGIN(NAME,CNUM)					      \
	class NAME : public objHeader {					      \
		friend inline std::ostream & operator << (std::ostream & os,  \
							  const NAME &   cl); \
									      \
		friend inline MemoryCursor & operator >> (MemoryCursor & mc,  \
							  NAME * &       cl); \
									      \
		friend inline MemoryCursor & operator << (MemoryCursor & mc,  \
							  const NAME &   cl); \
									      \
	public:								      \
		NAME(void) : objHeader(CNUM) { }			      \
		virtual ~NAME(void) { }					      \
									      \
		virtual uint8_t cType(void) const = 0;			      \
		virtual uint16_t length(void) const = 0;		      \
		virtual NAME * clone(void) const = 0;			      \
		virtual bool isConsistent(void) const = 0;		      \
									      \
	private:							      \
									      \
	};

#define DEFINE_CLASS_END(NAME)						     \
	inline std::ostream & operator << (std::ostream & os,		     \
					   const NAME &   cl)		     \
	{								     \
		os << "Object " # NAME            << std::endl		     \
		   << " cNum     " << cl.cNum()   << std::endl		     \
		   << " cType    " << cl.cType()  << std::endl		     \
		   << " length   " << cl.length() << std::endl;		     \
		NAME##_OBJ2STREAM(os, cl);				     \
		return os;						     \
	}								     \
									     \
									     \
	inline MemoryCursor & operator >> (MemoryCursor & mc,		     \
					   NAME * &       cl)		     \
	{								     \
		/* Check Object Header */				     \
		MemoryCursor mcObjHdr(mc);				     \
		mcObjHdr.resize(mc.start(),				     \
				mc.start() + (objHeader::objHdrLength - 1)); \
		assert(mcObjHdr.remainingSize() >= objHeader::objHdrLength); \
									     \
		uint8_t  cNum;						     \
		uint8_t  cType;						     \
		uint16_t len;						     \
									     \
		mcObjHdr >> len >> cNum >> cType;			     \
		assert(mcObjHdr.remainingSize() == 0);			     \
									     \
		int length;						     \
		length = len - objHeader::objHdrLength;			     \
									     \
		if (length <= 0) {					     \
			throw(std::string("Impossible to recv an object ("   \
					  #NAME ") with no data"));	     \
		}							     \
									     \
		MemoryCursor mcObj = mcObjHdr.neighbor(length);		     \
									     \
		NAME##_STREAM2OBJ(mcObj, cl, cType);			     \
		return mc;						     \
	}								     \
									     \
	inline MemoryCursor & operator << (MemoryCursor & mc,		     \
					   const NAME & cl)		     \
	{								     \
		/* Writes Object Header */				     \
									     \
		MemoryCursor mcObjHdr(mc);				     \
		mcObjHdr.resize(mc.start(),				     \
				mc.start() + (cl.objHdrLength - 1));	     \
		mcObjHdr << cl.length() << cl.cNum() << cl.cType();	     \
		assert(mcObjHdr.remainingSize() == 0);			     \
									     \
		int len;						     \
		len = mc.stop() - mc.start() + 1 - cl.objHdrLength;	     \
									     \
		if (len <= 0) {						     \
			throw(std::string("Impossible to send an object ("   \
					  #NAME ") with no data"));	     \
		}							     \
									     \
		MemoryCursor mcObj = mcObjHdr.neighbor(len);		     \
									     \
		/* Writes Object */					     \
		NAME##_OBJ2STREAM(mcObj, cl);				     \
		return mc;						     \
	}


	template <class FATHER, typename TYPE, uint8_t CTYPE>
	class Object : public FATHER {
	public:
		Object(void) : cType_(CTYPE) { }
		virtual ~Object(void) { }

		TYPE set(const TYPE & value) {
			TYPE t;

			t      = value_;
			value_ = value;

			return t;
		}

		TYPE & get(void) { return value_; }
		uint8_t cType(void) const { return cType_; }

		virtual bool isConsistent(void) const = 0;
		virtual uint16_t length(void) const = 0;
		virtual FATHER * clone(void) const = 0;

	protected:
		TYPE    value_;

	private:
		uint8_t cType_;
	};

#define _DEFINE_BASE_OBJECT(NAME,FATHERNAME,TYPE,CTYPE)	\
	typedef Object<FATHERNAME,TYPE,CTYPE> NAME

#define _DEFINE_DERIVED_OBJECT(NAME,FATHERNAME)		\
	class NAME : public FATHERNAME

#define DEFINE_OBJECT(NAME, CTYPE, TYPE, FATHER)		\
	_DEFINE_BASE_OBJECT(NAME##_##Base,FATHER,TYPE,CTYPE);	\
	_DEFINE_DERIVED_OBJECT(NAME,NAME##_##Base)

// Utils
#define OBJ_OPERATOR(NAME)						  \
	friend std::ostream & operator << (std::ostream &, const NAME &); \
	friend MemoryCursor & operator << (MemoryCursor &, const NAME &); \
	friend MemoryCursor & operator >> (MemoryCursor &, NAME &);

//////////////////////////////////////////////////////////////

// TLV Format (RFC 3471)
//
//           0             1              2             3
//    +-------------+-------------+-------------+-------------+
//    |           Type            |           Length          |
//    +-------------+-------------+-------------+-------------+
//    |                                                       |
//    ~                         Value                         ~
//    |                                                       |
//    +-------------+-------------+-------------+-------------+

	class TLV {
		friend MemoryCursor & operator >> (MemoryCursor & mc,
						   TLV * &        tlv);

		friend MemoryCursor & operator << (MemoryCursor & mc,
						   const TLV &    tlv);

		friend std::ostream & operator << (std::ostream & os,
						   const TLV &    tlv);
	public:
		TLV(uint16_t type) : type_(type){ }
		virtual ~TLV(void) { }

		virtual uint16_t length(void) const = 0;
		virtual TLV * clone(void) const = 0;
		virtual bool isConsistent(void) const = 0;

		uint16_t type(void) const { return type_; }
	private:
		uint16_t type_;
	public:
		static const uint16_t tlvHdrLength = 4; // Type & Length
	};

	//equivalent to	class Object' but type is uint16_t NOT uint8_t !!!
	template <typename TYPE, uint16_t TLV_TYPE>
	class TLV_template : public TLV {
	public:
		TLV_template(void) : TLV(TLV_TYPE) { }
		virtual ~TLV_template(void) { }

		TYPE set(const TYPE & value) {
			TYPE t;

			t      = value_;
			value_ = value;

			return t;
		}

		TYPE & get(void) { return value_; }

		virtual uint16_t length(void) const = 0;
		virtual bool isConsistent(void) const = 0;
		virtual TLV * clone(void) const = 0;

	protected:
		TYPE     value_;
	};


#define _DEFINE_BASE_TLV(NAME,DATA,TYPE)	\
	typedef TLV_template<DATA, TYPE> NAME

#define DEFINE_TLV(NAME, DATA, TYPE)			\
	_DEFINE_BASE_TLV(NAME##_##Base,DATA,TYPE);	\
	_DEFINE_DERIVED_OBJECT(NAME,NAME##_##Base)

/////////////////////////////////////////////////////////////////

// ERO SubObject Format (RFC 3209)
//
//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-------------//----------------+
//	   |L|    Type     |     Length    | (Subobject contents)          |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-------------//----------------+
//
//	      L
//
//	        The L bit is an attribute of the subobject.  The L bit is set
//	        if the subobject represents a loose hop in the explicit route.
//	        If the bit is not set, the subobject represents a strict hop in
//	        the explicit route.
//
//	      Type
//
//	        The Type indicates the type of contents of the subobject.
//	        Currently defined values are:
//
//	                  1   IPv4 prefix
//	                  2   IPv6 prefix
//	                 32   Autonomous system number
//
//	      Length
//
//	        The Length contains the total length of the subobject in bytes,
//	        including the L, Type and Length fields.  The Length MUST be at
//	        least 4, and MUST be a multiple of 4.

	class EroSubObject {
		friend MemoryCursor & operator >> (MemoryCursor &   mc,
						   EroSubObject * & obj);

		friend MemoryCursor & operator << (MemoryCursor &       mc,
						   const EroSubObject & obj);

		friend std::ostream & operator << (std::ostream &       os,
						   const EroSubObject & obj);
	public:
		EroSubObject(uint8_t type, bool loose = false) :
			type_(type), loose_flag_(loose) { }
		virtual ~EroSubObject(void) { }

		uint8_t type(void) const        { return type_;       }
		bool looseFlag(void) const      { return loose_flag_; }
		void looseFlag(const bool flag) { loose_flag_ = flag; }

		virtual uint8_t length(void) const = 0;
		virtual EroSubObject * clone(void) const = 0;
		virtual bool isConsistent(void) const = 0;

	private:
		uint8_t type_;
		bool    loose_flag_;
	public:
		static const uint16_t eroSubObjHdrLength = 2; // L/Type/Length
	};

	template <typename TYPE, uint8_t EROOBJ_TYPE>
	class EroSubObject_template : public EroSubObject {
	public:
		EroSubObject_template(void) : EroSubObject(EROOBJ_TYPE) { }
		virtual ~EroSubObject_template(void) { }

		TYPE set(const TYPE & content) {
			TYPE t;

			t = content_;
			content_ = content;

			return t;
		}

		TYPE & get(void) { return content_; }

		virtual bool isConsistent(void) const = 0;
		virtual uint8_t length(void) const = 0;
		virtual EroSubObject * clone(void) const = 0;

	protected:
		TYPE    content_;
	};


#define _DEFINE_BASE_EROSUBOBJ(NAME,DATA,TYPE)		\
	typedef EroSubObject_template<DATA, TYPE> NAME

#define DEFINE_EROSUBOBJ(NAME, DATA, TYPE)			\
	_DEFINE_BASE_EROSUBOBJ(NAME##_##Base,DATA,TYPE);	\
	_DEFINE_DERIVED_OBJECT(NAME,NAME##_##Base)

/////////////////////////////////////////////////////////////////

// RRO SubObject Format (RFC 3209)
//
//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-------------//----------------+
//	   |      Type     |     Length    | (Subobject contents)          |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-------------//----------------+

	class RroSubObject {
		friend MemoryCursor & operator >> (MemoryCursor &   mc,
						   RroSubObject * & obj);

		friend MemoryCursor & operator << (MemoryCursor &       mc,
						   const RroSubObject & obj);

		friend std::ostream & operator << (std::ostream &       os,
						   const RroSubObject & obj);
	public:
		RroSubObject(uint8_t type) : type_(type) { }
		virtual ~RroSubObject(void) { }

		uint8_t type(void) const { return type_; }

		virtual uint8_t length(void) const = 0;
		virtual RroSubObject * clone(void) const = 0;
		virtual bool isConsistent(void) const = 0;

	private:
		uint8_t type_;
	public:
		static const uint16_t rroSubObjHdrLength = 2; // L/Type/Length
	};

	template <typename TYPE, uint8_t RROOBJ_TYPE>
	class RroSubObject_template : public RroSubObject {
	public:
		RroSubObject_template(void): RroSubObject(RROOBJ_TYPE) { }
		virtual ~RroSubObject_template(void) { }

		TYPE set(const TYPE & content) {
			TYPE t;

			t = content_;
			content_ = content;

			return t;
		}

		TYPE & get(void) { return content_; }

		virtual bool isConsistent(void) const = 0;
		virtual uint8_t length(void) const = 0;
		virtual RroSubObject * clone(void) const = 0;

	protected:
		TYPE    content_;
	};


#define _DEFINE_BASE_RROSUBOBJ(NAME,DATA,TYPE)		\
	typedef RroSubObject_template<DATA, TYPE> NAME

#define DEFINE_RROSUBOBJ(NAME, DATA, TYPE)			\
	_DEFINE_BASE_RROSUBOBJ(NAME##_##Base,DATA,TYPE);	\
	_DEFINE_DERIVED_OBJECT(NAME,NAME##_##Base)

/////////////////////////////////////////////////////////////////

// SubObject Format
//
//       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |       Length                  |    Type       |  Sub-Type     |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      //                             Value                           //
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	class SubObject {
	public:
		SubObject(uint8_t type, uint8_t subtype) :
			type_(type),
			subType_(subtype) { }
		virtual ~SubObject(void) { }

		uint8_t type(void) const    { return type_;    }
		uint8_t subType(void) const { return subType_; }

	protected:
		uint8_t type_;
		uint8_t subType_;
	public:
		static const uint16_t subObjHdrLength = 4; // len/type/subType
	};

	class GuniSubObj : public SubObject {
		friend MemoryCursor & operator >> (MemoryCursor & mc,
						   GuniSubObj * & obj);

		friend MemoryCursor & operator << (MemoryCursor &     mc,
						   const GuniSubObj & obj);

		friend std::ostream & operator << (std::ostream &     os,
						   const GuniSubObj & obj);
	public:
		GuniSubObj(uint8_t type, uint8_t subtype) :
			SubObject(type, subtype) { }
		virtual ~GuniSubObj(void) { }

		virtual uint16_t length(void) const = 0;
		virtual GuniSubObj * clone(void) const = 0;
		virtual bool isConsistent(void) const = 0;

	protected:

	public:

	};

	class SpcLabel_SubObj;

	GuniSubObj * createGuniSrcSubObj(const g2mpls_addr_t & add);
	GuniSubObj * createGuniDstSubObj(const g2mpls_addr_t & add);
	GuniSubObj * createGuniSpcLblSubObj(const g2mpls_addr_t & data_link,
					    const label_id_t &    label);
	GuniSubObj * createGuniEgrLblSubObj(const g2mpls_addr_t & data_link,
					    const label_id_t &    label);
	EroSubObject * createEroSubObj(const lsp_ero_sobj_t & sObj);
	bool createSrcG2mplsAddr(GuniSubObj & guni, g2mpls_addr_t & addr);
	bool createDestG2mplsAddr(GuniSubObj & guni, g2mpls_addr_t & addr);
	bool createDlLabel(SpcLabel_SubObj * spcLbl,
			   g2mpls_addr_t &   dl,
			   label_id_t &      lbl);

	class GnsSubObj : public SubObject {
		friend MemoryCursor & operator >> (MemoryCursor & mc,
						   GnsSubObj * &  obj);

		friend MemoryCursor & operator << (MemoryCursor &    mc,
						   const GnsSubObj & obj);

		friend std::ostream & operator << (std::ostream &    os,
						   const GnsSubObj & obj);
	public:
		GnsSubObj(uint8_t type, uint8_t subtype) :
			SubObject(type, subtype) { }
		virtual ~GnsSubObj(void) { }

		virtual uint16_t length(void) const = 0;
		virtual GnsSubObj * clone(void) const = 0;
		virtual bool isConsistent(void) const = 0;

	protected:

	public:

	};

	template <typename TYPE, class FATHER,
		uint8_t SOBJ_TYPE, uint8_t SOBJ_SUBTYPE>
	class SubObject_template : public FATHER {
	public:
		SubObject_template(void) :
			FATHER(SOBJ_TYPE, SOBJ_SUBTYPE) { }
		virtual ~SubObject_template(void) { }

		TYPE set(const TYPE & value) {
			TYPE t;

			t = value_;
			value_ = value;

			return t;
		}

		TYPE & get(void) { return value_; }

		virtual bool isConsistent(void) const = 0;
		virtual uint16_t length(void) const = 0;
		virtual FATHER * clone(void) const = 0;

	protected:
		TYPE value_;
	};


#define _DEFINE_BASE_SUBOBJ(NAME,DATA,TYPE,SUBTYPE,FATHER)		\
	typedef SubObject_template<DATA, FATHER, TYPE, SUBTYPE> NAME

#define DEFINE_GUNI_SUBOBJ(NAME, TYPE, SUBTYPE, DATA)			 \
	_DEFINE_BASE_SUBOBJ(NAME##_##Base,DATA,TYPE,SUBTYPE,GuniSubObj); \
	_DEFINE_DERIVED_OBJECT(NAME,NAME##_##Base)

#define DEFINE_GNS_SUBOBJ(NAME, TYPE, SUBTYPE, DATA)			\
	_DEFINE_BASE_SUBOBJ(NAME##_##Base,DATA,TYPE,SUBTYPE,GnsSubObj);	\
	_DEFINE_DERIVED_OBJECT(NAME,NAME##_##Base)

/////////////////////////////////////////////////////////////////

//CLASS SESSION (RFC 2205)

	DEFINE_CLASS_BEGIN(Session_Class, G2RSVP_CLASS_SESSION);

//           +-------------+-------------+-------------+-------------+
//           |             IPv4 DestAddress (4 bytes)                |
//           +-------------+-------------+-------------+-------------+
//           | Protocol Id |    Flags    |          DstPort          |
//           +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv4_t  ipv4_addr;
		bit8_t	protoId;
		bit8_t	flags;
		bit16_t	port;
	} ipv4_address_t;
	DEFINE_OBJECT(IPv4Session_Object,
		      CTYPE_SESSION_IPV4,
		      ipv4_address_t,
		      Session_Class) {
		OBJ_OPERATOR(IPv4Session_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv4_address_t);
		}

		Session_Class * clone(void) const {
			IPv4Session_Object * obj = new IPv4Session_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//           +-------------+-------------+-------------+-------------+
//           |                                                       |
//           +                                                       +
//           |                                                       |
//           +               IPv6 DestAddress (16 bytes)             +
//           |                                                       |
//           +                                                       +
//           |                                                       |
//           +-------------+-------------+-------------+-------------+
//           | Protocol Id |     Flags   |          DstPort          |
//           +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv6_t	ipv6_addr;
		bit8_t	protoId;
		bit8_t	flags;
		bit16_t	port;
	} ipv6_address_t;
	DEFINE_OBJECT(IPv6Session_Object,
		      CTYPE_SESSION_IPV6,
		      ipv6_address_t,
		      Session_Class) {
		OBJ_OPERATOR(IPv6Session_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv6_address_t);
		}

		Session_Class * clone(void) const {
			IPv6Session_Object * obj = new IPv6Session_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};


//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                   IPv4 tunnel end point address               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |  MUST be zero                 |      Tunnel ID                |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                       Extended Tunnel ID                      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv4_t	ipv4_addr;   // ipv4 tunnel end point
		bit16_t zero_field;
		bit16_t	tunnelID;
		bit32_t	extendedTunnelID;
	} session_lsp_tunnel_ipv4_t;
	DEFINE_OBJECT(LSPTunnelIPv4Session_Object,
		      CTYPE_SESSION_LSP_TUNNEL_IPV4,
		      session_lsp_tunnel_ipv4_t,
		      Session_Class) {
		OBJ_OPERATOR(LSPTunnelIPv4Session_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength +
				sizeof(session_lsp_tunnel_ipv4_t);
		}

		Session_Class * clone(void) const {
			LSPTunnelIPv4Session_Object * obj;
			obj = new LSPTunnelIPv4Session_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};


//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   +                                                               +
//	   |                   IPv6 tunnel end point address               |
//	   +                                                               +
//	   |                            (16 bytes)                         |
//	   +                                                               +
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |  MUST be zero                 |      Tunnel ID                |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   +                                                               +
//	   |                       Extended Tunnel ID                      |
//	   +                                                               +
//	   |                            (16 bytes)                         |
//	   +                                                               +
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv6_t	ipv6_addr;
		bit16_t zero_field;
		bit16_t	tunnelID;
		bit32_t	extendedTunnelID[4];
	} session_lsp_tunnel_ipv6_t;
	DEFINE_OBJECT(LSPTunnelIPv6Session_Object,
		      CTYPE_SESSION_LSP_TUNNEL_IPV6,
		      session_lsp_tunnel_ipv6_t,
		      Session_Class) {
		OBJ_OPERATOR(LSPTunnelIPv6Session_Object);
	public:
		bool isConsistent(void) const {
			return false;
		}

		uint16_t length(void) const {
			return objHdrLength +
				sizeof(session_lsp_tunnel_ipv6_t);
		}

		Session_Class * clone(void) const {
			LSPTunnelIPv6Session_Object * obj;
			obj = new LSPTunnelIPv6Session_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                          IPv4 address                         |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |  MUST be zero                 |      Tunnel ID                |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                     Extended IPv4 address                     |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv4_t	ipv4_addr;
		bit16_t zero_field;
		bit16_t	tunnel_id;
		bit32_t	ext_IPv4_addr;
	} session_uni_ipv4_t;
	DEFINE_OBJECT(UNIIPv4Session_Object,
		      CTYPE_SESSION_UNI_IPV4,
		      session_uni_ipv4_t,
		      Session_Class) {
		OBJ_OPERATOR(UNIIPv4Session_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength +
				sizeof(session_uni_ipv4_t);
		}

		Session_Class * clone(void) const {
			UNIIPv4Session_Object * obj;
			obj = new UNIIPv4Session_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Session_Class);

//CLASS LABEL
	DEFINE_CLASS_BEGIN(Label_Class, G2RSVP_CLASS_LABEL);

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                             Label                             |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

#define LABEL_OBJ_BODY(OBJ)					\
	{							\
		OBJ_OPERATOR(OBJ);				\
	public:							\
		bool isConsistent(void) const {			\
			return true;				\
		}						\
								\
		uint16_t length(void) const {			\
			return objHdrLength + sizeof(label_t);	\
		}						\
								\
		Label_Class * clone(void) const {		\
			OBJ * obj;				\
			obj = new OBJ();			\
			if (obj) {				\
				obj->value_ = value_;		\
			}					\
			return obj;				\
		}						\
	};

	typedef struct {
		uint32_t label;
	} label_t;

	DEFINE_OBJECT(LabelMpls_Object,
		      CTYPE_LABEL,
		      label_t,
		      Label_Class)
	LABEL_OBJ_BODY(LabelMpls_Object);


//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                             Label                             |
//	   |                              ...                              |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

#define GLABEL_OBJ_BODY(OBJ,CLASS_OBJ)					      \
	{								      \
		OBJ_OPERATOR(OBJ);					      \
	public:								      \
		OBJ(void) {						      \
			value_.size  = 0;				      \
			value_.label = 0;				      \
		}							      \
									      \
		OBJ(uint32_t * label, uint32_t bytes_size) {		      \
			value_.size  = bytes_size;			      \
			value_.label = new uint32_t[bytes_size /	      \
						    sizeof(uint32_t)];	      \
			assert(label);					      \
			assert(value_.label);				      \
			memcpy(value_.label, label, bytes_size);	      \
		}							      \
									      \
		~OBJ(void) {						      \
			DELETE_ARR_OBJ(value_.label, value_.size / 4);	      \
		}							      \
									      \
		bool isConsistent(void) const {				      \
			return true;					      \
		}							      \
									      \
		uint16_t length(void) const {				      \
			return objHdrLength +				      \
				value_.size;				      \
		}							      \
									      \
		CLASS_OBJ * clone(void) const {				      \
			OBJ * obj;					      \
			obj = new OBJ();				      \
			if (obj) {					      \
				if (value_.size / sizeof(uint32_t) > 0 &&     \
				    value_.label) {			      \
					obj->value_.label = new		      \
						uint32_t[value_.size /	      \
							 sizeof(uint32_t)];   \
					assert(obj->value_.label);	      \
					memcpy(obj->value_.label,	      \
					       value_.label,		      \
					       value_.size);		      \
				} else {				      \
					obj->value_.label = 0;		      \
				}					      \
				obj->value_.size = value_.size;		      \
									      \
			}						      \
			return obj;					      \
		}							      \
	};

	typedef struct {
		uint32_t * label;
		uint32_t   size;  // label length in bytes
	} gen_label_t;

	DEFINE_OBJECT(GLabel_Object,
		      CTYPE_GENERALISED_LABEL,
		      gen_label_t,
		      Label_Class)
	GLABEL_OBJ_BODY(GLabel_Object, Label_Class);

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                          Waveband Id                          |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                          Start Label                          |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                           End Label                           |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

#define WBSLABEL_OBJ_BODY(OBJ)						 \
	{								 \
		OBJ_OPERATOR(OBJ);					 \
	public:								 \
		bool isConsistent(void) const {				 \
			return true;					 \
		}							 \
									 \
		uint16_t length(void) const {				 \
			return objHdrLength + sizeof(wb_switch_label_t); \
		}							 \
									 \
		Label_Class * clone(void) const {			 \
			OBJ * obj;					 \
			obj = new OBJ();				 \
			if (obj) {					 \
				obj->value_ = value_;			 \
			}						 \
			return obj;					 \
		}							 \
	};

	typedef struct {
		uint32_t waveband_id;
		uint32_t start_label;
		uint32_t end_label;
	} wb_switch_label_t;

	DEFINE_OBJECT(WBSLabel_Object,
		      CTYPE_WBS_LABEL,
		      wb_switch_label_t,
		      Label_Class)
	WBSLABEL_OBJ_BODY(WBSLabel_Object);

	DEFINE_CLASS_END(Label_Class);

#ifdef CRANKBACK_ON

//CLASS LSP ATTR
	DEFINE_CLASS_BEGIN(LSPAttr_Class, G2RSVP_CLASS_LSP_ATTRIBUTES);

//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                                                               |
//   |                                                               |
//   ~                             TLVs                              ~
//   |                                                               |
//   |                                                               |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		std::list<TLV *> tlvs;
	} lsp_attr_t;

	DEFINE_OBJECT(LSPAttr_Object,
		      CTYPE_LSP_ATTRIBUTES,
		      lsp_attr_t,
		      LSPAttr_Class) {
		OBJ_OPERATOR(LSPAttr_Object);
        public:
		~LSPAttr_Object(void) {
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.tlvs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength + len;
		}

		LSPAttr_Class * clone(void) const {
			LSPAttr_Object * obj;
			obj = new LSPAttr_Object();

			if (obj) {
				std::list<TLV *>::const_iterator it;
				for (it = value_.tlvs.begin();
				     it != value_.tlvs.end();
				     it++) {
					if (*it) {
						obj->value_.tlvs.push_back
							((*it)->clone());
					}
				}
			}

			return obj;
		}
	};

	// LSPAttr TLV

//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |E|B|S|                     Flags                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//
//      E: End-to-end re-routing desired.
//
//      B: Boundary re-routing desired.
//
//      S: Segment-based re-routing desired.

	typedef struct {
		bit1_t  e2e_cb;
		bit1_t  boundary_cb;
		bit1_t  segment_cb;
		bit29_t flags;
	} tlv_lsp_attr_t;
	DEFINE_TLV(LSP_ATTR_TLV,
		   tlv_lsp_attr_t,
		   TLV_LSP_ATTR) {
		OBJ_OPERATOR(LSP_ATTR_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return TLV_LSP_ATTR_LEN;
		}

		TLV * clone(void) const {
			LSP_ATTR_TLV * obj;
			obj = new LSP_ATTR_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(LSPAttr_Class);

#endif // CRANKBACK_ON

//CLASS SESSION ATTR
	DEFINE_CLASS_BEGIN(Session_Attr_Class, G2RSVP_CLASS_SESSION_ATTRIBUTE);

//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |   Setup Prio  | Holding Prio  |     Flags     |  Name Length  |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                                                               |
//   //          Session Name      (NULL padded display string)      //
//   |                                                               |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit8_t setup_prio;
		bit8_t hold_prio;
		bit8_t flags;
		bit8_t name_len;
		char * name;          // uint8_t * ???
	} session_attr_lsp_tunnel_t;

	DEFINE_OBJECT(LSPTunnelSessionAttr_Object,
		      CTYPE_SESSION_ATTRIBUTE_LSP_TUNNEL,
		      session_attr_lsp_tunnel_t,
		      Session_Attr_Class) {
		OBJ_OPERATOR(LSPTunnelSessionAttr_Object);
	public:
		LSPTunnelSessionAttr_Object(void) {
			value_.name     = 0;
			value_.name_len = 0;
		}

		~LSPTunnelSessionAttr_Object(void) {
			DELETE_ARR_OBJ(value_.name, value_.name_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength +
				sizeof(value_.setup_prio) +
				sizeof(value_.hold_prio)  +
				sizeof(value_.flags)      +
				sizeof(value_.name_len)   +
				sizeof(*value_.name) * value_.name_len;
		}

		Session_Attr_Class * clone(void) const {
			LSPTunnelSessionAttr_Object * obj;
			obj = new LSPTunnelSessionAttr_Object();
			if (obj) {
				obj->value_.setup_prio = value_.setup_prio;
				obj->value_.hold_prio  = value_.hold_prio;
				obj->value_.flags      = value_.flags;

				obj->value_.name_len   = value_.name_len;
				if (value_.name_len > 0 && value_.name) {
					obj->value_.name = new
						char[value_.name_len];
					assert(obj->value_.name);
					memcpy(obj->value_.name,
					       value_.name,
					       value_.name_len);
				} else {
					obj->value_.name = 0;
				}
			}
			return obj;
		}
	};

//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                         Exclude-any                           |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                         Include-any                           |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                         Include-all                           |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |   Setup Prio  | Holding Prio  |     Flags     |  Name Length  |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                                                               |
//   //          Session Name      (NULL padded display string)      //
//   |                                                               |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit32_t	exclude_any;
		bit32_t	include_any;
		bit32_t	include_all;
		bit8_t	setup_prio;
		bit8_t	hold_prio;
		bit8_t	flags;
		bit8_t	name_len;
		char *  name;          // uint8_t * ???
	} session_attr_lsp_tunnel_ra_t;

	DEFINE_OBJECT(LSPTunnelRASessionAttr_Object,
		      CTYPE_SESSION_ATTRIBUTE_LSP_TUNNEL_RA,
		      session_attr_lsp_tunnel_ra_t,
		      Session_Attr_Class) {
		OBJ_OPERATOR(LSPTunnelRASessionAttr_Object);

	public:
		LSPTunnelRASessionAttr_Object(void) {
			value_.name     = 0;
			value_.name_len = 0;
		}

		~LSPTunnelRASessionAttr_Object(void) {
			DELETE_ARR_OBJ(value_.name, value_.name_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength                +
				sizeof(value_.exclude_any) +
				sizeof(value_.include_any) +
				sizeof(value_.include_all) +
				sizeof(value_.setup_prio)  +
				sizeof(value_.hold_prio)   +
				sizeof(value_.flags)       +
				sizeof(value_.name_len)    +
				sizeof(*value_.name) * value_.name_len;
		}

		Session_Attr_Class * clone(void) const {
			LSPTunnelRASessionAttr_Object * obj;
			obj = new LSPTunnelRASessionAttr_Object();
			if (obj) {
				obj->value_.exclude_any = value_.exclude_any;
				obj->value_.include_any = value_.include_any;
				obj->value_.include_all = value_.include_all;
				obj->value_.setup_prio  = value_.setup_prio;
				obj->value_.hold_prio   = value_.hold_prio;
				obj->value_.flags       = value_.flags;

				obj->value_.name_len    = value_.name_len;
				if (value_.name_len > 0 && value_.name) {
					obj->value_.name =
						new char[value_.name_len];
					assert(obj->value_.name);
					memcpy(obj->value_.name,
					       value_.name,
					       value_.name_len);
				} else {
					obj->value_.name = 0;
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Session_Attr_Class);

//CLASS RSVP HOP
	DEFINE_CLASS_BEGIN(Rsvp_Hop_Class, G2RSVP_CLASS_RSVP_HOP);

	/////////////// TLVs definition /////////////
	typedef ipv4_addr_t tlv_ipv4_addr_t;
	DEFINE_TLV(IPv4_TLV,
		   tlv_ipv4_addr_t,
		   IF_ID_TLV_IPV4) {
		OBJ_OPERATOR(IPv4_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return IF_ID_TLV_IPV4_LEN;
		}

		TLV * clone(void) const {
			IPv4_TLV * obj;
			obj = new IPv4_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	typedef ipv6_addr_t tlv_ipv6_addr_t;
	DEFINE_TLV(IPv6_TLV,
		   tlv_ipv6_addr_t,
		   IF_ID_TLV_IPV6) {
		OBJ_OPERATOR(IPv6_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return IF_ID_TLV_IPV6_LEN;
		}

		TLV * clone(void) const {
			IPv6_TLV * obj;
			obj = new IPv6_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                            IP Address                         |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                           Interface ID                        |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	      IP Address: 32 bits
//
//	         The IP address field may carry either an IP address of a link
//	         or an IP address associated with the router, where associated
//	         address is the value carried in a router address TLV of
//	         routing.
//
//	      Interface ID: 32 bits
//
//	         For type 3 usage, the Interface ID carries an interface
//	         identifier.
//
//	         For types 4 and 5, the Interface ID indicates a bundled
//	         component link.  The special value 0xFFFFFFFF can be used to
//	         indicate the same label is to be valid across all component
//	         links.
//
	typedef struct {
		bit32_t ip_address;
		bit32_t interface_id;
	} tlv_if_idx_t, tlv_comp_if_down_t, tlv_comp_if_up_t;
	DEFINE_TLV(IF_IDX_TLV,
		   tlv_if_idx_t,
		   IF_ID_TLV_IF_INDEX) {
		OBJ_OPERATOR(IF_IDX_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return IF_ID_TLV_IF_INDEX_LEN;
		}

		TLV * clone(void) const {
			IF_IDX_TLV * obj;
			obj = new IF_IDX_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_TLV(COMP_IF_DOWN_TLV,
		   tlv_comp_if_down_t,
		   IF_ID_TLV_COMP_IF_DOWN) {
		OBJ_OPERATOR(COMP_IF_DOWN_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return IF_ID_TLV_COMP_IF_DOWN_LEN;
		}

		TLV * clone(void) const {
			COMP_IF_DOWN_TLV * obj;
			obj = new COMP_IF_DOWN_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_TLV(COMP_IF_UP_TLV,
		   tlv_comp_if_up_t,
		   IF_ID_TLV_COMP_IF_UP) {
		OBJ_OPERATOR(COMP_IF_UP_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return IF_ID_TLV_COMP_IF_UP_LEN;
		}

		TLV * clone(void) const {
			COMP_IF_UP_TLV * obj;
			obj = new COMP_IF_UP_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                             Label                             |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	DEFINE_TLV(DOWN_LABEL_TLV,
		   label_t,
		   IF_ID_TLV_DOWN_LABEL) {
		OBJ_OPERATOR(DOWN_LABEL_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return IF_ID_TLV_DOWN_LABEL_LEN;
		}

		TLV * clone(void) const {
			DOWN_LABEL_TLV * obj;
			obj = new DOWN_LABEL_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_TLV(UP_LABEL_TLV,
		   label_t,
		   IF_ID_TLV_UP_LABEL) {
		OBJ_OPERATOR(UP_LABEL_TLV);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return IF_ID_TLV_UP_LABEL_LEN;
		}

		TLV * clone(void) const {
			UP_LABEL_TLV * obj;
			obj = new UP_LABEL_TLV();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};
	/////////////////////////////////////

//           +-------------+-------------+-------------+-------------+
//           |             IPv4 Next/Previous Hop Address            |
//           +-------------+-------------+-------------+-------------+
//           |                 Logical Interface Handle              |
//           +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv4_t	ipv4_next_hop;
		bit32_t	if_handle;
	} rsvp_hop_ipv4_t;
	DEFINE_OBJECT(IPv4RsvpHop_Object,
		      CTYPE_RSVP_HOP_IPV4,
		      rsvp_hop_ipv4_t,
		      Rsvp_Hop_Class) {
		OBJ_OPERATOR(IPv4RsvpHop_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(rsvp_hop_ipv4_t);
		}

		Rsvp_Hop_Class * clone(void) const {
			IPv4RsvpHop_Object * obj;
			obj = new IPv4RsvpHop_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//           +-------------+-------------+-------------+-------------+
//           |                                                       |
//           +                                                       +
//           |                                                       |
//           +             IPv6 Next/Previous Hop Address            +
//           |                                                       |
//           +                                                       +
//           |                                                       |
//           +-------------+-------------+-------------+-------------+
//           |                Logical Interface Handle               |
//           +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv6_t  ipv6_next_hop;
		bit32_t if_handle;
	} rsvp_hop_ipv6_t;
	DEFINE_OBJECT(IPv6RsvpHop_Object,
		      CTYPE_RSVP_HOP_IPV6,
		      rsvp_hop_ipv6_t,
		      Rsvp_Hop_Class) {
		OBJ_OPERATOR(IPv6RsvpHop_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(rsvp_hop_ipv6_t);
		}

		Rsvp_Hop_Class * clone(void) const {
			IPv6RsvpHop_Object * obj;
			obj = new IPv6RsvpHop_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                 IPv4 Next/Previous Hop Address                |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                     Logical Interface Handle                  |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   ~                              TLVs                             ~
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv4_t           next_prev_ipv4_hop;
		bit32_t          logical_if_hdl;
		std::list<TLV *> tlvs;
	} ipv4_if_id_rsvp_hop_t;
	DEFINE_OBJECT(IFIDIPv4RsvpHop_Object,
		      CTYPE_RSVP_HOP_IF_ID_IPV4,
		      ipv4_if_id_rsvp_hop_t,
		      Rsvp_Hop_Class) {
		OBJ_OPERATOR(IFIDIPv4RsvpHop_Object);
	public:
		~IFIDIPv4RsvpHop_Object(void) {
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.tlvs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength                       +
				len                               +
				sizeof(value_.next_prev_ipv4_hop) +
				sizeof(value_.logical_if_hdl);
		}

		Rsvp_Hop_Class * clone(void) const {

			IFIDIPv4RsvpHop_Object * obj;
			obj = new IFIDIPv4RsvpHop_Object();

			if (obj) {
				obj->value_.next_prev_ipv4_hop =
					value_.next_prev_ipv4_hop;
				obj->value_.logical_if_hdl     =
					value_.logical_if_hdl;

				obj->value_.tlvs.clear();

				std::list<TLV *>::const_iterator it;
				for (it = value_.tlvs.begin();
				     it != value_.tlvs.end();
				     it++) {
					if (*it) {
						obj->value_.tlvs.push_back
							((*it)->clone());
					}
				}
			}

			return obj;
		}
	};

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   |                 IPv6 Next/Previous Hop Address                |
//	   |                                                               |
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                     Logical Interface Handle                  |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   ~                              TLVs                             ~
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv6_t           next_prev_ipv6_hop;
		bit32_t          logical_if_hdl;
		std::list<TLV *> tlvs;
	} ipv6_if_id_rsvp_hop_t;
	DEFINE_OBJECT(IFIDIPv6RsvpHop_Object,
		      CTYPE_RSVP_HOP_IF_ID_IPV6,
		      ipv6_if_id_rsvp_hop_t,
		      Rsvp_Hop_Class) {
		OBJ_OPERATOR(IFIDIPv6RsvpHop_Object);
	public:
		~IFIDIPv6RsvpHop_Object(void) {
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.tlvs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength                       +
				len                               +
				sizeof(value_.next_prev_ipv6_hop) +
				sizeof(value_.logical_if_hdl);
		}

		Rsvp_Hop_Class * clone(void) const {
			IFIDIPv6RsvpHop_Object * obj;
			obj = new IFIDIPv6RsvpHop_Object();
			if (obj) {
				obj->value_.next_prev_ipv6_hop =
					value_.next_prev_ipv6_hop;
				obj->value_.logical_if_hdl     =
					value_.logical_if_hdl;

				obj->value_.tlvs.clear();

				std::list<TLV *>::const_iterator it;
				for (it = value_.tlvs.begin();
				     it != value_.tlvs.end();
				     it++) {
					if (*it) {
						obj->value_.tlvs.push_back
							((*it)->clone());
					}
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Rsvp_Hop_Class);

//CLASS TIME VALUES
	DEFINE_CLASS_BEGIN(TimeValues_Class, G2RSVP_CLASS_TIME_VALUES);

	typedef struct {
		bit32_t refresh_period;
	} time_values_t;
	DEFINE_OBJECT(TimeValues_Object,
		      CTYPE_TIME_VALUES,
		      time_values_t,
		      TimeValues_Class) {
		OBJ_OPERATOR(TimeValues_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(time_values_t);
		}

		TimeValues_Class * clone(void) const {
			TimeValues_Object * obj;
			obj = new TimeValues_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(TimeValues_Class);

//CLASS ERROR SPEC
	DEFINE_CLASS_BEGIN(Error_Spec_Class, G2RSVP_CLASS_ERROR_SPEC);

//     +-------------+-------------+-------------+-------------+
//     |            IPv4 Error Node Address (4 bytes)          |
//     +-------------+-------------+-------------+-------------+
//     |    Flags    |  Error Code |        Error Value        |
//     +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv4_t	ipv4_err_node;
		bit8_t	flags;
		bit8_t	error_code;
		bit16_t error_value;
	} ipv4_error_spec_t;
	DEFINE_OBJECT(IPv4ErrorSpec_Object,
		      CTYPE_ERROR_SPEC_IPV4,
		      ipv4_error_spec_t,
		      Error_Spec_Class) {
		OBJ_OPERATOR(IPv4ErrorSpec_Object);
	public:
		IPv4ErrorSpec_Object(void) { }
		IPv4ErrorSpec_Object(uint32_t ipv4,
				     uint8_t  flags,
				     uint8_t  errCode,
				     uint16_t errVal) {
			value_.ipv4_err_node.s_addr = htonl(ipv4);
			value_.flags                = flags;
			value_.error_code           = errCode;
			value_.error_value          = errVal;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv4_error_spec_t);
		}

		Error_Spec_Class * clone(void) const {
			IPv4ErrorSpec_Object * obj;
			obj = new IPv4ErrorSpec_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//      +-------------+-------------+-------------+-------------+
//      |                                                       |
//      +                                                       +
//      |                                                       |
//      +           IPv6 Error Node Address (16 bytes)          +
//      |                                                       |
//      +                                                       +
//      |                                                       |
//      +-------------+-------------+-------------+-------------+
//      |    Flags    |  Error Code |        Error Value        |
//      +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv6_t	ipv6_err_node;
		bit8_t	flags;
		bit8_t	error_code;
		bit16_t error_value;
	} ipv6_error_spec_t;
	DEFINE_OBJECT(IPv6ErrorSpec_Object,
		      CTYPE_ERROR_SPEC_IPV6,
		      ipv6_error_spec_t,
		      Error_Spec_Class) {
		OBJ_OPERATOR(IPv6ErrorSpec_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv6_error_spec_t);
		}

		Error_Spec_Class * clone(void) const {
			IPv6ErrorSpec_Object * obj;
			obj = new IPv6ErrorSpec_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//     |            Length             | Class-Num (6) | C-Type (3)    |
//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//     |                     IPv4 Error Node Address                   |
//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//     |     Flags     |   Error Code  |          Error Value          |
//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//     |                                                               |
//     ~                              TLVs                             ~
//     |                                                               |
//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv4_t           ipv4_err_node;
		bit8_t           flags;
		bit8_t           error_code;
		bit16_t          error_value;
		std::list<TLV *> tlvs;
	} if_id_ipv4_error_spec_t;
	DEFINE_OBJECT(IFIDIPv4ErrorSpec_Object,
		      CTYPE_ERROR_SPEC_IF_ID_IPV4,
		      if_id_ipv4_error_spec_t,
		      Error_Spec_Class) {
		OBJ_OPERATOR(IFIDIPv4ErrorSpec_Object);
	public:
		IFIDIPv4ErrorSpec_Object(void) { }
		IFIDIPv4ErrorSpec_Object(uint32_t ipv4,
					 uint8_t  flags,
					 uint8_t  errCode,
					 uint16_t errVal) {
			value_.ipv4_err_node.s_addr = htonl(ipv4);
			value_.flags                = flags;
			value_.error_code           = errCode;
			value_.error_value          = errVal;
		}
		~IFIDIPv4ErrorSpec_Object(void) {
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.tlvs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength                +
				len                        +
				sizeof(value_.flags)       +
				sizeof(value_.error_code)  +
				sizeof(value_.error_value) +
				sizeof(value_.ipv4_err_node);
		}

		Error_Spec_Class * clone(void) const {
			IFIDIPv4ErrorSpec_Object * obj;
			obj = new IFIDIPv4ErrorSpec_Object();

			if (obj) {
				obj->value_.ipv4_err_node=value_.ipv4_err_node;
				obj->value_.flags        = value_.flags;
				obj->value_.error_code   = value_.error_code;
				obj->value_.error_value  = value_.error_value;

				obj->value_.tlvs.clear();

				std::list<TLV *>::const_iterator it;
				for (it = value_.tlvs.begin();
				     it != value_.tlvs.end();
				     it++) {
					if (*it) {
						obj->value_.tlvs.push_back
							((*it)->clone());
					}
				}
			}

			return obj;
		}
	};

//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |            Length             | Class-Num (6) | C-Type (4)    |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                                                               |
//      |                     IPv6 Error Node Address                   |
//      |                                                               |
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |     Flags     |   Error Code  |          Error Value          |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                                                               |
//      ~                              TLVs                             ~
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv6_t           ipv6_err_node;
		bit8_t           flags;
		bit8_t           error_code;
		bit16_t          error_value;
		std::list<TLV *> tlvs;
	} if_id_ipv6_error_spec_t;
	DEFINE_OBJECT(IFIDIPv6ErrorSpec_Object,
		      CTYPE_ERROR_SPEC_IF_ID_IPV6,
		      if_id_ipv6_error_spec_t,
		      Error_Spec_Class) {
		OBJ_OPERATOR(IFIDIPv6ErrorSpec_Object);
	public:
		~IFIDIPv6ErrorSpec_Object(void) {
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.tlvs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<TLV *>::const_iterator it;
			for (it = value_.tlvs.begin();
			     it != value_.tlvs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength                +
				len                        +
				sizeof(value_.flags)       +
				sizeof(value_.error_code)  +
				sizeof(value_.error_value) +
				sizeof(value_.ipv6_err_node);
		}

		Error_Spec_Class * clone(void) const {
			IFIDIPv6ErrorSpec_Object * obj;
			obj = new IFIDIPv6ErrorSpec_Object();

			if (obj) {
				obj->value_.ipv6_err_node=value_.ipv6_err_node;
				obj->value_.flags        = value_.flags;
				obj->value_.error_code   = value_.error_code;
				obj->value_.error_value  = value_.error_value;

				obj->value_.tlvs.clear();

				std::list<TLV *>::const_iterator it;
				for (it = value_.tlvs.begin();
				     it != value_.tlvs.end();
				     it++) {
					if (*it) {
						obj->value_.tlvs.push_back
							((*it)->clone());
					}
				}
			}

			return obj;
		}
	};

	DEFINE_CLASS_END(Error_Spec_Class);

//CLASS_SCOPE
	DEFINE_CLASS_BEGIN(Scope_Class, G2RSVP_CLASS_SCOPE);

	typedef struct {
		std::list<ipv4_t *> subObjs;
	} ipv4_scope_t;
	DEFINE_OBJECT(IPv4Scope_Object,
		      CTYPE_IPV4_SCOPE,
		      ipv4_scope_t,
		      Scope_Class) {
		OBJ_OPERATOR(IPv4Scope_Object);
	public:
		~IPv4Scope_Object(void) {
			std::list<ipv4_t *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.subObjs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<ipv4_t *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it) {
					len += sizeof(*(*it));
				}
			}
			// equivalent to
			//len = sizeof(ipv4_t) * value_.subObjs.size()

			return objHdrLength + len;
		}

		Scope_Class * clone(void) const {
			IPv4Scope_Object * obj;
			obj = new IPv4Scope_Object();

			if (obj) {
				obj->value_.subObjs.clear();

				std::list<ipv4_t *>::const_iterator it;
				for (it = value_.subObjs.begin();
				     it != value_.subObjs.end();
				     it++) {
					if (*it) {
						ipv4_t * ip = new ipv4_t();
						*ip = *(*it);
						obj->value_.
							subObjs.push_back(ip);
					}
				}
			}

			return obj;
		}
	};

	typedef struct {
		std::list<ipv6_t *> subObjs;
	} ipv6_scope_t;
	DEFINE_OBJECT(IPv6Scope_Object,
		      CTYPE_IPV6_SCOPE,
		      ipv6_scope_t,
		      Scope_Class){
		OBJ_OPERATOR(IPv6Scope_Object);
	public:
		~IPv6Scope_Object(void) {
			std::list<ipv6_t *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.subObjs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<ipv6_t *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it) {
					len += sizeof(*(*it));
				}
			}
			// equivalent to
			//len = sizeof(ipv6_t) * subObjs.size()

			return objHdrLength + len;
		}

		Scope_Class * clone(void) const {
			IPv6Scope_Object * obj;
			obj = new IPv6Scope_Object();

			if (obj) {
				obj->value_.subObjs.clear();

				std::list<ipv6_t *>::const_iterator it;
				for (it = value_.subObjs.begin();
				     it != value_.subObjs.end();
				     it++) {
					if (*it) {
						ipv6_t * ip = new ipv6_t();
						*ip = *(*it);
						obj->value_.
							subObjs.push_back(ip);
					}
				}
			}

			return obj;
		}
	};

	DEFINE_CLASS_END(Scope_Class);

//CLASS FLOWSPEC
	DEFINE_CLASS_BEGIN(FlowSpec_Class, G2RSVP_CLASS_FLOWSPEC);

// FLOWSPEC object format (RFC 2210/2211)
//
// ... when requesting Controlled-Load service
//        31           24 23           16 15            8 7             0
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   1   | 0 (a) |    reserved           |             7 (b)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   2   |    5  (c)     |0| reserved    |             6 (d)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   3   |   127 (e)     |    0 (f)      |             5 (g)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   4   |  Token Bucket Rate [r] (32-bit IEEE floating point number)    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   5   |  Token Bucket Size [b] (32-bit IEEE floating point number)    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   6   |  Peak Data Rate [p] (32-bit IEEE floating point number)       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   7   |  Minimum Policed Unit [m] (32-bit integer)                    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   8   |  Maximum Packet Size [M]  (32-bit integer)                    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//     (a) - Message format version number (0)
//     (b) - Overall length (7 words not including header)
//     (c) - Service header, service number 5 (Controlled-Load)
//     (d) - Length of controlled-load data, 6 words not including
//           per-service header
//     (e) - Parameter ID, parameter 127 (Token Bucket TSpec)
//     (f) - Parameter 127 flags (none set)
//     (g) - Parameter 127 length, 5 words not including per-service
//           header
//
// ... when requesting Garanteed service
//        31           24 23           16 15            8 7             0
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   1   | 0 (a) |    Unused             |            10 (b)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   2   |    2  (c)     |0| reserved    |             9 (d)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   3   |   127 (e)     |    0 (f)      |             5 (g)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   4   |  Token Bucket Rate [r] (32-bit IEEE floating point number)    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   5   |  Token Bucket Size [b] (32-bit IEEE floating point number)    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   6   |  Peak Data Rate [p] (32-bit IEEE floating point number)       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   7   |  Minimum Policed Unit [m] (32-bit integer)                    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   8   |  Maximum Packet Size [M]  (32-bit integer)                    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   9   |     130 (h)   |    0 (i)      |            2 (j)              |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   10  |  Rate [R]  (32-bit IEEE floating point number)                |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   11  |  Slack Term [S]  (32-bit integer)                             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//     (a) - Message format version number (0)
//     (b) - Overall length (9 words not including header)
//     (c) - Service header, service number 2 (Guaranteed)
//     (d) - Length of per-service data, 9 words not including per-service
//           header
//     (e) - Parameter ID, parameter 127 (Token Bucket TSpec)
//     (f) - Parameter 127 flags (none set)
//     (g) - Parameter 127 length, 5 words not including parameter header
//     (h) - Parameter ID, parameter 130 (Guaranteed Service RSpec)
//     (i) - Parameter 130 flags (none set)
//     (j) - Parameter 130 length, 2 words not including parameter header

	typedef struct {
		bit4_t  msg_version;         //a
		bit12_t reserved12;
		bit16_t length;              //b
		bit8_t  service_header;      //c
		bit1_t  zero_bit;
		bit7_t  reserved7;
		bit16_t service_data_length; //d
		bit8_t  param_id_127;        //e
		bit8_t  param127_flags;      //f
		bit16_t param127_length;     //g
		bit32_t token_bucket_rate;   //[r] 32-bit IEEE floating point
		bit32_t token_bucket_size;   //[b] 32-bit IEEE floating point
		bit32_t peak_data_rate;      //[p] 32-bit IEEE floating point
		bit32_t min_policed_unit;    //[m] integer
		bit32_t max_policed_unit;    //[M] integer
	} flow_spec_t, sender_spec_t;

	typedef struct {
		flow_spec_t data;
		bit8_t *    param_id_130;      //h
		bit8_t *    param130_flags;    //i
		bit16_t *   param130_length;   //j
		bit32_t *   rate;              //[R] 32-bit IEEE floating point
		bit32_t *   slack_term;        //[S] 32-bit integer
	} int_serv_flow_spec_t;
	DEFINE_OBJECT(IntServFlowSpec_Object,
		      CTYPE_INT_SERV_FLOWSPEC,
		      int_serv_flow_spec_t,
		      FlowSpec_Class) {
		OBJ_OPERATOR(IntServFlowSpec_Object);
	public:
		IntServFlowSpec_Object(void) { }
		IntServFlowSpec_Object(flow_spec_t data) {
			value_.data = data;
		}
		~IntServFlowSpec_Object(void) {
			DELETE_OBJ(value_.param_id_130);
			DELETE_OBJ(value_.param130_flags);
			DELETE_OBJ(value_.param130_length);
			DELETE_OBJ(value_.rate);
			DELETE_OBJ(value_.slack_term);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength          +
				sizeof(uint32_t)     + // msg_vers/resvd/length
				value_.data.length * sizeof(uint32_t);
		}

		FlowSpec_Class * clone(void) const {
			IntServFlowSpec_Object * obj;
			obj = new IntServFlowSpec_Object();

			if (!obj) return 0;

			obj->value_.data = value_.data;

			if (value_.data.service_header ==
			    GUARANTEED_SERVICE) {
				if (value_.param_id_130) {
					obj->value_.param_id_130 = new uint8_t;
					assert(obj->value_.param_id_130);
					*obj->value_.param_id_130 =
						*value_.param_id_130;
				}
				if (value_.param130_flags) {
					obj->value_.param130_flags =
						new uint8_t;
					assert(obj->value_.param130_flags);
					*obj->value_.param130_flags =
						*value_.param130_flags;
				}
				if (value_.param130_length) {
					obj->value_.param130_length =
						new uint16_t;
					assert(obj->value_.param130_length);
					*obj->value_.param130_length =
						*value_.param130_length;
				}
				if (value_.rate) {
					obj->value_.rate = new uint32_t;
					assert(obj->value_.rate);
					*obj->value_.rate = *value_.rate;
				}
				if (value_.slack_term) {
					obj->value_.slack_term = new uint32_t;
					assert(obj->value_.slack_term);
					*obj->value_.slack_term =
						*value_.slack_term;
				}
			}

			return obj;
		}
	};

	DEFINE_CLASS_END(FlowSpec_Class);

//CLASS_MESSAGE_ID
	DEFINE_CLASS_BEGIN(MessageId_Class, G2RSVP_CLASS_MESSAGE_ID);

	typedef struct {
		bit8_t  flags;
		bit24_t epoch;
		bit32_t message_id;
	} message_id_ack_nack_t;
	DEFINE_OBJECT(MessageId_Object,
		      CTYPE_MESSAGE_ID,
		      message_id_ack_nack_t,
		      MessageId_Class) {
		OBJ_OPERATOR(MessageId_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength          +
				sizeof(uint32_t)     + // flags + epoch
				sizeof(value_.message_id);
		}

		MessageId_Class * clone(void) const {
			MessageId_Object * obj = new MessageId_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(MessageId_Class);

//CLASS MESSAGE_ID_ACK
	DEFINE_CLASS_BEGIN(MessageIdAckNack_Class,
			   G2RSVP_CLASS_MESSAGE_ID_ACK_NACK);

	DEFINE_OBJECT(MessageIdAck_Object,
		      CTYPE_MESSAGE_ID_ACK,
		      message_id_ack_nack_t,
		      MessageIdAckNack_Class){
		OBJ_OPERATOR(MessageIdAck_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength          +
				sizeof(uint32_t)     + // flags + epoch
				sizeof(value_.message_id);
		}

		MessageIdAckNack_Class * clone(void) const {
			MessageIdAck_Object * obj = new MessageIdAck_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_OBJECT(MessageIdNack_Object,
		      CTYPE_MESSAGE_ID_NACK,
		      message_id_ack_nack_t,
		      MessageIdAckNack_Class){
		OBJ_OPERATOR(MessageIdNack_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength          +
				sizeof(uint32_t)     + // flags + epoch
				sizeof(value_.message_id);
		}

		MessageIdAckNack_Class * clone(void) const {
			MessageIdNack_Object* obj = new MessageIdNack_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(MessageIdAckNack_Class);

//MESSAGE_ID_LIST
	DEFINE_CLASS_BEGIN(MessageIdList_Class, G2RSVP_CLASS_MESSAGE_ID_LIST);
	typedef struct {
		bit8_t  flags;
		bit24_t epoch;
		std::list<bit32_t> msg_id_list;
	} msg_id_list_t;

	DEFINE_OBJECT(MessageIdList_Object,
		      CTYPE_MSG_ID_LIST,	//CTYPE 1
		      msg_id_list_t,
		      MessageIdList_Class){
		bool isConsistent(void) const {
			return true;
		}
	};

	//intermediate data structure
	typedef struct {
		bit32_t msg_id;
		ipv4_t  src_ipv4_addr;
	} ipv4_source_tuple;

	typedef struct {
		bit8_t  flags;
		bit24_t epoch;
		std::list<ipv4_source_tuple *> source_list;
	} msg_id_ipv4_src_list_t;

	DEFINE_OBJECT(MessageIdIPv4SrcList_Object,
		      CTYPE_MSG_ID_IPV4_SOURCE_LIST,	//CTYPE 2
		      msg_id_ipv4_src_list_t,
		      MessageIdList_Class){
		bool isConsistent(void) const {
			return true;
		}
	};

	typedef struct {
		bit32_t msg_id;
		ipv4_t  src_ipv4_addr;
		ipv4_t	dst_ipv4_addr;
	} mcast_ipv4_msg_id_tuple;

	typedef struct {
		bit8_t  flags;
		bit24_t epoch;
		std::list<mcast_ipv4_msg_id_tuple *> source_list;
	} mcast_ipv4_msg_id_list_t;


	DEFINE_OBJECT(McastMsgIDIPv4List_Object,
		      CTYPE_MSG_ID_IPV4_MCAST_LIST,	//CTYPE 4
		      mcast_ipv4_msg_id_list_t,
		      MessageIdList_Class) {
		bool isConsistent(void) const {
			return true;
		}
	};
	//TODO IPV6 C-types asap

	DEFINE_CLASS_END(MessageIdList_Class);

//CLASS INTEGRITY
	DEFINE_CLASS_BEGIN(Integrity_Class, G2RSVP_CLASS_INTEGRITY);

//       +-------------+-------------+-------------+-------------+
//       |    Flags    | 0 (Reserved)|                           |
//       +-------------+-------------+                           +
//       |                    Key Identifier                     |
//       +-------------+-------------+-------------+-------------+
//       |                    Sequence Number                    |
//       |                                                       |
//       +-------------+-------------+-------------+-------------+
//       |                                                       |
//       +                                                       +
//       |                                                       |
//       +                  Keyed Message Digest                 |
//       |                                                       |
//       +                                                       +
//       |                                                       |
//       +-------------+-------------+-------------+-------------+
//
//   o    Flags: An 8-bit field with the following format:
//
//                                      Flags
//
//                          0   1   2   3   4   5   6   7
//                        +---+---+---+---+---+---+---+---+
//                        | H |                           |
//                        | F |             0             |
//                        +---+---+---+---+---+---+---+---+
//
//          Currently only one flag (HF) is defined.  The remaining flags
//          are reserved for future use and MUST be set to 0.
//
//          o    Bit 0: Handshake Flag (HF) concerns the integrity
//               handshake mechanism (Section 4.3).  Message senders
//               willing to respond to integrity handshake messages SHOULD
//               set this flag to 1 whereas those that will reject
//               integrity handshake messages SHOULD set this to 0.
//
//     o    Key Identifier: An unsigned 48-bit number that MUST be unique
//          for a given sender.  Locally unique Key Identifiers can be
//          generated using some combination of the address (IP or MAC or
//          LIH) of the sending interface and the key number.  The
//          combination of the Key Identifier and the sending system's IP
//          address uniquely identifies the security association (Section
//          2.2).
//
//     o    Sequence Number: An unsigned 64-bit monotonically increasing,
//          unique sequence number.
//
//          Sequence Number values may be any monotonically increasing
//          sequence that provides the INTEGRITY object [of each RSVP
//          message] with a tag that is unique for the associated key's
//          lifetime.  Details on sequence number generation are presented
//          in Section 3.
//
//     o    Keyed Message Digest: The digest MUST be a multiple of 4
//          octets long.  For HMAC-MD5, it will be 16 bytes long.

#define MSG_DIGEST_LEN 16
	typedef struct {
		bit8_t	flags;
		bit8_t	reserved;
		bit48_t	key_id;
		bit64_t seq_number;
		bit8_t	digest[MSG_DIGEST_LEN];
	} integrity_t;
	DEFINE_OBJECT(Integrity_Object,
		      CTYPE_INTEGRITY,
		      integrity_t,
		      Integrity_Class) {
		OBJ_OPERATOR(Integrity_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength                +
				sizeof(value_.key_id)      + //flags/rsvd/keyId
				sizeof(value_.seq_number)  +
				sizeof(value_.digest);
		}

		Integrity_Class * clone(void) const {
			Integrity_Object * obj = new Integrity_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Integrity_Class);

//CLASS EXPLICIT ROUTE
	DEFINE_CLASS_BEGIN(ExplicitRoute_Class, G2RSVP_CLASS_EXPLICIT_ROUTE);

	/////////////// ERO subobjects definition /////////////

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |L|    Type     |     Length    | IPv4 address (4 bytes)        |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv4 address (continued)      | Prefix Length |      Resvd    |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv4_t ipv4_addr;
		bit8_t preflen;
		bit8_t reserved;
	} ipv4_prefix_t;
	DEFINE_EROSUBOBJ(IPv4Pref_SubObj,
			 ipv4_prefix_t,
			 IPV4_PREF_SUBOBJECT) {
		OBJ_OPERATOR(IPv4Pref_SubObj);
	public:
		IPv4Pref_SubObj(void) { }
		IPv4Pref_SubObj(ipv4_t addr, uint8_t preflen) {
			content_.ipv4_addr = addr;
			content_.preflen   = preflen;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return eroSubObjHdrLength           +
				sizeof(content_.ipv4_addr)  +
				sizeof(content_.preflen)    +
				sizeof(content_.reserved);
		}

		EroSubObject * clone(void) const {
			IPv4Pref_SubObj * obj = new IPv4Pref_SubObj();
			if (obj) {
				obj->content_ = content_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |L|    Type     |     Length    | IPv6 address (16 bytes)       |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)                                      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)                                      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)                                      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)      | Prefix Length |      Resvd    |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv6_t ipv6_addr;
		bit8_t preflen;
		bit8_t reserved;
	} ipv6_prefix_t;
	DEFINE_EROSUBOBJ(IPv6Pref_SubObj,
			 ipv6_prefix_t,
			 IPV6_PREF_SUBOBJECT) {
		OBJ_OPERATOR(IPv6Pref_SubObj);
	public:
		IPv6Pref_SubObj(void) { }
		IPv6Pref_SubObj(ipv6_t addr, uint8_t preflen) {
			content_.ipv6_addr = addr;
			content_.preflen   = preflen;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return eroSubObjHdrLength           +
				sizeof(content_.ipv6_addr)  +
				sizeof(content_.preflen)    +
				sizeof(content_.reserved);
		}

		EroSubObject * clone(void) const {
			IPv6Pref_SubObj * obj = new IPv6Pref_SubObj();
			if (obj) {
				obj->content_ = content_;
			}
			return obj;
		}
	};

	typedef struct {
		bit16_t as_number;
	} as_number_t;
	DEFINE_EROSUBOBJ(ASNum_SubObj,
			 as_number_t,
			 AS_SUBOBJECT) {
		OBJ_OPERATOR(ASNum_SubObj);
	public:
		ASNum_SubObj(void) { }
		ASNum_SubObj(uint16_t as_number) {
			content_.as_number = as_number;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return eroSubObjHdrLength + sizeof(as_number_t);
		}

		EroSubObject * clone(void) const {
			ASNum_SubObj * obj = new ASNum_SubObj();
			if (obj) {
				obj->content_ = content_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |L|    Type     |     Length    |    Reserved (MUST be zero)    |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                           Router ID                           |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                     Interface ID (32 bits)                    |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit16_t reserved;
		bit32_t router_id;
		bit32_t interface_id;
	} unnum_link_t;
	DEFINE_EROSUBOBJ(Unnum_SubObj,
			 unnum_link_t,
			 UNNUM_LINK_SUBOBJECT) {
		OBJ_OPERATOR(Unnum_SubObj);
	public:
		Unnum_SubObj(void) { }
		Unnum_SubObj(uint32_t router_id, uint32_t interface_id) {
			content_.router_id    = router_id;
			content_.interface_id = interface_id;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return eroSubObjHdrLength             +
				sizeof(content_.reserved)     +
				sizeof(content_.router_id)    +
				sizeof(content_.interface_id);
		}

		EroSubObject * clone(void) const {
			Unnum_SubObj * obj = new Unnum_SubObj();
			if (obj) {
				obj->content_ = content_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |L|    Type     |     Length    |U|   Reserved  |   C-Type      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                             Label                             |
//	   |                              ...                              |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	  C-Type
//
//	      The C-Type of the included Label Object.  Copied from the Label
//	      Object.
//
//        ....
//
//	   The Label Explicit and Record Routes contains:
//
//	      L: 1 bit
//
//	        This bit must be set to 0.
//
//	      U: 1 bit
//
//	        This bit indicates the direction of the label.  It is 0 for the
//	        downstream label.  It is set to 1 for the upstream label and is
//	        only used on bidirectional LSPs.

	typedef struct {
		bit1_t        upstream;
		bit7_t        reserved;
		bit8_t        c_type;
		Label_Class * label;
	} label_ero_t;
	DEFINE_EROSUBOBJ(LabelERO_SubObj,
			 label_ero_t,
			 LABEL_ERO_SUBOBJECT) {
		OBJ_OPERATOR(LabelERO_SubObj);
	public:
		LabelERO_SubObj(void) {
			content_.label = 0;
		}

		LabelERO_SubObj(uint32_t label, bool upstream) {
			GLabel_Object * lbl;
			lbl = new RSVP::GLabel_Object(&label, sizeof(label));
			assert(lbl);

			content_.label    = lbl;
			content_.upstream = upstream;
			content_.c_type   = CTYPE_GENERALISED_LABEL;
		}

		~LabelERO_SubObj(void) {
			DELETE_OBJ(content_.label);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return eroSubObjHdrLength        +
				sizeof(content_.c_type)  +
				sizeof(bit8_t)           + // Upstream/reserved
				(content_.label ?
				(content_.label->length() -
				 objHeader::objHdrLength) : 0);
		}

		EroSubObject * clone(void) const {
			LabelERO_SubObj * obj = new LabelERO_SubObj();
			if (obj) {
				obj->content_.upstream = content_.upstream;
				obj->content_.c_type   = content_.c_type;
				obj->content_.reserved = content_.reserved;
				if (content_.label) {
					obj->content_.label =
						content_.label->clone();
				}
			}
			return obj;
		}
	};

	//
	// TODO!!!!!!
	//

#define COMP_IF_ID_IPv4_SUBOBJECT        10
#define COMP_IF_ID_IPv4_SUBOBJECT_LEN   100
	typedef struct {
		// TODO !!!!!!!!
	} compIFIDIPv4_t;
	DEFINE_EROSUBOBJ(CompIFIDIPv4_SubObj,
			 compIFIDIPv4_t,
			 COMP_IF_ID_IPv4_SUBOBJECT) {
			 //COMP_IF_ID_IPv4_SUBOBJECT_LEN){
		bool isConsistent(void) const {
			return true;
		}
	};

#define COMP_IF_ID_IPv6_SUBOBJECT        11
#define COMP_IF_ID_IPv6_SUBOBJECT_LEN   111
	typedef struct {
		// TODO !!!!!!!!
	} compIFIDIPv6_t;
	DEFINE_EROSUBOBJ(CompIFIDIPv6_SubObj,
			 compIFIDIPv6_t,
			 COMP_IF_ID_IPv6_SUBOBJECT) {
			 //COMP_IF_ID_IPv6_SUBOBJECT_LEN){
		bool isConsistent(void) const {
			return true;
		}
	};

#define COMP_IF_ID_UNNUM_SUBOBJECT       12
#define COMP_IF_ID_UNNUM_SUBOBJECT_LEN  122
	typedef struct {
		// TODO !!!!!!!!
	} compIFIDUnnum_t;
	DEFINE_EROSUBOBJ(CompIFIDUnnum_SubObj,
			 compIFIDUnnum_t,
			 COMP_IF_ID_UNNUM_SUBOBJECT) {
			 //COMP_IF_ID_UNNUM_SUBOBJECT_LEN){
		bool isConsistent(void) const {
			return true;
		}
	};
	/////////////////////////////////////

// ERO format (RFC 3209)
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   //                        (Subobjects)                          //
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		std::list<EroSubObject *> subObjs;
	} ero_t;
	DEFINE_OBJECT(ExplicitRoute_Object,
		      CTYPE_ERO,
		      ero_t,
		      ExplicitRoute_Class) {
		OBJ_OPERATOR(ExplicitRoute_Object);
	public:
		~ExplicitRoute_Object(void) {
			std::list<EroSubObject *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.subObjs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		bool insertEroSubObj(EroSubObject * obj, bool insertTail) {

			if (!obj) {
				G2RSVPTE_WRN("Invalid EROsubObj, null pointer");
				return false;
			}

			std::list<EroSubObject *>::iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it == obj) {
					G2RSVPTE_WRN("ERO subObj %p is "
						     "already present", obj);
					return false;
				}
			}

			if (insertTail) {
				value_.subObjs.push_back(obj);
			} else {
				value_.subObjs.push_front(obj);
			}

			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<EroSubObject *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength + len;
		}

		ExplicitRoute_Class * clone(void) const {
			ExplicitRoute_Object* obj = new ExplicitRoute_Object();
			if (obj) {
				obj->value_.subObjs.clear();

				std::list<EroSubObject *>::const_iterator it;
				for (it = value_.subObjs.begin();
				     it != value_.subObjs.end();
				     it++) {
					if (*it) {
						obj->value_.subObjs.push_back
							((*it)->clone());
					}
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(ExplicitRoute_Class);

//LABEL REQUEST
	DEFINE_CLASS_BEGIN(LabelRequest_Class, G2RSVP_CLASS_LABEL_REQ);

	typedef struct {
		bit8_t	enc_type;
		bit8_t	switch_type;
		bit16_t g_pid;
	} glabel_request_t;

	DEFINE_OBJECT(GLabelRequest_Object,
		      CTYPE_GENERALISED_LABEL_REQ,
		      glabel_request_t,
		      LabelRequest_Class) {
		OBJ_OPERATOR(GLabelRequest_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(glabel_request_t);
		}

		LabelRequest_Class * clone(void) const {
			GLabelRequest_Object* obj = new GLabelRequest_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(LabelRequest_Class);

//CLASS CALL OPS
	DEFINE_CLASS_BEGIN(CallOptions_Class, G2RSVP_CLASS_CALL_OPS);

//       0                   1                   2                   3
//       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |            Length             |Class-Num (228)|  C-Type (1)   |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                Reserved                       | Call ops flag |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//        Two flags are currently defined for the "call ops flag":
//        0x01: call without connection
//        0x02: synchronizing a call (for restart mechanism)

	typedef struct {
		bit24_t reserved;
		bit8_t	call_ops_flags;
	} call_ops_t;
	DEFINE_OBJECT(CallOptions_Object,
		      CTYPE_CALL_OPS,
		      call_ops_t,
		      CallOptions_Class) {
		OBJ_OPERATOR(CallOptions_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(uint32_t);
		}

		CallOptions_Class * clone(void) const {
			CallOptions_Object * obj = new CallOptions_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(CallOptions_Class);

//CLASS CALL ID
	DEFINE_CLASS_BEGIN(CallId_Class, G2RSVP_CLASS_CALL_ID);

//       0                   1                   2                   3
//       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |            Length             |Class-Num (230)|  C-Type  (1)  |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |     Type      |                     Resv                      |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                      Source LSR address                       |
//      ~                              ...                              ~
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                       Local identifier                        |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                  Local identifier  (continued)                |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//       'Type' is defined to indicate the type of format used for the src LSR
//
//           Type=0x01, the source LSR address is 4 bytes
//           Type=0x02, the source LSR address is 16 bytes
//           Type=0x03, the source LSR address is 20 bytes
//           type=0x04, the source LSR address is 6 bytes
//           type=0x7f, the source LSR address has length defined by the vendor
//
//       'Source LSR address':
//           An address of the LSR controlled by the source network.
//
//       'Local identifier':
//           A 64-bit identifier constant over the life of the call.

	typedef struct {
		bit8_t	addr_type;
		bit24_t reserved;
		bit8_t  src_lsr_addr_len;  // just for internal use
		char *	src_lsr_addr;      // variable length
		bit64_t	local_id;
	} call_id_op_spec_t;

	DEFINE_OBJECT(CallIdOpSpec_Object,
		      CTYPE_CALL_ID_OPERATOR_SPECIFIC,
		      call_id_op_spec_t,
		      CallId_Class) {
		OBJ_OPERATOR(CallIdOpSpec_Object);
	public:
		CallIdOpSpec_Object(void) {
			value_.src_lsr_addr     = 0;
			value_.src_lsr_addr_len = 0;
		}

		CallIdOpSpec_Object(bit8_t  type,
				    bit8_t  addr_len,
				    char *  addr,
				    bit64_t loc_id) {
			value_.addr_type        = type;
			value_.src_lsr_addr     = addr;
			value_.src_lsr_addr_len = addr_len;
			value_.local_id         = loc_id;
		}

		~CallIdOpSpec_Object(void) {
			DELETE_ARR_OBJ(value_.src_lsr_addr,
				       value_.src_lsr_addr_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength                   +
				sizeof(uint32_t)              +
				(sizeof(*value_.src_lsr_addr) *
				 value_.src_lsr_addr_len)     +
				sizeof(value_.local_id);
		}

		CallId_Class * clone(void) const {
			CallIdOpSpec_Object * obj = new CallIdOpSpec_Object();
			if (obj) {
				obj->value_.addr_type = value_.addr_type;
				obj->value_.reserved  = value_.reserved;
				obj->value_.local_id  = value_.local_id;
				obj->value_.src_lsr_addr_len =
					value_.src_lsr_addr_len;

				if (value_.src_lsr_addr_len > 0 &&
				    value_.src_lsr_addr) {
					obj->value_.src_lsr_addr = new
						char[value_.src_lsr_addr_len];
					memcpy(obj->value_.src_lsr_addr,
					       value_.src_lsr_addr,
					       value_.src_lsr_addr_len);
				} else {
					obj->value_.src_lsr_addr = 0;
				}
			}
			return obj;
		}
	};

//       0                   1                   2                   3
//       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |            Length             |Class-Num (230)|  C-Type  (2)  |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |     Type      |        ITU international country code         |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                       ITU carrier code                        |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      | ITU carrie code (cont)        |  Unique access point code     |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |               Unique access point code (continued)            |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                      Source LSR address                       |
//      ~                              ...                              ~
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                       Local identifier                        |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                  Local identifier  (continued)                |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//       'Type' is defined to indicate the type of format used for the src LSR
//
//           Type=0x01, the source LSR address is 4 bytes
//           Type=0x02, the source LSR address is 16 bytes
//           Type=0x03, the source LSR address is 20 bytes
//           type=0x04, the source LSR address is 6 bytes
//           type=0x7f, the source LSR address has length defined by the vendor
//
//       'Source LSR address':
//           An address of the LSR controlled by the source network.
//
//       'Local identifier':
//           A 64-bit identifier constant over the life of the call.

	typedef struct {
		bit8_t	addr_type;
		bit24_t itu_country_code;    // International segment
		bit48_t itu_carrier_code;    // National segment - part 1
		bit48_t unique_access_point; // National segment - part 2
		bit8_t  src_lsr_addr_len;    // just for internal use
		char *	src_lsr_addr;        // variable length
		bit64_t	local_id;
	} call_id_glb_t;
	DEFINE_OBJECT(CallIdGlobal_Object,
		      CTYPE_CALL_ID_GLOBALLY_UNIQUE,
		      call_id_glb_t,
		      CallId_Class) {
		OBJ_OPERATOR(CallIdGlobal_Object);
	public:
		CallIdGlobal_Object(void) {
			value_.src_lsr_addr     = 0;
			value_.src_lsr_addr_len = 0;
		}

		CallIdGlobal_Object(bit8_t  type,
				    bit24_t country_code,
				    bit48_t carrier_code,
				    bit48_t unique_ap,
				    bit8_t  addr_len,
				    char *  addr,
				    bit64_t loc_id) {
			value_.addr_type           = type;
			value_.itu_country_code    = country_code;
			value_.itu_carrier_code    = carrier_code;
			value_.unique_access_point = unique_ap;
			value_.src_lsr_addr        = addr;
			value_.src_lsr_addr_len    = addr_len;
			value_.local_id            = loc_id;
		}

		~CallIdGlobal_Object(void) {
			DELETE_ARR_OBJ(value_.src_lsr_addr,
				       value_.src_lsr_addr_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength +
				sizeof(uint32_t) +  // Type + Intl Seg.
				sizeof(uint32_t) +  // Nat Seg. - word 1
				sizeof(uint32_t) +  // Nat Seg. - word 2
				sizeof(uint32_t) +  // Nat Seg. - word 3
				(sizeof(*value_.src_lsr_addr) *
				 value_.src_lsr_addr_len) +
				sizeof(value_.local_id);
		}

		CallId_Class * clone(void) const {
			CallIdGlobal_Object * obj = new CallIdGlobal_Object();
			if (obj) {
				obj->value_.addr_type = value_.addr_type;
				obj->value_.local_id  = value_.local_id;
				obj->value_.itu_country_code    =
					value_.itu_country_code;
				obj->value_.itu_carrier_code =
					value_.itu_carrier_code;
				obj->value_.unique_access_point =
					value_.unique_access_point;
				obj->value_.src_lsr_addr_len =
					value_.src_lsr_addr_len;

				if (obj->value_.src_lsr_addr_len > 0 &&
				    obj->value_.src_lsr_addr) {
					obj->value_.src_lsr_addr =
					      new char[value_.src_lsr_addr_len];
					assert(obj->value_.src_lsr_addr);
					memcpy(obj->value_.src_lsr_addr,
					       value_.src_lsr_addr,
					       value_.src_lsr_addr_len);
				} else {
					obj->value_.src_lsr_addr = 0;
				}
			}
			return obj;
		}
	};

	//

	typedef call_id_op_spec_t call_id_null_t;

	DEFINE_OBJECT(CallIdNull_Object,
		      CTYPE_CALL_ID_NULL,
		      call_id_null_t,
		      CallId_Class) {
		OBJ_OPERATOR(CallIdNull_Object);
	public:
		CallIdNull_Object(void) {
			value_.src_lsr_addr     = 0;
			value_.src_lsr_addr_len = 0;
		}

		CallIdNull_Object(bit8_t  type,
				  bit8_t  addr_len,
				  char *  addr,
				  bit64_t loc_id) {
			value_.addr_type        = type;
			value_.src_lsr_addr     = addr;
			value_.src_lsr_addr_len = addr_len;
			value_.local_id         = loc_id;
		}

		~CallIdNull_Object(void) {
			DELETE_ARR_OBJ(value_.src_lsr_addr,
				       value_.src_lsr_addr_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength                   +
				sizeof(uint32_t)              +
				(sizeof(*value_.src_lsr_addr) *
				 value_.src_lsr_addr_len)     +
				sizeof(value_.local_id);
		}

		CallId_Class * clone(void) const {
			CallIdNull_Object * obj = new CallIdNull_Object();
			if (obj) {
				obj->value_.addr_type = value_.addr_type;
				obj->value_.reserved  = value_.reserved;
				obj->value_.local_id  = value_.local_id;
				obj->value_.src_lsr_addr_len =
					value_.src_lsr_addr_len;

				if (value_.src_lsr_addr_len > 0 &&
				    value_.src_lsr_addr) {
					obj->value_.src_lsr_addr = new
						char[value_.src_lsr_addr_len];
					memcpy(obj->value_.src_lsr_addr,
					       value_.src_lsr_addr,
					       value_.src_lsr_addr_len);
				} else {
					obj->value_.src_lsr_addr = 0;
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(CallId_Class);

//CLASS ADMIN STATUS
	DEFINE_CLASS_BEGIN(AdminStatus_Class, G2RSVP_CLASS_ADMIN_STATUS);

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |R|                        Reserved                       |T|A|D|
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	      Reflect (R): 1 bit
//	         When set, indicates that the edge node SHOULD reflect the
//	         object/TLV back in the appropriate message.  This bit MUST NOT
//	         be set in state change request, i.e., Notify, messages.
//
//	      Reserved: 28 bits
//	         This field is reserved. It MUST be set to zero on transmission
//	         and MUST be ignored on receipt.  These bits SHOULD be pass
//	         through unmodified by transit nodes.
//
//	      Testing (T): 1 bit
//	         When set, indicates that the local actions related to the
//	         "testing" mode should be taken.
//
//	      Administratively down (A): 1 bit
//	         When set, indicates that the local actions related to the
//	         "administratively down" state should be taken.
//
//	      Deletion in progress (D): 1 bit
//	         When set, indicates that that the local actions related to LSP
//	         teardown should be taken.  Edge nodes may use this flag to
//	         control connection teardown.

	typedef struct {
		bit1_t	R;
		bit28_t reserved;
		bit1_t	T;
		bit1_t	A;
		bit1_t	D;
	} admin_status_t;
	DEFINE_OBJECT(AdminStatus_Object,
		      CTYPE_ADMIN_STATUS,
		      admin_status_t,
		      AdminStatus_Class) {
		OBJ_OPERATOR(AdminStatus_Object);
	public:
		AdminStatus_Object(void) {
			value_.R = 0;
			value_.T = 0;
			value_.A = 0;
			value_.D = 0;
		}

		AdminStatus_Object(bool ref, bool test, bool admin, bool del) {
			value_.R = ref   ? 1 : 0;
			value_.T = test  ? 1 : 0;
			value_.A = admin ? 1 : 0;
			value_.D = del   ? 1 : 0;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(uint32_t); // R/rsvd/TAD
		}

		AdminStatus_Class * clone(void) const {
			AdminStatus_Object * obj = new AdminStatus_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(AdminStatus_Class);

//CLASS PROTECTION
	DEFINE_CLASS_BEGIN(Protection_Class, G2RSVP_CLASS_PROTECTION);

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |S|                  Reserved                       | Link Flags|
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	     Secondary (S): 1 bit
//	         When set, indicates that the requested LSP is a secondary LSP.
//
//	      Reserved: 25 bits
//	         This field is reserved. It MUST be set to zero on transmission
//	         and MUST be ignored on receipt.  These bits SHOULD be pass
//	         through unmodified by transit nodes.
//
//	      Link Flags: 6 bits
//	         Indicates desired link protection type.  As previously
//	         mentioned, protection capabilities of a link may be advertised
//	         in routing.  A value of 0 implies that any, including no, link
//	         protection may be used.  More than one bit may be set to
//	         indicate when multiple protection types are acceptable.  When
//	         multiple bits are set and multiple protection types are
//	         available, the choice of protection type is a local (policy)
//	         decision.
//
//	         The following flags are defined:
//
//	         0x20  Enhanced
//	            Indicates that a protection scheme that is more reliable
//	            than Dedicated 1+1 should be used, e.g., 4 fiber
//                  BLSR/MS-SPRING.
//
//	         0x10  Dedicated 1+1
//	            Indicates that a dedicated link layer protection scheme,
//	            i.e., 1+1 protection, should be used to support the LSP.
//
//	         0x08  Dedicated 1:1
//	            Indicates that a dedicated link layer protection scheme,
//	            i.e., 1:1 protection, should be used to support the LSP.
//
//	         0x04  Shared
//	            Indicates that a shared link layer protection scheme, such
//	            as 1:N protection, should be used to support the LSP.
//
//	         0x02  Unprotected
//	            Indicates that the LSP should not use any link layer
//	            protection.
//
//	         0x01  Extra Traffic
//	            Indicates that the LSP should use links that are protecting
//	            other (primary) traffic.  Such LSPs may be preempted when
//	            the links carrying the (primary) traffic being protected
//	            fail.

	typedef struct {
		bit1_t	secondary;
		bit25_t reserved;
		bit6_t	link_flags;
	} protection_t;
	DEFINE_OBJECT(Protection_Object,
		      CTYPE_PROTECTION,
		      protection_t,
		      Protection_Class) {
		OBJ_OPERATOR(Protection_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(uint32_t); // S/rsvd/flags
		}

		Protection_Class * clone(void) const {
			Protection_Object * obj = new Protection_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Protection_Class);

//CLASS STYLE
	DEFINE_CLASS_BEGIN(Style_Class, G2RSVP_CLASS_STYLE);

//   +-------------+-------------+-------------+-------------+
//   |   Flags     |              Option Vector              |
//   +-------------+-------------+-------------+-------------+
//
//      Flags: 8 bits
//
//      Option Vector: 24 bits
//           A set of bit fields giving values for the reservation
//           options.  If new options are added in the future,
//           corresponding fields in the option vector will be assigned
//           from the least-significant end.  If a node does not recognize
//           a style ID, it may interpret as much of the option vector as
//           it can, ignoring new fields that may have been defined.
//
//           The option vector bits are assigned (from the left) as
//           follows:
//
//           19 bits: Reserved
//
//           2 bits: Sharing control
//                00b: Reserved
//                01b: Distinct reservations
//                10b: Shared reservations
//                11b: Reserved
//
//           3 bits: Sender selection control
//                000b: Reserved
//                001b: Wildcard
//                010b: Explicit
//                011b - 111b: Reserved
//      The low order bits of the option vector are determined by the style, as
//      follows: WF 10 001, FF 01 010, SE 10 010

#define G2RSVPTE_SHARING_CTRL_DISTINCT_RESV             0x1
#define G2RSVPTE_SHARING_CTRL_SHARED_RESV               0x2
#define G2RSVPTE_SENDER_SELECTION_CTRL_WILDCARD         0x1
#define G2RSVPTE_SENDER_SELECTION_CTRL_EXPLICIT         0x2

	typedef struct {
		bit8_t  flags;
		//bit24_t option_vect;
		bit2_t  sharing;
		bit3_t  sender_selection;
	} style_t;
	DEFINE_OBJECT(Style_Object,
		      CTYPE_STYLE,
		      style_t,
		      Style_Class) {
		OBJ_OPERATOR(Style_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		bool isFF(void) {
			if ((value_.sharing          &
			     G2RSVPTE_SHARING_CTRL_DISTINCT_RESV)     &&
			    (value_.sender_selection &
			     G2RSVPTE_SENDER_SELECTION_CTRL_EXPLICIT)) {
				return true;
			}
			return false;
		}

		void setFFstyle(void) {
			value_.sharing = G2RSVPTE_SHARING_CTRL_DISTINCT_RESV;
			value_.sender_selection =
				G2RSVPTE_SENDER_SELECTION_CTRL_EXPLICIT;
		}

		bit8_t flags(void)         { return value_.flags;  }
		void   flags(bit8_t flags) { value_.flags = flags; }

		uint16_t length(void) const {
			return objHdrLength + sizeof(uint32_t);
		}

		Style_Class * clone(void) const {
			Style_Object * obj = new Style_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Style_Class);

//CLASS FILTER SPEC
	DEFINE_CLASS_BEGIN(FilterSpec_Class, G2RSVP_CLASS_FILTER_SPEC);

//         +-------------+-------------+-------------+-------------+
//         |                                                       |
//         +                                                       +
//         |                                                       |
//         +               IPv6 SrcAddress (16 bytes)              +
//         |                                                       |
//         +                                                       +
//         |                                                       |
//         +-------------+-------------+-------------+-------------+
//         |   ///////   |         Flow Label (24 bits)            |
//         +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv6_t  ipv6_addr;
		bit8_t	reserved;
		bit24_t	flow_label;
	} ipv6_flow_label_t;
	DEFINE_OBJECT(IPV6FlowFilterSpec_Object,
		      CTYPE_FLOW_LABEL_FILTER_SPEC,
		      ipv6_flow_label_t,
		      FilterSpec_Class) {
		OBJ_OPERATOR(IPV6FlowFilterSpec_Object);
	public:
		IPV6FlowFilterSpec_Object(void) { }
		IPV6FlowFilterSpec_Object(ipv6_flow_label_t data) {
			value_ = data;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength              +
				sizeof(value_.ipv6_addr) +
				sizeof(value_.flow_label);  // flowLabel/resvd
		}

		FilterSpec_Class * clone(void) const {
			IPV6FlowFilterSpec_Object * obj;
			obj = new IPV6FlowFilterSpec_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//         +-------------+-------------+-------------+-------------+
//         |               IPv4 SrcAddress (4 bytes)               |
//         +-------------+-------------+-------------+-------------+
//         |    //////   |    //////   |          SrcPort          |
//         +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv4_t	ipv4_addr;
		bit16_t reserved;
		bit16_t	src_port;
	} ipv4_filter_spec_t;
	DEFINE_OBJECT(IPV4FilterSpec_Object,
		      CTYPE_IPV4_FILTER_SPEC,
		      ipv4_filter_spec_t,
		      FilterSpec_Class) {
		OBJ_OPERATOR(IPV4FilterSpec_Object);
	public:
		IPV4FilterSpec_Object(void) { }
		IPV4FilterSpec_Object(ipv4_filter_spec_t data) {
			value_ = data;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv4_filter_spec_t);
		}

		FilterSpec_Class * clone(void) const {
			IPV4FilterSpec_Object * obj;
			obj = new IPV4FilterSpec_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//         +-------------+-------------+-------------+-------------+
//         |                                                       |
//         +                                                       +
//         |                                                       |
//         +               IPv6 SrcAddress (16 bytes)              +
//         |                                                       |
//         +                                                       +
//         |                                                       |
//         +-------------+-------------+-------------+-------------+
//         |    //////   |    //////   |          SrcPort          |
//         +-------------+-------------+-------------+-------------+

	typedef struct {
		ipv6_t	ipv6_addr;
		bit16_t reserved;
		bit16_t src_port;
	} ipv6_filter_spec_t;
	DEFINE_OBJECT(IPV6FilterSpec_Object,
		      CTYPE_IPV6_FILTER_SPEC,
		      ipv6_filter_spec_t,
		      FilterSpec_Class) {
		OBJ_OPERATOR(IPV6FilterSpec_Object);
	public:
		IPV6FilterSpec_Object(void) { }
		IPV6FilterSpec_Object(ipv6_filter_spec_t data) {
			value_ = data;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv6_filter_spec_t);
		}

		FilterSpec_Class * clone(void) const {
			IPV6FilterSpec_Object * obj;
			obj = new IPV6FilterSpec_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                   IPv4 tunnel sender address                  |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |  MUST be zero                 |            LSP ID             |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	// The format of the LSP_TUNNEL_IPv4 FILTER_SPEC object is identical to
	// the LSP_TUNNEL_IPv4 SENDER_TEMPLATE object.
	typedef struct {
		ipv4_t  ipv4_addr;
		bit16_t reserved;
		bit16_t lsp_id;
	} ipv4_lsp_tunnel_t;
	DEFINE_OBJECT(IPV4LSPFilterSpec_Object,
		      CTYPE_LSP_IPV4_FILTER_SPEC,
		      ipv4_lsp_tunnel_t,
		      FilterSpec_Class) {
		OBJ_OPERATOR(IPV4LSPFilterSpec_Object);
	public:
		IPV4LSPFilterSpec_Object(void) { }
		IPV4LSPFilterSpec_Object(ipv4_lsp_tunnel_t data) {
			value_ = data;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv4_lsp_tunnel_t);
		}

		FilterSpec_Class * clone(void) const {
			IPV4LSPFilterSpec_Object * obj;
			obj = new IPV4LSPFilterSpec_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   +                                                               +
//	   |                   IPv6 tunnel sender address                  |
//	   +                                                               +
//	   |                            (16 bytes)                         |
//	   +                                                               +
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |  MUST be zero                 |            LSP ID             |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv6_t  ipv6_addr;
		bit16_t reserved;
		bit16_t lsp_id;
	} ipv6_lsp_tunnel_t;
	DEFINE_OBJECT(IPV6LSPFilterSpec_Object,
		      CTYPE_LSP_IPV6_FILTER_SPEC,
		      ipv6_lsp_tunnel_t,
		      FilterSpec_Class) {
		OBJ_OPERATOR(IPV6LSPFilterSpec_Object);
	public:
		IPV6LSPFilterSpec_Object(void) { }
		IPV6LSPFilterSpec_Object(ipv6_lsp_tunnel_t data) {
			value_ = data;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv6_lsp_tunnel_t);
		}

		FilterSpec_Class * clone(void) const {
			IPV6LSPFilterSpec_Object * obj;
			obj = new IPV6LSPFilterSpec_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(FilterSpec_Class);

//CLASS POLICY DATA
	DEFINE_CLASS_BEGIN(Policy_Class, G2RSVP_CLASS_POLICY);

//    Policy Element Information
//     +-------------+-------------+-------------+-------------+
//     |  Length                   |   P-Type                  |
//     +---------------------------+---------------------------+
//     |                                                       |
//     // Policy information  (Opaque to RSVP)                //
//     |                                                       |
//     +-------------------------------------------------------+

// We consider length = header length + Policy Information length (in BYTES!!!)

	class Policy_Object;
	class PolicyInfo {
		OBJ_OPERATOR(PolicyInfo);

		friend MemoryCursor & operator >> (MemoryCursor &,
						   Policy_Object &);
	public:
		PolicyInfo(uint16_t     obj_type,
			   uint16_t     data_len,
			   const char * data);
		~PolicyInfo(void);

		bool         isConsistent(void) const;
		uint16_t     length(void) const;
		PolicyInfo * clone(void) const;

	private:

	protected:
		PolicyInfo(void);
	public:
		static const uint16_t pInfoHdrLength = 4; // Type & Length
		uint16_t              type;
		uint16_t              opaque_data_len; // Length of PolInf only
		char *                opaque_data;
	};

//      +---------------------------+-------------+-------------+
//      |  Data Offset              | 0 (reserved)              |
//      +---------------------------+-------------+-------------+
//      |                                                       |
//      // Option List                                         //
//      |                                                       |
//      +-------------------------------------------------------+
//      |                                                       |
//      // Policy Element List                                 //
//      |                                                       |
//      +-------------------------------------------------------+
//
//    Data Offset: 16 bits
//      The offset in bytes of the data portion (first byte of the obj header)
//
//    Reserved: 16 bits
//      Always 0.
//
//    Option List: Variable length
//      The list of options and their usage is defined in RFC 2750.
//
//    Policy Element List: Variable length
//      The contents of policy elements is opaque to RSVP.

	typedef struct {
		bit16_t                       data_offset_;
		// std::list<FilterSpec_Class *> filters;
		Scope_Class *                 scope;
		Rsvp_Hop_Class *              orig_rsvp_hop;
		Rsvp_Hop_Class *              dest_rsvp_hop;
		Integrity_Class *             integrity;
		TimeValues_Class *            time_values;
		std::list<PolicyInfo *>       policy_elems;
	} policy_t;
	DEFINE_OBJECT(Policy_Object,
		      CTYPE_POLICY,
		      policy_t,
		      Policy_Class) {
		OBJ_OPERATOR(Policy_Object);
	public:
		Policy_Object(void) {
			value_.scope         = 0;
			value_.orig_rsvp_hop = 0;
			value_.dest_rsvp_hop = 0;
			value_.integrity     = 0;
			value_.time_values   = 0;
		}

		~Policy_Object(void) {
			DELETE_OBJ(value_.scope);
			DELETE_OBJ(value_.orig_rsvp_hop);
			DELETE_OBJ(value_.dest_rsvp_hop);
			DELETE_OBJ(value_.integrity);
			DELETE_OBJ(value_.time_values);

			std::list<PolicyInfo *>::const_iterator it;
			for (it = value_.policy_elems.begin();
			     it != value_.policy_elems.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.policy_elems.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t data_offset(void) const {
			return sizeof(uint32_t) + // data len && reserved
				(value_.scope ?
				 value_.scope->length() : 0) +
				(value_.orig_rsvp_hop ?
				 value_.orig_rsvp_hop->length() : 0) +
				(value_.dest_rsvp_hop ?
				 value_.dest_rsvp_hop->length() : 0) +
				(value_.integrity ?
				 value_.integrity->length() : 0) +
				(value_.time_values ?
				 value_.time_values->length() : 0);
		}

		uint16_t length(void) const {

			uint16_t policy_info_len = 0;

			std::list<PolicyInfo *>::const_iterator it;
			for (it = value_.policy_elems.begin();
			     it != value_.policy_elems.end();
			     it++) {
				if (*it) {
					policy_info_len += (*it)->length();
				}
			}

			return objHdrLength + data_offset() + policy_info_len;
		}

		Policy_Class * clone(void) const {
			Policy_Object * obj = new Policy_Object();
			if (obj) {
				obj->value_.policy_elems.clear();

				std::list<PolicyInfo *>::const_iterator it;
				for (it = value_.policy_elems.begin();
				     it != value_.policy_elems.end();
				     it++) {
					if (*it) {
						obj->value_.policy_elems.
						     push_back((*it)->clone());
					}
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Policy_Class);

//CLASS SENDER TEMPLATE
	DEFINE_CLASS_BEGIN(SenderTemplate_Class, G2RSVP_CLASS_SENDER_TEMPLATE);

	DEFINE_OBJECT(IPV6FlowSenderTemplate_Object,
		      CTYPE_FLOW_LABEL_SENDER_TEMPLATE,
		      ipv6_flow_label_t,
		      SenderTemplate_Class) {
		OBJ_OPERATOR(IPV6FlowSenderTemplate_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength              +
				sizeof(value_.ipv6_addr) +
				sizeof(value_.flow_label);  // flowLabel/resvd
		}

		SenderTemplate_Class * clone(void) const {
			IPV6FlowSenderTemplate_Object * obj;
			obj = new IPV6FlowSenderTemplate_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_OBJECT(IPV4SenderTemplate_Object,
		      CTYPE_IPV4_SENDER_TEMPLATE,
		      ipv4_filter_spec_t,
		      SenderTemplate_Class) {
		OBJ_OPERATOR(IPV4SenderTemplate_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv4_filter_spec_t);
		}

		SenderTemplate_Class * clone(void) const {
			IPV4SenderTemplate_Object * obj;
			obj = new IPV4SenderTemplate_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_OBJECT(IPV6SenderTemplate_Object,
		      CTYPE_IPV6_SENDER_TEMPLATE,
		      ipv6_filter_spec_t,
		      SenderTemplate_Class) {
		OBJ_OPERATOR(IPV6SenderTemplate_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv6_filter_spec_t);
		}

		SenderTemplate_Class * clone(void) const {
			IPV6SenderTemplate_Object * obj;
			obj = new IPV6SenderTemplate_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_OBJECT(IPV4LSPSenderTemplate_Object,
		      CTYPE_LSP_IPV4_SENDER_TEMPLATE,
		      ipv4_lsp_tunnel_t,
		      SenderTemplate_Class) {
		OBJ_OPERATOR(IPV4LSPSenderTemplate_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv4_lsp_tunnel_t);
		}

		SenderTemplate_Class * clone(void) const {
			IPV4LSPSenderTemplate_Object * obj;
			obj = new IPV4LSPSenderTemplate_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_OBJECT(IPV6LSPSenderTemplate_Object,
		      CTYPE_LSP_IPV6_SENDER_TEMPLATE,
		      ipv6_lsp_tunnel_t,
		      SenderTemplate_Class) {
		OBJ_OPERATOR(IPV6LSPSenderTemplate_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv6_lsp_tunnel_t);
		}

		SenderTemplate_Class * clone(void) const {
			IPV6LSPSenderTemplate_Object * obj;
			obj = new IPV6LSPSenderTemplate_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(SenderTemplate_Class);

//CLASS NOTIFY_REQUEST
	DEFINE_CLASS_BEGIN(NotifyRequest_Class, G2RSVP_CLASS_NOTIFY_REQUEST);

	typedef ipv4_addr_t ipv4_notify_request_t;
	DEFINE_OBJECT(IPV4NotifyRequest_Object,
		      CTYPE_IPV4_NOTIFY_REQUEST,
		      ipv4_notify_request_t,
		      NotifyRequest_Class) {
		OBJ_OPERATOR(IPV4NotifyRequest_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv4_notify_request_t);
		}

		NotifyRequest_Class * clone(void) const {
			IPV4NotifyRequest_Object * obj;
			obj = new IPV4NotifyRequest_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	typedef ipv6_addr_t ipv6_notify_request_t;
	DEFINE_OBJECT(IPV6NotifyRequest_Object,
		      CTYPE_IPV6_NOTIFY_REQUEST,
		      ipv6_notify_request_t,
		      NotifyRequest_Class) {
		OBJ_OPERATOR(IPV6NotifyRequest_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(ipv6_notify_request_t);
		}

		NotifyRequest_Class * clone(void) const {
			IPV6NotifyRequest_Object * obj;
			obj = new IPV6NotifyRequest_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(NotifyRequest_Class);

//CLASS LABEL SET (RFC 3471)
	DEFINE_CLASS_BEGIN(LabelSet_Class, G2RSVP_CLASS_LABEL_SET);

//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |    Action     |      Reserved     |        Label Type         |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                          Subchannel 1                         |
//	   |                              ...                              |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   :                               :                               :
//	   :                               :                               :
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                          Subchannel N                         |
//	   |                              ...                              |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	  Action: 8 bits
//
//	      0 - Inclusive List
//
//	         Indicates that the object/TLV contains one or more subchannel
//	         elements that are included in the Label Set.
//
//	      1 - Exclusive List
//
//	         Indicates that the object/TLV contains one or more subchannel
//	         elements that are excluded from the Label Set.
//
//	      2 - Inclusive Range
//
//	         Indicates that the object/TLV contains a range of labels.  The
//	         object/TLV contains two subchannel elements. The first element
//	         indicates the start of the range. The second element indicates
//	         the end of the range.  A value of zero indicates that there is
//	         no bound on the corresponding portion of the range.
//
//	      3 - Exclusive Range
//
//	         Indicates that the object/TLV contains a range of labels that
//	         are excluded from the Label Set.  The object/TLV contains two
//	         subchannel elements.  The first element indicates the start of
//	         the range.  The second element indicates the end of the range.
//	         A value of zero indicates that there is no bound on the
//	         corresponding portion of the range.
//
//	   Reserved: 10 bits
//
//	      This field is reserved. It MUST be set to zero on transmission
//	      and MUST be ignored on receipt.
//
//	   Label Type: 14 bits
//
//	      Indicates the type and format of the labels carried in the
//	      object/TLV.  Values are signaling protocol specific.
//
//	   Subchannel:
//
//	      The subchannel represents the label (wavelength, fiber ... )which
//	      is eligible for allocation.  This field has the same format as
//	      described for labels under section 3.2.
//
//	      Note that subchannel to local channel identifiers (e.g.,
//	      wavelength) mappings are a local matter.
//
	typedef struct {
		bit8_t	                 action;
		bit10_t	                 reserved;    // useless ...
		bit14_t                  label_type;
		std::list<Label_Class *> labels;
	} label_set_t, acceptable_label_set_t;

	//////////////////////////////////////////////////////////////////////
	// NB: if LabelSet obj has the abow format, and labels are GenLabel //
	//     obj (label_type == 2) -> where is the length of each label?  //
	//                                                                  //
	// IMPORTANT:   temp assumption => each label is a 4-byte label !!! //
	//////////////////////////////////////////////////////////////////////

	DEFINE_OBJECT(LabelSet_Object,
		      CTYPE_LABEL_SET,
		      label_set_t,
		      LabelSet_Class) {
		OBJ_OPERATOR(LabelSet_Object);
	public:
		~LabelSet_Object(void) {
			std::list<Label_Class *>::const_iterator it;
			for (it = value_.labels.begin();
			     it != value_.labels.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.labels.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<Label_Class *>::const_iterator it;
			for (it = value_.labels.begin();
			     it != value_.labels.end();
			     it++) {
				if (*it) {
					len += (*it)->length() - objHdrLength;
				}
			}

			return objHdrLength + len +
				sizeof(uint32_t);  // act_res_lblType length
		}

		LabelSet_Class * clone(void) const {
			LabelSet_Object * obj = new LabelSet_Object();

			if (obj) {
				obj->value_.action     = value_.action;
				obj->value_.reserved   = value_.reserved;
				obj->value_.label_type = value_.label_type;

				obj->value_.labels.clear();

				std::list<Label_Class *>::const_iterator it;
				for (it = value_.labels.begin();
				     it != value_.labels.end();
				     it++) {
					if (*it) {
						obj->value_.labels.push_back
							((*it)->clone());
					}
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(LabelSet_Class);

//CLASS ACCEPTABLE LABEL SET
	DEFINE_CLASS_BEGIN(AcceptableLabelSet_Class,
			   G2RSVP_CLASS_ACCEPTABLE_LABEL_SET);

	DEFINE_OBJECT(AcceptableLabelSet_Object,
		      CTYPE_ACCEPTABLE_LABEL_SET,
		      acceptable_label_set_t,
		      AcceptableLabelSet_Class) {
		OBJ_OPERATOR(AcceptableLabelSet_Object);
	public:
		~AcceptableLabelSet_Object(void) {
			std::list<Label_Class *>::const_iterator it;
			for (it = value_.labels.begin();
			     it != value_.labels.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.labels.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<Label_Class *>::const_iterator it;
			for (it = value_.labels.begin();
			     it != value_.labels.end();
			     it++) {
				if (*it) {
					len += (*it)->length() - objHdrLength;
				}
			}

			return objHdrLength + len +
				sizeof(uint32_t);  // act_res_lblType length
		}

		AcceptableLabelSet_Class * clone(void) const {
			AcceptableLabelSet_Object * obj;
			obj = new AcceptableLabelSet_Object();

			if (obj) {
				obj->value_.action     = value_.action;
				obj->value_.reserved   = value_.reserved;
				obj->value_.label_type = value_.label_type;

				obj->value_.labels.clear();

				std::list<Label_Class *>::const_iterator it;
				for (it = value_.labels.begin();
				     it != value_.labels.end();
				     it++) {
					if (*it) {
						obj->value_.labels.push_back
							((*it)->clone());
					}
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(AcceptableLabelSet_Class);

//CLASS SUGGESTED LABEL
	DEFINE_CLASS_BEGIN(SuggestedLabel_Class, G2RSVP_CLASS_SUGGESTED_LABEL);

	DEFINE_OBJECT(SuggestedLabel_Object,
		      CTYPE_SUGGESTED_GENERALISED_LABEL,
		      gen_label_t,
		      SuggestedLabel_Class)
	GLABEL_OBJ_BODY(SuggestedLabel_Object, SuggestedLabel_Class);

	DEFINE_CLASS_END(SuggestedLabel_Class);

//CLASS UPSTREAM LABEL
	DEFINE_CLASS_BEGIN(UpstreamLabel_Class, G2RSVP_CLASS_UPSTREAM_LABEL);

	DEFINE_OBJECT(UpstreamLabel_Object,
		      CTYPE_UPSTREAM_GENERALISED_LABEL,
		      gen_label_t,
		      UpstreamLabel_Class)
	GLABEL_OBJ_BODY(UpstreamLabel_Object, UpstreamLabel_Class);

	DEFINE_CLASS_END(UpstreamLabel_Class);

//CLASS RECOVERY LABEL
	DEFINE_CLASS_BEGIN(RecoveryLabel_Class, G2RSVP_CLASS_RECOVERY_LABEL);

	DEFINE_OBJECT(RecoveryLabel_Object,
		      CTYPE_RECOVERY_GENERALISED_LABEL,
		      gen_label_t,
		      RecoveryLabel_Class)
	GLABEL_OBJ_BODY(RecoveryLabel_Object, RecoveryLabel_Class);

	DEFINE_CLASS_END(RecoveryLabel_Class);

//CLASS ADSPEC
	DEFINE_CLASS_BEGIN(ADSpec_Class, G2RSVP_CLASS_ADSPEC);

//       31           24 23            16 15            8 7             0
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       | 0 (a) |      reserved         |  Msg length - 1 (b)           |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                                                               |
//       |    Default General Parameters fragment (Service 1)  (c)       |
//       |    (Always Present)                                           |
//       |                                                               |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                                                               |
//       |    Guaranteed Service Fragment (Service 2)    (d)             |
//       |    (Present if application might use Guaranteed Service)      |
//       |                                                               |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                                                               |
//       |    Controlled-Load Service Fragment (Service 5)  (e)          |
//       |    (Present if application might use Controlled-Load Service) |
//       |                                                               |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//     (a) - Message format version number (0)
//     (b) - Overall message length not including header word
//     (c, d, e) - Data fragments

	typedef struct {
		bit4_t  msg_version;
		bit12_t reserved;
		bit16_t msg_len;
		//default general params always present
		bit8_t  service_number;    // = 1 for the default parameters
		bit1_t  gl_break_bitx;
		bit7_t  resverved;
		bit16_t gen_params_length;
		bit8_t  param_id4;         // = 4
		bit8_t  param4_flags;
		bit16_t param4_length;
		bit32_t is_hop_count;
		bit8_t  param6_id;         // = 6
		bit8_t  param6_flags;
		bit16_t param6_length;
		bit32_t path_bw_estimate;  //32-bit IEEE floating point number
		bit8_t  param_id8;         // = 8
		bit8_t  param8_flags;
		bit16_t param8_length;
		bit32_t minimum_path_latency;
		bit8_t  param_id10;        // = 10
		bit8_t  param10_flags;
		bit16_t param10_length;
		bit32_t composed_mtu;
		// service_params can be used if guaranteed service or
		// controlled load service is being requested
		int_serv_flow_spec_t * service_params;
	} adspec_t;
	DEFINE_OBJECT(ADSpec_Object,
		      CTYPE_ADSPEC,
		      adspec_t,
		      ADSpec_Class) {
		bool isConsistent(void) const {
			return true;
		}
	};

	DEFINE_CLASS_END(ADSpec_Class);

//SENDER TSPEC
	DEFINE_CLASS_BEGIN(SenderTSpec_Class, G2RSVP_CLASS_SENDER_TSPEC);

//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   1   | 0 (a) |    reserved           |             7 (b)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   2   |    1  (c)     |0| reserved    |             6 (d)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   3   |   127 (e)     |    0 (f)      |             5 (g)             |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   4   |  Token Bucket Rate [r] (32-bit IEEE floating point number)    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   5   |  Token Bucket Size [b] (32-bit IEEE floating point number)    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   6   |  Peak Data Rate [p] (32-bit IEEE floating point number)       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   7   |  Minimum Policed Unit [m] (32-bit integer)                    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   8   |  Maximum Packet Size [M]  (32-bit integer)                    |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//
//     (a) - Message format version number (0)
//     (b) - Overall length (7 words not including header)
//     (c) - Service header, service number 1 (default/global information)
//     (d) - Length of service 1 data, 6 words not including header
//     (e) - Parameter ID, parameter 127 (Token_Bucket_TSpec)
//     (f) - Parameter 127 flags (none set)
//     (g) - Parameter 127 length, 5 words not including header
//

	DEFINE_OBJECT(SenderTSpec_Object,
		      CTYPE_SENDER_TSPEC,
		      sender_spec_t,
		      SenderTSpec_Class) {
		OBJ_OPERATOR(SenderTSpec_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength          +
				sizeof(uint32_t)     + // msg_vers/resvd/length
				value_.length * sizeof(uint32_t);
		}

		SenderTSpec_Class * clone(void) const {
			SenderTSpec_Object * obj = new SenderTSpec_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(SenderTSpec_Class);

//CLASS RECORDED ROUTE
	DEFINE_CLASS_BEGIN(RecordedRoute_Class, G2RSVP_CLASS_RECORDED_ROUTE);

	/////////////// RRO subobjects definition /////////////

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |      Type     |     Length    | IPv4 address (4 bytes)        |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv4 address (continued)      | Prefix Length |      Flags    |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	      IPv4 address
//
//	         A 32-bit unicast, host address.  Any network-reachable
//	         interface address is allowed here.  Illegal addresses, such as
//	         certain loopback addresses, SHOULD NOT be used.
//
//	      Prefix length
//
//	         32
//
//	      Flags
//
//	         0x01  Local protection available
//
//	               Indicates that the link downstream of this node is
//	               protected via a local repair mechanism.  This flag can
//	               only be set if the Local protection flag was set in the
//	               SESSION_ATTRIBUTE object of the corresponding Path
//	               message.
//
//	         0x02  Local protection in use
//
//	               Indicates that a local repair mechanism is in use to
//	               maintain this tunnel (usually in the face of an outage
//	               of the link it was previously routed over).
//

	//SubObject:
	typedef struct {
		ipv4_t ipv4_addr;
		bit8_t preflen;
		bit8_t flags;
	} ipv4_subobj_t;
	DEFINE_RROSUBOBJ(IPv4Addr_SubObj,
			 ipv4_subobj_t,
			 IPV4_ADDR_SUBOBJ) {
		OBJ_OPERATOR(IPv4Addr_SubObj);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return rroSubObjHdrLength           +
				sizeof(content_.ipv4_addr)  +
				sizeof(content_.preflen)    +
				sizeof(content_.flags);
		}

		RroSubObject * clone(void) const {
			IPv4Addr_SubObj * obj = new IPv4Addr_SubObj();
			if (obj) {
				obj->content_ = content_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |      Type     |     Length    | IPv6 address (16 bytes)       |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)                                      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)                                      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)                                      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   | IPv6 address (continued)      | Prefix Length |      Flags    |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		ipv6_t ipv6_addr;
		bit8_t preflen;
		bit8_t flags;
	} ipv6_subobj_t;
	DEFINE_RROSUBOBJ(IPv6Addr_SubObj,
			 ipv6_subobj_t,
			 IPV6_ADDR_SUBOBJ) {
		OBJ_OPERATOR(IPv6Addr_SubObj);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return rroSubObjHdrLength           +
				sizeof(content_.ipv6_addr)  +
				sizeof(content_.preflen)    +
				sizeof(content_.flags);
		}

		RroSubObject * clone(void) const {
			IPv6Addr_SubObj * obj = new IPv6Addr_SubObj();
			if (obj) {
				obj->content_ = content_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |      Type     |     Length    |     Flags     | Reserved (MBZ)|
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                           Router ID                           |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                     Interface ID (32 bits)                    |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	   Flags:
//
//	    0x01  Local protection available
//
//	      Indicates that the link downstream of this node is protected via
//	      a local repair mechanism.  This flag can only be set if the Local
//	      protection flag was set in the SESSION_ATTRIBUTE object of the
//	      corresponding Path message.
//
//	    0x02  Local protection in use
//
//	      Indicates that a local repair mechanism is in use to maintain
//	      this tunnel (usually in the face of an outage of the link it was
//	      previously routed over).

	typedef struct {
		bit8_t  flags;
		bit8_t  reserved;
		bit32_t router_id;
		bit32_t interface_id;
	} unnum_subobj_t;
	DEFINE_RROSUBOBJ(Unnumbered_SubObj,
			 unnum_subobj_t,
			 UNNUM_LINK_SUBOBJ) {
		OBJ_OPERATOR(Unnumbered_SubObj);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return rroSubObjHdrLength             +
				sizeof(content_.flags)        +
				sizeof(content_.reserved)     +
				sizeof(content_.router_id)    +
				sizeof(content_.interface_id);
		}

		RroSubObject * clone(void) const {
			Unnumbered_SubObj * obj = new Unnumbered_SubObj();
			if (obj) {
				obj->content_ = content_;
			}
			return obj;
		}
	};

//	    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |     Type      |     Length    |    Flags      |   C-Type      |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |       Contents of Label Object                                |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	      Flags
//
//	         0x01 = Global label
//	           This flag indicates that the label will be understood
//	           if received on any interface.
//
//	      C-Type
//
//	         The C-Type of the included Label Object. Copied from the Label
//	         Object.
//
//	      Contents of Label Object
//
//	         The contents of the Label Object. Copied from the Label Object

	typedef struct {
		bit8_t        flags;
		bit8_t        c_type;
		Label_Class * label;
	} label_subobj_t;
	DEFINE_RROSUBOBJ(Label_SubObj,
			 label_subobj_t,
			 LABEL_SUBOBJ) {
		OBJ_OPERATOR(Label_SubObj);
	public:
		Label_SubObj(void) {
			content_.label = 0;
		}

		~Label_SubObj(void) {
			DELETE_OBJ(content_.label);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint8_t length(void) const {
			return rroSubObjHdrLength        +
				sizeof(content_.flags)   +
				sizeof(content_.c_type)  +
				(content_.label ?
				(content_.label->length() -
				 objHeader::objHdrLength) : 0);
		}

		RroSubObject * clone(void) const {
			Label_SubObj * obj = new Label_SubObj();
			if (obj) {
				obj->content_.flags  = content_.flags;
				obj->content_.c_type = content_.c_type;

				if (content_.label) {
					obj->content_.label =
						content_.label->clone();
				}
			}
			return obj;
		}
	};
	/////////////////////////////////////

// RRO format (RFC 3209)
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   //                        (Subobjects)                          //
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		std::list<RroSubObject *> subObjs;
	} rro_t;
	DEFINE_OBJECT(RecordedRoute_Object,
		      CTYPE_RECORDED_ROUTE,
		      rro_t,
		      RecordedRoute_Class) {
		OBJ_OPERATOR(RecordedRoute_Object);
	public:
		~RecordedRoute_Object(void) {
			std::list<RroSubObject *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.subObjs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<RroSubObject *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength + len;
		}

		RecordedRoute_Class * clone(void) const {
			RecordedRoute_Object* obj = new RecordedRoute_Object();
			if (obj) {
				obj->value_.subObjs.clear();

				std::list<RroSubObject *>::const_iterator it;
				for (it = value_.subObjs.begin();
				     it != value_.subObjs.end();
				     it++) {
					if (*it) {
						obj->value_.subObjs.push_back
							((*it)->clone());
					}
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(RecordedRoute_Class);

//CLASS RESV CONFIRM CLASS
	DEFINE_CLASS_BEGIN(ResvConfirm_Class, G2RSVP_CLASS_RESV_CONFIRM);

//      +-------------+-------------+-------------+-------------+
//      |            IPv4 Receiver Address (4 bytes)            |
//      +-------------+-------------+-------------+-------------+

	typedef ipv4_addr_t ipv4_resv_conf_t;
	DEFINE_OBJECT(ResvConfirmIPv4_Object,
		      CTYPE_RESVCONF_IPV4,
		      ipv4_resv_conf_t,
		      ResvConfirm_Class) {
		OBJ_OPERATOR(ResvConfirmIPv4_Object);
	public:
		ResvConfirmIPv4_Object(void) { }
		ResvConfirmIPv4_Object(ipv4_t ipv4_addr) {
			value_.ipv4_addr = ipv4_addr;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength +
				sizeof(ipv4_resv_conf_t);
		}

		ResvConfirm_Class * clone(void) const {
			ResvConfirmIPv4_Object * obj;
			obj = new ResvConfirmIPv4_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};



//      +-------------+-------------+-------------+-------------+
//      |                                                       |
//      +                                                       +
//      |                                                       |
//      +            IPv6 Receiver Address (16 bytes)           +
//      |                                                       |
//      +                                                       +
//      |                                                       |
//      +-------------+-------------+-------------+-------------+

	typedef ipv6_addr_t ipv6_resv_conf_t;
	DEFINE_OBJECT(ResvConfirmIPv6_Object,
		      CTYPE_RESVCONF_IPV6,
		      ipv6_resv_conf_t,
		      ResvConfirm_Class) {
		OBJ_OPERATOR(ResvConfirmIPv6_Object);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength +
				sizeof(ipv6_resv_conf_t);
		}

		ResvConfirm_Class * clone(void) const {
			ResvConfirmIPv6_Object * obj;
			obj = new ResvConfirmIPv6_Object();

			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(ResvConfirm_Class);

////////FLOW DESCRIPTION LISTS//////////////////////////

	// FF FLOW DESC
	typedef struct ff_flow_item_desc ff_flow_item_desc_t;
	typedef struct ff_flow_item_desc {
		FlowSpec_Class *      flow_spec;   // mandatory only for 1 elem
		FilterSpec_Class *    filter_spec;
		Label_Class *         label;
		RecordedRoute_Class * rro;         // optional

		OBJ_OPERATOR(ff_flow_item_desc);

		ff_flow_item_desc(void) {
			flow_spec   = 0;
			filter_spec = 0;
			label       = 0;
			rro         = 0;
		}

		~ff_flow_item_desc(void) {
			DELETE_OBJ(flow_spec);
			DELETE_OBJ(filter_spec);
			DELETE_OBJ(label);
			DELETE_OBJ(rro);
		}

		uint16_t length(void) const
		{
			uint16_t len = 0;

			ADD_LENGTH(flow_spec, len);
			ADD_LENGTH(filter_spec, len);
			ADD_LENGTH(label, len);
			ADD_LENGTH(rro, len);

			return len;
		}

		ff_flow_item_desc * clone(void) const
		{
			ff_flow_item_desc * item = new ff_flow_item_desc();
			assert(item);

			CLONE_OBJ_INTO_MSG(flow_spec, item);
			CLONE_OBJ_INTO_MSG(filter_spec, item);
			CLONE_OBJ_INTO_MSG(label, item);
			CLONE_OBJ_INTO_MSG(rro, item);

			return item;
		}

	} ff_flow_item_desc_t;

	typedef struct FF_FlowDesc FF_FlowDesc_t;
	typedef struct FF_FlowDesc {
		std::list<ff_flow_item_desc_t *> filters;

		OBJ_OPERATOR(FF_FlowDesc);

		~FF_FlowDesc(void) {
			std::list<ff_flow_item_desc_t *>::const_iterator it;
			for (it = filters.begin(); it != filters.end(); it++) {
				DELETE_OBJ(*it);
			}
			filters.clear();
		}

		uint16_t length(void) const
		{
			uint16_t len = 0;
			std::list<ff_flow_item_desc_t *>::const_iterator it;

			for (it = filters.begin(); it != filters.end(); it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return len;
		}

		FF_FlowDesc * clone(void) const
		{
			FF_FlowDesc * ff = new FF_FlowDesc();
			assert(ff);

			std::list<ff_flow_item_desc_t *>::const_iterator it;
			for (it = filters.begin(); it != filters.end(); it++) {
				if (*it) {
					ff_flow_item_desc_t * tmp;
					tmp = (*it)->clone();
					assert(tmp);
					ff->filters.push_back(tmp);
				}
			}

			return ff;
		}
	} FF_FlowDesc_t;

	// SE FLOW DESC - TODO!!!!!!!!!!!!!
	typedef struct {
		FilterSpec_Class *    filter_spec;
		Label_Class *         label;
		RecordedRoute_Class * rro;         // optional
	} se_flow_item_desc_t;
	typedef struct {
		FlowSpec_Class *                 flow_spec;
		std::list<se_flow_item_desc_t *> filters;
	} SE_FlowDesc_t;

	//TMP ??? to be discussed before parser/formatter implementation
	class Flow_Desc_List;
	class Flow_Desc_List {
		OBJ_OPERATOR(Flow_Desc_List);

		FF_FlowDesc_t * ff_desc;
		SE_FlowDesc_t * se_desc;

	public:
		Flow_Desc_List(FF_FlowDesc_t *);
		Flow_Desc_List(SE_FlowDesc_t *);
		~Flow_Desc_List(void);

		Flow_Desc_List * clone(void) const;
		bool             isConsistent(void) const;
		uint16_t         length(void) const;

		bool setLabelObj(Label_Class * label);
		bool setRROobj(RecordedRoute_Class * rro);
		bool setFlowSpecObj(FlowSpec_Class * flow_spec);
		bool setFilterSpecObj(FilterSpec_Class * filter_spec);

		Label_Class *         getLabelObj(void);
		RecordedRoute_Class * getRROobj(void);
		FlowSpec_Class *      getFlowSpecObj(void);
		FilterSpec_Class *    getFilterSpecObj(void);

	private:
		Flow_Desc_List(void);
	};
////////////////////////////////////////////////////////


//CLASS HELLO
	DEFINE_CLASS_BEGIN(Hello_Class, G2RSVP_CLASS_HELLO); //22

	typedef struct {
		bit32_t src_instance;
		bit32_t dst_instance;
	} hello_src_dst_t;

	DEFINE_OBJECT(Hello_Object,
		      CTYPE_HELLO,
		      hello_src_dst_t,
		      Hello_Class) {
		bool isConsistent(void) const {
			return true;
		}
	};

	DEFINE_OBJECT(HelloAck_Object,
		      CTYPE_HELLO_ACK,
		      hello_src_dst_t,
		      Hello_Class) {
		bool isConsistent(void) const {
			return true;
		}
	};

	DEFINE_CLASS_END(Hello_Class);

//CLASS RESTART CAP
	DEFINE_CLASS_BEGIN(RestartCap_Class, G2RSVP_CLASS_RESTARTCAP);

	typedef struct {
		bit32_t restart_time;
		bit32_t recovery_time;
	} restart_cap_t;
	DEFINE_OBJECT(RestartCap_Object,
		      CTYPE_RESTART_CAP,
		      restart_cap_t,
		      RestartCap_Class) {
		bool isConsistent(void) const {
			return true;
		}
	};

	DEFINE_CLASS_END(RestartCap_Class);

//CLASS GNS_CALL_EXT (c-num = 248)
	DEFINE_CLASS_BEGIN(GnsCallExt_Class, G2RSVP_CLASS_GNS_CALL_EXT);

//        0                   1                   2                   3
//       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                                                               |
//      |                parent GNS transaction identifier              |
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                         Call Type                             |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                         Start time                            |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                          Duration                             |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      | Job Name Len  |                                               |
//	|                                                               |
//      //             Job name (NULL padded display string)           //
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      | Job Proj Len  |                                               |
//	|                                                               |
//      //             Job project (NULL padded display string)        //
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//      Call Type
//      0x00000000: CALL_TYPE_SPC
//      0x00000001: CALL_TYPE_PC
//      0x00000002: CALL_TYPE_SC
//      0x00000003: CALL_TYPE_AUTO
//      0x00000004: CALL_TYPE_aUGWzUGW
//      0x00000005: CALL_TYPE_aMGTzEGW
//      0x00000006: CALL_TYPE_aUGWzEGW
//      0x00000007: CALL_TYPE_aEGWzMGT
//      0x00000008: CALL_TYPE_aEGWzUGW
//      0x00000009: CALL_TYPE_aEGWzEGW

	typedef struct {
		bit64_t gns_trans_id;
		bit32_t call_type;
		bit32_t start_time;
		bit32_t duration;
		bit8_t	job_name_len;
		char *  job_name;
		bit8_t	job_proj_len;
		char *  job_project;
	} gns_call_ext_t;

	DEFINE_OBJECT(GnsCallExt_Object,
		      CTYPE_GNS_CALL,
		      gns_call_ext_t,
		      GnsCallExt_Class) {
		OBJ_OPERATOR(GnsCallExt_Object);
	public:
		GnsCallExt_Object(void) {
			value_.call_type    = 0;
			value_.job_name     = 0;
			value_.job_name_len = 0;
			value_.job_project  = 0;
			value_.job_proj_len = 0;
		}

		GnsCallExt_Object(call_info_t call_info) {
			value_.gns_trans_id = 0;
			switch (call_info.call_type) {
				case CALL_TYPE_SPC:
					value_.call_type = 0x00000000;
					break;
				case CALL_TYPE_PC:
					value_.call_type = 0x00000001;
					break;
				case CALL_TYPE_SC:
					value_.call_type = 0x00000002;
					break;
				case CALL_TYPE_AUTO:
					value_.call_type = 0x00000003;
					break;
				case CALL_TYPE_aUGWzUGW:
					value_.call_type = 0x00000004;
					break;
				case CALL_TYPE_aMGTzEGW:
					value_.call_type = 0x00000005;
					break;
				case CALL_TYPE_aUGWzEGW:
					value_.call_type = 0x00000006;
					break;
				case CALL_TYPE_aEGWzMGT:
					value_.call_type = 0x00000007;
					break;
				case CALL_TYPE_aEGWzUGW:
					value_.call_type = 0x00000008;
					break;
				case CALL_TYPE_aEGWzEGW:
					value_.call_type = 0x00000009;
					break;
				default:
					value_.call_type = 0x00000003;
					break;
			}
			value_.start_time   = call_info.times.start_time;
			value_.duration     = call_info.times.end_time -
				call_info.times.start_time;
			value_.job_name_len = call_info.job_name->length();
			value_.job_name     = new char[value_.job_name_len];
			strncpy(value_.job_name,
				call_info.job_name->c_str(),
				value_.job_name_len);
			value_.job_proj_len = call_info.job_project->length();
			value_.job_project  = new char[value_.job_proj_len];
			strncpy(value_.job_project,
				call_info.job_project->c_str(),
				value_.job_proj_len);
		}

		~GnsCallExt_Object(void) {
			DELETE_ARR_OBJ(value_.job_name, value_.job_name_len);
			DELETE_ARR_OBJ(value_.job_project, value_.job_proj_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength                    +
				sizeof(value_.gns_trans_id)    +
				sizeof(value_.call_type)       +
				sizeof(value_.start_time)      +
				sizeof(value_.duration)        +
				sizeof(value_.job_name_len)    +
				sizeof(*value_.job_name) *
				value_.job_name_len            +
				sizeof(value_.job_proj_len) +
				sizeof(*value_.job_project) *
				value_.job_proj_len;
		}

		GnsCallExt_Class * clone(void) const {
			GnsCallExt_Object * obj = new GnsCallExt_Object();

			if (obj) {
				obj->value_.gns_trans_id = value_.gns_trans_id;
				obj->value_.call_type    = value_.call_type;
				obj->value_.start_time   = value_.start_time;
				obj->value_.duration     = value_.duration;

				obj->value_.job_name_len = value_.job_name_len;
				if (value_.job_name_len>0 && value_.job_name) {
					obj->value_.job_name =
						new char[value_.job_name_len];
					assert(obj->value_.job_name);
					memcpy(obj->value_.job_name,
					       value_.job_name,
					       value_.job_name_len);
				} else {
					obj->value_.job_name = 0;
				}

				obj->value_.job_proj_len = value_.job_proj_len;
				if (value_.job_proj_len > 0 &&
				    value_.job_project) {
					obj->value_.job_project =
						new char[value_.job_proj_len];
					assert(obj->value_.job_project);
					memcpy(obj->value_.job_project,
					       value_.job_project,
					       value_.job_proj_len);
				} else {
					obj->value_.job_project = 0;
				}
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(GnsCallExt_Class);

//CLASS GNS_UNI (c-num = 249)
	DEFINE_CLASS_BEGIN(GnsUni_Class, G2RSVP_CLASS_GNS_UNI);

	/////////////// Subobjects definition /////////////

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (4)              |    Type (1)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |      Application Type         |       Application Version     |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit16_t app_type;
		bit16_t app_version;
	} application_t;
	DEFINE_GNS_SUBOBJ(Application_SubObj,
			  TYPE_APPLICATION,
			  SUBTYPE_APPLICATION,
			  application_t) {
		OBJ_OPERATOR(Application_SubObj);
	public:
		Application_SubObj(void) {}

		Application_SubObj(grid_application_t appl) {
			version_t version;
			version = appl.version;
			value_.app_type    = appl.type;
			value_.app_version = ((version.major_revision & 0x000F) << 10) |
				             ((version.minor_revision & 0x000F) << 6 ) |
				              (version.build_fix      & 0x003F);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(application_t);
		}

		GnsSubObj * clone(void) const {
			Application_SubObj * obj = new Application_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//       0                   1                   2                   3
//       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |       Length (4)              |    Type (2)   |  Sub-Type (1) |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |                      IPv4 address                             |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef ipv4_addr_t ipv4_host_t;
	DEFINE_GNS_SUBOBJ(IPv4CandidateHost_SubObj,
			  TYPE_CANDIDATE_HOSTS,
			  SUBTYPE_IPv4_CANDIDATE_HOSTS,
			  ipv4_host_t) {
		OBJ_OPERATOR(IPv4CandidateHost_SubObj);
	public:
		IPv4CandidateHost_SubObj(void) {}

		IPv4CandidateHost_SubObj(ipv4_t addr) {
			value_.ipv4_addr = addr;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(ipv4_host_t);
		}

		GnsSubObj * clone(void) const {
			IPv4CandidateHost_SubObj * obj;
			obj = new IPv4CandidateHost_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (16)             |    Type (2)   |  Sub-Type (2) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                                                               |
//       |                       IPv6 address                            |
//       |                                                               |
//       |                                                               |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef ipv6_addr_t ipv6_host_t;
	DEFINE_GNS_SUBOBJ(IPv6CandidateHost_SubObj,
			  TYPE_CANDIDATE_HOSTS,
			  SUBTYPE_IPv6_CANDIDATE_HOSTS,
			  ipv6_host_t) {
		OBJ_OPERATOR(IPv6CandidateHost_SubObj);
	public:
		IPv6CandidateHost_SubObj(void) {}

		IPv6CandidateHost_SubObj(ipv6_t addr) {
			value_.ipv6_addr = addr;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(ipv6_host_t);
		}

		GnsSubObj * clone(void) const {
			IPv6CandidateHost_SubObj * obj;
			obj = new IPv6CandidateHost_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (20)             |    Type (2)   |  Sub-Type (3) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |  NSAP  length |                  Reserved (0)                 |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                                                               |
//       //             NSAP address (Zero padded for alignment)        //
//       |                                                               |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit8_t  nsap_len;     // the length of the next field
		bit24_t reserved;
		char *  nsap_addr;    // char to be considered as byte
	} nsap_host_t;
	DEFINE_GNS_SUBOBJ(NsapCandidateHost_SubObj,
			  TYPE_CANDIDATE_HOSTS,
			  SUBTYPE_NSAP_CANDIDATE_HOSTS,
			  nsap_host_t) {
		OBJ_OPERATOR(NsapCandidateHost_SubObj);
	public:
		NsapCandidateHost_SubObj(void) {
			value_.nsap_addr = 0;
			value_.nsap_len  = 0;
		}

		NsapCandidateHost_SubObj(nsap__t addr) {
			memcpy(value_.nsap_addr, &addr, sizeof(addr));
			value_.nsap_len  = strlen(value_.nsap_addr);
		}

		~NsapCandidateHost_SubObj(void) {
			DELETE_ARR_OBJ(value_.nsap_addr, value_.nsap_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(uint32_t) +
				sizeof(*value_.nsap_addr) * value_.nsap_len;
		}

		GnsSubObj * clone(void) const {
			NsapCandidateHost_SubObj * obj;
			obj = new NsapCandidateHost_SubObj();
			if (obj) {
				obj->value_.reserved  = value_.reserved; //???
				obj->value_.nsap_len  = value_.nsap_len;

				if (value_.nsap_len > 0 && value_.nsap_addr) {
					obj->value_.nsap_addr =
						new char[value_.nsap_len];
					assert(obj->value_.nsap_addr);
					memcpy(obj->value_.nsap_addr,
					       value_.nsap_addr,
					       value_.nsap_len);
				} else {
					obj->value_.nsap_addr = 0;
				}
			}
			return obj;
		}
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length                  |    Type (3)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |    FS Name    |    FS Type    |         Range Flags           |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                    Disk Space Lower Bound                     |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                    Disk Space Upper Bound                     |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       | Mount Pt Len  |                                               |
//       +-+-+-+-+-+-+-+-+                                               +
//       |                                                               |
//       //           Mount Point  (NULL padded display string)         //
//       |                                                               |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       | Mount Src Len |                                               |
//       +-+-+-+-+-+-+-+-+                                               +
//       |                                                               |
//       //           Mount Source (NULL padded display string)         //
//       |                                                               |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit8_t	fs_name;
		bit8_t	fs_type;
		bit16_t range_flags;
		bit32_t disk_space_lb;	//disk space lower bound
		bit32_t disk_space_ub;	//disk space upper bound
		bit8_t	mount_point_len;
		char *  mount_point;
		bit8_t	mount_source_len;
		char *  mount_source;
	} fs_resources_t;
	DEFINE_GNS_SUBOBJ(FsResources_SubObj,
			  TYPE_FILE_SYSTEM_RES,
			  SUBTYPE_FILE_SYSTEM_RES,
			  fs_resources_t) {
		OBJ_OPERATOR(FsResources_SubObj);
	public:
		FsResources_SubObj(void) {
			value_.mount_point      = 0;
			value_.mount_point_len  = 0;
			value_.mount_source     = 0;
			value_.mount_source_len = 0;
		}

		FsResources_SubObj(grid_file_system_t fs_resources) {
			value_.fs_name          = fs_resources.fs_name;
			value_.fs_type          = fs_resources.fs_type;
			value_.range_flags      = (fs_resources.disk_space.ub_included << 1) |
				                  (fs_resources.disk_space.lb_included);
			value_.disk_space_lb    = fs_resources.disk_space.lower_bound;
			value_.disk_space_ub    = fs_resources.disk_space.upper_bound;
			value_.mount_point_len  = fs_resources.mount_point->length();
			value_.mount_point      = new char[value_.mount_point_len];
			strncpy(value_.mount_point,
				fs_resources.mount_point->c_str(),
				value_.mount_point_len);
			value_.mount_source_len = fs_resources.mount_source->length();
			value_.mount_source     = new char[value_.mount_source_len];
			strncpy(value_.mount_source,
				fs_resources.mount_source->c_str(),
				value_.mount_source_len);
		}

		~FsResources_SubObj(void) {
			DELETE_ARR_OBJ(value_.mount_point,
				       value_.mount_point_len);
			DELETE_ARR_OBJ(value_.mount_source,
				       value_.mount_source_len);
		}


		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength +
				sizeof(value_.fs_name) +
				sizeof(value_.fs_type) +
				sizeof(value_.range_flags) +
				sizeof(value_.disk_space_lb) +
				sizeof(value_.disk_space_ub) +
				sizeof(value_.mount_point_len) +
				(sizeof(*value_.mount_point) *
				 value_.mount_point_len) +
				sizeof(value_.mount_source_len) +
				(sizeof(*value_.mount_source) *
				 value_.mount_source_len);
		}

		GnsSubObj * clone(void) const {
			FsResources_SubObj * obj;
			obj = new FsResources_SubObj();
			if (obj) {
				obj->value_.fs_name      = value_.fs_name;
				obj->value_.fs_type      = value_.fs_type;
				obj->value_.range_flags  = value_.range_flags;
				obj->value_.disk_space_lb=value_.disk_space_lb;
				obj->value_.disk_space_ub=value_.disk_space_ub;

				obj->value_.mount_point_len =
					value_.mount_point_len;
				if (value_.mount_point_len > 0 &&
				    value_.mount_point) {
					obj->value_.mount_point =
					       new char[value_.mount_point_len];
					assert(obj->value_.mount_point);
					memcpy(obj->value_.mount_point,
					       value_.mount_point,
					       value_.mount_point_len);
				} else {
					obj->value_.mount_point = 0;
				}

				obj->value_.mount_source_len =
					value_.mount_source_len;
				if (value_.mount_source_len > 0 &&
				    value_.mount_source) {
					obj->value_.mount_source =
					      new char[value_.mount_source_len];
					assert(obj->value_.mount_source);
					memcpy(obj->value_.mount_source,
					       value_.mount_source,
					       value_.mount_source_len);
				} else {
					obj->value_.mount_source = 0;
				}
			}
			return obj;
		}
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (8)              |    Type (4)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |           OS Type             |             OS  Version       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |   CPU Arch    |E|                reserved                     |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
	typedef struct {
		bit16_t os_type;
		bit16_t os_version;
		bit8_t  cpu_arch;
		bit1_t  E;
		bit24_t reserved24;
	} sys_capabilities_t;
	DEFINE_GNS_SUBOBJ(SystemCaps_SubObj,
			  TYPE_SYSTEM_CAPS,
			  SUBTYPE_SYSTEM_CAPS,
			  sys_capabilities_t) {
		OBJ_OPERATOR(SystemCaps_SubObj);
	public:
		SystemCaps_SubObj(void) {}

		SystemCaps_SubObj(grid_system_caps_t sys_caps) {
			version_t version;
			version = sys_caps.os_version;
			value_.os_version    = ((version.major_revision & 0x000F) << 10) |
				               ((version.minor_revision & 0x000F) << 6 ) |
						(version.build_fix      & 0x003F);
			value_.os_type       = sys_caps.os_type;
			value_.cpu_arch      = sys_caps.cpu_arch;
			value_.E             = sys_caps.exclusive_access;
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength +
				sizeof(value_.os_type) +
				sizeof(value_.os_version) +
				sizeof(uint32_t);
		}

		GnsSubObj * clone(void) const {
			SystemCaps_SubObj * obj = new SystemCaps_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};


	typedef struct {
		bit32_t lower_bound;
		bit32_t upper_bound;
		bit2_t  SF;	  // Speed Flags (LB incl.=0x01, UB incl.=0x10)
		bit30_t reserved30;
	} range_t;

#define LENGTH()				\
	uint16_t length(void) const {		\
		return subObjHdrLength +	\
			sizeof(uint32_t) +	\
			sizeof(uint32_t) +	\
			sizeof(uint32_t);	\
	}

#define GNS_CLONE(OBJ)				\
	GnsSubObj * clone(void) const {		\
		OBJ * obj = new OBJ();		\
		if (obj) {			\
			obj->value_ = value_;	\
		}				\
		return obj;			\
	}

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (5)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU speed Lower Bound                       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU speed Upper Bound                       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t indv_cpu_speed_t;
	DEFINE_GNS_SUBOBJ(IndvCpuSpeed_SubObj,
			  TYPE_INDIV_CPU_SPEED,
			  SUBTYPE_INDIV_CPU_SPEED,
			  indv_cpu_speed_t) {
		OBJ_OPERATOR(IndvCpuSpeed_SubObj);
	public:
		IndvCpuSpeed_SubObj(void) {}

		IndvCpuSpeed_SubObj(range_set_t	ind_cpu_speed) {
			value_.lower_bound   = ind_cpu_speed.lower_bound;
			value_.upper_bound   = ind_cpu_speed.upper_bound;
			value_.SF            = (ind_cpu_speed.ub_included << 1) |
				               (ind_cpu_speed.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(IndvCpuSpeed_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (6)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Time  Lower Bound                       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Time  Upper Bound                       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t indv_cpu_time_t;
	DEFINE_GNS_SUBOBJ(IndvCpuTime_SubObj,
			  TYPE_INDIV_CPU_TIME,
			  SUBTYPE_INDIV_CPU_TIME,
			  indv_cpu_time_t) {
		OBJ_OPERATOR(IndvCpuTime_SubObj);
	public:
		IndvCpuTime_SubObj(void) {}

		IndvCpuTime_SubObj(range_set_t ind_cpu_time) {
			value_.lower_bound   = ind_cpu_time.lower_bound;
			value_.upper_bound   = ind_cpu_time.upper_bound;
			value_.SF            = (ind_cpu_time.ub_included << 1) |
				               (ind_cpu_time.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(IndvCpuTime_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (7)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Count  Lower Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Count  Upper Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t indv_cpu_count_t;
	DEFINE_GNS_SUBOBJ(IndvCpuCount_SubObj,
			  TYPE_INDIV_CPU_COUNT,
			  SUBTYPE_INDIV_CPU_COUNT,
			  indv_cpu_count_t) {
		OBJ_OPERATOR(IndvCpuCount_SubObj);
	public:
		IndvCpuCount_SubObj(void) {}

		IndvCpuCount_SubObj(range_set_t ind_cpu_count) {
			value_.lower_bound   = ind_cpu_count.lower_bound;
			value_.upper_bound   = ind_cpu_count.upper_bound;
			value_.SF            = (ind_cpu_count.ub_included << 1) |
				               (ind_cpu_count.lb_included);
		}
		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(IndvCpuCount_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (8)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   NET BW     Lower Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   NET BW     Upper Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t indv_net_BW_t;
	DEFINE_GNS_SUBOBJ(IndvNetBw_SubObj,
			  TYPE_INDIV_NET_BW,
			  SUBTYPE_INDIV_NET_BW,
			  indv_net_BW_t) {
		OBJ_OPERATOR(IndvNetBw_SubObj);
	public:
		IndvNetBw_SubObj(void) {}

		IndvNetBw_SubObj(range_set_t ind_net_bw) {
			value_.lower_bound   = ind_net_bw.lower_bound;
			value_.upper_bound   = ind_net_bw.upper_bound;
			value_.SF            = (ind_net_bw.ub_included << 1) |
				               (ind_net_bw.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(IndvNetBw_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (9)   |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Physical Memory Lower Bound                 |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Physical Memory Upper Bound                 |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t indv_phy_mem_t;
	DEFINE_GNS_SUBOBJ(IndvPhyMem_SubObj,
			  TYPE_INDIV_PHY_MEM,
			  SUBTYPE_INDIV_PHY_MEM,
			  indv_phy_mem_t) {
		OBJ_OPERATOR(IndvPhyMem_SubObj);
	public:
		IndvPhyMem_SubObj(void) {}

		IndvPhyMem_SubObj(range_set_t ind_phy_mem) {
		        value_.lower_bound   = ind_phy_mem.lower_bound;
			value_.upper_bound   = ind_phy_mem.upper_bound;
			value_.SF            = (ind_phy_mem.ub_included << 1) |
				               (ind_phy_mem.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(IndvPhyMem_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (10)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Virtual Memory Lower Bound                  |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Virtual Memory Upper Bound                  |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t indv_vir_mem_t;
	DEFINE_GNS_SUBOBJ(IndvVirMem_SubObj,
			  TYPE_INDIV_VIR_MEM,
			  SUBTYPE_INDIV_VIR_MEM,
			  indv_vir_mem_t) {
		OBJ_OPERATOR(IndvVirMem_SubObj);
	public:
		IndvVirMem_SubObj(void) {}

		IndvVirMem_SubObj(range_set_t ind_vir_mem) {
			value_.lower_bound   = ind_vir_mem.lower_bound;
			value_.upper_bound   = ind_vir_mem.upper_bound;
			value_.SF            = (ind_vir_mem.ub_included << 1) |
				               (ind_vir_mem.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(IndvVirMem_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (11)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Disk Space Lower Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Disk Space Upper Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t indv_disk_space_t;
	DEFINE_GNS_SUBOBJ(IndvDiskSpace_SubObj,
			  TYPE_INDIV_DISK_SPACE,
			  SUBTYPE_INDIV_DISK_SPACE,
			  indv_disk_space_t) {
		OBJ_OPERATOR(IndvDiskSpace_SubObj);
	public:
		IndvDiskSpace_SubObj(void) {}

		IndvDiskSpace_SubObj(range_set_t ind_disk_space) {
			value_.lower_bound   = ind_disk_space.lower_bound;
			value_.upper_bound   = ind_disk_space.upper_bound;
			value_.SF            = (ind_disk_space.ub_included << 1) |
				               (ind_disk_space.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(IndvDiskSpace_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (12)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Time  Lower Bound                       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Time  Upper Bound                       |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t total_cpu_time_t;
	DEFINE_GNS_SUBOBJ(TotCpuTime_SubObj,
			  TYPE_TOTAL_CPU_TIME,
			  SUBTYPE_TOTAL_CPU_TIME,
			  total_cpu_time_t) {
		OBJ_OPERATOR(TotCpuTime_SubObj);
	public:
		TotCpuTime_SubObj(void) {}

		TotCpuTime_SubObj(range_set_t tot_cpu_time) {
			value_.lower_bound   = tot_cpu_time.lower_bound;
		        value_.upper_bound   = tot_cpu_time.upper_bound;
			value_.SF            = (tot_cpu_time.ub_included << 1) |
				               (tot_cpu_time.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(TotCpuTime_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (13)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Count  Lower Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   CPU Count  Upper Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t total_cpu_count_t;
	DEFINE_GNS_SUBOBJ(TotCpuCount_SubObj,
			  TYPE_TOTAL_CPU_COUNT,
			  SUBTYPE_TOTAL_CPU_COUNT,
			  total_cpu_count_t) {
		OBJ_OPERATOR(TotCpuCount_SubObj);
	public:
		TotCpuCount_SubObj(void) {}

		TotCpuCount_SubObj(range_set_t tot_cpu_count) {
			value_.lower_bound   = tot_cpu_count.lower_bound;
		        value_.upper_bound   = tot_cpu_count.upper_bound;
			value_.SF            = (tot_cpu_count.ub_included << 1) |
				               (tot_cpu_count.lb_included);
		}
		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(TotCpuCount_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (14)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Physical Memory Lower Bound                 |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Physical Memory Upper Bound                 |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t total_phy_mem_t;
	DEFINE_GNS_SUBOBJ(TotPhyMem_SubObj,
			  TYPE_TOTAL_PHY_MEM,
			  SUBTYPE_TOTAL_PHY_MEM,
			  total_phy_mem_t) {
		OBJ_OPERATOR(TotPhyMem_SubObj);
	public:
		TotPhyMem_SubObj(void) {}

		TotPhyMem_SubObj(range_set_t tot_phy_mem) {
			value_.lower_bound   = tot_phy_mem.lower_bound;
			value_.upper_bound   = tot_phy_mem.upper_bound;
			value_.SF            = (tot_phy_mem.ub_included << 1) |
				               (tot_phy_mem.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(TotPhyMem_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (15)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Virtual Memory Lower Bound                  |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Virtual Memory Upper Bound                  |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t total_vir_mem_t;
	DEFINE_GNS_SUBOBJ(TotVirMem_SubObj,
			  TYPE_TOTAL_VIR_MEM,
			  SUBTYPE_TOTAL_VIR_MEM,
			  total_vir_mem_t) {
		OBJ_OPERATOR(TotVirMem_SubObj);
	public:
		TotVirMem_SubObj(void) {}

		TotVirMem_SubObj(range_set_t tot_vir_mem) {
			value_.lower_bound   = tot_vir_mem.lower_bound;
			value_.upper_bound   = tot_vir_mem.upper_bound;
			value_.SF            = (tot_vir_mem.ub_included << 1) |
				               (tot_vir_mem.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(TotVirMem_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (16)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Disk Space Lower Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Disk Space Upper Bound                      |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t total_disk_space_t;
	DEFINE_GNS_SUBOBJ(TotDiskSpace_SubObj,
			  TYPE_TOTAL_DISK_SPACE,
			  SUBTYPE_TOTAL_DISK_SPACE,
			  total_disk_space_t) {
		OBJ_OPERATOR(TotDiskSpace_SubObj);
	public:
		TotDiskSpace_SubObj(void) {}

		TotDiskSpace_SubObj(range_set_t tot_disk_space) {
			value_.lower_bound   = tot_disk_space.lower_bound;
			value_.upper_bound   = tot_disk_space.upper_bound;
			value_.SF            = (tot_disk_space.ub_included << 1) |
				               (tot_disk_space.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(TotDiskSpace_SubObj);
	};

//        0                   1                   2                   3
//        0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |       Length (12)             |    Type (17)  |  Sub-Type (1) |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Total Resource Count Lower Bound            |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |                   Total Resource Count Upper Bound            |
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//       |SF |                      reserved                            //
//       +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef range_t total_res_count_t;
	DEFINE_GNS_SUBOBJ(TotResCount_SubObj,
			  TYPE_TOTAL_RES_COUNT,
			  SUBTYPE_TOTAL_RES_COUNT,
			  total_res_count_t) {
		OBJ_OPERATOR(TotResCount_SubObj);
	public:
		TotResCount_SubObj(void) {}

		TotResCount_SubObj(range_set_t tot_res_count) {
			value_.lower_bound   = tot_res_count.lower_bound;
			value_.upper_bound   = tot_res_count.upper_bound;
			value_.SF            = (tot_res_count.ub_included << 1) |
				               (tot_res_count.lb_included);
		}

		bool isConsistent(void) const {
			return true;
		}

		LENGTH();

		GNS_CLONE(TotResCount_SubObj);
	};

//       0                   1                   2                   3
//       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |       Length                  |    Type (18)  |  Sub-Type (1) |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      |    FS Name    |  CF |D|  res. | File Name Len |               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+               +
//      |                                                               |
//      //             File Name   (NULL padded display string)        //
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      | Source   Len  |                                               |
//      +-+-+-+-+-+-+-+-+                                               +
//      |                                                               |
//      //               Source   (NULL padded display string)         //
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//      | Target   Len  |                                               |
//      +-+-+-+-+-+-+-+-+                                               +
//      |                                                               |
//      //               Target   (NULL padded display string)         //
//      |                                                               |
//      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit8_t FSname;
		bit3_t creation_flag;
		bit1_t delete_on_termination;
		bit8_t file_name_len;
		char * file_name;
		bit8_t source_len;
		char * source;
		bit8_t target_len;
		char * target;
	} data_staging_t;
	DEFINE_GNS_SUBOBJ(DataStaging_SubObj,
			  TYPE_DATA_STAGING,
			  SUBTYPE_DATA_STAGING,
			  data_staging_t) {
		OBJ_OPERATOR(DataStaging_SubObj);
	public:
		DataStaging_SubObj(void) {
			value_.file_name     = 0;
			value_.file_name_len = 0;
			value_.source        = 0;
			value_.source_len    = 0;
			value_.target        = 0;
			value_.target_len    = 0;
		}

		DataStaging_SubObj(grid_data_staging_t data_staging) {
			value_.FSname                = data_staging.fs_name;
			value_.creation_flag         = data_staging.creation_flag;
			value_.delete_on_termination = data_staging.del_on_termination;
			value_.file_name_len         = data_staging.filename->length();
			value_.file_name             = new char[value_.file_name_len];
			strncpy(value_.file_name,
				data_staging.filename->c_str(),
				value_.file_name_len);
			value_.source_len            = data_staging.source->length();
			value_.source                = new char[value_.source_len];
			strncpy(value_.source,
				data_staging.source->c_str(),
				value_.source_len);
			value_.target_len            = data_staging.target->length();
			value_.target                = new char [value_.target_len];
			strncpy(value_.target,
				data_staging.target->c_str(),
			        value_.target_len);
		}

		~DataStaging_SubObj(void) {
			DELETE_ARR_OBJ(value_.file_name, value_.file_name_len);
			DELETE_ARR_OBJ(value_.source, value_.source_len);
			DELETE_ARR_OBJ(value_.target, value_.target_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength +
				sizeof(value_.FSname) +
				sizeof(uint8_t) +
				sizeof(value_.file_name_len) +
				(sizeof(*value_.file_name) *
				 value_.file_name_len) +
				sizeof(value_.source_len) +
				(sizeof(*value_.source) *
				 value_.source_len) +
				sizeof(value_.target_len) +
				(sizeof(*value_.target) *
				 value_.target_len);
		}

		GnsSubObj * clone(void) const {
			DataStaging_SubObj * obj = new DataStaging_SubObj();
			if (obj) {
				obj->value_.FSname        = value_.FSname;
				obj->value_.creation_flag =
					value_.creation_flag;
				obj->value_.delete_on_termination =
					value_.delete_on_termination;


				obj->value_.file_name_len =
					value_.file_name_len;
				if (obj->value_.file_name_len > 0 &&
				    obj->value_.file_name) {
					obj->value_.file_name =
						new char[value_.file_name_len];
					assert(obj->value_.file_name);
					memcpy(obj->value_.file_name,
					       value_.file_name,
					       value_.file_name_len);
				} else {
					obj->value_.file_name = 0;
				}

				obj->value_.source_len = value_.source_len;
				if (obj->value_.source_len > 0 &&
				    obj->value_.source) {
					obj->value_.source =
						new char[value_.source_len];
					assert(obj->value_.source);
					memcpy(obj->value_.source,
					       value_.source,
					       value_.source_len);
				} else {
					obj->value_.source = 0;
				}

				obj->value_.target_len = value_.target_len;
				if (obj->value_.target_len > 0 &&
				    obj->value_.target) {
					obj->value_.target =
						new char[value_.target_len];
					assert(obj->value_.target);
					memcpy(obj->value_.target,
					       value_.target,
					       value_.target_len);
				} else {
					obj->value_.target = 0;
				}
			}
			return obj;
		}
	};

// GNS UNI format
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   //                        (Subobjects)                          //
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		std::list<GnsSubObj *> subObjs;
	} gns_uni_t;
	DEFINE_OBJECT(GnsUni_Object,
		      CTYPE_GNS_UNI,
		      gns_uni_t,
		      GnsUni_Class) {
		OBJ_OPERATOR(GnsUni_Object);
	public:
		~GnsUni_Object(void) {
			std::list<GnsSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.subObjs.clear();
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<GnsSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength + len;
		}

		GnsUni_Class * clone(void) const {
			GnsUni_Object * obj = new GnsUni_Object();

			if (obj) {
				obj->value_.subObjs.clear();

				std::list<GnsSubObj *>::const_iterator it;
				for (it = value_.subObjs.begin();
				     it != value_.subObjs.end();
				     it++) {
					if (*it) {
						obj->value_.subObjs.push_back
							((*it)->clone());
					}
				}
			}

			return obj;
		}
	};

	DEFINE_CLASS_END(GnsUni_Class);

//CLASS GENERALIZED UNI
	DEFINE_CLASS_BEGIN(GeneralisedUni_Class, G2RSVP_CLASS_GENERALIZED_UNI);

	/////////////// Subobjects definition /////////////

//     0                   1                   2                   3
//    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |        Source UNI-C's IPv4 TNA address (4 bytes)              |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef ipv4_addr_t ipv4_src_tna_addr_t;
	DEFINE_GUNI_SUBOBJ(IPv4SourceTnaAddr_SubObj,
			   TYPE_SRC_TNA,
			   SUBTYPE_IPv4_SRC_TNA,
			   ipv4_src_tna_addr_t) {
		OBJ_OPERATOR(IPv4SourceTnaAddr_SubObj);
	public:
		IPv4SourceTnaAddr_SubObj(ipv4_t ipv4) {
			value_.ipv4_addr = ipv4;
		}
		IPv4SourceTnaAddr_SubObj(void) { }
		~IPv4SourceTnaAddr_SubObj(void) { }

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(ipv4_src_tna_addr_t);
		}

		GuniSubObj * clone(void) const {
			IPv4SourceTnaAddr_SubObj * obj;
			obj = new IPv4SourceTnaAddr_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//     0                   1                   2                   3
//    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                                                               |
//   //        Source UNI-C's IPv6 TNA address (16 bytes)           //
//   |                                                               |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef ipv6_addr_t ipv6_src_tna_addr_t;
	DEFINE_GUNI_SUBOBJ(IPv6SourceTnaAddr_SubObj,
			   TYPE_SRC_TNA,
			   SUBTYPE_IPv6_SRC_TNA,
			   ipv6_src_tna_addr_t) {
		OBJ_OPERATOR(IPv6SourceTnaAddr_SubObj);
	public:
		IPv6SourceTnaAddr_SubObj(ipv6_t ipv6) {
			value_.ipv6_addr = ipv6;
		}
		IPv6SourceTnaAddr_SubObj(void) { }
		~IPv6SourceTnaAddr_SubObj(void) { }

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(ipv6_src_tna_addr_t);
		}

		GuniSubObj * clone(void) const {
			IPv6SourceTnaAddr_SubObj * obj;
			obj = new IPv6SourceTnaAddr_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

//     0                   1                   2                   3
//    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |  NSAP  length |                  Reserved (0)                 |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                                                               |
//   //            Source UNI-C's NSAP-format TNA address           //
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   | NSAP address continue. ...  Zero padded for alignment         |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit8_t  nsap_len;     // the length of the next field
		bit24_t reserved;
		char *  nsap_addr;    // char to be considered as byte
	} nsap_src_tna_addr_t;
	DEFINE_GUNI_SUBOBJ(NSAPSourceTnaAddr_SubObj,
			   TYPE_SRC_TNA,
			   SUBTYPE_NSAP_SRC_TNA,
			   nsap_src_tna_addr_t) {
		OBJ_OPERATOR(NSAPSourceTnaAddr_SubObj);
	public:
		NSAPSourceTnaAddr_SubObj(nsap__t nsap) {
			value_.nsap_len  = sizeof(nsap); // should be 20!!!
			value_.nsap_addr = new char[value_.nsap_len];
			assert(value_.nsap_addr);

			for (int i = 0; i < 20; i++) {
				value_.nsap_addr[i] = nsap.nsap_addr8[i];
			}
		}

		NSAPSourceTnaAddr_SubObj(void) {
			value_.nsap_addr = 0;
			value_.nsap_len  = 0;
		}

		~NSAPSourceTnaAddr_SubObj(void) {
			DELETE_ARR_OBJ(value_.nsap_addr, value_.nsap_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(uint32_t) + //len/resvd
				sizeof(*value_.nsap_addr) * value_.nsap_len;
		}

		GuniSubObj * clone(void) const {
			NSAPSourceTnaAddr_SubObj * obj;
			obj = new NSAPSourceTnaAddr_SubObj();

			if (obj) {
				obj->value_.reserved = value_.reserved;
				obj->value_.nsap_len = value_.nsap_len;

				if (value_.nsap_len > 0 && value_.nsap_addr) {
					obj->value_.nsap_addr =
						new char[value_.nsap_len];
					assert(obj->value_.nsap_addr);
					memcpy(obj->value_.nsap_addr,
					       value_.nsap_addr,
					       value_.nsap_len);
				} else {
					obj->value_.nsap_addr = 0;
				}
			}

			return obj;
		}
	};

	typedef ipv4_addr_t ipv4_dst_tna_addr_t;
	DEFINE_GUNI_SUBOBJ(IPv4DestTnaAddr_SubObj,
			   TYPE_DST_TNA,
			   SUBTYPE_IPv4_DST_TNA,
			   ipv4_dst_tna_addr_t) {
		OBJ_OPERATOR(IPv4DestTnaAddr_SubObj);
	public:
		IPv4DestTnaAddr_SubObj(ipv4_t ipv4) {
			value_.ipv4_addr = ipv4;
		}
		IPv4DestTnaAddr_SubObj(void) { }
		~IPv4DestTnaAddr_SubObj(void) { }

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(ipv4_dst_tna_addr_t);
		}

		GuniSubObj * clone(void) const {
			IPv4DestTnaAddr_SubObj * obj;
			obj = new IPv4DestTnaAddr_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	typedef ipv6_addr_t ipv6_dst_tna_addr_t;
	DEFINE_GUNI_SUBOBJ(IPv6DestTnaAddr_SubObj,
			   TYPE_DST_TNA,
			   SUBTYPE_IPv6_DST_TNA,
			   ipv6_dst_tna_addr_t) {
		OBJ_OPERATOR(IPv6DestTnaAddr_SubObj);
	public:
		IPv6DestTnaAddr_SubObj(ipv6_t ipv6) {
			value_.ipv6_addr = ipv6;
		}
		IPv6DestTnaAddr_SubObj(void) { }
		~IPv6DestTnaAddr_SubObj(void) { }

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(ipv6_dst_tna_addr_t);
		}

		GuniSubObj * clone(void) const {
			IPv6DestTnaAddr_SubObj * obj;
			obj = new IPv6DestTnaAddr_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	typedef nsap_src_tna_addr_t nsap_dst_tna_addr_t;
	DEFINE_GUNI_SUBOBJ(NSAPDestTnaAddr_SubObj,
			   TYPE_DST_TNA,
			   SUBTYPE_NSAP_DST_TNA,
			   nsap_dst_tna_addr_t) {
		OBJ_OPERATOR(NSAPDestTnaAddr_SubObj);
	public:
		NSAPDestTnaAddr_SubObj(nsap__t nsap) {
			value_.nsap_len  = sizeof(nsap); // should be 20!!!
			value_.nsap_addr = new char[value_.nsap_len];
			assert(value_.nsap_addr);

			for (int i = 0; i < 20; i++) {
				value_.nsap_addr[i] = nsap.nsap_addr8[i];
			}
		}

		NSAPDestTnaAddr_SubObj(void) {
			value_.nsap_addr = 0;
			value_.nsap_len  = 0;
		}

		~NSAPDestTnaAddr_SubObj(void) {
			DELETE_ARR_OBJ(value_.nsap_addr, value_.nsap_len);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength + sizeof(uint32_t) + //len/resvd
				sizeof(*value_.nsap_addr) * value_.nsap_len;
		}

		GuniSubObj * clone(void) const {
			NSAPDestTnaAddr_SubObj * obj;
			obj = new NSAPDestTnaAddr_SubObj();

			if (obj) {
				obj->value_.reserved = value_.reserved;
				obj->value_.nsap_len = value_.nsap_len;

				if (value_.nsap_len > 0 && value_.nsap_addr) {
					obj->value_.nsap_addr =
						new char[value_.nsap_len];
					assert(obj->value_.nsap_addr);
					memcpy(obj->value_.nsap_addr,
					       value_.nsap_addr,
					       value_.nsap_len);
				} else {
					obj->value_.nsap_addr = 0;
				}
			}

			return obj;
		}
	};

//     0                   1                   2                   3
//    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |   DS  |                    Reserved (0)                       |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   //        UNI_IPv4_SESSION (includes RSVP object header)       //
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   //                LSP_TUNNEL_IPv4_SENDER_TEMPLATE              //
//   //                (includes RSVP object header)                //
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//    DS indicates the diversity requirement as following:
//	1 - Node diverse. The new connection SHALL NOT use any nodes that
//	    are on the path of the listed connection.
//	2 - Link Diverse. The new connection SHALL NOT use any links that
//	    are on the path of the listed connection.
//	3 - Shared Risk Link Group diverse. The new connection SHALL NOT
//	    use any link that have the same SRLG of the listed connection.
//	4 - Shared Path. The new connection SHALL use the same links as
//	    the listed connection.
//	0 and 5-15, Reserved.


	typedef struct {
		bit4_t                         DS;
		bit28_t                        reserved;
		UNIIPv4Session_Object *        uni_ipv4_session;
		IPV4LSPSenderTemplate_Object * lsp_tun_ipv4_send_temp;
	} diversity_t;
	DEFINE_GUNI_SUBOBJ(Diversity_SubObj,
			   TYPE_DIVERSITY,
			   SUBTYPE_DIVERSITY,
			   diversity_t) {
		OBJ_OPERATOR(Diversity_SubObj);
	public:
		Diversity_SubObj(void) {
			value_.uni_ipv4_session       = 0;
			value_.lsp_tun_ipv4_send_temp = 0;
		}

		~Diversity_SubObj(void) {
			DELETE_OBJ(value_.uni_ipv4_session);
			DELETE_OBJ(value_.lsp_tun_ipv4_send_temp);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength                   +
				sizeof(uint32_t)                 + // DS/resvd
				(value_.uni_ipv4_session ?
				 value_.uni_ipv4_session->length() : 0) +
				(value_.lsp_tun_ipv4_send_temp ?
				 value_.lsp_tun_ipv4_send_temp->length() : 0);
		}

		GuniSubObj * clone(void) const {
			Diversity_SubObj * obj = new Diversity_SubObj();

			if (obj) {
				obj->value_.DS       = value_.DS;
				obj->value_.reserved = value_.reserved;

				if (value_.uni_ipv4_session) {
					obj->value_.uni_ipv4_session =
					   static_cast<UNIIPv4Session_Object *>
					   (value_.uni_ipv4_session->clone());
				}

				if (value_.lsp_tun_ipv4_send_temp) {
					obj->value_.lsp_tun_ipv4_send_temp =
						static_cast<
						IPV4LSPSenderTemplate_Object *>
						(value_.lsp_tun_ipv4_send_temp
						 ->clone());
				}
			}

			return obj;
		}
	};

//     0                   1                   2                   3
//    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |U|        Reserved (0)                         |   Label Type  |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                    Logical Port Identifier                    |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                         Egress Label                          |
//   |                              ...                              |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	Logical Port Identifier: This field indicates a logical port identifier
//	assigned at the destination UNI-C. It is used to select a link at the
//	destination UNI. The identified link MUST be used to allocate resources
//	for the connection.
//
//	U-bit: This bit indicates the direction of the label and the logical
//	port ID.  It is 0 for the downstream label and port ID and 1 for the
//	upstream label and port ID. It is only used on bi-direct connections.
//
//	Egress label: This field identifies the label to be used.  The format
//	of this field is identical to the one used by the Label field in GMPLS
//	SONET/SDH Label [RFC3946], or GMPLS G.709 Label [RFC4328].

	typedef struct {
		bit1_t        U;
		bit23_t       reserved;
		bit8_t        label_type;
		bit32_t       logical_port_id;
		Label_Class * egress_label;
	} egress_label_t;
	DEFINE_GUNI_SUBOBJ(EgressLabel_SubObj,
			   TYPE_EGR_LABEL,
			   SUBTYPE_EGR_LABEL,
			   egress_label_t) {
		OBJ_OPERATOR(EgressLabel_SubObj);
	public:
		EgressLabel_SubObj(uint32_t      logPortId,
				   Label_Class * egrLabel,
				   bool          upStream = false) {
			assert(egrLabel);
			value_.logical_port_id = logPortId;
			value_.egress_label    = egrLabel;
			value_.label_type      = egrLabel->cType();
			value_.reserved        = 0;
			value_.U               = upStream ? 1 : 0;
		}

		EgressLabel_SubObj(void) {
			value_.egress_label = 0;
		}

		~EgressLabel_SubObj(void) {
			DELETE_OBJ(value_.egress_label);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength                 +
				sizeof(bit32_t)                + //U/resvd/Type
				sizeof(value_.logical_port_id) +
				(value_.egress_label ?
				 (value_.egress_label->length() -
				  objHeader::objHdrLength) : 0);
		}

		GuniSubObj * clone(void) const {
			EgressLabel_SubObj * obj = new EgressLabel_SubObj();

			if (obj) {
				obj->value_.U          = value_.U;
				obj->value_.reserved   = value_.reserved;
				obj->value_.label_type = value_.label_type;
				obj->value_.logical_port_id =
					value_.logical_port_id;

				CLONE_OBJ_INTO_MSG(value_.egress_label, obj);
			}

			return obj;
		}
	};

//     0                   1                   2                   3
//    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |U|        Reserved (0)                         |   Label Type  |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                    Logical Port Identifier                    |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   |                          SPC Label                            |
//   |                              ...                              |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//
//	Logical Port Identifier: This field indicates a logical port identifier
//	assigned at the destination UNI-C. It is used to select a link at the
//	destination UNI. The identified link MUST be used to allocate resources
//	for the connection.
//	U-bit: This bit indicates the direction of the label and the logical
//	port ID.  It is 0 for the downstream label and port ID and 1 for the
//	upstream label and port ID. It is only used on bi-directional
//	connections.
//
//	SPC label: This field identifies the label to be used.  The format of
//	this field is identical to the one used by the Label field in GMPLS
//	SONET/SDH Label [RFC3946], or GMPLS G.709 Label [RFC4328].

	typedef struct {
		bit1_t        U;
		bit23_t       reserved;
		bit8_t        label_type;
		bit32_t       logical_port_id;
		Label_Class * spc_label;
	} spc_label_t;
	DEFINE_GUNI_SUBOBJ(SpcLabel_SubObj,
			   TYPE_EGR_LABEL,
			   SUBTYPE_EGR_SPC_LABEL,
			   spc_label_t) {
		OBJ_OPERATOR(SpcLabel_SubObj);
	public:
		SpcLabel_SubObj(uint32_t      logPortId,
				Label_Class * spcLabel,
				bool          upStream = false) {
			assert(spcLabel);
			value_.logical_port_id = logPortId;
			value_.spc_label       = spcLabel;
			value_.label_type      = spcLabel->cType();
			value_.reserved        = 0;
			value_.U               = upStream ? 1 : 0;
		}

		SpcLabel_SubObj(void) {
			value_.spc_label = 0;
		}

		~SpcLabel_SubObj(void) {
			DELETE_OBJ(value_.spc_label);
		}

		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength                 +
				sizeof(bit32_t)                + //U/resvd/Type
				sizeof(value_.logical_port_id) +
				(value_.spc_label ?
				 (value_.spc_label->length() -
				  objHeader::objHdrLength) : 0);
		}

		GuniSubObj * clone(void) const {
			SpcLabel_SubObj * obj = new SpcLabel_SubObj();

			if (obj) {
				obj->value_.U          = value_.U;
				obj->value_.reserved   = value_.reserved;
				obj->value_.label_type = value_.label_type;
				obj->value_.logical_port_id =
					value_.logical_port_id;

				CLONE_OBJ_INTO_MSG(value_.spc_label, obj);
			}

			return obj;
		}
	};

//     0                   1                   2                   3
//    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//   | Service Level |                Reserved (0)                   |
//   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		bit8_t  service_level;
		bit24_t reserved;
	} service_level_t;
	DEFINE_GUNI_SUBOBJ(ServiceLevel_SubObj,
			   TYPE_SERVICE_LVL,
			   SUBTYPE_SERVICE_LVL,
			   service_level_t) {
		OBJ_OPERATOR(ServiceLevel_SubObj);
	public:
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return subObjHdrLength  +
				sizeof(uint32_t);  // service_level/resvd
		}

		GuniSubObj * clone(void) const {
			ServiceLevel_SubObj * obj = new ServiceLevel_SubObj();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	///////////////////////////////////////////////

// Generalised UNI format
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//	   |                                                               |
//	   //                        (Subobjects)                          //
//	   |                                                               |
//	   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

	typedef struct {
		std::list<GuniSubObj *> subObjs;
	} generalised_uni_t;
	DEFINE_OBJECT(GeneralisedUni_Object,
		      CTYPE_GENERALIZED_UNI,
		      generalised_uni_t,
		      GeneralisedUni_Class) {
		OBJ_OPERATOR(GeneralisedUni_Object);
	public:
		GeneralisedUni_Object(net_res_spec_t iRes, net_res_spec_t eRes){
			assert(isNetResSpecValid(iRes));
			assert(isNetResSpecValid(eRes));

			GuniSubObj * subObj;

			subObj = createGuniDstSubObj(eRes.tna);
			if (!insertGuniSubObj(subObj)) {
				throw(std::string("Impossible to ins DST_TNA"));
			}

			subObj = createGuniSrcSubObj(iRes.tna);
			if (!insertGuniSubObj(subObj)) {
				throw(std::string("Impossible to ins SRC_TNA"));
			}

			if (BITMASK_BITTEST(eRes.mask_, data_link) &&
			    BITMASK_BITTEST(eRes.mask_, label)) {

				subObj = createGuniSpcLblSubObj(eRes.data_link,
								eRes.label);
				if (!insertGuniSubObj(subObj)) {
					throw(std::string("Impossible to ins "
							  "SPC_LABEL"));
				}
			}

		}

		GeneralisedUni_Object(net_res_spec_t iRes,
				      net_res_spec_t eRes,
				      call_type_t    type) {
			assert(isNetResSpecValid(iRes));
			assert(isNetResSpecValid(eRes));

			GuniSubObj * subObj;

			subObj = createGuniDstSubObj(eRes.tna);
			if (!insertGuniSubObj(subObj)) {
				throw(std::string("Impossible to ins DST_TNA"));
			}

			subObj = createGuniSrcSubObj(iRes.tna);
			if (!insertGuniSubObj(subObj)) {
				throw(std::string("Impossible to ins SRC_TNA"));
			}

			if (BITMASK_BITTEST(eRes.mask_, data_link) &&
			    BITMASK_BITTEST(eRes.mask_, label)) {
				if (type == CALL_TYPE_SPC) {

					subObj = createGuniSpcLblSubObj(eRes.data_link,
									eRes.label);
					if (!insertGuniSubObj(subObj)) {
						throw(std::string("Impossible to ins "
								  "SPC_LABEL"));
					}
				} else if (type == CALL_TYPE_SC || type == CALL_TYPE_AUTO) {
					subObj = createGuniEgrLblSubObj(eRes.data_link,
									eRes.label);
					if (!insertGuniSubObj(subObj)) {
						throw(std::string("Impossible to ins "
								  "EGRESS_LABEL"));
					}
				} else {
					throw(std::string("Call type 0x%x not supported", type));
				}
			}

		}

		GeneralisedUni_Object(g2mpls_addr_t iTna, g2mpls_addr_t eTna) {
			GuniSubObj * subObj;

			subObj = createGuniDstSubObj(eTna);
			if (!insertGuniSubObj(subObj)) {
				throw(std::string("Impossible to ins DST_TNA"));
			}

			subObj = createGuniSrcSubObj(iTna);
			if (!insertGuniSubObj(subObj)) {
				throw(std::string("Impossible to ins SRC_TNA"));
			}
		}

		GeneralisedUni_Object(void) { }

		~GeneralisedUni_Object(void) {
			std::list<GuniSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				DELETE_OBJ(*it);
			}
			value_.subObjs.clear();
		}

		bool isConsistent(void) const {

			// According to OIF-UNI-02.0-RSVP
			if (value_.subObjs.empty()) {
				return true;
			}

			std::list<GuniSubObj *>::const_iterator it;
			GuniSubObj * subObj;

			//it = value_.subObjs.begin();
			//
			//// First subObj MUST be destination_tna
			//GuniSubObj * subObj;
			//subObj = *it;
			//if (!subObj || !subObj->isConsistent() ||
			//    subObj->type() != TYPE_DST_TNA) {
			//	return false;
			//}
			//
			// Second subObj MUST be source_tna
			//it++;
			//subObj = (it != value_.subObjs.end()) ? (*it) : 0;
			//if (!subObj || !subObj->isConsistent() ||
			//    subObj->type() != TYPE_SRC_TNA) {
			//	return false;
			//}
			//
			// Check all the other subObjs consistency
			//for (it++; it != value_.subObjs.end(); it++) {
			//	subObj = *it;
			//	if (!subObj || !subObj->isConsistent()) {
			//		return false;
			//	}
			//}

			// Check all the subObjs consistency
			for (it = value_.subObjs.begin(); it != value_.subObjs.end(); it++) {
				subObj = *it;
				if (!subObj || !subObj->isConsistent()) {
					return false;
				}
			}

			return true;
		}

		uint16_t length(void) const {
			int len = 0;
			std::list<GuniSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it) {
					len += (*it)->length();
				}
			}

			return objHdrLength + len;
		}

		GeneralisedUni_Class * clone(void) const {
			GeneralisedUni_Object * obj;
			obj = new GeneralisedUni_Object();

			if (obj) {
				obj->value_.subObjs.clear();

				std::list<GuniSubObj *>::const_iterator it;
				for (it = value_.subObjs.begin();
				     it != value_.subObjs.end();
				     it++) {
					if (*it) {
						obj->value_.subObjs.push_back
							((*it)->clone());
					}
				}
			}

			return obj;
		}

		bool insertGuniSubObj(GuniSubObj* obj, bool insertTail = true) {

			if (!obj) {
				G2RSVPTE_WRN("Invalid GunisubObj,null pointer");
				return false;
			}

			std::list<GuniSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				if (*it == obj) {
					G2RSVPTE_WRN("Guni subObj %p is "
						     "already present", obj);
					return false;
				}
			}

			if (insertTail) {
				value_.subObjs.push_back(obj);
			} else {
				value_.subObjs.push_front(obj);
			}

			return true;
		}

		GuniSubObj * getSrcTNA(void) {

			if (value_.subObjs.empty()) {
				return 0;
			}

			GuniSubObj * subObj;
			std::list<GuniSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				subObj = *it;
				if (subObj                      &&
				    subObj->type() == TYPE_SRC_TNA) {
					return subObj;
				}
			}

			return 0;
		}

		GuniSubObj * getDestTNA(void) {

			if (value_.subObjs.empty()) {
				return 0;
			}

			GuniSubObj * subObj;
			std::list<GuniSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				subObj = *it;
				if (subObj                      &&
				    subObj->type() == TYPE_DST_TNA) {
					return subObj;
				}
			}

			return 0;
		}

		GuniSubObj * getSpcLabel(void) {

			if (value_.subObjs.empty()) {
				return 0;
			}

			GuniSubObj * subObj;
			std::list<GuniSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				subObj = *it;
				if (subObj                                  &&
				    subObj->type()    == TYPE_EGR_LABEL     &&
				    subObj->subType() == SUBTYPE_EGR_SPC_LABEL){
					return subObj;
				}
			}

			return 0;
		}

		GuniSubObj * getEgressLabel(void) {

			if (value_.subObjs.empty()) {
				return 0;
			}

			GuniSubObj * subObj;
			std::list<GuniSubObj *>::const_iterator it;
			for (it = value_.subObjs.begin();
			     it != value_.subObjs.end();
			     it++) {
				subObj = *it;
				if (subObj                                  &&
				    subObj->type()    == TYPE_EGR_LABEL     &&
				    subObj->subType() == SUBTYPE_EGR_LABEL){
					return subObj;
				}
			}

			return 0;
		}
	};

	DEFINE_CLASS_END(GeneralisedUni_Class);


//CLASS TIMES

//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//     |            Length             |  C-Num (250)  |  C-Type (1)   |
//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//     |                           Start Time                          |
//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
//     |                            End Time                           |
//     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


	DEFINE_CLASS_BEGIN(Times_Class, G2RSVP_CLASS_TIMES);

	DEFINE_OBJECT(Times_Object,
		      CTYPE_TIMES,
		      times_t,
		      Times_Class) {
		OBJ_OPERATOR(Times_Object);
	public:
		Times_Object(void) { };
		Times_Object(uint32_t start_time,
			     uint32_t end_time) {
			value_.start_time = start_time;
			value_.end_time   = end_time;
		}
		bool isConsistent(void) const {
			return true;
		}

		uint16_t length(void) const {
			return objHdrLength + sizeof(times_t);
		}

		Times_Class * clone(void) const {
			Times_Object * obj;
			obj = new Times_Object();
			if (obj) {
				obj->value_ = value_;
			}
			return obj;
		}
	};

	DEFINE_CLASS_END(Times_Class);
};


#endif // G2RSVPTE_MESSAGE_OBJECTS_H
